/* Generated by Cython 0.28.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "brainpy/_c/"
        ],
        "name": "brainpy._c.isotopic_distribution",
        "sources": [
            "brainpy/_c/isotopic_distribution.pyx"
        ]
    },
    "module_name": "brainpy._c.isotopic_distribution"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__brainpy___c__isotopic_distribution
#define __PYX_HAVE_API__brainpy___c__isotopic_distribution
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "brainpy\\_c\\isotopic_distribution.pyx",
  "brainpy\\_c\\isotopic_distribution.pxd",
  "brainpy\\_c\\composition.pxd",
};

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef int count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
typedef int __pyx_t_7brainpy_2_c_11composition_count_type;

/*--- Type declarations ---*/
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */
struct __pyx_t_7brainpy_2_c_11composition_Isotope {
  double mass;
  double abundance;
  int neutrons;
  int neutron_shift;
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;
  size_t size;
};

/* "brainpy/_c/composition.pxd":35
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */
struct __pyx_t_7brainpy_2_c_11composition_Element {
  char *symbol;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;
  int monoisotopic_isotope_index;
};

/* "brainpy/_c/composition.pxd":54
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":63
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;
  size_t size;
};

/* "brainpy/_c/composition.pxd":92
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */
struct __pyx_t_7brainpy_2_c_11composition_Composition {
  char **elements;
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;
  size_t size;
  size_t used;
  int max_variants;
};

/* "brainpy/_c/composition.pxd":124
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double mass(self)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {
  int __pyx_n;
  int free_existing;
};
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector;

/* "brainpy/_c/double_vector.pxd":1
 * cdef struct DoubleVector:             # <<<<<<<<<<<<<<
 *     double* v
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector {
  double *v;
  size_t used;
  size_t size;
};
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants;
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants;

/* "brainpy/_c/isotopic_constants.pxd":6
 * from brainpy._c.double_vector cimport DoubleVector
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_18isotopic_constants_dvec;

/* "brainpy/_c/isotopic_constants.pxd":9
 * 
 * 
 * cdef struct PolynomialParameters:             # <<<<<<<<<<<<<<
 *     dvec* elementary_symmetric_polynomial
 *     dvec* power_sum
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters {
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *elementary_symmetric_polynomial;
  __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *power_sum;
};

/* "brainpy/_c/isotopic_constants.pxd":14
 * 
 * 
 * cdef struct PhiConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     Element* element
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_11composition_Element *element;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *element_coefficients;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PolynomialParameters *mass_coefficients;
};

/* "brainpy/_c/isotopic_constants.pxd":21
 * 
 * 
 * cdef struct IsotopicConstants:             # <<<<<<<<<<<<<<
 *     int order
 *     PhiConstants** constants
 */
struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants {
  int order;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **constants;
  size_t size;
  size_t used;
};
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants;
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants;
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap;

/* "brainpy/_c/isotopic_distribution.pxd":7
 * 
 * 
 * cdef struct Peak:             # <<<<<<<<<<<<<<
 *     double mz
 *     double intensity
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak {
  double mz;
  double intensity;
  int charge;
};

/* "brainpy/_c/isotopic_distribution.pxd":13
 * 
 * 
 * cdef struct PeakList:             # <<<<<<<<<<<<<<
 *     Peak* peaks
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *peaks;
  size_t used;
  size_t size;
};

/* "brainpy/_c/isotopic_distribution.pxd":35
 * 
 * 
 * cdef struct ElementCache:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *source;
};

/* "brainpy/_c/isotopic_distribution.pxd":49
 * 
 * 
 * cdef struct IsotopicDistribution:             # <<<<<<<<<<<<<<
 *     Composition* composition
 *     IsotopicConstants* _isotopic_constants
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *composition;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *_isotopic_constants;
  int order;
  double average_mass;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *monoisotopic_peak;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":58
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":77
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil             # <<<<<<<<<<<<<<
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks {
  int __pyx_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *cache;
};

/* "brainpy/_c/isotopic_distribution.pxd":79
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=*) nogil
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=*, int charge=*, double charge_carrier=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants {
  int __pyx_n;
  PyObject *npeaks;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pxd":82
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=*, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants {
  int __pyx_n;
  int charge;
  double charge_carrier;
};

/* "brainpy/_c/isotopic_distribution.pyx":39
 * #     double_vector_append, print_double_vector)
 * 
 * ctypedef DoubleVector dvec             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec;

/* "brainpy/_c/isotopic_distribution.pyx":50
 * # ElementPolynomialMap Declaration and Methods
 * 
 * cdef struct ElementPolynomialMap:             # <<<<<<<<<<<<<<
 *     char** elements
 *     dvec** polynomials
 */
struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap {
  char **elements;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec **polynomials;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":117
 * cdef Composition* dict_to_composition(dict comp_dict)
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;
  double cached_mass;
  int _clean;
};


/* "brainpy/_c/isotopic_distribution.pxd":63
 * 
 * @freelist(100000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double mz
 */
struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtab;
  double mz;
  double intensity;
  int charge;
};



/* "brainpy/_c/composition.pxd":117
 * cdef Composition* dict_to_composition(dict comp_dict)
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */

struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;


/* "brainpy/_c/isotopic_distribution.pyx":773
 * 
 * @cython.freelist(1000000)
 * cdef class TheoreticalPeak(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, mz, intensity, charge):
 *         self.mz = mz
 */

struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*_create)(double, double, int);
  int (*_eq)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*clone)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__eq(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_clone(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__create(double __pyx_v_mz, double __pyx_v_intensity, int __pyx_v_charge); /* proto*/

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyObject **__pyx_vp_7brainpy_2_c_11composition_nist_mass = 0;
#define __pyx_v_7brainpy_2_c_11composition_nist_mass (*__pyx_vp_7brainpy_2_c_11composition_nist_mass)
static double *__pyx_vp_7brainpy_2_c_11composition_PROTON = 0;
#define __pyx_v_7brainpy_2_c_11composition_PROTON (*__pyx_vp_7brainpy_2_c_11composition_PROTON)
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable **__pyx_vp_7brainpy_2_c_11composition__ElementTable = 0;
#define __pyx_v_7brainpy_2_c_11composition__ElementTable (*__pyx_vp_7brainpy_2_c_11composition__ElementTable)
static double (*__pyx_f_7brainpy_2_c_11composition_mass_charge_ratio)(double, int, struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio *__pyx_optional_args); /*proto*/
static char *(*__pyx_f_7brainpy_2_c_11composition__parse_isotope_string)(char *, int *, char *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Isotope *(*__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift)(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int); /*proto*/
static double (*__pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass)(struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static int (*__pyx_f_7brainpy_2_c_11composition_element_max_neutron_shift)(struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Element *(*__pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element)(struct __pyx_t_7brainpy_2_c_11composition_Element *, int); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *(*__pyx_f_7brainpy_2_c_11composition_make_element_hash_table)(size_t); /*proto*/
static int (*__pyx_f_7brainpy_2_c_11composition_element_hash_table_get)(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **); /*proto*/
static int (*__pyx_f_7brainpy_2_c_11composition_element_hash_table_put)(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static size_t (*__pyx_f_7brainpy_2_c_11composition_free_element_hash_table)(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_11composition_print_composition)(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static int (*__pyx_f_7brainpy_2_c_11composition_composition_get_element_count)(struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type *); /*proto*/
static double (*__pyx_f_7brainpy_2_c_11composition_composition_mass)(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_11composition_free_composition)(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Composition *(*__pyx_f_7brainpy_2_c_11composition_dict_to_composition)(PyObject *); /*proto*/

/* Module declarations from 'brainpy._c.double_vector' */
static struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(*__pyx_f_7brainpy_2_c_13double_vector_make_double_vector_with_size)(size_t); /*proto*/
static struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(*__pyx_f_7brainpy_2_c_13double_vector_make_double_vector)(void); /*proto*/
static int (*__pyx_f_7brainpy_2_c_13double_vector_double_vector_append)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double); /*proto*/
static void (*__pyx_f_7brainpy_2_c_13double_vector_free_double_vector)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_13double_vector_print_double_vector)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_13double_vector_reset_double_vector)(struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *); /*proto*/

/* Module declarations from 'brainpy._c.isotopic_constants' */
static size_t *__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE = 0;
#define __pyx_v_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE (*__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE)
static void (*__pyx_f_7brainpy_2_c_18isotopic_constants_newton)(__pyx_t_7brainpy_2_c_18isotopic_constants_dvec *, __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *, int); /*proto*/
static struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *(*__pyx_f_7brainpy_2_c_18isotopic_constants_make_isotopic_constants)(void); /*proto*/
static int (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_resize)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_18isotopic_constants_free_isotopic_constants)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *); /*proto*/
static void (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_add_element)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *); /*proto*/
static int (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_get)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **); /*proto*/
static void (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_update_coefficients)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *); /*proto*/
static double (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, int); /*proto*/
static double (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_modified_element_power_sum)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, int); /*proto*/
static double (*__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum_by_index)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, size_t, int); /*proto*/
static void (*__pyx_f_7brainpy_2_c_18isotopic_constants_print_isotopic_constants)(struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc' */

/* Module declarations from 'brainpy._c.isotopic_distribution' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list(void); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_peak_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_resize_peak_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *); /*proto*/
static PyObject *__pyx_f_7brainpy_2_c_21isotopic_distribution_peaklist_to_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *); /*proto*/
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_cache(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_cache(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_resize_element_cache(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_put(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *, struct __pyx_t_7brainpy_2_c_11composition_Element **); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **); /*proto*/
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(struct __pyx_t_7brainpy_2_c_11composition_Composition *, int, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution *__pyx_optional_args); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks(struct __pyx_t_7brainpy_2_c_11composition_Composition *, size_t, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args); /*proto*/
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_polynomial_map(size_t); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_set(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *, char *, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_get(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *, char *, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec **); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_polynomial_map(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_validate_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_isotopic_constants(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_order(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, int); /*proto*/
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_make_monoisotopic_peak(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *); /*proto*/
static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_phi_values(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *); /*proto*/
static double __pyx_f_7brainpy_2_c_21isotopic_distribution__id_phi_value(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, int); /*proto*/
static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_modified_phi_values(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, char *, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *); /*proto*/
static double __pyx_f_7brainpy_2_c_21isotopic_distribution__id_modified_phi_value(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, char *, int); /*proto*/
static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_probability_vector(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *); /*proto*/
static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_center_mass_vector(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *); /*proto*/
static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_aggregated_isotopic_variants(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *, int, double); /*proto*/
static void __pyx_f_7brainpy_2_c_21isotopic_distribution_sort_by_mz(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_compare_by_mz(void const *, void const *); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(PyObject *, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "brainpy._c.isotopic_distribution"
extern int __pyx_module_is_main_brainpy___c__isotopic_distribution;
int __pyx_module_is_main_brainpy___c__isotopic_distribution = 0;

/* Implementation of 'brainpy._c.isotopic_distribution' */
static PyObject *__pyx_builtin_range;
static const char __pyx_k_H[] = "H";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_eq[] = "_eq";
static const char __pyx_k_mz[] = "mz";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_Done[] = "Done";
static const char __pyx_k_dist[] = "dist";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "test";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_main_2[] = "main";
static const char __pyx_k_npeaks[] = "npeaks";
static const char __pyx_k_test_2[] = "__test__";
static const char __pyx_k_peak_set[] = "peak_set";
static const char __pyx_k_peaklist[] = "peaklist";
static const char __pyx_k_Free_Done[] = "Free Done";
static const char __pyx_k_comp_dict[] = "comp_dict";
static const char __pyx_k_intensity[] = "intensity";
static const char __pyx_k_max_npeaks[] = "max_npeaks";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_Really_done[] = "Really done";
static const char __pyx_k_Second_Free[] = "Second Free";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_distribution[] = "distribution";
static const char __pyx_k_distribution2[] = "distribution2";
static const char __pyx_k_Trying_to_free[] = "Trying to free";
static const char __pyx_k_charge_carrier[] = "charge_carrier";
static const char __pyx_k_max_n_variants[] = "max_n_variants";
static const char __pyx_k_Guessed_of_Peaks[] = "Guessed # of Peaks: ";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_composition_struct[] = "composition_struct";
static const char __pyx_k_Seconc_construction[] = "Seconc construction";
static const char __pyx_k_pyisotopic_variants[] = "pyisotopic_variants";
static const char __pyx_k_Actual_of_Peaks_Returned[] = "Actual # of Peaks Returned:";
static const char __pyx_k_Going_to_print_constants[] = "Going to print constants";
static const char __pyx_k_Size_of_probability_vector[] = "Size of probability vector:";
static const char __pyx_k_Peak_mz_f_intensity_f_charge_d[] = "Peak(mz=%f, intensity=%f, charge=%d)";
static const char __pyx_k_brainpy__c_isotopic_distribution[] = "brainpy\\_c\\isotopic_distribution.pyx";
static const char __pyx_k_brainpy__c_isotopic_distribution_2[] = "brainpy._c.isotopic_distribution";
static PyObject *__pyx_kp_s_Actual_of_Peaks_Returned;
static PyObject *__pyx_n_s_Done;
static PyObject *__pyx_kp_s_Free_Done;
static PyObject *__pyx_kp_s_Going_to_print_constants;
static PyObject *__pyx_kp_s_Guessed_of_Peaks;
static PyObject *__pyx_n_s_H;
static PyObject *__pyx_n_s_O;
static PyObject *__pyx_kp_s_Peak_mz_f_intensity_f_charge_d;
static PyObject *__pyx_kp_s_Really_done;
static PyObject *__pyx_kp_s_Seconc_construction;
static PyObject *__pyx_kp_s_Second_Free;
static PyObject *__pyx_kp_s_Size_of_probability_vector;
static PyObject *__pyx_kp_s_Trying_to_free;
static PyObject *__pyx_kp_s_brainpy__c_isotopic_distribution;
static PyObject *__pyx_n_s_brainpy__c_isotopic_distribution_2;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_charge_carrier;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_comp_dict;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_composition_struct;
static PyObject *__pyx_n_s_dist;
static PyObject *__pyx_n_s_distribution;
static PyObject *__pyx_n_s_distribution2;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_intensity;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_main_2;
static PyObject *__pyx_n_s_max_n_variants;
static PyObject *__pyx_n_s_max_npeaks;
static PyObject *__pyx_n_s_mz;
static PyObject *__pyx_n_s_npeaks;
static PyObject *__pyx_n_s_peak_set;
static PyObject *__pyx_n_s_peaklist;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyisotopic_variants;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_test_2;
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_2pyisotopic_variants(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_npeaks, int __pyx_v_charge, double __pyx_v_charge_carrier); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_4_isotopic_variants(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_npeaks, int __pyx_v_charge, double __pyx_v_charge_carrier); /* proto */
static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak___init__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_mz, PyObject *__pyx_v_intensity, PyObject *__pyx_v_charge); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2__repr__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_4_eq(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6__hash__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_8__richcmp__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_10__reduce__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_12clone(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self); /* proto */
static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_6main(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_8test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, int __pyx_v_max_npeaks); /* proto */
static PyObject *__pyx_tp_new_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static double __pyx_k_;
static double __pyx_k__2;
static double __pyx_k__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
/* Late includes */

/* "brainpy/_c/isotopic_distribution.pyx":56
 *     size_t size
 * 
 * cdef ElementPolynomialMap* make_element_polynomial_map(size_t sizehint) nogil:             # <<<<<<<<<<<<<<
 *     cdef ElementPolynomialMap* result
 *     result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_polynomial_map(size_t __pyx_v_sizehint) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_v_result;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_r;

  /* "brainpy/_c/isotopic_distribution.pyx":58
 * cdef ElementPolynomialMap* make_element_polynomial_map(size_t sizehint) nogil:
 *     cdef ElementPolynomialMap* result
 *     result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))             # <<<<<<<<<<<<<<
 *     result.elements = <char**>malloc(sizeof(char*) * sizehint)
 *     result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)
 */
  __pyx_v_result = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap))));

  /* "brainpy/_c/isotopic_distribution.pyx":59
 *     cdef ElementPolynomialMap* result
 *     result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))
 *     result.elements = <char**>malloc(sizeof(char*) * sizehint)             # <<<<<<<<<<<<<<
 *     result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)
 *     result.size = sizehint
 */
  __pyx_v_result->elements = ((char **)malloc(((sizeof(char *)) * __pyx_v_sizehint)));

  /* "brainpy/_c/isotopic_distribution.pyx":60
 *     result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))
 *     result.elements = <char**>malloc(sizeof(char*) * sizehint)
 *     result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)             # <<<<<<<<<<<<<<
 *     result.size = sizehint
 *     result.used = 0
 */
  __pyx_v_result->polynomials = ((__pyx_t_7brainpy_2_c_21isotopic_distribution_dvec **)malloc(((sizeof(__pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *)) * __pyx_v_sizehint)));

  /* "brainpy/_c/isotopic_distribution.pyx":61
 *     result.elements = <char**>malloc(sizeof(char*) * sizehint)
 *     result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)
 *     result.size = sizehint             # <<<<<<<<<<<<<<
 *     result.used = 0
 * 
 */
  __pyx_v_result->size = __pyx_v_sizehint;

  /* "brainpy/_c/isotopic_distribution.pyx":62
 *     result.polynomials = <dvec**>malloc(sizeof(dvec*) * sizehint)
 *     result.size = sizehint
 *     result.used = 0             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->used = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":64
 *     result.used = 0
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef int element_polynomial_map_set(ElementPolynomialMap* ep_map, char* element, dvec* polynomial) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":56
 *     size_t size
 * 
 * cdef ElementPolynomialMap* make_element_polynomial_map(size_t sizehint) nogil:             # <<<<<<<<<<<<<<
 *     cdef ElementPolynomialMap* result
 *     result = <ElementPolynomialMap*>malloc(sizeof(ElementPolynomialMap))
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":66
 *     return result
 * 
 * cdef int element_polynomial_map_set(ElementPolynomialMap* ep_map, char* element, dvec* polynomial) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_set(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_v_ep_map, char *__pyx_v_element, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_polynomial) {
  size_t __pyx_v_i;
  int __pyx_v_done;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":72
 *         bint done
 * 
 *     done = False             # <<<<<<<<<<<<<<
 *     i = 0
 *     while (i < ep_map.used):
 */
  __pyx_v_done = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":73
 * 
 *     done = False
 *     i = 0             # <<<<<<<<<<<<<<
 *     while (i < ep_map.used):
 *         if strcmp(element, ep_map.elements[i]) == 0:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":74
 *     done = False
 *     i = 0
 *     while (i < ep_map.used):             # <<<<<<<<<<<<<<
 *         if strcmp(element, ep_map.elements[i]) == 0:
 *             done = True
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_ep_map->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/isotopic_distribution.pyx":75
 *     i = 0
 *     while (i < ep_map.used):
 *         if strcmp(element, ep_map.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             ep_map.polynomials[i] = polynomial
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element, (__pyx_v_ep_map->elements[__pyx_v_i])) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":76
 *     while (i < ep_map.used):
 *         if strcmp(element, ep_map.elements[i]) == 0:
 *             done = True             # <<<<<<<<<<<<<<
 *             ep_map.polynomials[i] = polynomial
 *             return 0
 */
      __pyx_v_done = 1;

      /* "brainpy/_c/isotopic_distribution.pyx":77
 *         if strcmp(element, ep_map.elements[i]) == 0:
 *             done = True
 *             ep_map.polynomials[i] = polynomial             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
      (__pyx_v_ep_map->polynomials[__pyx_v_i]) = __pyx_v_polynomial;

      /* "brainpy/_c/isotopic_distribution.pyx":78
 *             done = True
 *             ep_map.polynomials[i] = polynomial
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 *     if not done:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/isotopic_distribution.pyx":75
 *     i = 0
 *     while (i < ep_map.used):
 *         if strcmp(element, ep_map.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             ep_map.polynomials[i] = polynomial
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":79
 *             ep_map.polynomials[i] = polynomial
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 *     if not done:
 *         ep_map.used += 1
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":80
 *             return 0
 *         i += 1
 *     if not done:             # <<<<<<<<<<<<<<
 *         ep_map.used += 1
 *         if ep_map.used >= ep_map.size:
 */
  __pyx_t_1 = ((!(__pyx_v_done != 0)) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":81
 *         i += 1
 *     if not done:
 *         ep_map.used += 1             # <<<<<<<<<<<<<<
 *         if ep_map.used >= ep_map.size:
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
 */
    __pyx_v_ep_map->used = (__pyx_v_ep_map->used + 1);

    /* "brainpy/_c/isotopic_distribution.pyx":82
 *     if not done:
 *         ep_map.used += 1
 *         if ep_map.used >= ep_map.size:             # <<<<<<<<<<<<<<
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
 *             return -1
 */
    __pyx_t_1 = ((__pyx_v_ep_map->used >= __pyx_v_ep_map->size) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":83
 *         ep_map.used += 1
 *         if ep_map.used >= ep_map.size:
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)             # <<<<<<<<<<<<<<
 *             return -1
 *         ep_map.elements[i] = element
 */
      (void)(printf(((char const *)"Overloaded ElementPolynomialMap\n %d, %d\n"), __pyx_v_i, __pyx_v_ep_map->size));

      /* "brainpy/_c/isotopic_distribution.pyx":84
 *         if ep_map.used >= ep_map.size:
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
 *             return -1             # <<<<<<<<<<<<<<
 *         ep_map.elements[i] = element
 *         ep_map.polynomials[i] = polynomial
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "brainpy/_c/isotopic_distribution.pyx":82
 *     if not done:
 *         ep_map.used += 1
 *         if ep_map.used >= ep_map.size:             # <<<<<<<<<<<<<<
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
 *             return -1
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":85
 *             printf("Overloaded ElementPolynomialMap\n %d, %d\n", i, ep_map.size)
 *             return -1
 *         ep_map.elements[i] = element             # <<<<<<<<<<<<<<
 *         ep_map.polynomials[i] = polynomial
 *         return 0
 */
    (__pyx_v_ep_map->elements[__pyx_v_i]) = __pyx_v_element;

    /* "brainpy/_c/isotopic_distribution.pyx":86
 *             return -1
 *         ep_map.elements[i] = element
 *         ep_map.polynomials[i] = polynomial             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
    (__pyx_v_ep_map->polynomials[__pyx_v_i]) = __pyx_v_polynomial;

    /* "brainpy/_c/isotopic_distribution.pyx":87
 *         ep_map.elements[i] = element
 *         ep_map.polynomials[i] = polynomial
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":80
 *             return 0
 *         i += 1
 *     if not done:             # <<<<<<<<<<<<<<
 *         ep_map.used += 1
 *         if ep_map.used >= ep_map.size:
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":88
 *         ep_map.polynomials[i] = polynomial
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef int element_polynomial_map_get(ElementPolynomialMap* ep_map, char* element, dvec** polynomial) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":66
 *     return result
 * 
 * cdef int element_polynomial_map_set(ElementPolynomialMap* ep_map, char* element, dvec* polynomial) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":90
 *     return 1
 * 
 * cdef int element_polynomial_map_get(ElementPolynomialMap* ep_map, char* element, dvec** polynomial) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_get(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_v_ep_map, char *__pyx_v_element, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec **__pyx_v_polynomial) {
  size_t __pyx_v_i;
  CYTHON_UNUSED int __pyx_v_done;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":96
 *         bint done
 * 
 *     done = False             # <<<<<<<<<<<<<<
 *     i = 0
 *     while i < ep_map.used:
 */
  __pyx_v_done = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":97
 * 
 *     done = False
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < ep_map.used:
 *         if strcmp(ep_map.elements[i], element) == 0:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":98
 *     done = False
 *     i = 0
 *     while i < ep_map.used:             # <<<<<<<<<<<<<<
 *         if strcmp(ep_map.elements[i], element) == 0:
 *             polynomial[0] = ep_map.polynomials[i]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_ep_map->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/isotopic_distribution.pyx":99
 *     i = 0
 *     while i < ep_map.used:
 *         if strcmp(ep_map.elements[i], element) == 0:             # <<<<<<<<<<<<<<
 *             polynomial[0] = ep_map.polynomials[i]
 *             return 0
 */
    __pyx_t_1 = ((strcmp((__pyx_v_ep_map->elements[__pyx_v_i]), __pyx_v_element) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":100
 *     while i < ep_map.used:
 *         if strcmp(ep_map.elements[i], element) == 0:
 *             polynomial[0] = ep_map.polynomials[i]             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
      (__pyx_v_polynomial[0]) = (__pyx_v_ep_map->polynomials[__pyx_v_i]);

      /* "brainpy/_c/isotopic_distribution.pyx":101
 *         if strcmp(ep_map.elements[i], element) == 0:
 *             polynomial[0] = ep_map.polynomials[i]
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 *     return 1
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/isotopic_distribution.pyx":99
 *     i = 0
 *     while i < ep_map.used:
 *         if strcmp(ep_map.elements[i], element) == 0:             # <<<<<<<<<<<<<<
 *             polynomial[0] = ep_map.polynomials[i]
 *             return 0
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":102
 *             polynomial[0] = ep_map.polynomials[i]
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":103
 *             return 0
 *         i += 1
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef void free_element_polynomial_map(ElementPolynomialMap* ep_map) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":90
 *     return 1
 * 
 * cdef int element_polynomial_map_get(ElementPolynomialMap* ep_map, char* element, dvec** polynomial) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":105
 *     return 1
 * 
 * cdef void free_element_polynomial_map(ElementPolynomialMap* ep_map) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_polynomial_map(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_v_ep_map) {
  size_t __pyx_v_i;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":108
 *     cdef:
 *         size_t i
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < ep_map.used:
 *         free_double_vector(ep_map.polynomials[i])
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":109
 *         size_t i
 *     i = 0
 *     while i < ep_map.used:             # <<<<<<<<<<<<<<
 *         free_double_vector(ep_map.polynomials[i])
 *         i += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_ep_map->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/isotopic_distribution.pyx":110
 *     i = 0
 *     while i < ep_map.used:
 *         free_double_vector(ep_map.polynomials[i])             # <<<<<<<<<<<<<<
 *         i += 1
 *     free(ep_map.elements)
 */
    __pyx_f_7brainpy_2_c_13double_vector_free_double_vector((__pyx_v_ep_map->polynomials[__pyx_v_i]));

    /* "brainpy/_c/isotopic_distribution.pyx":111
 *     while i < ep_map.used:
 *         free_double_vector(ep_map.polynomials[i])
 *         i += 1             # <<<<<<<<<<<<<<
 *     free(ep_map.elements)
 *     free(ep_map.polynomials)
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":112
 *         free_double_vector(ep_map.polynomials[i])
 *         i += 1
 *     free(ep_map.elements)             # <<<<<<<<<<<<<<
 *     free(ep_map.polynomials)
 *     free(ep_map)
 */
  free(__pyx_v_ep_map->elements);

  /* "brainpy/_c/isotopic_distribution.pyx":113
 *         i += 1
 *     free(ep_map.elements)
 *     free(ep_map.polynomials)             # <<<<<<<<<<<<<<
 *     free(ep_map)
 * 
 */
  free(__pyx_v_ep_map->polynomials);

  /* "brainpy/_c/isotopic_distribution.pyx":114
 *     free(ep_map.elements)
 *     free(ep_map.polynomials)
 *     free(ep_map)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_ep_map);

  /* "brainpy/_c/isotopic_distribution.pyx":105
 *     return 1
 * 
 * cdef void free_element_polynomial_map(ElementPolynomialMap* ep_map) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":120
 * # Peak Methods
 * 
 * cdef void print_peak(Peak* peak) nogil:             # <<<<<<<<<<<<<<
 *     printf("Peak: %f, %f, %d\n", peak.mz, peak.intensity, peak.charge)
 * 
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_print_peak(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_v_peak) {

  /* "brainpy/_c/isotopic_distribution.pyx":121
 * 
 * cdef void print_peak(Peak* peak) nogil:
 *     printf("Peak: %f, %f, %d\n", peak.mz, peak.intensity, peak.charge)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(printf(((char const *)"Peak: %f, %f, %d\n"), __pyx_v_peak->mz, __pyx_v_peak->intensity, __pyx_v_peak->charge));

  /* "brainpy/_c/isotopic_distribution.pyx":120
 * # Peak Methods
 * 
 * cdef void print_peak(Peak* peak) nogil:             # <<<<<<<<<<<<<<
 *     printf("Peak: %f, %f, %d\n", peak.mz, peak.intensity, peak.charge)
 * 
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":124
 * 
 * 
 * cdef Peak* make_peak(double mz, double intensity, int charge) nogil:             # <<<<<<<<<<<<<<
 *     cdef Peak* peak
 *     peak = <Peak*>malloc(sizeof(Peak))
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak(double __pyx_v_mz, double __pyx_v_intensity, int __pyx_v_charge) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_v_peak;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_r;

  /* "brainpy/_c/isotopic_distribution.pyx":126
 * cdef Peak* make_peak(double mz, double intensity, int charge) nogil:
 *     cdef Peak* peak
 *     peak = <Peak*>malloc(sizeof(Peak))             # <<<<<<<<<<<<<<
 *     peak.mz = mz
 *     peak.intensity = intensity
 */
  __pyx_v_peak = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak))));

  /* "brainpy/_c/isotopic_distribution.pyx":127
 *     cdef Peak* peak
 *     peak = <Peak*>malloc(sizeof(Peak))
 *     peak.mz = mz             # <<<<<<<<<<<<<<
 *     peak.intensity = intensity
 *     peak.charge = charge
 */
  __pyx_v_peak->mz = __pyx_v_mz;

  /* "brainpy/_c/isotopic_distribution.pyx":128
 *     peak = <Peak*>malloc(sizeof(Peak))
 *     peak.mz = mz
 *     peak.intensity = intensity             # <<<<<<<<<<<<<<
 *     peak.charge = charge
 *     return peak
 */
  __pyx_v_peak->intensity = __pyx_v_intensity;

  /* "brainpy/_c/isotopic_distribution.pyx":129
 *     peak.mz = mz
 *     peak.intensity = intensity
 *     peak.charge = charge             # <<<<<<<<<<<<<<
 *     return peak
 * 
 */
  __pyx_v_peak->charge = __pyx_v_charge;

  /* "brainpy/_c/isotopic_distribution.pyx":130
 *     peak.intensity = intensity
 *     peak.charge = charge
 *     return peak             # <<<<<<<<<<<<<<
 * 
 * # -----------------------------------------------------------------------------
 */
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":124
 * 
 * 
 * cdef Peak* make_peak(double mz, double intensity, int charge) nogil:             # <<<<<<<<<<<<<<
 *     cdef Peak* peak
 *     peak = <Peak*>malloc(sizeof(Peak))
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":135
 * # PeakList Methods
 * 
 * cdef PeakList* make_peak_list() nogil:             # <<<<<<<<<<<<<<
 *     cdef PeakList* result
 * 
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list(void) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_result;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_r;

  /* "brainpy/_c/isotopic_distribution.pyx":138
 *     cdef PeakList* result
 * 
 *     result = <PeakList*>malloc(sizeof(PeakList))             # <<<<<<<<<<<<<<
 *     result.peaks = <Peak*>malloc(sizeof(Peak) * 10)
 *     result.size = 10
 */
  __pyx_v_result = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList))));

  /* "brainpy/_c/isotopic_distribution.pyx":139
 * 
 *     result = <PeakList*>malloc(sizeof(PeakList))
 *     result.peaks = <Peak*>malloc(sizeof(Peak) * 10)             # <<<<<<<<<<<<<<
 *     result.size = 10
 *     result.used = 0
 */
  __pyx_v_result->peaks = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak)) * 10)));

  /* "brainpy/_c/isotopic_distribution.pyx":140
 *     result = <PeakList*>malloc(sizeof(PeakList))
 *     result.peaks = <Peak*>malloc(sizeof(Peak) * 10)
 *     result.size = 10             # <<<<<<<<<<<<<<
 *     result.used = 0
 * 
 */
  __pyx_v_result->size = 10;

  /* "brainpy/_c/isotopic_distribution.pyx":141
 *     result.peaks = <Peak*>malloc(sizeof(Peak) * 10)
 *     result.size = 10
 *     result.used = 0             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->used = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":143
 *     result.used = 0
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef void free_peak_list(PeakList* peaklist) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":135
 * # PeakList Methods
 * 
 * cdef PeakList* make_peak_list() nogil:             # <<<<<<<<<<<<<<
 *     cdef PeakList* result
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":145
 *     return result
 * 
 * cdef void free_peak_list(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     free(peaklist.peaks)
 *     free(peaklist)
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_peak_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist) {

  /* "brainpy/_c/isotopic_distribution.pyx":146
 * 
 * cdef void free_peak_list(PeakList* peaklist) nogil:
 *     free(peaklist.peaks)             # <<<<<<<<<<<<<<
 *     free(peaklist)
 * 
 */
  free(__pyx_v_peaklist->peaks);

  /* "brainpy/_c/isotopic_distribution.pyx":147
 * cdef void free_peak_list(PeakList* peaklist) nogil:
 *     free(peaklist.peaks)
 *     free(peaklist)             # <<<<<<<<<<<<<<
 * 
 * cdef int resize_peak_list(PeakList* peaklist) nogil:
 */
  free(__pyx_v_peaklist);

  /* "brainpy/_c/isotopic_distribution.pyx":145
 *     return result
 * 
 * cdef void free_peak_list(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     free(peaklist.peaks)
 *     free(peaklist)
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":149
 *     free(peaklist)
 * 
 * cdef int resize_peak_list(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Peak* peaks
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_resize_peak_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_v_peaks;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":152
 *     cdef:
 *         Peak* peaks
 *     peaks = <Peak*>realloc(peaklist.peaks, sizeof(Peak) * peaklist.size * 2)             # <<<<<<<<<<<<<<
 *     if peaks == NULL:
 *         printf("realloc peaklist returned NULL\n")
 */
  __pyx_v_peaks = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)realloc(__pyx_v_peaklist->peaks, (((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak)) * __pyx_v_peaklist->size) * 2)));

  /* "brainpy/_c/isotopic_distribution.pyx":153
 *         Peak* peaks
 *     peaks = <Peak*>realloc(peaklist.peaks, sizeof(Peak) * peaklist.size * 2)
 *     if peaks == NULL:             # <<<<<<<<<<<<<<
 *         printf("realloc peaklist returned NULL\n")
 *         return -1
 */
  __pyx_t_1 = ((__pyx_v_peaks == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":154
 *     peaks = <Peak*>realloc(peaklist.peaks, sizeof(Peak) * peaklist.size * 2)
 *     if peaks == NULL:
 *         printf("realloc peaklist returned NULL\n")             # <<<<<<<<<<<<<<
 *         return -1
 *     peaklist.peaks = peaks
 */
    (void)(printf(((char const *)"realloc peaklist returned NULL\n")));

    /* "brainpy/_c/isotopic_distribution.pyx":155
 *     if peaks == NULL:
 *         printf("realloc peaklist returned NULL\n")
 *         return -1             # <<<<<<<<<<<<<<
 *     peaklist.peaks = peaks
 *     peaklist.size *= 2
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":153
 *         Peak* peaks
 *     peaks = <Peak*>realloc(peaklist.peaks, sizeof(Peak) * peaklist.size * 2)
 *     if peaks == NULL:             # <<<<<<<<<<<<<<
 *         printf("realloc peaklist returned NULL\n")
 *         return -1
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":156
 *         printf("realloc peaklist returned NULL\n")
 *         return -1
 *     peaklist.peaks = peaks             # <<<<<<<<<<<<<<
 *     peaklist.size *= 2
 *     return 0
 */
  __pyx_v_peaklist->peaks = __pyx_v_peaks;

  /* "brainpy/_c/isotopic_distribution.pyx":157
 *         return -1
 *     peaklist.peaks = peaks
 *     peaklist.size *= 2             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_peaklist->size = (__pyx_v_peaklist->size * 2);

  /* "brainpy/_c/isotopic_distribution.pyx":158
 *     peaklist.peaks = peaks
 *     peaklist.size *= 2
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":149
 *     free(peaklist)
 * 
 * cdef int resize_peak_list(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Peak* peaks
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":160
 *     return 0
 * 
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:             # <<<<<<<<<<<<<<
 *     if peaklist.used == peaklist.size:
 *         resize_peak_list(peaklist)
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_v_peak) {
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":161
 * 
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:
 *     if peaklist.used == peaklist.size:             # <<<<<<<<<<<<<<
 *         resize_peak_list(peaklist)
 *     peaklist.peaks[peaklist.used] = peak[0]
 */
  __pyx_t_1 = ((__pyx_v_peaklist->used == __pyx_v_peaklist->size) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":162
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:
 *     if peaklist.used == peaklist.size:
 *         resize_peak_list(peaklist)             # <<<<<<<<<<<<<<
 *     peaklist.peaks[peaklist.used] = peak[0]
 *     peaklist.used += 1
 */
    (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_resize_peak_list(__pyx_v_peaklist));

    /* "brainpy/_c/isotopic_distribution.pyx":161
 * 
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:
 *     if peaklist.used == peaklist.size:             # <<<<<<<<<<<<<<
 *         resize_peak_list(peaklist)
 *     peaklist.peaks[peaklist.used] = peak[0]
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":163
 *     if peaklist.used == peaklist.size:
 *         resize_peak_list(peaklist)
 *     peaklist.peaks[peaklist.used] = peak[0]             # <<<<<<<<<<<<<<
 *     peaklist.used += 1
 * 
 */
  (__pyx_v_peaklist->peaks[__pyx_v_peaklist->used]) = (__pyx_v_peak[0]);

  /* "brainpy/_c/isotopic_distribution.pyx":164
 *         resize_peak_list(peaklist)
 *     peaklist.peaks[peaklist.used] = peak[0]
 *     peaklist.used += 1             # <<<<<<<<<<<<<<
 * 
 * cdef void peak_list_reset(PeakList* peaklist) nogil:
 */
  __pyx_v_peaklist->used = (__pyx_v_peaklist->used + 1);

  /* "brainpy/_c/isotopic_distribution.pyx":160
 *     return 0
 * 
 * cdef void peak_list_append(PeakList* peaklist, Peak* peak) nogil:             # <<<<<<<<<<<<<<
 *     if peaklist.used == peaklist.size:
 *         resize_peak_list(peaklist)
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":166
 *     peaklist.used += 1
 * 
 * cdef void peak_list_reset(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     peaklist.used = 0
 * 
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_reset(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist) {

  /* "brainpy/_c/isotopic_distribution.pyx":167
 * 
 * cdef void peak_list_reset(PeakList* peaklist) nogil:
 *     peaklist.used = 0             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision
 */
  __pyx_v_peaklist->used = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":166
 *     peaklist.used += 1
 * 
 * cdef void peak_list_reset(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     peaklist.used = 0
 * 
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":170
 * 
 * @cython.cdivision
 * cdef PeakList* peak_list_ignore_below(PeakList* peaklist, double ignore_below, PeakList* result) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double total
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_ignore_below(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist, double __pyx_v_ignore_below, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_result) {
  double __pyx_v_total;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak __pyx_v_p;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  size_t __pyx_t_6;

  /* "brainpy/_c/isotopic_distribution.pyx":177
 *         Peak p
 * 
 *     total = 0             # <<<<<<<<<<<<<<
 *     n = peaklist.used
 * 
 */
  __pyx_v_total = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":178
 * 
 *     total = 0
 *     n = peaklist.used             # <<<<<<<<<<<<<<
 * 
 *     if result == NULL:
 */
  __pyx_t_1 = __pyx_v_peaklist->used;
  __pyx_v_n = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":180
 *     n = peaklist.used
 * 
 *     if result == NULL:             # <<<<<<<<<<<<<<
 *         result = make_peak_list()
 * 
 */
  __pyx_t_2 = ((__pyx_v_result == NULL) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/isotopic_distribution.pyx":181
 * 
 *     if result == NULL:
 *         result = make_peak_list()             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
    __pyx_v_result = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list();

    /* "brainpy/_c/isotopic_distribution.pyx":180
 *     n = peaklist.used
 * 
 *     if result == NULL:             # <<<<<<<<<<<<<<
 *         result = make_peak_list()
 * 
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":183
 *         result = make_peak_list()
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         p = result.peaks[i]
 *         if (p.intensity < ignore_below) and (i > 1):
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":184
 * 
 *     for i in range(n):
 *         p = result.peaks[i]             # <<<<<<<<<<<<<<
 *         if (p.intensity < ignore_below) and (i > 1):
 *             continue
 */
    __pyx_v_p = (__pyx_v_result->peaks[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":185
 *     for i in range(n):
 *         p = result.peaks[i]
 *         if (p.intensity < ignore_below) and (i > 1):             # <<<<<<<<<<<<<<
 *             continue
 *         else:
 */
    __pyx_t_5 = ((__pyx_v_p.intensity < __pyx_v_ignore_below) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = ((__pyx_v_i > 1) != 0);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_2) {

      /* "brainpy/_c/isotopic_distribution.pyx":186
 *         p = result.peaks[i]
 *         if (p.intensity < ignore_below) and (i > 1):
 *             continue             # <<<<<<<<<<<<<<
 *         else:
 *             total += p.intensity
 */
      goto __pyx_L4_continue;

      /* "brainpy/_c/isotopic_distribution.pyx":185
 *     for i in range(n):
 *         p = result.peaks[i]
 *         if (p.intensity < ignore_below) and (i > 1):             # <<<<<<<<<<<<<<
 *             continue
 *         else:
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":188
 *             continue
 *         else:
 *             total += p.intensity             # <<<<<<<<<<<<<<
 *         peak_list_append(result, &p)
 *     n = result.used
 */
    /*else*/ {
      __pyx_v_total = (__pyx_v_total + __pyx_v_p.intensity);
    }

    /* "brainpy/_c/isotopic_distribution.pyx":189
 *         else:
 *             total += p.intensity
 *         peak_list_append(result, &p)             # <<<<<<<<<<<<<<
 *     n = result.used
 *     for i in range(n):
 */
    __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append(__pyx_v_result, (&__pyx_v_p));
    __pyx_L4_continue:;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":190
 *             total += p.intensity
 *         peak_list_append(result, &p)
 *     n = result.used             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         result.peaks[i].intensity /= total
 */
  __pyx_t_1 = __pyx_v_result->used;
  __pyx_v_n = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":191
 *         peak_list_append(result, &p)
 *     n = result.used
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         result.peaks[i].intensity /= total
 *     return result
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":192
 *     n = result.used
 *     for i in range(n):
 *         result.peaks[i].intensity /= total             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_t_6 = __pyx_v_i;
    (__pyx_v_result->peaks[__pyx_t_6]).intensity = ((__pyx_v_result->peaks[__pyx_t_6]).intensity / __pyx_v_total);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":193
 *     for i in range(n):
 *         result.peaks[i].intensity /= total
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":170
 * 
 * @cython.cdivision
 * cdef PeakList* peak_list_ignore_below(PeakList* peaklist, double ignore_below, PeakList* result) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double total
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":196
 * 
 * @cython.cdivision
 * cdef PeakList* peak_list_truncate_after(PeakList* peaklist, double truncate_after, PeakList* result) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double cumsum
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_truncate_after(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist, double __pyx_v_truncate_after, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_result) {
  double __pyx_v_cumsum;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak __pyx_v_peak;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;

  /* "brainpy/_c/isotopic_distribution.pyx":202
 *         size_t i, n
 * 
 *     cumsum = 0             # <<<<<<<<<<<<<<
 *     n = peaklist.used
 *     if result == NULL:
 */
  __pyx_v_cumsum = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":203
 * 
 *     cumsum = 0
 *     n = peaklist.used             # <<<<<<<<<<<<<<
 *     if result == NULL:
 *         result = make_peak_list()
 */
  __pyx_t_1 = __pyx_v_peaklist->used;
  __pyx_v_n = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":204
 *     cumsum = 0
 *     n = peaklist.used
 *     if result == NULL:             # <<<<<<<<<<<<<<
 *         result = make_peak_list()
 * 
 */
  __pyx_t_2 = ((__pyx_v_result == NULL) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/isotopic_distribution.pyx":205
 *     n = peaklist.used
 *     if result == NULL:
 *         result = make_peak_list()             # <<<<<<<<<<<<<<
 * 
 *     for i in range(n):
 */
    __pyx_v_result = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list();

    /* "brainpy/_c/isotopic_distribution.pyx":204
 *     cumsum = 0
 *     n = peaklist.used
 *     if result == NULL:             # <<<<<<<<<<<<<<
 *         result = make_peak_list()
 * 
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":207
 *         result = make_peak_list()
 * 
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peak = peaklist.peaks[i]
 *         cumsum += peak.intensity
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":208
 * 
 *     for i in range(n):
 *         peak = peaklist.peaks[i]             # <<<<<<<<<<<<<<
 *         cumsum += peak.intensity
 *         peak_list_append(result, &peak)
 */
    __pyx_v_peak = (__pyx_v_peaklist->peaks[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":209
 *     for i in range(n):
 *         peak = peaklist.peaks[i]
 *         cumsum += peak.intensity             # <<<<<<<<<<<<<<
 *         peak_list_append(result, &peak)
 *         if cumsum >= truncate_after:
 */
    __pyx_v_cumsum = (__pyx_v_cumsum + __pyx_v_peak.intensity);

    /* "brainpy/_c/isotopic_distribution.pyx":210
 *         peak = peaklist.peaks[i]
 *         cumsum += peak.intensity
 *         peak_list_append(result, &peak)             # <<<<<<<<<<<<<<
 *         if cumsum >= truncate_after:
 *             break
 */
    __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append(__pyx_v_result, (&__pyx_v_peak));

    /* "brainpy/_c/isotopic_distribution.pyx":211
 *         cumsum += peak.intensity
 *         peak_list_append(result, &peak)
 *         if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    __pyx_t_2 = ((__pyx_v_cumsum >= __pyx_v_truncate_after) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/isotopic_distribution.pyx":212
 *         peak_list_append(result, &peak)
 *         if cumsum >= truncate_after:
 *             break             # <<<<<<<<<<<<<<
 * 
 *     n = result.used
 */
      goto __pyx_L5_break;

      /* "brainpy/_c/isotopic_distribution.pyx":211
 *         cumsum += peak.intensity
 *         peak_list_append(result, &peak)
 *         if cumsum >= truncate_after:             # <<<<<<<<<<<<<<
 *             break
 * 
 */
    }
  }
  __pyx_L5_break:;

  /* "brainpy/_c/isotopic_distribution.pyx":214
 *             break
 * 
 *     n = result.used             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         result.peaks[i].intensity /= cumsum
 */
  __pyx_t_1 = __pyx_v_result->used;
  __pyx_v_n = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":215
 * 
 *     n = result.used
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         result.peaks[i].intensity /= cumsum
 *     return result
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":216
 *     n = result.used
 *     for i in range(n):
 *         result.peaks[i].intensity /= cumsum             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_t_5 = __pyx_v_i;
    (__pyx_v_result->peaks[__pyx_t_5]).intensity = ((__pyx_v_result->peaks[__pyx_t_5]).intensity / __pyx_v_cumsum);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":217
 *     for i in range(n):
 *         result.peaks[i].intensity /= cumsum
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef void peak_list_shift(PeakList* peaklist, double shift) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":196
 * 
 * @cython.cdivision
 * cdef PeakList* peak_list_truncate_after(PeakList* peaklist, double truncate_after, PeakList* result) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double cumsum
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":219
 *     return result
 * 
 * cdef void peak_list_shift(PeakList* peaklist, double shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_shift(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist, double __pyx_v_shift) {
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  double __pyx_v_delta;
  size_t __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;

  /* "brainpy/_c/isotopic_distribution.pyx":224
 *         double delta
 * 
 *     n = peaklist.used             # <<<<<<<<<<<<<<
 *     if n == 0:
 *         return
 */
  __pyx_t_1 = __pyx_v_peaklist->used;
  __pyx_v_n = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":225
 * 
 *     n = peaklist.used
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return
 *     delta = shift - peaklist.peaks[0].mz
 */
  __pyx_t_2 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/isotopic_distribution.pyx":226
 *     n = peaklist.used
 *     if n == 0:
 *         return             # <<<<<<<<<<<<<<
 *     delta = shift - peaklist.peaks[0].mz
 *     for i in range(n):
 */
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":225
 * 
 *     n = peaklist.used
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return
 *     delta = shift - peaklist.peaks[0].mz
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":227
 *     if n == 0:
 *         return
 *     delta = shift - peaklist.peaks[0].mz             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         peaklist.peaks[i].mz += delta
 */
  __pyx_v_delta = (__pyx_v_shift - (__pyx_v_peaklist->peaks[0]).mz);

  /* "brainpy/_c/isotopic_distribution.pyx":228
 *         return
 *     delta = shift - peaklist.peaks[0].mz
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         peaklist.peaks[i].mz += delta
 * 
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":229
 *     delta = shift - peaklist.peaks[0].mz
 *     for i in range(n):
 *         peaklist.peaks[i].mz += delta             # <<<<<<<<<<<<<<
 * 
 * # -----------------------------------------------------------------------------
 */
    __pyx_t_5 = __pyx_v_i;
    (__pyx_v_peaklist->peaks[__pyx_t_5]).mz = ((__pyx_v_peaklist->peaks[__pyx_t_5]).mz + __pyx_v_delta);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":219
 *     return result
 * 
 * cdef void peak_list_shift(PeakList* peaklist, double shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i, n
 */

  /* function exit code */
  __pyx_L0:;
}

/* "brainpy/_c/isotopic_distribution.pyx":234
 * # ElementCache Methods
 * 
 * cdef ElementCache* make_element_cache(ElementHashTable* source) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         ElementCache* cache
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_cache(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_source) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_r;

  /* "brainpy/_c/isotopic_distribution.pyx":237
 *     cdef:
 *         ElementCache* cache
 *     cache = <ElementCache*>malloc(sizeof(ElementCache))             # <<<<<<<<<<<<<<
 *     cache.source = source
 *     cache.elements = <Element**>malloc(sizeof(Element*) * 10)
 */
  __pyx_v_cache = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache))));

  /* "brainpy/_c/isotopic_distribution.pyx":238
 *         ElementCache* cache
 *     cache = <ElementCache*>malloc(sizeof(ElementCache))
 *     cache.source = source             # <<<<<<<<<<<<<<
 *     cache.elements = <Element**>malloc(sizeof(Element*) * 10)
 *     cache.used = 0
 */
  __pyx_v_cache->source = __pyx_v_source;

  /* "brainpy/_c/isotopic_distribution.pyx":239
 *     cache = <ElementCache*>malloc(sizeof(ElementCache))
 *     cache.source = source
 *     cache.elements = <Element**>malloc(sizeof(Element*) * 10)             # <<<<<<<<<<<<<<
 *     cache.used = 0
 *     cache.size = 10
 */
  __pyx_v_cache->elements = ((struct __pyx_t_7brainpy_2_c_11composition_Element **)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element *)) * 10)));

  /* "brainpy/_c/isotopic_distribution.pyx":240
 *     cache.source = source
 *     cache.elements = <Element**>malloc(sizeof(Element*) * 10)
 *     cache.used = 0             # <<<<<<<<<<<<<<
 *     cache.size = 10
 *     return cache
 */
  __pyx_v_cache->used = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":241
 *     cache.elements = <Element**>malloc(sizeof(Element*) * 10)
 *     cache.used = 0
 *     cache.size = 10             # <<<<<<<<<<<<<<
 *     return cache
 * 
 */
  __pyx_v_cache->size = 10;

  /* "brainpy/_c/isotopic_distribution.pyx":242
 *     cache.used = 0
 *     cache.size = 10
 *     return cache             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_cache;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":234
 * # ElementCache Methods
 * 
 * cdef ElementCache* make_element_cache(ElementHashTable* source) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         ElementCache* cache
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":245
 * 
 * 
 * cdef void free_element_cache(ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     free(cache.elements)
 *     free(cache)
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_cache(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache) {

  /* "brainpy/_c/isotopic_distribution.pyx":246
 * 
 * cdef void free_element_cache(ElementCache* cache) nogil:
 *     free(cache.elements)             # <<<<<<<<<<<<<<
 *     free(cache)
 * 
 */
  free(__pyx_v_cache->elements);

  /* "brainpy/_c/isotopic_distribution.pyx":247
 * cdef void free_element_cache(ElementCache* cache) nogil:
 *     free(cache.elements)
 *     free(cache)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_cache);

  /* "brainpy/_c/isotopic_distribution.pyx":245
 * 
 * 
 * cdef void free_element_cache(ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     free(cache.elements)
 *     free(cache)
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":250
 * 
 * 
 * cdef int resize_element_cache(ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element** values
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_resize_element_cache(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache) {
  struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_values;
  size_t __pyx_v_new_size;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":254
 *         Element** values
 *         size_t new_size
 *     new_size = cache.size * 10             # <<<<<<<<<<<<<<
 *     values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)
 *     if values == NULL:
 */
  __pyx_v_new_size = (__pyx_v_cache->size * 10);

  /* "brainpy/_c/isotopic_distribution.pyx":255
 *         size_t new_size
 *     new_size = cache.size * 10
 *     values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)             # <<<<<<<<<<<<<<
 *     if values == NULL:
 *         printf("resize_element_cache returned -1\n")
 */
  __pyx_v_values = ((struct __pyx_t_7brainpy_2_c_11composition_Element **)realloc(__pyx_v_cache->elements, ((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element *)) * __pyx_v_new_size)));

  /* "brainpy/_c/isotopic_distribution.pyx":256
 *     new_size = cache.size * 10
 *     values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)
 *     if values == NULL:             # <<<<<<<<<<<<<<
 *         printf("resize_element_cache returned -1\n")
 *         return -1
 */
  __pyx_t_1 = ((__pyx_v_values == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":257
 *     values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)
 *     if values == NULL:
 *         printf("resize_element_cache returned -1\n")             # <<<<<<<<<<<<<<
 *         return -1
 *     cache.elements = values
 */
    (void)(printf(((char const *)"resize_element_cache returned -1\n")));

    /* "brainpy/_c/isotopic_distribution.pyx":258
 *     if values == NULL:
 *         printf("resize_element_cache returned -1\n")
 *         return -1             # <<<<<<<<<<<<<<
 *     cache.elements = values
 *     cache.size = new_size
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":256
 *     new_size = cache.size * 10
 *     values = <Element**>realloc(cache.elements, sizeof(Element*) * new_size)
 *     if values == NULL:             # <<<<<<<<<<<<<<
 *         printf("resize_element_cache returned -1\n")
 *         return -1
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":259
 *         printf("resize_element_cache returned -1\n")
 *         return -1
 *     cache.elements = values             # <<<<<<<<<<<<<<
 *     cache.size = new_size
 *     return 0
 */
  __pyx_v_cache->elements = __pyx_v_values;

  /* "brainpy/_c/isotopic_distribution.pyx":260
 *         return -1
 *     cache.elements = values
 *     cache.size = new_size             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_cache->size = __pyx_v_new_size;

  /* "brainpy/_c/isotopic_distribution.pyx":261
 *     cache.elements = values
 *     cache.size = new_size
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":250
 * 
 * 
 * cdef int resize_element_cache(ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element** values
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":264
 * 
 * 
 * cdef int element_cache_put(ElementCache* cache, Element** element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_put(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache, struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_element) {
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":267
 *     cdef:
 *         size_t i
 *     if (cache.used + 1) == cache.size:             # <<<<<<<<<<<<<<
 *         resize_element_cache(cache)
 *     cache.elements[cache.used] = element[0]
 */
  __pyx_t_1 = (((__pyx_v_cache->used + 1) == __pyx_v_cache->size) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":268
 *         size_t i
 *     if (cache.used + 1) == cache.size:
 *         resize_element_cache(cache)             # <<<<<<<<<<<<<<
 *     cache.elements[cache.used] = element[0]
 *     cache.used += 1
 */
    (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_resize_element_cache(__pyx_v_cache));

    /* "brainpy/_c/isotopic_distribution.pyx":267
 *     cdef:
 *         size_t i
 *     if (cache.used + 1) == cache.size:             # <<<<<<<<<<<<<<
 *         resize_element_cache(cache)
 *     cache.elements[cache.used] = element[0]
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":269
 *     if (cache.used + 1) == cache.size:
 *         resize_element_cache(cache)
 *     cache.elements[cache.used] = element[0]             # <<<<<<<<<<<<<<
 *     cache.used += 1
 *     return 0
 */
  (__pyx_v_cache->elements[__pyx_v_cache->used]) = (__pyx_v_element[0]);

  /* "brainpy/_c/isotopic_distribution.pyx":270
 *         resize_element_cache(cache)
 *     cache.elements[cache.used] = element[0]
 *     cache.used += 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_cache->used = (__pyx_v_cache->used + 1);

  /* "brainpy/_c/isotopic_distribution.pyx":271
 *     cache.elements[cache.used] = element[0]
 *     cache.used += 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":264
 * 
 * 
 * cdef int element_cache_put(ElementCache* cache, Element** element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":274
 * 
 * 
 * cdef int element_cache_get(ElementCache* cache, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache, char *__pyx_v_symbol, struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_out) {
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  int __pyx_r;
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "brainpy/_c/isotopic_distribution.pyx":279
 *         Element* element
 * 
 *     if cache == NULL:             # <<<<<<<<<<<<<<
 *         return element_hash_table_get(_ElementTable, symbol, out)
 * 
 */
  __pyx_t_1 = ((__pyx_v_cache == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":280
 * 
 *     if cache == NULL:
 *         return element_hash_table_get(_ElementTable, symbol, out)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(cache.used):
 */
    __pyx_r = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_symbol, __pyx_v_out);
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":279
 *         Element* element
 * 
 *     if cache == NULL:             # <<<<<<<<<<<<<<
 *         return element_hash_table_get(_ElementTable, symbol, out)
 * 
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":282
 *         return element_hash_table_get(_ElementTable, symbol, out)
 * 
 *     for i in range(cache.used):             # <<<<<<<<<<<<<<
 *         element = cache.elements[i]
 *         if strcmp(element.symbol, symbol) == 0:
 */
  __pyx_t_2 = __pyx_v_cache->used;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":283
 * 
 *     for i in range(cache.used):
 *         element = cache.elements[i]             # <<<<<<<<<<<<<<
 *         if strcmp(element.symbol, symbol) == 0:
 *             out[0] = element
 */
    __pyx_v_element = (__pyx_v_cache->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":284
 *     for i in range(cache.used):
 *         element = cache.elements[i]
 *         if strcmp(element.symbol, symbol) == 0:             # <<<<<<<<<<<<<<
 *             out[0] = element
 *             return 0
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element->symbol, __pyx_v_symbol) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":285
 *         element = cache.elements[i]
 *         if strcmp(element.symbol, symbol) == 0:
 *             out[0] = element             # <<<<<<<<<<<<<<
 *             return 0
 *     element_hash_table_get(cache.source, symbol, out)
 */
      (__pyx_v_out[0]) = __pyx_v_element;

      /* "brainpy/_c/isotopic_distribution.pyx":286
 *         if strcmp(element.symbol, symbol) == 0:
 *             out[0] = element
 *             return 0             # <<<<<<<<<<<<<<
 *     element_hash_table_get(cache.source, symbol, out)
 *     element_cache_put(cache, out)
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/isotopic_distribution.pyx":284
 *     for i in range(cache.used):
 *         element = cache.elements[i]
 *         if strcmp(element.symbol, symbol) == 0:             # <<<<<<<<<<<<<<
 *             out[0] = element
 *             return 0
 */
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":287
 *             out[0] = element
 *             return 0
 *     element_hash_table_get(cache.source, symbol, out)             # <<<<<<<<<<<<<<
 *     element_cache_put(cache, out)
 *     return 1
 */
  (void)(__pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_cache->source, __pyx_v_symbol, __pyx_v_out));

  /* "brainpy/_c/isotopic_distribution.pyx":288
 *             return 0
 *     element_hash_table_get(cache.source, symbol, out)
 *     element_cache_put(cache, out)             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
  (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_put(__pyx_v_cache, __pyx_v_out));

  /* "brainpy/_c/isotopic_distribution.pyx":289
 *     element_hash_table_get(cache.source, symbol, out)
 *     element_cache_put(cache, out)
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":274
 * 
 * 
 * cdef int element_cache_get(ElementCache* cache, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":295
 * # Support Functions
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache) {
  size_t __pyx_v_i;
  int __pyx_v_max_variants;
  double __pyx_v_count;
  char *__pyx_v_symbol;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  int __pyx_r;
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;

  /* "brainpy/_c/isotopic_distribution.pyx":303
 *         Element* element
 * 
 *     if composition.max_variants != 0:             # <<<<<<<<<<<<<<
 *         return composition.max_variants
 * 
 */
  __pyx_t_1 = ((__pyx_v_composition->max_variants != 0) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":304
 * 
 *     if composition.max_variants != 0:
 *         return composition.max_variants             # <<<<<<<<<<<<<<
 * 
 *     max_variants = 0
 */
    __pyx_r = __pyx_v_composition->max_variants;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":303
 *         Element* element
 * 
 *     if composition.max_variants != 0:             # <<<<<<<<<<<<<<
 *         return composition.max_variants
 * 
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":306
 *         return composition.max_variants
 * 
 *     max_variants = 0             # <<<<<<<<<<<<<<
 *     for i in range(composition.used):
 *         symbol = composition.elements[i]
 */
  __pyx_v_max_variants = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":307
 * 
 *     max_variants = 0
 *     for i in range(composition.used):             # <<<<<<<<<<<<<<
 *         symbol = composition.elements[i]
 *         count = composition.counts[i]
 */
  __pyx_t_2 = __pyx_v_composition->used;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":308
 *     max_variants = 0
 *     for i in range(composition.used):
 *         symbol = composition.elements[i]             # <<<<<<<<<<<<<<
 *         count = composition.counts[i]
 *         if cache == NULL:
 */
    __pyx_v_symbol = (__pyx_v_composition->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":309
 *     for i in range(composition.used):
 *         symbol = composition.elements[i]
 *         count = composition.counts[i]             # <<<<<<<<<<<<<<
 *         if cache == NULL:
 *             element_hash_table_get(_ElementTable, symbol, &element)
 */
    __pyx_v_count = (__pyx_v_composition->counts[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":310
 *         symbol = composition.elements[i]
 *         count = composition.counts[i]
 *         if cache == NULL:             # <<<<<<<<<<<<<<
 *             element_hash_table_get(_ElementTable, symbol, &element)
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_cache == NULL) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":311
 *         count = composition.counts[i]
 *         if cache == NULL:
 *             element_hash_table_get(_ElementTable, symbol, &element)             # <<<<<<<<<<<<<<
 *         else:
 *             element_cache_get(cache, symbol, &element)
 */
      (void)(__pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_symbol, (&__pyx_v_element)));

      /* "brainpy/_c/isotopic_distribution.pyx":310
 *         symbol = composition.elements[i]
 *         count = composition.counts[i]
 *         if cache == NULL:             # <<<<<<<<<<<<<<
 *             element_hash_table_get(_ElementTable, symbol, &element)
 *         else:
 */
      goto __pyx_L6;
    }

    /* "brainpy/_c/isotopic_distribution.pyx":313
 *             element_hash_table_get(_ElementTable, symbol, &element)
 *         else:
 *             element_cache_get(cache, symbol, &element)             # <<<<<<<<<<<<<<
 *         max_variants += <int>(count) * element_max_neutron_shift(element)
 *     composition.max_variants = max_variants
 */
    /*else*/ {
      (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get(__pyx_v_cache, __pyx_v_symbol, (&__pyx_v_element)));
    }
    __pyx_L6:;

    /* "brainpy/_c/isotopic_distribution.pyx":314
 *         else:
 *             element_cache_get(cache, symbol, &element)
 *         max_variants += <int>(count) * element_max_neutron_shift(element)             # <<<<<<<<<<<<<<
 *     composition.max_variants = max_variants
 *     return max_variants
 */
    __pyx_v_max_variants = (__pyx_v_max_variants + (((int)__pyx_v_count) * __pyx_f_7brainpy_2_c_11composition_element_max_neutron_shift(__pyx_v_element)));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":315
 *             element_cache_get(cache, symbol, &element)
 *         max_variants += <int>(count) * element_max_neutron_shift(element)
 *     composition.max_variants = max_variants             # <<<<<<<<<<<<<<
 *     return max_variants
 * 
 */
  __pyx_v_composition->max_variants = __pyx_v_max_variants;

  /* "brainpy/_c/isotopic_distribution.pyx":316
 *         max_variants += <int>(count) * element_max_neutron_shift(element)
 *     composition.max_variants = max_variants
 *     return max_variants             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_max_variants;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":295
 * # Support Functions
 * 
 * cdef int max_variants(Composition* composition, ElementCache* cache) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":319
 * 
 * 
 * cdef void validate_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_validate_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  size_t __pyx_v_i;
  char *__pyx_v_element_symbol;
  char *__pyx_v_element_buffer;
  int __pyx_v_status;
  int __pyx_v_isotope_number;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/isotopic_distribution.pyx":327
 *         Element* element
 * 
 *     for i in range(composition.used):             # <<<<<<<<<<<<<<
 *         element_symbol = composition.elements[i]
 *         status = element_hash_table_get(_ElementTable, element_symbol, &element)
 */
  __pyx_t_1 = __pyx_v_composition->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":328
 * 
 *     for i in range(composition.used):
 *         element_symbol = composition.elements[i]             # <<<<<<<<<<<<<<
 *         status = element_hash_table_get(_ElementTable, element_symbol, &element)
 *         # printf("Element %s, Status %d\n", element_symbol, status)
 */
    __pyx_v_element_symbol = (__pyx_v_composition->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":329
 *     for i in range(composition.used):
 *         element_symbol = composition.elements[i]
 *         status = element_hash_table_get(_ElementTable, element_symbol, &element)             # <<<<<<<<<<<<<<
 *         # printf("Element %s, Status %d\n", element_symbol, status)
 *         if status == -1:
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_element_symbol, (&__pyx_v_element));

    /* "brainpy/_c/isotopic_distribution.pyx":331
 *         status = element_hash_table_get(_ElementTable, element_symbol, &element)
 *         # printf("Element %s, Status %d\n", element_symbol, status)
 *         if status == -1:             # <<<<<<<<<<<<<<
 *             element_buffer = <char*>malloc(sizeof(char) * 10)
 *             # printf("Could not resolve element %s, attempting to generate fixed isotope\n", element_symbol)
 */
    __pyx_t_4 = ((__pyx_v_status == -1L) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/isotopic_distribution.pyx":332
 *         # printf("Element %s, Status %d\n", element_symbol, status)
 *         if status == -1:
 *             element_buffer = <char*>malloc(sizeof(char) * 10)             # <<<<<<<<<<<<<<
 *             # printf("Could not resolve element %s, attempting to generate fixed isotope\n", element_symbol)
 *             _parse_isotope_string(element_symbol, &isotope_number, element_buffer)
 */
      __pyx_v_element_buffer = ((char *)malloc(((sizeof(char)) * 10)));

      /* "brainpy/_c/isotopic_distribution.pyx":334
 *             element_buffer = <char*>malloc(sizeof(char) * 10)
 *             # printf("Could not resolve element %s, attempting to generate fixed isotope\n", element_symbol)
 *             _parse_isotope_string(element_symbol, &isotope_number, element_buffer)             # <<<<<<<<<<<<<<
 *             # printf("Element: %s, Isotope: %d\n", element_buffer, isotope_number)
 *             if isotope_number != 0:
 */
      (void)(__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(__pyx_v_element_symbol, (&__pyx_v_isotope_number), __pyx_v_element_buffer));

      /* "brainpy/_c/isotopic_distribution.pyx":336
 *             _parse_isotope_string(element_symbol, &isotope_number, element_buffer)
 *             # printf("Element: %s, Isotope: %d\n", element_buffer, isotope_number)
 *             if isotope_number != 0:             # <<<<<<<<<<<<<<
 *                 status = element_hash_table_get(_ElementTable, element_buffer, &element)
 *                 # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
 */
      __pyx_t_4 = ((__pyx_v_isotope_number != 0) != 0);
      if (__pyx_t_4) {

        /* "brainpy/_c/isotopic_distribution.pyx":337
 *             # printf("Element: %s, Isotope: %d\n", element_buffer, isotope_number)
 *             if isotope_number != 0:
 *                 status = element_hash_table_get(_ElementTable, element_buffer, &element)             # <<<<<<<<<<<<<<
 *                 # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
 *                 element = make_fixed_isotope_element(element, isotope_number)
 */
        __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_element_buffer, (&__pyx_v_element));

        /* "brainpy/_c/isotopic_distribution.pyx":339
 *                 status = element_hash_table_get(_ElementTable, element_buffer, &element)
 *                 # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
 *                 element = make_fixed_isotope_element(element, isotope_number)             # <<<<<<<<<<<<<<
 *                 element_hash_table_put(_ElementTable, element)
 *                 free(element_buffer)
 */
        __pyx_v_element = __pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(__pyx_v_element, __pyx_v_isotope_number);

        /* "brainpy/_c/isotopic_distribution.pyx":340
 *                 # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
 *                 element = make_fixed_isotope_element(element, isotope_number)
 *                 element_hash_table_put(_ElementTable, element)             # <<<<<<<<<<<<<<
 *                 free(element_buffer)
 *             else:
 */
        (void)(__pyx_f_7brainpy_2_c_11composition_element_hash_table_put(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_element));

        /* "brainpy/_c/isotopic_distribution.pyx":341
 *                 element = make_fixed_isotope_element(element, isotope_number)
 *                 element_hash_table_put(_ElementTable, element)
 *                 free(element_buffer)             # <<<<<<<<<<<<<<
 *             else:
 *                 # printf("Could not resolve element %s\n", element_symbol)
 */
        free(__pyx_v_element_buffer);

        /* "brainpy/_c/isotopic_distribution.pyx":336
 *             _parse_isotope_string(element_symbol, &isotope_number, element_buffer)
 *             # printf("Element: %s, Isotope: %d\n", element_buffer, isotope_number)
 *             if isotope_number != 0:             # <<<<<<<<<<<<<<
 *                 status = element_hash_table_get(_ElementTable, element_buffer, &element)
 *                 # printf("Retreived Base Element: %s. Status: %d\n", element_buffer, status)
 */
        goto __pyx_L6;
      }

      /* "brainpy/_c/isotopic_distribution.pyx":344
 *             else:
 *                 # printf("Could not resolve element %s\n", element_symbol)
 *                 free(element_buffer)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      /*else*/ {
        free(__pyx_v_element_buffer);

        /* "brainpy/_c/isotopic_distribution.pyx":345
 *                 # printf("Could not resolve element %s\n", element_symbol)
 *                 free(element_buffer)
 *                 return             # <<<<<<<<<<<<<<
 * 
 * # -----------------------------------------------------------------------------
 */
        goto __pyx_L0;
      }
      __pyx_L6:;

      /* "brainpy/_c/isotopic_distribution.pyx":331
 *         status = element_hash_table_get(_ElementTable, element_symbol, &element)
 *         # printf("Element %s, Status %d\n", element_symbol, status)
 *         if status == -1:             # <<<<<<<<<<<<<<
 *             element_buffer = <char*>malloc(sizeof(char) * 10)
 *             # printf("Could not resolve element %s, attempting to generate fixed isotope\n", element_symbol)
 */
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":319
 * 
 * 
 * cdef void validate_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
}

/* "brainpy/_c/isotopic_distribution.pyx":350
 * # IsotopicDistribution Methods
 * 
 * cdef void isotopic_distribution_update_isotopic_constants(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_isotopic_constants(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution) {
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *__pyx_v_isotopes;
  char *__pyx_v_symbol;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_t_1;
  struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *__pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;

  /* "brainpy/_c/isotopic_distribution.pyx":357
 *         char* symbol
 * 
 *     composition = distribution.composition             # <<<<<<<<<<<<<<
 *     isotopes = distribution._isotopic_constants
 * 
 */
  __pyx_t_1 = __pyx_v_distribution->composition;
  __pyx_v_composition = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":358
 * 
 *     composition = distribution.composition
 *     isotopes = distribution._isotopic_constants             # <<<<<<<<<<<<<<
 * 
 *     for i in range(composition.used):
 */
  __pyx_t_2 = __pyx_v_distribution->_isotopic_constants;
  __pyx_v_isotopes = __pyx_t_2;

  /* "brainpy/_c/isotopic_distribution.pyx":360
 *     isotopes = distribution._isotopic_constants
 * 
 *     for i in range(composition.used):             # <<<<<<<<<<<<<<
 *         symbol = composition.elements[i]
 *         isotopic_constants_add_element(isotopes, symbol)
 */
  __pyx_t_3 = __pyx_v_composition->used;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "brainpy/_c/isotopic_distribution.pyx":361
 * 
 *     for i in range(composition.used):
 *         symbol = composition.elements[i]             # <<<<<<<<<<<<<<
 *         isotopic_constants_add_element(isotopes, symbol)
 *     isotopes.order = distribution.order
 */
    __pyx_v_symbol = (__pyx_v_composition->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":362
 *     for i in range(composition.used):
 *         symbol = composition.elements[i]
 *         isotopic_constants_add_element(isotopes, symbol)             # <<<<<<<<<<<<<<
 *     isotopes.order = distribution.order
 *     isotopic_constants_update_coefficients(isotopes)
 */
    __pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_add_element(__pyx_v_isotopes, __pyx_v_symbol);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":363
 *         symbol = composition.elements[i]
 *         isotopic_constants_add_element(isotopes, symbol)
 *     isotopes.order = distribution.order             # <<<<<<<<<<<<<<
 *     isotopic_constants_update_coefficients(isotopes)
 * 
 */
  __pyx_t_6 = __pyx_v_distribution->order;
  __pyx_v_isotopes->order = __pyx_t_6;

  /* "brainpy/_c/isotopic_distribution.pyx":364
 *         isotopic_constants_add_element(isotopes, symbol)
 *     isotopes.order = distribution.order
 *     isotopic_constants_update_coefficients(isotopes)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_update_coefficients(__pyx_v_isotopes);

  /* "brainpy/_c/isotopic_distribution.pyx":350
 * # IsotopicDistribution Methods
 * 
 * cdef void isotopic_distribution_update_isotopic_constants(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":367
 * 
 * 
 * cdef void isotopic_distribution_update_order(IsotopicDistribution* distribution, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int possible_variants
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_order(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, int __pyx_v_order) {
  int __pyx_v_possible_variants;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/isotopic_distribution.pyx":371
 *         int possible_variants
 * 
 *     possible_variants = max_variants(distribution.composition, NULL)             # <<<<<<<<<<<<<<
 *     if order == -1:
 *         distribution.order = possible_variants
 */
  __pyx_v_possible_variants = __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(__pyx_v_distribution->composition, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":372
 * 
 *     possible_variants = max_variants(distribution.composition, NULL)
 *     if order == -1:             # <<<<<<<<<<<<<<
 *         distribution.order = possible_variants
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_order == -1L) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":373
 *     possible_variants = max_variants(distribution.composition, NULL)
 *     if order == -1:
 *         distribution.order = possible_variants             # <<<<<<<<<<<<<<
 *     else:
 *         distribution.order = min(order, possible_variants)
 */
    __pyx_v_distribution->order = __pyx_v_possible_variants;

    /* "brainpy/_c/isotopic_distribution.pyx":372
 * 
 *     possible_variants = max_variants(distribution.composition, NULL)
 *     if order == -1:             # <<<<<<<<<<<<<<
 *         distribution.order = possible_variants
 *     else:
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":375
 *         distribution.order = possible_variants
 *     else:
 *         distribution.order = min(order, possible_variants)             # <<<<<<<<<<<<<<
 * 
 *     isotopic_distribution_update_isotopic_constants(distribution)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_possible_variants;
    __pyx_t_3 = __pyx_v_order;
    if (((__pyx_t_2 < __pyx_t_3) != 0)) {
      __pyx_t_4 = __pyx_t_2;
    } else {
      __pyx_t_4 = __pyx_t_3;
    }
    __pyx_v_distribution->order = __pyx_t_4;
  }
  __pyx_L3:;

  /* "brainpy/_c/isotopic_distribution.pyx":377
 *         distribution.order = min(order, possible_variants)
 * 
 *     isotopic_distribution_update_isotopic_constants(distribution)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_isotopic_constants(__pyx_v_distribution);

  /* "brainpy/_c/isotopic_distribution.pyx":367
 * 
 * 
 * cdef void isotopic_distribution_update_order(IsotopicDistribution* distribution, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int possible_variants
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":380
 * 
 * 
 * cdef Peak* isotopic_distribution_make_monoisotopic_peak(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Peak* peak
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_make_monoisotopic_peak(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_v_peak;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  CYTHON_UNUSED int __pyx_v_status;
  double __pyx_v_intensity;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;

  /* "brainpy/_c/isotopic_distribution.pyx":388
 *         double intensity
 * 
 *     peak = <Peak*>malloc(sizeof(Peak))             # <<<<<<<<<<<<<<
 *     peak.mz = composition_mass(distribution.composition)
 * 
 */
  __pyx_v_peak = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak))));

  /* "brainpy/_c/isotopic_distribution.pyx":389
 * 
 *     peak = <Peak*>malloc(sizeof(Peak))
 *     peak.mz = composition_mass(distribution.composition)             # <<<<<<<<<<<<<<
 * 
 *     intensity = 0
 */
  __pyx_v_peak->mz = __pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_distribution->composition);

  /* "brainpy/_c/isotopic_distribution.pyx":391
 *     peak.mz = composition_mass(distribution.composition)
 * 
 *     intensity = 0             # <<<<<<<<<<<<<<
 *     for i in range(distribution.composition.used):
 *         status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)
 */
  __pyx_v_intensity = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":392
 * 
 *     intensity = 0
 *     for i in range(distribution.composition.used):             # <<<<<<<<<<<<<<
 *         status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)
 *         intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)
 */
  __pyx_t_1 = __pyx_v_distribution->composition->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":393
 *     intensity = 0
 *     for i in range(distribution.composition.used):
 *         status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)             # <<<<<<<<<<<<<<
 *         intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)
 *     intensity = exp(intensity)
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get(__pyx_v_distribution->cache, (__pyx_v_distribution->composition->elements[__pyx_v_i]), (&__pyx_v_element));

    /* "brainpy/_c/isotopic_distribution.pyx":394
 *     for i in range(distribution.composition.used):
 *         status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)
 *         intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)             # <<<<<<<<<<<<<<
 *     intensity = exp(intensity)
 *     peak.intensity = intensity
 */
    __pyx_v_intensity = (__pyx_v_intensity + log(__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift(__pyx_v_element->isotopes, 0)->abundance));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":395
 *         status = element_cache_get(distribution.cache, distribution.composition.elements[i], &element)
 *         intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)
 *     intensity = exp(intensity)             # <<<<<<<<<<<<<<
 *     peak.intensity = intensity
 *     peak.charge = 0
 */
  __pyx_v_intensity = exp(__pyx_v_intensity);

  /* "brainpy/_c/isotopic_distribution.pyx":396
 *         intensity += log(get_isotope_by_neutron_shift(element.isotopes, 0).abundance)
 *     intensity = exp(intensity)
 *     peak.intensity = intensity             # <<<<<<<<<<<<<<
 *     peak.charge = 0
 *     return peak
 */
  __pyx_v_peak->intensity = __pyx_v_intensity;

  /* "brainpy/_c/isotopic_distribution.pyx":397
 *     intensity = exp(intensity)
 *     peak.intensity = intensity
 *     peak.charge = 0             # <<<<<<<<<<<<<<
 *     return peak
 * 
 */
  __pyx_v_peak->charge = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":398
 *     peak.intensity = intensity
 *     peak.charge = 0
 *     return peak             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_peak;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":380
 * 
 * 
 * cdef Peak* isotopic_distribution_make_monoisotopic_peak(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Peak* peak
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":401
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=NULL) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopicDistribution* result
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, int __pyx_v_order, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution *__pyx_optional_args) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)NULL);
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_result;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_r;
  int __pyx_t_1;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cache = __pyx_optional_args->cache;
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":404
 *     cdef:
 *         IsotopicDistribution* result
 *     if cache == NULL:             # <<<<<<<<<<<<<<
 *         cache = make_element_cache(_ElementTable)
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
 */
  __pyx_t_1 = ((__pyx_v_cache == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":405
 *         IsotopicDistribution* result
 *     if cache == NULL:
 *         cache = make_element_cache(_ElementTable)             # <<<<<<<<<<<<<<
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
 *     result.composition = composition
 */
    __pyx_v_cache = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_cache(__pyx_v_7brainpy_2_c_11composition__ElementTable);

    /* "brainpy/_c/isotopic_distribution.pyx":404
 *     cdef:
 *         IsotopicDistribution* result
 *     if cache == NULL:             # <<<<<<<<<<<<<<
 *         cache = make_element_cache(_ElementTable)
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":406
 *     if cache == NULL:
 *         cache = make_element_cache(_ElementTable)
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))             # <<<<<<<<<<<<<<
 *     result.composition = composition
 *     result.cache = cache
 */
  __pyx_v_result = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution))));

  /* "brainpy/_c/isotopic_distribution.pyx":407
 *         cache = make_element_cache(_ElementTable)
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
 *     result.composition = composition             # <<<<<<<<<<<<<<
 *     result.cache = cache
 *     result.order = 0
 */
  __pyx_v_result->composition = __pyx_v_composition;

  /* "brainpy/_c/isotopic_distribution.pyx":408
 *     result = <IsotopicDistribution*>malloc(sizeof(IsotopicDistribution))
 *     result.composition = composition
 *     result.cache = cache             # <<<<<<<<<<<<<<
 *     result.order = 0
 *     result._isotopic_constants = make_isotopic_constants()
 */
  __pyx_v_result->cache = __pyx_v_cache;

  /* "brainpy/_c/isotopic_distribution.pyx":409
 *     result.composition = composition
 *     result.cache = cache
 *     result.order = 0             # <<<<<<<<<<<<<<
 *     result._isotopic_constants = make_isotopic_constants()
 *     isotopic_distribution_update_order(result, order)
 */
  __pyx_v_result->order = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":410
 *     result.cache = cache
 *     result.order = 0
 *     result._isotopic_constants = make_isotopic_constants()             # <<<<<<<<<<<<<<
 *     isotopic_distribution_update_order(result, order)
 *     result.average_mass = 0
 */
  __pyx_v_result->_isotopic_constants = __pyx_f_7brainpy_2_c_18isotopic_constants_make_isotopic_constants();

  /* "brainpy/_c/isotopic_distribution.pyx":411
 *     result.order = 0
 *     result._isotopic_constants = make_isotopic_constants()
 *     isotopic_distribution_update_order(result, order)             # <<<<<<<<<<<<<<
 *     result.average_mass = 0
 *     result.monoisotopic_peak = isotopic_distribution_make_monoisotopic_peak(result)
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_update_order(__pyx_v_result, __pyx_v_order);

  /* "brainpy/_c/isotopic_distribution.pyx":412
 *     result._isotopic_constants = make_isotopic_constants()
 *     isotopic_distribution_update_order(result, order)
 *     result.average_mass = 0             # <<<<<<<<<<<<<<
 *     result.monoisotopic_peak = isotopic_distribution_make_monoisotopic_peak(result)
 *     return result
 */
  __pyx_v_result->average_mass = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":413
 *     isotopic_distribution_update_order(result, order)
 *     result.average_mass = 0
 *     result.monoisotopic_peak = isotopic_distribution_make_monoisotopic_peak(result)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->monoisotopic_peak = __pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_distribution_make_monoisotopic_peak(__pyx_v_result);

  /* "brainpy/_c/isotopic_distribution.pyx":414
 *     result.average_mass = 0
 *     result.monoisotopic_peak = isotopic_distribution_make_monoisotopic_peak(result)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":401
 * 
 * 
 * cdef IsotopicDistribution* make_isotopic_distribution(Composition* composition, int order, ElementCache* cache=NULL) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopicDistribution* result
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":417
 * 
 * 
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil:             # <<<<<<<<<<<<<<
 *     free(distribution.monoisotopic_peak)
 *     free_isotopic_constants(distribution._isotopic_constants)
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution) {
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":418
 * 
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil:
 *     free(distribution.monoisotopic_peak)             # <<<<<<<<<<<<<<
 *     free_isotopic_constants(distribution._isotopic_constants)
 *     if distribution.cache != NULL:
 */
  free(__pyx_v_distribution->monoisotopic_peak);

  /* "brainpy/_c/isotopic_distribution.pyx":419
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil:
 *     free(distribution.monoisotopic_peak)
 *     free_isotopic_constants(distribution._isotopic_constants)             # <<<<<<<<<<<<<<
 *     if distribution.cache != NULL:
 *         free_element_cache(distribution.cache)
 */
  __pyx_f_7brainpy_2_c_18isotopic_constants_free_isotopic_constants(__pyx_v_distribution->_isotopic_constants);

  /* "brainpy/_c/isotopic_distribution.pyx":420
 *     free(distribution.monoisotopic_peak)
 *     free_isotopic_constants(distribution._isotopic_constants)
 *     if distribution.cache != NULL:             # <<<<<<<<<<<<<<
 *         free_element_cache(distribution.cache)
 *     free(distribution)
 */
  __pyx_t_1 = ((__pyx_v_distribution->cache != NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":421
 *     free_isotopic_constants(distribution._isotopic_constants)
 *     if distribution.cache != NULL:
 *         free_element_cache(distribution.cache)             # <<<<<<<<<<<<<<
 *     free(distribution)
 * 
 */
    __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_cache(__pyx_v_distribution->cache);

    /* "brainpy/_c/isotopic_distribution.pyx":420
 *     free(distribution.monoisotopic_peak)
 *     free_isotopic_constants(distribution._isotopic_constants)
 *     if distribution.cache != NULL:             # <<<<<<<<<<<<<<
 *         free_element_cache(distribution.cache)
 *     free(distribution)
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":422
 *     if distribution.cache != NULL:
 *         free_element_cache(distribution.cache)
 *     free(distribution)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_distribution);

  /* "brainpy/_c/isotopic_distribution.pyx":417
 * 
 * 
 * cdef void free_isotopic_distribution(IsotopicDistribution* distribution)  nogil:             # <<<<<<<<<<<<<<
 *     free(distribution.monoisotopic_peak)
 *     free_isotopic_constants(distribution._isotopic_constants)
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":425
 * 
 * 
 * cdef dvec* id_phi_values(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* power_sum
 */

static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_phi_values(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution) {
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_power_sum;
  size_t __pyx_v_i;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  size_t __pyx_t_3;

  /* "brainpy/_c/isotopic_distribution.pyx":429
 *         dvec* power_sum
 *         size_t i
 *     power_sum = make_double_vector_with_size(distribution.order)             # <<<<<<<<<<<<<<
 *     double_vector_append(power_sum, 0.)
 *     for i in range(1, distribution.order + 1):
 */
  __pyx_v_power_sum = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector_with_size(__pyx_v_distribution->order);

  /* "brainpy/_c/isotopic_distribution.pyx":430
 *         size_t i
 *     power_sum = make_double_vector_with_size(distribution.order)
 *     double_vector_append(power_sum, 0.)             # <<<<<<<<<<<<<<
 *     for i in range(1, distribution.order + 1):
 *         double_vector_append(power_sum, _id_phi_value(distribution, i))
 */
  (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_power_sum, 0.));

  /* "brainpy/_c/isotopic_distribution.pyx":431
 *     power_sum = make_double_vector_with_size(distribution.order)
 *     double_vector_append(power_sum, 0.)
 *     for i in range(1, distribution.order + 1):             # <<<<<<<<<<<<<<
 *         double_vector_append(power_sum, _id_phi_value(distribution, i))
 *     return power_sum
 */
  __pyx_t_1 = (__pyx_v_distribution->order + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 1; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":432
 *     double_vector_append(power_sum, 0.)
 *     for i in range(1, distribution.order + 1):
 *         double_vector_append(power_sum, _id_phi_value(distribution, i))             # <<<<<<<<<<<<<<
 *     return power_sum
 * 
 */
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_power_sum, __pyx_f_7brainpy_2_c_21isotopic_distribution__id_phi_value(__pyx_v_distribution, __pyx_v_i)));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":433
 *     for i in range(1, distribution.order + 1):
 *         double_vector_append(power_sum, _id_phi_value(distribution, i))
 *     return power_sum             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_power_sum;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":425
 * 
 * 
 * cdef dvec* id_phi_values(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* power_sum
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":436
 * 
 * 
 * cdef double _id_phi_value(IsotopicDistribution* distribution, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double phi
 */

static double __pyx_f_7brainpy_2_c_21isotopic_distribution__id_phi_value(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, int __pyx_v_order) {
  double __pyx_v_phi;
  char *__pyx_v_element;
  double __pyx_v_count;
  size_t __pyx_v_i;
  double __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":443
 *         size_t i
 * 
 *     phi = 0             # <<<<<<<<<<<<<<
 *     i = 0
 *     while i < distribution.composition.used:
 */
  __pyx_v_phi = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":444
 * 
 *     phi = 0
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":445
 *     phi = 0
 *     i = 0
 *     while i < distribution.composition.used:             # <<<<<<<<<<<<<<
 *         element = distribution.composition.elements[i]
 *         count = distribution.composition.counts[i]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_distribution->composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/isotopic_distribution.pyx":446
 *     i = 0
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]             # <<<<<<<<<<<<<<
 *         count = distribution.composition.counts[i]
 *         phi += isotopic_constants_nth_element_power_sum(
 */
    __pyx_v_element = (__pyx_v_distribution->composition->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":447
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]
 *         count = distribution.composition.counts[i]             # <<<<<<<<<<<<<<
 *         phi += isotopic_constants_nth_element_power_sum(
 *             distribution._isotopic_constants, element, order) * count
 */
    __pyx_v_count = (__pyx_v_distribution->composition->counts[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":448
 *         element = distribution.composition.elements[i]
 *         count = distribution.composition.counts[i]
 *         phi += isotopic_constants_nth_element_power_sum(             # <<<<<<<<<<<<<<
 *             distribution._isotopic_constants, element, order) * count
 *         i += 1
 */
    __pyx_v_phi = (__pyx_v_phi + (__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum(__pyx_v_distribution->_isotopic_constants, __pyx_v_element, __pyx_v_order) * __pyx_v_count));

    /* "brainpy/_c/isotopic_distribution.pyx":450
 *         phi += isotopic_constants_nth_element_power_sum(
 *             distribution._isotopic_constants, element, order) * count
 *         i += 1             # <<<<<<<<<<<<<<
 *     return phi
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":451
 *             distribution._isotopic_constants, element, order) * count
 *         i += 1
 *     return phi             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_phi;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":436
 * 
 * 
 * cdef double _id_phi_value(IsotopicDistribution* distribution, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double phi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":454
 * 
 * 
 * cdef dvec* id_modified_phi_values(IsotopicDistribution* distribution, char* element, dvec* power_sum) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_modified_phi_values(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, char *__pyx_v_element, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_power_sum) {
  size_t __pyx_v_i;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  size_t __pyx_t_3;

  /* "brainpy/_c/isotopic_distribution.pyx":458
 *         size_t i
 * 
 *     double_vector_append(power_sum, 0.)             # <<<<<<<<<<<<<<
 *     for i in range(1, distribution.order + 1):
 *         double_vector_append(power_sum,
 */
  (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_power_sum, 0.));

  /* "brainpy/_c/isotopic_distribution.pyx":459
 * 
 *     double_vector_append(power_sum, 0.)
 *     for i in range(1, distribution.order + 1):             # <<<<<<<<<<<<<<
 *         double_vector_append(power_sum,
 *                              _id_modified_phi_value(distribution, element, i))
 */
  __pyx_t_1 = (__pyx_v_distribution->order + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 1; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":460
 *     double_vector_append(power_sum, 0.)
 *     for i in range(1, distribution.order + 1):
 *         double_vector_append(power_sum,             # <<<<<<<<<<<<<<
 *                              _id_modified_phi_value(distribution, element, i))
 *     return power_sum
 */
    (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_power_sum, __pyx_f_7brainpy_2_c_21isotopic_distribution__id_modified_phi_value(__pyx_v_distribution, __pyx_v_element, __pyx_v_i)));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":462
 *         double_vector_append(power_sum,
 *                              _id_modified_phi_value(distribution, element, i))
 *     return power_sum             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_power_sum;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":454
 * 
 * 
 * cdef dvec* id_modified_phi_values(IsotopicDistribution* distribution, char* element, dvec* power_sum) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":465
 * 
 * 
 * cdef double _id_modified_phi_value(IsotopicDistribution* distribution, char* symbol, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double phi
 */

static double __pyx_f_7brainpy_2_c_21isotopic_distribution__id_modified_phi_value(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, char *__pyx_v_symbol, int __pyx_v_order) {
  double __pyx_v_phi;
  char *__pyx_v_element;
  double __pyx_v_count;
  size_t __pyx_v_i;
  double __pyx_v_coef;
  double __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":473
 *         double coef
 * 
 *     phi = 0             # <<<<<<<<<<<<<<
 *     i = 0
 *     while i < distribution.composition.used:
 */
  __pyx_v_phi = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":474
 * 
 *     phi = 0
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":475
 *     phi = 0
 *     i = 0
 *     while i < distribution.composition.used:             # <<<<<<<<<<<<<<
 *         element = distribution.composition.elements[i]
 *         count = distribution.composition.counts[i]
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_distribution->composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/isotopic_distribution.pyx":476
 *     i = 0
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]             # <<<<<<<<<<<<<<
 *         count = distribution.composition.counts[i]
 * 
 */
    __pyx_v_element = (__pyx_v_distribution->composition->elements[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":477
 *     while i < distribution.composition.used:
 *         element = distribution.composition.elements[i]
 *         count = distribution.composition.counts[i]             # <<<<<<<<<<<<<<
 * 
 *         if strcmp(element, symbol) == 0:
 */
    __pyx_v_count = (__pyx_v_distribution->composition->counts[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":479
 *         count = distribution.composition.counts[i]
 * 
 *         if strcmp(element, symbol) == 0:             # <<<<<<<<<<<<<<
 *             coef = count - 1
 *         else:
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element, __pyx_v_symbol) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/isotopic_distribution.pyx":480
 * 
 *         if strcmp(element, symbol) == 0:
 *             coef = count - 1             # <<<<<<<<<<<<<<
 *         else:
 *             coef = count
 */
      __pyx_v_coef = (__pyx_v_count - 1.0);

      /* "brainpy/_c/isotopic_distribution.pyx":479
 *         count = distribution.composition.counts[i]
 * 
 *         if strcmp(element, symbol) == 0:             # <<<<<<<<<<<<<<
 *             coef = count - 1
 *         else:
 */
      goto __pyx_L5;
    }

    /* "brainpy/_c/isotopic_distribution.pyx":482
 *             coef = count - 1
 *         else:
 *             coef = count             # <<<<<<<<<<<<<<
 * 
 *         phi += isotopic_constants_nth_element_power_sum(
 */
    /*else*/ {
      __pyx_v_coef = __pyx_v_count;
    }
    __pyx_L5:;

    /* "brainpy/_c/isotopic_distribution.pyx":484
 *             coef = count
 * 
 *         phi += isotopic_constants_nth_element_power_sum(             # <<<<<<<<<<<<<<
 *             distribution._isotopic_constants, element, order) * coef
 *         i += 1
 */
    __pyx_v_phi = (__pyx_v_phi + (__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum(__pyx_v_distribution->_isotopic_constants, __pyx_v_element, __pyx_v_order) * __pyx_v_coef));

    /* "brainpy/_c/isotopic_distribution.pyx":486
 *         phi += isotopic_constants_nth_element_power_sum(
 *             distribution._isotopic_constants, element, order) * coef
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     phi += isotopic_constants_nth_modified_element_power_sum(
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":488
 *         i += 1
 * 
 *     phi += isotopic_constants_nth_modified_element_power_sum(             # <<<<<<<<<<<<<<
 *         distribution._isotopic_constants, symbol, order)
 * 
 */
  __pyx_v_phi = (__pyx_v_phi + __pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_modified_element_power_sum(__pyx_v_distribution->_isotopic_constants, __pyx_v_symbol, __pyx_v_order));

  /* "brainpy/_c/isotopic_distribution.pyx":491
 *         distribution._isotopic_constants, symbol, order)
 * 
 *     return phi             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_phi;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":465
 * 
 * 
 * cdef double _id_modified_phi_value(IsotopicDistribution* distribution, char* symbol, int order) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double phi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":494
 * 
 * 
 * cdef dvec* id_probability_vector(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* phi_values
 */

static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_probability_vector(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution) {
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_phi_values;
  int __pyx_v_max_variant_count;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_probability_vector;
  size_t __pyx_v_i;
  int __pyx_v_sign;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  size_t __pyx_t_5;

  /* "brainpy/_c/isotopic_distribution.pyx":502
 *         int sign
 * 
 *     phi_values = id_phi_values(distribution)             # <<<<<<<<<<<<<<
 *     probability_vector = make_double_vector_with_size(phi_values.size)
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 */
  __pyx_v_phi_values = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_phi_values(__pyx_v_distribution);

  /* "brainpy/_c/isotopic_distribution.pyx":503
 * 
 *     phi_values = id_phi_values(distribution)
 *     probability_vector = make_double_vector_with_size(phi_values.size)             # <<<<<<<<<<<<<<
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 * 
 */
  __pyx_v_probability_vector = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector_with_size(__pyx_v_phi_values->size);

  /* "brainpy/_c/isotopic_distribution.pyx":504
 *     phi_values = id_phi_values(distribution)
 *     probability_vector = make_double_vector_with_size(phi_values.size)
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)             # <<<<<<<<<<<<<<
 * 
 *     newton(phi_values, probability_vector, max_variant_count)
 */
  __pyx_v_max_variant_count = __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(__pyx_v_distribution->composition, __pyx_v_distribution->cache);

  /* "brainpy/_c/isotopic_distribution.pyx":506
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 * 
 *     newton(phi_values, probability_vector, max_variant_count)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, probability_vector.used):
 */
  __pyx_f_7brainpy_2_c_18isotopic_constants_newton(__pyx_v_phi_values, __pyx_v_probability_vector, __pyx_v_max_variant_count);

  /* "brainpy/_c/isotopic_distribution.pyx":508
 *     newton(phi_values, probability_vector, max_variant_count)
 * 
 *     for i in range(0, probability_vector.used):             # <<<<<<<<<<<<<<
 *         sign = 1 if i % 2 == 0 else -1
 *         probability_vector.v[i] *= distribution.monoisotopic_peak.intensity * sign
 */
  __pyx_t_1 = __pyx_v_probability_vector->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":509
 * 
 *     for i in range(0, probability_vector.used):
 *         sign = 1 if i % 2 == 0 else -1             # <<<<<<<<<<<<<<
 *         probability_vector.v[i] *= distribution.monoisotopic_peak.intensity * sign
 * 
 */
    if ((((__pyx_v_i % 2) == 0) != 0)) {
      __pyx_t_4 = 1;
    } else {
      __pyx_t_4 = -1;
    }
    __pyx_v_sign = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":510
 *     for i in range(0, probability_vector.used):
 *         sign = 1 if i % 2 == 0 else -1
 *         probability_vector.v[i] *= distribution.monoisotopic_peak.intensity * sign             # <<<<<<<<<<<<<<
 * 
 *     free_double_vector(phi_values)
 */
    __pyx_t_5 = __pyx_v_i;
    (__pyx_v_probability_vector->v[__pyx_t_5]) = ((__pyx_v_probability_vector->v[__pyx_t_5]) * (__pyx_v_distribution->monoisotopic_peak->intensity * __pyx_v_sign));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":512
 *         probability_vector.v[i] *= distribution.monoisotopic_peak.intensity * sign
 * 
 *     free_double_vector(phi_values)             # <<<<<<<<<<<<<<
 *     return probability_vector
 * 
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_phi_values);

  /* "brainpy/_c/isotopic_distribution.pyx":513
 * 
 *     free_double_vector(phi_values)
 *     return probability_vector             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_probability_vector;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":494
 * 
 * 
 * cdef dvec* id_probability_vector(IsotopicDistribution* distribution) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* phi_values
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":517
 * 
 * @cython.cdivision
 * cdef dvec* id_center_mass_vector(IsotopicDistribution* distribution, dvec* probability_vector) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* mass_vector
 */

static __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_center_mass_vector(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_probability_vector) {
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_mass_vector;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_power_sum;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_ele_sym_poly;
  int __pyx_v_max_variant_count;
  int __pyx_v_sign;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element_struct;
  char *__pyx_v_element;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v__element_count;
  double __pyx_v_center;
  double __pyx_v_temp;
  double __pyx_v_polynomial_term;
  double __pyx_v__monoisotopic_mass;
  double __pyx_v_base_intensity;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_k;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementPolynomialMap *__pyx_v_ep_map;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_r;
  double __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;

  /* "brainpy/_c/isotopic_distribution.pyx":531
 *         ElementPolynomialMap* ep_map
 * 
 *     mass_vector = make_double_vector_with_size(probability_vector.size + 3)             # <<<<<<<<<<<<<<
 *     power_sum = make_double_vector()
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 */
  __pyx_v_mass_vector = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector_with_size((__pyx_v_probability_vector->size + 3));

  /* "brainpy/_c/isotopic_distribution.pyx":532
 * 
 *     mass_vector = make_double_vector_with_size(probability_vector.size + 3)
 *     power_sum = make_double_vector()             # <<<<<<<<<<<<<<
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 *     base_intensity = distribution.monoisotopic_peak.intensity
 */
  __pyx_v_power_sum = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

  /* "brainpy/_c/isotopic_distribution.pyx":533
 *     mass_vector = make_double_vector_with_size(probability_vector.size + 3)
 *     power_sum = make_double_vector()
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)             # <<<<<<<<<<<<<<
 *     base_intensity = distribution.monoisotopic_peak.intensity
 *     ep_map = make_element_polynomial_map(distribution.composition.size)
 */
  __pyx_v_max_variant_count = __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(__pyx_v_distribution->composition, __pyx_v_distribution->cache);

  /* "brainpy/_c/isotopic_distribution.pyx":534
 *     power_sum = make_double_vector()
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 *     base_intensity = distribution.monoisotopic_peak.intensity             # <<<<<<<<<<<<<<
 *     ep_map = make_element_polynomial_map(distribution.composition.size)
 * 
 */
  __pyx_t_1 = __pyx_v_distribution->monoisotopic_peak->intensity;
  __pyx_v_base_intensity = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":535
 *     max_variant_count = max_variants(distribution.composition, distribution.cache)
 *     base_intensity = distribution.monoisotopic_peak.intensity
 *     ep_map = make_element_polynomial_map(distribution.composition.size)             # <<<<<<<<<<<<<<
 * 
 *     j = 0
 */
  __pyx_v_ep_map = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_polynomial_map(__pyx_v_distribution->composition->size);

  /* "brainpy/_c/isotopic_distribution.pyx":537
 *     ep_map = make_element_polynomial_map(distribution.composition.size)
 * 
 *     j = 0             # <<<<<<<<<<<<<<
 *     while j < distribution.composition.used:
 *         element = distribution.composition.elements[j]
 */
  __pyx_v_j = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":538
 * 
 *     j = 0
 *     while j < distribution.composition.used:             # <<<<<<<<<<<<<<
 *         element = distribution.composition.elements[j]
 *         reset_double_vector(power_sum)
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_j < __pyx_v_distribution->composition->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/isotopic_distribution.pyx":539
 *     j = 0
 *     while j < distribution.composition.used:
 *         element = distribution.composition.elements[j]             # <<<<<<<<<<<<<<
 *         reset_double_vector(power_sum)
 *         power_sum = id_modified_phi_values(distribution, element, power_sum)
 */
    __pyx_v_element = (__pyx_v_distribution->composition->elements[__pyx_v_j]);

    /* "brainpy/_c/isotopic_distribution.pyx":540
 *     while j < distribution.composition.used:
 *         element = distribution.composition.elements[j]
 *         reset_double_vector(power_sum)             # <<<<<<<<<<<<<<
 *         power_sum = id_modified_phi_values(distribution, element, power_sum)
 *         ele_sym_poly = make_double_vector()
 */
    __pyx_f_7brainpy_2_c_13double_vector_reset_double_vector(__pyx_v_power_sum);

    /* "brainpy/_c/isotopic_distribution.pyx":541
 *         element = distribution.composition.elements[j]
 *         reset_double_vector(power_sum)
 *         power_sum = id_modified_phi_values(distribution, element, power_sum)             # <<<<<<<<<<<<<<
 *         ele_sym_poly = make_double_vector()
 *         newton(power_sum, ele_sym_poly, max_variant_count)
 */
    __pyx_v_power_sum = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_modified_phi_values(__pyx_v_distribution, __pyx_v_element, __pyx_v_power_sum);

    /* "brainpy/_c/isotopic_distribution.pyx":542
 *         reset_double_vector(power_sum)
 *         power_sum = id_modified_phi_values(distribution, element, power_sum)
 *         ele_sym_poly = make_double_vector()             # <<<<<<<<<<<<<<
 *         newton(power_sum, ele_sym_poly, max_variant_count)
 *         element_polynomial_map_set(ep_map, element, ele_sym_poly)
 */
    __pyx_v_ele_sym_poly = __pyx_f_7brainpy_2_c_13double_vector_make_double_vector();

    /* "brainpy/_c/isotopic_distribution.pyx":543
 *         power_sum = id_modified_phi_values(distribution, element, power_sum)
 *         ele_sym_poly = make_double_vector()
 *         newton(power_sum, ele_sym_poly, max_variant_count)             # <<<<<<<<<<<<<<
 *         element_polynomial_map_set(ep_map, element, ele_sym_poly)
 *         j += 1
 */
    __pyx_f_7brainpy_2_c_18isotopic_constants_newton(__pyx_v_power_sum, __pyx_v_ele_sym_poly, __pyx_v_max_variant_count);

    /* "brainpy/_c/isotopic_distribution.pyx":544
 *         ele_sym_poly = make_double_vector()
 *         newton(power_sum, ele_sym_poly, max_variant_count)
 *         element_polynomial_map_set(ep_map, element, ele_sym_poly)             # <<<<<<<<<<<<<<
 *         j += 1
 *     free_double_vector(power_sum)
 */
    (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_set(__pyx_v_ep_map, __pyx_v_element, __pyx_v_ele_sym_poly));

    /* "brainpy/_c/isotopic_distribution.pyx":545
 *         newton(power_sum, ele_sym_poly, max_variant_count)
 *         element_polynomial_map_set(ep_map, element, ele_sym_poly)
 *         j += 1             # <<<<<<<<<<<<<<
 *     free_double_vector(power_sum)
 *     i = 0
 */
    __pyx_v_j = (__pyx_v_j + 1);
  }

  /* "brainpy/_c/isotopic_distribution.pyx":546
 *         element_polynomial_map_set(ep_map, element, ele_sym_poly)
 *         j += 1
 *     free_double_vector(power_sum)             # <<<<<<<<<<<<<<
 *     i = 0
 *     for i in range(distribution.order + 1):
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_power_sum);

  /* "brainpy/_c/isotopic_distribution.pyx":547
 *         j += 1
 *     free_double_vector(power_sum)
 *     i = 0             # <<<<<<<<<<<<<<
 *     for i in range(distribution.order + 1):
 *         sign = 1 if i % 2 == 0 else -1
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":548
 *     free_double_vector(power_sum)
 *     i = 0
 *     for i in range(distribution.order + 1):             # <<<<<<<<<<<<<<
 *         sign = 1 if i % 2 == 0 else -1
 *         center = 0.0
 */
  __pyx_t_3 = (__pyx_v_distribution->order + 1);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "brainpy/_c/isotopic_distribution.pyx":549
 *     i = 0
 *     for i in range(distribution.order + 1):
 *         sign = 1 if i % 2 == 0 else -1             # <<<<<<<<<<<<<<
 *         center = 0.0
 *         k = 0
 */
    if ((((__pyx_v_i % 2) == 0) != 0)) {
      __pyx_t_6 = 1;
    } else {
      __pyx_t_6 = -1;
    }
    __pyx_v_sign = __pyx_t_6;

    /* "brainpy/_c/isotopic_distribution.pyx":550
 *     for i in range(distribution.order + 1):
 *         sign = 1 if i % 2 == 0 else -1
 *         center = 0.0             # <<<<<<<<<<<<<<
 *         k = 0
 *         while k < ep_map.used:
 */
    __pyx_v_center = 0.0;

    /* "brainpy/_c/isotopic_distribution.pyx":551
 *         sign = 1 if i % 2 == 0 else -1
 *         center = 0.0
 *         k = 0             # <<<<<<<<<<<<<<
 *         while k < ep_map.used:
 *             element = ep_map.elements[k]
 */
    __pyx_v_k = 0;

    /* "brainpy/_c/isotopic_distribution.pyx":552
 *         center = 0.0
 *         k = 0
 *         while k < ep_map.used:             # <<<<<<<<<<<<<<
 *             element = ep_map.elements[k]
 *             element_polynomial_map_get(ep_map, element, &ele_sym_poly)
 */
    while (1) {
      __pyx_t_2 = ((__pyx_v_k < __pyx_v_ep_map->used) != 0);
      if (!__pyx_t_2) break;

      /* "brainpy/_c/isotopic_distribution.pyx":553
 *         k = 0
 *         while k < ep_map.used:
 *             element = ep_map.elements[k]             # <<<<<<<<<<<<<<
 *             element_polynomial_map_get(ep_map, element, &ele_sym_poly)
 * 
 */
      __pyx_v_element = (__pyx_v_ep_map->elements[__pyx_v_k]);

      /* "brainpy/_c/isotopic_distribution.pyx":554
 *         while k < ep_map.used:
 *             element = ep_map.elements[k]
 *             element_polynomial_map_get(ep_map, element, &ele_sym_poly)             # <<<<<<<<<<<<<<
 * 
 *             composition_get_element_count(distribution.composition, element, &_element_count)
 */
      (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_element_polynomial_map_get(__pyx_v_ep_map, __pyx_v_element, (&__pyx_v_ele_sym_poly)));

      /* "brainpy/_c/isotopic_distribution.pyx":556
 *             element_polynomial_map_get(ep_map, element, &ele_sym_poly)
 * 
 *             composition_get_element_count(distribution.composition, element, &_element_count)             # <<<<<<<<<<<<<<
 *             element_cache_get(distribution.cache, element, &element_struct)
 * 
 */
      (void)(__pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_distribution->composition, __pyx_v_element, (&__pyx_v__element_count)));

      /* "brainpy/_c/isotopic_distribution.pyx":557
 * 
 *             composition_get_element_count(distribution.composition, element, &_element_count)
 *             element_cache_get(distribution.cache, element, &element_struct)             # <<<<<<<<<<<<<<
 * 
 *             _monoisotopic_mass = element_monoisotopic_mass(element_struct)
 */
      (void)(__pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get(__pyx_v_distribution->cache, __pyx_v_element, (&__pyx_v_element_struct)));

      /* "brainpy/_c/isotopic_distribution.pyx":559
 *             element_cache_get(distribution.cache, element, &element_struct)
 * 
 *             _monoisotopic_mass = element_monoisotopic_mass(element_struct)             # <<<<<<<<<<<<<<
 * 
 *             polynomial_term = ele_sym_poly.v[i]
 */
      __pyx_v__monoisotopic_mass = __pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass(__pyx_v_element_struct);

      /* "brainpy/_c/isotopic_distribution.pyx":561
 *             _monoisotopic_mass = element_monoisotopic_mass(element_struct)
 * 
 *             polynomial_term = ele_sym_poly.v[i]             # <<<<<<<<<<<<<<
 *             temp = _element_count
 *             temp *= sign * polynomial_term
 */
      __pyx_v_polynomial_term = (__pyx_v_ele_sym_poly->v[__pyx_v_i]);

      /* "brainpy/_c/isotopic_distribution.pyx":562
 * 
 *             polynomial_term = ele_sym_poly.v[i]
 *             temp = _element_count             # <<<<<<<<<<<<<<
 *             temp *= sign * polynomial_term
 *             temp *= base_intensity * _monoisotopic_mass
 */
      __pyx_v_temp = __pyx_v__element_count;

      /* "brainpy/_c/isotopic_distribution.pyx":563
 *             polynomial_term = ele_sym_poly.v[i]
 *             temp = _element_count
 *             temp *= sign * polynomial_term             # <<<<<<<<<<<<<<
 *             temp *= base_intensity * _monoisotopic_mass
 *             center += temp
 */
      __pyx_v_temp = (__pyx_v_temp * (__pyx_v_sign * __pyx_v_polynomial_term));

      /* "brainpy/_c/isotopic_distribution.pyx":564
 *             temp = _element_count
 *             temp *= sign * polynomial_term
 *             temp *= base_intensity * _monoisotopic_mass             # <<<<<<<<<<<<<<
 *             center += temp
 *             k += 1
 */
      __pyx_v_temp = (__pyx_v_temp * (__pyx_v_base_intensity * __pyx_v__monoisotopic_mass));

      /* "brainpy/_c/isotopic_distribution.pyx":565
 *             temp *= sign * polynomial_term
 *             temp *= base_intensity * _monoisotopic_mass
 *             center += temp             # <<<<<<<<<<<<<<
 *             k += 1
 *         if probability_vector.v[i] == 0:
 */
      __pyx_v_center = (__pyx_v_center + __pyx_v_temp);

      /* "brainpy/_c/isotopic_distribution.pyx":566
 *             temp *= base_intensity * _monoisotopic_mass
 *             center += temp
 *             k += 1             # <<<<<<<<<<<<<<
 *         if probability_vector.v[i] == 0:
 *             double_vector_append(mass_vector, 0)
 */
      __pyx_v_k = (__pyx_v_k + 1);
    }

    /* "brainpy/_c/isotopic_distribution.pyx":567
 *             center += temp
 *             k += 1
 *         if probability_vector.v[i] == 0:             # <<<<<<<<<<<<<<
 *             double_vector_append(mass_vector, 0)
 *         else:
 */
    __pyx_t_2 = (((__pyx_v_probability_vector->v[__pyx_v_i]) == 0.0) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/isotopic_distribution.pyx":568
 *             k += 1
 *         if probability_vector.v[i] == 0:
 *             double_vector_append(mass_vector, 0)             # <<<<<<<<<<<<<<
 *         else:
 *             double_vector_append(mass_vector, center / probability_vector.v[i])
 */
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_mass_vector, 0.0));

      /* "brainpy/_c/isotopic_distribution.pyx":567
 *             center += temp
 *             k += 1
 *         if probability_vector.v[i] == 0:             # <<<<<<<<<<<<<<
 *             double_vector_append(mass_vector, 0)
 *         else:
 */
      goto __pyx_L9;
    }

    /* "brainpy/_c/isotopic_distribution.pyx":570
 *             double_vector_append(mass_vector, 0)
 *         else:
 *             double_vector_append(mass_vector, center / probability_vector.v[i])             # <<<<<<<<<<<<<<
 * 
 *     free_element_polynomial_map(ep_map)
 */
    /*else*/ {
      (void)(__pyx_f_7brainpy_2_c_13double_vector_double_vector_append(__pyx_v_mass_vector, (__pyx_v_center / (__pyx_v_probability_vector->v[__pyx_v_i]))));
    }
    __pyx_L9:;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":572
 *             double_vector_append(mass_vector, center / probability_vector.v[i])
 * 
 *     free_element_polynomial_map(ep_map)             # <<<<<<<<<<<<<<
 * 
 *     return mass_vector
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_polynomial_map(__pyx_v_ep_map);

  /* "brainpy/_c/isotopic_distribution.pyx":574
 *     free_element_polynomial_map(ep_map)
 * 
 *     return mass_vector             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_mass_vector;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":517
 * 
 * @cython.cdivision
 * cdef dvec* id_center_mass_vector(IsotopicDistribution* distribution, dvec* probability_vector) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* mass_vector
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":578
 * 
 * @cython.cdivision
 * cdef PeakList* id_aggregated_isotopic_variants(IsotopicDistribution* distribution, int charge, double charge_carrier) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* probability_vector
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_id_aggregated_isotopic_variants(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution, int __pyx_v_charge, double __pyx_v_charge_carrier) {
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_probability_vector;
  __pyx_t_7brainpy_2_c_21isotopic_distribution_dvec *__pyx_v_center_mass_vector;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peak_set;
  double __pyx_v_average_mass;
  double __pyx_v_adjusted_mz;
  double __pyx_v_total;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak __pyx_v_peak;
  double __pyx_v_center_mass_i;
  double __pyx_v_intensity_i;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  double __pyx_t_7;
  struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio __pyx_t_8;

  /* "brainpy/_c/isotopic_distribution.pyx":589
 *         double center_mass_i, intensity_i
 * 
 *     probability_vector = id_probability_vector(distribution)             # <<<<<<<<<<<<<<
 *     center_mass_vector = id_center_mass_vector(distribution, probability_vector)
 * 
 */
  __pyx_v_probability_vector = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_probability_vector(__pyx_v_distribution);

  /* "brainpy/_c/isotopic_distribution.pyx":590
 * 
 *     probability_vector = id_probability_vector(distribution)
 *     center_mass_vector = id_center_mass_vector(distribution, probability_vector)             # <<<<<<<<<<<<<<
 * 
 *     peak_set = make_peak_list()
 */
  __pyx_v_center_mass_vector = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_center_mass_vector(__pyx_v_distribution, __pyx_v_probability_vector);

  /* "brainpy/_c/isotopic_distribution.pyx":592
 *     center_mass_vector = id_center_mass_vector(distribution, probability_vector)
 * 
 *     peak_set = make_peak_list()             # <<<<<<<<<<<<<<
 * 
 *     average_mass = 0
 */
  __pyx_v_peak_set = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list();

  /* "brainpy/_c/isotopic_distribution.pyx":594
 *     peak_set = make_peak_list()
 * 
 *     average_mass = 0             # <<<<<<<<<<<<<<
 *     total = 0
 * 
 */
  __pyx_v_average_mass = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":595
 * 
 *     average_mass = 0
 *     total = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(probability_vector.used):
 */
  __pyx_v_total = 0.0;

  /* "brainpy/_c/isotopic_distribution.pyx":597
 *     total = 0
 * 
 *     for i in range(probability_vector.used):             # <<<<<<<<<<<<<<
 *         total += probability_vector.v[i]
 * 
 */
  __pyx_t_1 = __pyx_v_probability_vector->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/isotopic_distribution.pyx":598
 * 
 *     for i in range(probability_vector.used):
 *         total += probability_vector.v[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(distribution.order + 1):
 */
    __pyx_v_total = (__pyx_v_total + (__pyx_v_probability_vector->v[__pyx_v_i]));
  }

  /* "brainpy/_c/isotopic_distribution.pyx":600
 *         total += probability_vector.v[i]
 * 
 *     for i in range(distribution.order + 1):             # <<<<<<<<<<<<<<
 *         center_mass_i = center_mass_vector.v[i]
 *         intensity_i = probability_vector.v[i]
 */
  __pyx_t_4 = (__pyx_v_distribution->order + 1);
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_5; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "brainpy/_c/isotopic_distribution.pyx":601
 * 
 *     for i in range(distribution.order + 1):
 *         center_mass_i = center_mass_vector.v[i]             # <<<<<<<<<<<<<<
 *         intensity_i = probability_vector.v[i]
 * 
 */
    __pyx_v_center_mass_i = (__pyx_v_center_mass_vector->v[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":602
 *     for i in range(distribution.order + 1):
 *         center_mass_i = center_mass_vector.v[i]
 *         intensity_i = probability_vector.v[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_intensity_i = (__pyx_v_probability_vector->v[__pyx_v_i]);

    /* "brainpy/_c/isotopic_distribution.pyx":605
 * 
 * 
 *         if charge != 0:             # <<<<<<<<<<<<<<
 *             adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)
 *         else:
 */
    __pyx_t_6 = ((__pyx_v_charge != 0) != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/isotopic_distribution.pyx":606
 * 
 *         if charge != 0:
 *             adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)             # <<<<<<<<<<<<<<
 *         else:
 *             adjusted_mz = center_mass_i
 */
      __pyx_t_8.__pyx_n = 1;
      __pyx_t_8.charge_carrier = __pyx_v_charge_carrier;
      __pyx_t_7 = __pyx_f_7brainpy_2_c_11composition_mass_charge_ratio(__pyx_v_center_mass_i, __pyx_v_charge, &__pyx_t_8); 
      __pyx_v_adjusted_mz = __pyx_t_7;

      /* "brainpy/_c/isotopic_distribution.pyx":605
 * 
 * 
 *         if charge != 0:             # <<<<<<<<<<<<<<
 *             adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)
 *         else:
 */
      goto __pyx_L7;
    }

    /* "brainpy/_c/isotopic_distribution.pyx":608
 *             adjusted_mz = mass_charge_ratio(center_mass_i, charge, charge_carrier)
 *         else:
 *             adjusted_mz = center_mass_i             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_v_adjusted_mz = __pyx_v_center_mass_i;
    }
    __pyx_L7:;

    /* "brainpy/_c/isotopic_distribution.pyx":611
 * 
 * 
 *         peak.mz = adjusted_mz             # <<<<<<<<<<<<<<
 *         peak.intensity = intensity_i / total
 *         peak.charge = charge
 */
    __pyx_v_peak.mz = __pyx_v_adjusted_mz;

    /* "brainpy/_c/isotopic_distribution.pyx":612
 * 
 *         peak.mz = adjusted_mz
 *         peak.intensity = intensity_i / total             # <<<<<<<<<<<<<<
 *         peak.charge = charge
 * 
 */
    __pyx_v_peak.intensity = (__pyx_v_intensity_i / __pyx_v_total);

    /* "brainpy/_c/isotopic_distribution.pyx":613
 *         peak.mz = adjusted_mz
 *         peak.intensity = intensity_i / total
 *         peak.charge = charge             # <<<<<<<<<<<<<<
 * 
 *         if peak.intensity < 1e-10:
 */
    __pyx_v_peak.charge = __pyx_v_charge;

    /* "brainpy/_c/isotopic_distribution.pyx":615
 *         peak.charge = charge
 * 
 *         if peak.intensity < 1e-10:             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
    __pyx_t_6 = ((__pyx_v_peak.intensity < 1e-10) != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/isotopic_distribution.pyx":616
 * 
 *         if peak.intensity < 1e-10:
 *             continue             # <<<<<<<<<<<<<<
 * 
 *         peak_list_append(peak_set, &peak)
 */
      goto __pyx_L5_continue;

      /* "brainpy/_c/isotopic_distribution.pyx":615
 *         peak.charge = charge
 * 
 *         if peak.intensity < 1e-10:             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":618
 *             continue
 * 
 *         peak_list_append(peak_set, &peak)             # <<<<<<<<<<<<<<
 * 
 *         average_mass += adjusted_mz * intensity_i
 */
    __pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append(__pyx_v_peak_set, (&__pyx_v_peak));

    /* "brainpy/_c/isotopic_distribution.pyx":620
 *         peak_list_append(peak_set, &peak)
 * 
 *         average_mass += adjusted_mz * intensity_i             # <<<<<<<<<<<<<<
 * 
 *     average_mass /= total
 */
    __pyx_v_average_mass = (__pyx_v_average_mass + (__pyx_v_adjusted_mz * __pyx_v_intensity_i));
    __pyx_L5_continue:;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":622
 *         average_mass += adjusted_mz * intensity_i
 * 
 *     average_mass /= total             # <<<<<<<<<<<<<<
 * 
 *     free_double_vector(probability_vector)
 */
  __pyx_v_average_mass = (__pyx_v_average_mass / __pyx_v_total);

  /* "brainpy/_c/isotopic_distribution.pyx":624
 *     average_mass /= total
 * 
 *     free_double_vector(probability_vector)             # <<<<<<<<<<<<<<
 *     free_double_vector(center_mass_vector)
 * 
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_probability_vector);

  /* "brainpy/_c/isotopic_distribution.pyx":625
 * 
 *     free_double_vector(probability_vector)
 *     free_double_vector(center_mass_vector)             # <<<<<<<<<<<<<<
 * 
 *     distribution.average_mass = average_mass
 */
  __pyx_f_7brainpy_2_c_13double_vector_free_double_vector(__pyx_v_center_mass_vector);

  /* "brainpy/_c/isotopic_distribution.pyx":627
 *     free_double_vector(center_mass_vector)
 * 
 *     distribution.average_mass = average_mass             # <<<<<<<<<<<<<<
 * 
 *     sort_by_mz(peak_set)
 */
  __pyx_v_distribution->average_mass = __pyx_v_average_mass;

  /* "brainpy/_c/isotopic_distribution.pyx":629
 *     distribution.average_mass = average_mass
 * 
 *     sort_by_mz(peak_set)             # <<<<<<<<<<<<<<
 *     return peak_set
 * 
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_sort_by_mz(__pyx_v_peak_set);

  /* "brainpy/_c/isotopic_distribution.pyx":630
 * 
 *     sort_by_mz(peak_set)
 *     return peak_set             # <<<<<<<<<<<<<<
 * 
 * cdef void sort_by_mz(PeakList* peaklist) nogil:
 */
  __pyx_r = __pyx_v_peak_set;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":578
 * 
 * @cython.cdivision
 * cdef PeakList* id_aggregated_isotopic_variants(IsotopicDistribution* distribution, int charge, double charge_carrier) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         dvec* probability_vector
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":632
 *     return peak_set
 * 
 * cdef void sort_by_mz(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)
 * 
 */

static void __pyx_f_7brainpy_2_c_21isotopic_distribution_sort_by_mz(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist) {

  /* "brainpy/_c/isotopic_distribution.pyx":633
 * 
 * cdef void sort_by_mz(PeakList* peaklist) nogil:
 *     qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)             # <<<<<<<<<<<<<<
 * 
 * cdef int compare_by_mz(const void * a, const void * b) nogil:
 */
  qsort(__pyx_v_peaklist->peaks, __pyx_v_peaklist->used, (sizeof(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak)), __pyx_f_7brainpy_2_c_21isotopic_distribution_compare_by_mz);

  /* "brainpy/_c/isotopic_distribution.pyx":632
 *     return peak_set
 * 
 * cdef void sort_by_mz(PeakList* peaklist) nogil:             # <<<<<<<<<<<<<<
 *     qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)
 * 
 */

  /* function exit code */
}

/* "brainpy/_c/isotopic_distribution.pyx":635
 *     qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)
 * 
 * cdef int compare_by_mz(const void * a, const void * b) nogil:             # <<<<<<<<<<<<<<
 *     if (<Peak*>a).mz < (<Peak*>b).mz:
 *         return -1
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_compare_by_mz(void const *__pyx_v_a, void const *__pyx_v_b) {
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":636
 * 
 * cdef int compare_by_mz(const void * a, const void * b) nogil:
 *     if (<Peak*>a).mz < (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return -1
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 */
  __pyx_t_1 = ((((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_a)->mz < ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_b)->mz) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":637
 * cdef int compare_by_mz(const void * a, const void * b) nogil:
 *     if (<Peak*>a).mz < (<Peak*>b).mz:
 *         return -1             # <<<<<<<<<<<<<<
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 *         return 0
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":636
 * 
 * cdef int compare_by_mz(const void * a, const void * b) nogil:
 *     if (<Peak*>a).mz < (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return -1
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":638
 *     if (<Peak*>a).mz < (<Peak*>b).mz:
 *         return -1
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return 0
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:
 */
  __pyx_t_1 = ((((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_a)->mz == ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_b)->mz) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":639
 *         return -1
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 *         return 0             # <<<<<<<<<<<<<<
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:
 *         return 1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":638
 *     if (<Peak*>a).mz < (<Peak*>b).mz:
 *         return -1
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return 0
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":640
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 *         return 0
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_a)->mz > ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)__pyx_v_b)->mz) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":641
 *         return 0
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":640
 *     elif (<Peak*>a).mz == (<Peak*>b).mz:
 *         return 0
 *     elif (<Peak*>a).mz > (<Peak*>b).mz:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  }

  /* "brainpy/_c/isotopic_distribution.pyx":635
 *     qsort(peaklist.peaks, peaklist.used, sizeof(Peak), compare_by_mz)
 * 
 * cdef int compare_by_mz(const void * a, const void * b) nogil:             # <<<<<<<<<<<<<<
 *     if (<Peak*>a).mz < (<Peak*>b).mz:
 *         return -1
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":644
 * 
 * 
 * cpdef bint check_composition_non_negative(dict composition):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint negative_element
 */

static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_1check_composition_non_negative(PyObject *__pyx_self, PyObject *__pyx_v_composition); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(PyObject *__pyx_v_composition, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_negative_element;
  CYTHON_UNUSED PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_v = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("check_composition_non_negative", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":650
 *         object v
 * 
 *     negative_element = False             # <<<<<<<<<<<<<<
 *     for k, v in composition.items():
 *         if v < 0:
 */
  __pyx_v_negative_element = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":651
 * 
 *     negative_element = False
 *     for k, v in composition.items():             # <<<<<<<<<<<<<<
 *         if v < 0:
 *             negative_element = True
 */
  if (unlikely(__pyx_v_composition == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 651, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Items(__pyx_v_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 651, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 651, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 651, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 651, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 651, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(PyString_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_k, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "brainpy/_c/isotopic_distribution.pyx":652
 *     negative_element = False
 *     for k, v in composition.items():
 *         if v < 0:             # <<<<<<<<<<<<<<
 *             negative_element = True
 *             break
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_v, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {

      /* "brainpy/_c/isotopic_distribution.pyx":653
 *     for k, v in composition.items():
 *         if v < 0:
 *             negative_element = True             # <<<<<<<<<<<<<<
 *             break
 *     return negative_element
 */
      __pyx_v_negative_element = 1;

      /* "brainpy/_c/isotopic_distribution.pyx":654
 *         if v < 0:
 *             negative_element = True
 *             break             # <<<<<<<<<<<<<<
 *     return negative_element
 * 
 */
      goto __pyx_L4_break;

      /* "brainpy/_c/isotopic_distribution.pyx":652
 *     negative_element = False
 *     for k, v in composition.items():
 *         if v < 0:             # <<<<<<<<<<<<<<
 *             negative_element = True
 *             break
 */
    }

    /* "brainpy/_c/isotopic_distribution.pyx":651
 * 
 *     negative_element = False
 *     for k, v in composition.items():             # <<<<<<<<<<<<<<
 *         if v < 0:
 *             negative_element = True
 */
  }
  __pyx_L4_break:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":655
 *             negative_element = True
 *             break
 *     return negative_element             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_negative_element;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":644
 * 
 * 
 * cpdef bint check_composition_non_negative(dict composition):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint negative_element
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("brainpy._c.isotopic_distribution.check_composition_non_negative", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_1check_composition_non_negative(PyObject *__pyx_self, PyObject *__pyx_v_composition); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative[] = "check_composition_non_negative(dict composition) -> bool";
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_1check_composition_non_negative(PyObject *__pyx_self, PyObject *__pyx_v_composition) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_composition_non_negative (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_composition), (&PyDict_Type), 1, "composition", 1))) __PYX_ERR(0, 644, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(__pyx_self, ((PyObject*)__pyx_v_composition));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("check_composition_non_negative", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative(__pyx_v_composition, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.check_composition_non_negative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":658
 * 
 * 
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=NULL) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int max_n_variants, npeaks
 */

static int __pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_struct, size_t __pyx_v_max_npeaks, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks *__pyx_optional_args) {
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *__pyx_v_cache = ((struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)NULL);
  int __pyx_v_max_n_variants;
  int __pyx_v_npeaks;
  int __pyx_r;
  size_t __pyx_t_1;
  long __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  size_t __pyx_t_5;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cache = __pyx_optional_args->cache;
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":661
 *     cdef:
 *         int max_n_variants, npeaks
 *     max_n_variants = max_variants(composition_struct, cache)             # <<<<<<<<<<<<<<
 *     npeaks = <int>sqrt(max_n_variants) - 2
 *     npeaks = min(max(npeaks, 3), max_npeaks)
 */
  __pyx_v_max_n_variants = __pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants(__pyx_v_composition_struct, __pyx_v_cache);

  /* "brainpy/_c/isotopic_distribution.pyx":662
 *         int max_n_variants, npeaks
 *     max_n_variants = max_variants(composition_struct, cache)
 *     npeaks = <int>sqrt(max_n_variants) - 2             # <<<<<<<<<<<<<<
 *     npeaks = min(max(npeaks, 3), max_npeaks)
 *     return npeaks
 */
  __pyx_v_npeaks = (((int)sqrt(__pyx_v_max_n_variants)) - 2);

  /* "brainpy/_c/isotopic_distribution.pyx":663
 *     max_n_variants = max_variants(composition_struct, cache)
 *     npeaks = <int>sqrt(max_n_variants) - 2
 *     npeaks = min(max(npeaks, 3), max_npeaks)             # <<<<<<<<<<<<<<
 *     return npeaks
 * 
 */
  __pyx_t_1 = __pyx_v_max_npeaks;
  __pyx_t_2 = 3;
  __pyx_t_3 = __pyx_v_npeaks;
  if (((__pyx_t_2 > __pyx_t_3) != 0)) {
    __pyx_t_4 = __pyx_t_2;
  } else {
    __pyx_t_4 = __pyx_t_3;
  }
  __pyx_t_2 = __pyx_t_4;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_5 = __pyx_t_1;
  } else {
    __pyx_t_5 = __pyx_t_2;
  }
  __pyx_v_npeaks = __pyx_t_5;

  /* "brainpy/_c/isotopic_distribution.pyx":664
 *     npeaks = <int>sqrt(max_n_variants) - 2
 *     npeaks = min(max(npeaks, 3), max_npeaks)
 *     return npeaks             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_npeaks;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":658
 * 
 * 
 * cdef int guess_npeaks(Composition* composition_struct, size_t max_npeaks, ElementCache* cache=NULL) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int max_n_variants, npeaks
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":667
 * 
 * 
 * def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,             # <<<<<<<<<<<<<<
 *                         double charge_carrier=PROTON):
 *     '''
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_3pyisotopic_variants(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_2pyisotopic_variants[] = "pyisotopic_variants(composition, npeaks=None, int charge=0, double charge_carrier=PROTON)\n\n    Compute a peak list representing the theoretical isotopic cluster for `composition`.\n\n    Parameters\n    ----------\n    composition : Mapping\n        Any Mapping type where keys are element symbols and values are integers\n    n_peaks: int\n        The number of peaks to include in the isotopic cluster, starting from the monoisotopic peak.\n        If given a number below 1 or above the maximum number of isotopic variants, the maximum will\n        be used. If `None`, a \"reasonable\" value is chosen by `int(sqrt(max_variants(composition)))`.\n    charge: int\n        The charge state of the isotopic cluster to produce. Defaults to 0, theoretical neutral mass.\n    charge_carrier: double\n        The mass of the molecule contributing the ion's charge\n\n    Returns\n    -------\n    list of TheoreticalPeak\n    ";
static PyMethodDef __pyx_mdef_7brainpy_2_c_21isotopic_distribution_3pyisotopic_variants = {"pyisotopic_variants", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_3pyisotopic_variants, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_2pyisotopic_variants};
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_3pyisotopic_variants(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_npeaks = 0;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pyisotopic_variants (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_npeaks,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_npeaks);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pyisotopic_variants") < 0)) __PYX_ERR(0, 667, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = values[0];
    __pyx_v_npeaks = values[1];
    if (values[2]) {
      __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 667, __pyx_L3_error)
    } else {
      __pyx_v_charge = ((int)0);
    }
    if (values[3]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k_;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pyisotopic_variants", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 667, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.pyisotopic_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_composition) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "composition"); __PYX_ERR(0, 667, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_2pyisotopic_variants(__pyx_self, __pyx_v_composition, __pyx_v_npeaks, __pyx_v_charge, __pyx_v_charge_carrier);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_2pyisotopic_variants(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_npeaks, int __pyx_v_charge, double __pyx_v_charge_carrier) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants __pyx_t_2;
  __Pyx_RefNannySetupContext("pyisotopic_variants", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":689
 *     list of TheoreticalPeak
 *     '''
 *     return _isotopic_variants(composition, npeaks, charge, charge_carrier)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.npeaks = __pyx_v_npeaks;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(__pyx_v_composition, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":667
 * 
 * 
 * def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,             # <<<<<<<<<<<<<<
 *                         double charge_carrier=PROTON):
 *     '''
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.pyisotopic_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":692
 * 
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=None, int charge=0, double charge_carrier=PROTON):             # <<<<<<<<<<<<<<
 *     '''
 *     Compute a peak list representing the theoretical isotopic cluster for `composition`.
 */

static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_5_isotopic_variants(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(PyObject *__pyx_v_composition, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args) {
  PyObject *__pyx_v_npeaks = ((PyObject *)Py_None);
  int __pyx_v_charge = ((int)0);
  double __pyx_v_charge_carrier = __pyx_k__2;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_struct;
  PyObject *__pyx_v_peaklist = 0;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peak_set;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_dist;
  int __pyx_v__npeaks;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_isotopic_variants", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_npeaks = __pyx_optional_args->npeaks;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge = __pyx_optional_args->charge;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
        }
      }
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":720
 *         int _npeaks, max_n_variants
 * 
 *     composition_struct = dict_to_composition(dict(composition))             # <<<<<<<<<<<<<<
 *     validate_composition(composition_struct)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_composition_struct = __pyx_f_7brainpy_2_c_11composition_dict_to_composition(((PyObject*)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":721
 * 
 *     composition_struct = dict_to_composition(dict(composition))
 *     validate_composition(composition_struct)             # <<<<<<<<<<<<<<
 * 
 *     if npeaks is None:
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_validate_composition(__pyx_v_composition_struct);

  /* "brainpy/_c/isotopic_distribution.pyx":723
 *     validate_composition(composition_struct)
 * 
 *     if npeaks is None:             # <<<<<<<<<<<<<<
 *         _npeaks = guess_npeaks(composition_struct, 300)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_npeaks == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "brainpy/_c/isotopic_distribution.pyx":724
 * 
 *     if npeaks is None:
 *         _npeaks = guess_npeaks(composition_struct, 300)             # <<<<<<<<<<<<<<
 *     else:
 *         # The npeaks variable is left as a Python-level variable to
 */
    __pyx_v__npeaks = __pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks(__pyx_v_composition_struct, 0x12C, NULL);

    /* "brainpy/_c/isotopic_distribution.pyx":723
 *     validate_composition(composition_struct)
 * 
 *     if npeaks is None:             # <<<<<<<<<<<<<<
 *         _npeaks = guess_npeaks(composition_struct, 300)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":728
 *         # The npeaks variable is left as a Python-level variable to
 *         # allow it to be any Python numeric type
 *         _npeaks = npeaks - 1             # <<<<<<<<<<<<<<
 * 
 *     dist = make_isotopic_distribution(composition_struct, _npeaks)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_npeaks, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v__npeaks = __pyx_t_4;
  }
  __pyx_L3:;

  /* "brainpy/_c/isotopic_distribution.pyx":730
 *         _npeaks = npeaks - 1
 * 
 *     dist = make_isotopic_distribution(composition_struct, _npeaks)             # <<<<<<<<<<<<<<
 *     peak_set = id_aggregated_isotopic_variants(dist, charge, charge_carrier)
 * 
 */
  __pyx_v_dist = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(__pyx_v_composition_struct, __pyx_v__npeaks, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":731
 * 
 *     dist = make_isotopic_distribution(composition_struct, _npeaks)
 *     peak_set = id_aggregated_isotopic_variants(dist, charge, charge_carrier)             # <<<<<<<<<<<<<<
 * 
 *     peaklist = peaklist_to_list(peak_set)
 */
  __pyx_v_peak_set = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_aggregated_isotopic_variants(__pyx_v_dist, __pyx_v_charge, __pyx_v_charge_carrier);

  /* "brainpy/_c/isotopic_distribution.pyx":733
 *     peak_set = id_aggregated_isotopic_variants(dist, charge, charge_carrier)
 * 
 *     peaklist = peaklist_to_list(peak_set)             # <<<<<<<<<<<<<<
 * 
 *     free_peak_list(peak_set)
 */
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution_peaklist_to_list(__pyx_v_peak_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peaklist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":735
 *     peaklist = peaklist_to_list(peak_set)
 * 
 *     free_peak_list(peak_set)             # <<<<<<<<<<<<<<
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_peak_list(__pyx_v_peak_set);

  /* "brainpy/_c/isotopic_distribution.pyx":736
 * 
 *     free_peak_list(peak_set)
 *     free_isotopic_distribution(dist)             # <<<<<<<<<<<<<<
 *     free_composition(composition_struct)
 *     return peaklist
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(__pyx_v_dist);

  /* "brainpy/_c/isotopic_distribution.pyx":737
 *     free_peak_list(peak_set)
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)             # <<<<<<<<<<<<<<
 *     return peaklist
 * 
 */
  __pyx_f_7brainpy_2_c_11composition_free_composition(__pyx_v_composition_struct);

  /* "brainpy/_c/isotopic_distribution.pyx":738
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)
 *     return peaklist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_peaklist);
  __pyx_r = __pyx_v_peaklist;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":692
 * 
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=None, int charge=0, double charge_carrier=PROTON):             # <<<<<<<<<<<<<<
 *     '''
 *     Compute a peak list representing the theoretical isotopic cluster for `composition`.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution._isotopic_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_peaklist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_5_isotopic_variants(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_4_isotopic_variants[] = "_isotopic_variants(composition, npeaks=None, int charge=0, double charge_carrier=PROTON) -> list\n\n    Compute a peak list representing the theoretical isotopic cluster for `composition`.\n\n    Parameters\n    ----------\n    composition : Mapping\n        Any Mapping type where keys are element symbols and values are integers\n    n_peaks: int\n        The number of peaks to include in the isotopic cluster, starting from the monoisotopic peak.\n        If given a number below 1 or above the maximum number of isotopic variants, the maximum will\n        be used. If `None`, a \"reasonable\" value is chosen by `int(sqrt(max_variants(composition)))`.\n    charge: int\n        The charge state of the isotopic cluster to produce. Defaults to 0, theoretical neutral mass.\n    charge_carrier: double\n        The mass of the molecule contributing the ion's charge\n\n    Returns\n    -------\n    list of TheoreticalPeak\n    ";
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_5_isotopic_variants(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_npeaks = 0;
  int __pyx_v_charge;
  double __pyx_v_charge_carrier;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_isotopic_variants (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_npeaks,&__pyx_n_s_charge,&__pyx_n_s_charge_carrier,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_npeaks);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge_carrier);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_isotopic_variants") < 0)) __PYX_ERR(0, 692, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = values[0];
    __pyx_v_npeaks = values[1];
    if (values[2]) {
      __pyx_v_charge = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_charge == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 692, __pyx_L3_error)
    } else {
      __pyx_v_charge = ((int)0);
    }
    if (values[3]) {
      __pyx_v_charge_carrier = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_charge_carrier == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 692, __pyx_L3_error)
    } else {
      __pyx_v_charge_carrier = __pyx_k__2;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_isotopic_variants", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 692, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution._isotopic_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_4_isotopic_variants(__pyx_self, __pyx_v_composition, __pyx_v_npeaks, __pyx_v_charge, __pyx_v_charge_carrier);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_4_isotopic_variants(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, PyObject *__pyx_v_npeaks, int __pyx_v_charge, double __pyx_v_charge_carrier) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants __pyx_t_2;
  __Pyx_RefNannySetupContext("_isotopic_variants", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.npeaks = __pyx_v_npeaks;
  __pyx_t_2.charge = __pyx_v_charge;
  __pyx_t_2.charge_carrier = __pyx_v_charge_carrier;
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants(__pyx_v_composition, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution._isotopic_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":741
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=0, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopicDistribution* dist
 */

static struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_variants(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, int __pyx_v_npeaks, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants *__pyx_optional_args) {
  int __pyx_v_charge = ((int)0);
  double __pyx_v_charge_carrier = __pyx_k__3;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_dist;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_r;
  int __pyx_t_1;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge = __pyx_optional_args->charge;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
      }
    }
  }

  /* "brainpy/_c/isotopic_distribution.pyx":747
 *         int max_n_variants
 * 
 *     validate_composition(composition)             # <<<<<<<<<<<<<<
 * 
 *     if npeaks == 0:
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_validate_composition(__pyx_v_composition);

  /* "brainpy/_c/isotopic_distribution.pyx":749
 *     validate_composition(composition)
 * 
 *     if npeaks == 0:             # <<<<<<<<<<<<<<
 *         npeaks = guess_npeaks(composition, 300)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_npeaks == 0) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/isotopic_distribution.pyx":750
 * 
 *     if npeaks == 0:
 *         npeaks = guess_npeaks(composition, 300)             # <<<<<<<<<<<<<<
 *     else:
 *         npeaks = npeaks - 1
 */
    __pyx_v_npeaks = __pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks(__pyx_v_composition, 0x12C, NULL);

    /* "brainpy/_c/isotopic_distribution.pyx":749
 *     validate_composition(composition)
 * 
 *     if npeaks == 0:             # <<<<<<<<<<<<<<
 *         npeaks = guess_npeaks(composition, 300)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":752
 *         npeaks = guess_npeaks(composition, 300)
 *     else:
 *         npeaks = npeaks - 1             # <<<<<<<<<<<<<<
 * 
 *     dist = make_isotopic_distribution(composition, npeaks)
 */
  /*else*/ {
    __pyx_v_npeaks = (__pyx_v_npeaks - 1);
  }
  __pyx_L3:;

  /* "brainpy/_c/isotopic_distribution.pyx":754
 *         npeaks = npeaks - 1
 * 
 *     dist = make_isotopic_distribution(composition, npeaks)             # <<<<<<<<<<<<<<
 *     peaklist = id_aggregated_isotopic_variants(dist, charge, charge_carrier)
 *     free_isotopic_distribution(dist)
 */
  __pyx_v_dist = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(__pyx_v_composition, __pyx_v_npeaks, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":755
 * 
 *     dist = make_isotopic_distribution(composition, npeaks)
 *     peaklist = id_aggregated_isotopic_variants(dist, charge, charge_carrier)             # <<<<<<<<<<<<<<
 *     free_isotopic_distribution(dist)
 * 
 */
  __pyx_v_peaklist = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_aggregated_isotopic_variants(__pyx_v_dist, __pyx_v_charge, __pyx_v_charge_carrier);

  /* "brainpy/_c/isotopic_distribution.pyx":756
 *     dist = make_isotopic_distribution(composition, npeaks)
 *     peaklist = id_aggregated_isotopic_variants(dist, charge, charge_carrier)
 *     free_isotopic_distribution(dist)             # <<<<<<<<<<<<<<
 * 
 *     return peaklist
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(__pyx_v_dist);

  /* "brainpy/_c/isotopic_distribution.pyx":758
 *     free_isotopic_distribution(dist)
 * 
 *     return peaklist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_peaklist;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":741
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=0, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopicDistribution* dist
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":761
 * 
 * 
 * cdef list peaklist_to_list(PeakList* peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list pypeaklist
 */

static PyObject *__pyx_f_7brainpy_2_c_21isotopic_distribution_peaklist_to_list(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peaklist) {
  PyObject *__pyx_v_pypeaklist = 0;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("peaklist_to_list", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":765
 *         list pypeaklist
 *         size_t i
 *     pypeaklist = []             # <<<<<<<<<<<<<<
 *     for i in range(peaklist.used):
 *         pypeaklist.append(TheoreticalPeak._create(
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pypeaklist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":766
 *         size_t i
 *     pypeaklist = []
 *     for i in range(peaklist.used):             # <<<<<<<<<<<<<<
 *         pypeaklist.append(TheoreticalPeak._create(
 *             peaklist.peaks[i].mz, peaklist.peaks[i].intensity, peaklist.peaks[i].charge))
 */
  __pyx_t_2 = __pyx_v_peaklist->used;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "brainpy/_c/isotopic_distribution.pyx":767
 *     pypeaklist = []
 *     for i in range(peaklist.used):
 *         pypeaklist.append(TheoreticalPeak._create(             # <<<<<<<<<<<<<<
 *             peaklist.peaks[i].mz, peaklist.peaks[i].intensity, peaklist.peaks[i].charge))
 *     return pypeaklist
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__create((__pyx_v_peaklist->peaks[__pyx_v_i]).mz, (__pyx_v_peaklist->peaks[__pyx_v_i]).intensity, (__pyx_v_peaklist->peaks[__pyx_v_i]).charge)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_pypeaklist, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":769
 *         pypeaklist.append(TheoreticalPeak._create(
 *             peaklist.peaks[i].mz, peaklist.peaks[i].intensity, peaklist.peaks[i].charge))
 *     return pypeaklist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pypeaklist);
  __pyx_r = __pyx_v_pypeaklist;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":761
 * 
 * 
 * cdef list peaklist_to_list(PeakList* peaklist):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list pypeaklist
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.peaklist_to_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pypeaklist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":774
 * @cython.freelist(1000000)
 * cdef class TheoreticalPeak(object):
 *     def __init__(self, mz, intensity, charge):             # <<<<<<<<<<<<<<
 *         self.mz = mz
 *         self.intensity = intensity
 */

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mz = 0;
  PyObject *__pyx_v_intensity = 0;
  PyObject *__pyx_v_charge = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mz,&__pyx_n_s_intensity,&__pyx_n_s_charge,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intensity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 774, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 774, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 774, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_mz = values[0];
    __pyx_v_intensity = values[1];
    __pyx_v_charge = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 774, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak___init__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), __pyx_v_mz, __pyx_v_intensity, __pyx_v_charge);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak___init__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_mz, PyObject *__pyx_v_intensity, PyObject *__pyx_v_charge) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":775
 * cdef class TheoreticalPeak(object):
 *     def __init__(self, mz, intensity, charge):
 *         self.mz = mz             # <<<<<<<<<<<<<<
 *         self.intensity = intensity
 *         self.charge = charge
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_mz); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 775, __pyx_L1_error)
  __pyx_v_self->mz = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":776
 *     def __init__(self, mz, intensity, charge):
 *         self.mz = mz
 *         self.intensity = intensity             # <<<<<<<<<<<<<<
 *         self.charge = charge
 * 
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_intensity); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_v_self->intensity = __pyx_t_1;

  /* "brainpy/_c/isotopic_distribution.pyx":777
 *         self.mz = mz
 *         self.intensity = intensity
 *         self.charge = charge             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_charge); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L1_error)
  __pyx_v_self->charge = __pyx_t_2;

  /* "brainpy/_c/isotopic_distribution.pyx":774
 * @cython.freelist(1000000)
 * cdef class TheoreticalPeak(object):
 *     def __init__(self, mz, intensity, charge):             # <<<<<<<<<<<<<<
 *         self.mz = mz
 *         self.intensity = intensity
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":779
 *         self.charge = charge
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2__repr__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2__repr__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":780
 * 
 *     def __repr__(self):
 *         return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint _eq(self, TheoreticalPeak other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Peak_mz_f_intensity_f_charge_d, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":779
 *         self.charge = charge
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":782
 *         return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)
 * 
 *     cpdef bint _eq(self, TheoreticalPeak other):             # <<<<<<<<<<<<<<
 *         cdef bint val
 *         val = (abs(self.mz - other.mz) < 1e-10 and\
 */

static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__eq(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_eq", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_5_eq)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":784
 *     cpdef bint _eq(self, TheoreticalPeak other):
 *         cdef bint val
 *         val = (abs(self.mz - other.mz) < 1e-10 and\             # <<<<<<<<<<<<<<
 *                abs(self.intensity - other.intensity) < 1e-10 and\
 *                self.charge == other.charge)
 */
  __pyx_t_7 = ((fabs((__pyx_v_self->mz - __pyx_v_other->mz)) < 1e-10) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":785
 *         cdef bint val
 *         val = (abs(self.mz - other.mz) < 1e-10 and\
 *                abs(self.intensity - other.intensity) < 1e-10 and\             # <<<<<<<<<<<<<<
 *                self.charge == other.charge)
 *         return val
 */
  __pyx_t_7 = ((fabs((__pyx_v_self->intensity - __pyx_v_other->intensity)) < 1e-10) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":786
 *         val = (abs(self.mz - other.mz) < 1e-10 and\
 *                abs(self.intensity - other.intensity) < 1e-10 and\
 *                self.charge == other.charge)             # <<<<<<<<<<<<<<
 *         return val
 * 
 */
  __pyx_t_7 = ((__pyx_v_self->charge == __pyx_v_other->charge) != 0);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_val = __pyx_t_6;

  /* "brainpy/_c/isotopic_distribution.pyx":787
 *                abs(self.intensity - other.intensity) < 1e-10 and\
 *                self.charge == other.charge)
 *         return val             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":782
 *         return "Peak(mz=%f, intensity=%f, charge=%d)" % (self.mz, self.intensity, self.charge)
 * 
 *     cpdef bint _eq(self, TheoreticalPeak other):             # <<<<<<<<<<<<<<
 *         cdef bint val
 *         val = (abs(self.mz - other.mz) < 1e-10 and\
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("brainpy._c.isotopic_distribution.TheoreticalPeak._eq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_4_eq[] = "TheoreticalPeak._eq(self, TheoreticalPeak other) -> bool";
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_5_eq(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_eq (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak, 1, "other", 0))) __PYX_ERR(0, 782, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_4_eq(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_4_eq(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_eq", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__eq(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak._eq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":789
 *         return val
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.mz, self.intensity, self.charge))
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_7__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_7__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6__hash__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6__hash__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_hash_t __pyx_t_5;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":790
 * 
 *     def __hash__(self):
 *         return hash((self.mz, self.intensity, self.charge))             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, int code):
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = PyObject_Hash(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_hash_t)-1))) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":789
 *         return val
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.mz, self.intensity, self.charge))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":792
 *         return hash((self.mz, self.intensity, self.charge))
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_8__richcmp__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_8__richcmp__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":793
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */
  switch (__pyx_v_code) {
    case 2:

    /* "brainpy/_c/isotopic_distribution.pyx":794
 *     def __richcmp__(self, other, int code):
 *         if code == 2:
 *             return self._eq(other)             # <<<<<<<<<<<<<<
 *         elif code == 3:
 *             return not self._eq(other)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak))))) __PYX_ERR(0, 794, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":793
 * 
 *     def __richcmp__(self, other, int code):
 *         if code == 2:             # <<<<<<<<<<<<<<
 *             return self._eq(other)
 *         elif code == 3:
 */
    break;

    /* "brainpy/_c/isotopic_distribution.pyx":795
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not self._eq(other)
 * 
 */
    case 3:

    /* "brainpy/_c/isotopic_distribution.pyx":796
 *             return self._eq(other)
 *         elif code == 3:
 *             return not self._eq(other)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak))))) __PYX_ERR(0, 796, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self->__pyx_vtab)->_eq(__pyx_v_self, ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "brainpy/_c/isotopic_distribution.pyx":795
 *         if code == 2:
 *             return self._eq(other)
 *         elif code == 3:             # <<<<<<<<<<<<<<
 *             return not self._eq(other)
 * 
 */
    break;
    default: break;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":792
 *         return hash((self.mz, self.intensity, self.charge))
 * 
 *     def __richcmp__(self, other, int code):             # <<<<<<<<<<<<<<
 *         if code == 2:
 *             return self._eq(other)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":798
 *             return not self._eq(other)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return TheoreticalPeak, (self.mz, self.intensity, self.charge)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_10__reduce__[] = "TheoreticalPeak.__reduce__(self)";
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_10__reduce__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_10__reduce__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":799
 * 
 *     def __reduce__(self):
 *         return TheoreticalPeak, (self.mz, self.intensity, self.charge)             # <<<<<<<<<<<<<<
 * 
 *     cpdef TheoreticalPeak clone(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak));
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":798
 *             return not self._eq(other)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return TheoreticalPeak, (self.mz, self.intensity, self.charge)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":801
 *         return TheoreticalPeak, (self.mz, self.intensity, self.charge)
 * 
 *     cpdef TheoreticalPeak clone(self):             # <<<<<<<<<<<<<<
 *         return TheoreticalPeak._create(self.mz, self.intensity, self.charge)
 * 
 */

static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_13clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_clone(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_13clone)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak))))) __PYX_ERR(0, 801, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/isotopic_distribution.pyx":802
 * 
 *     cpdef TheoreticalPeak clone(self):
 *         return TheoreticalPeak._create(self.mz, self.intensity, self.charge)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__create(__pyx_v_self->mz, __pyx_v_self->intensity, __pyx_v_self->charge)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":801
 *         return TheoreticalPeak, (self.mz, self.intensity, self.charge)
 * 
 *     cpdef TheoreticalPeak clone(self):             # <<<<<<<<<<<<<<
 *         return TheoreticalPeak._create(self.mz, self.intensity, self.charge)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_13clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_12clone[] = "TheoreticalPeak.clone(self) -> TheoreticalPeak";
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_13clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_12clone(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_12clone(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":805
 * 
 *     @staticmethod
 *     cdef TheoreticalPeak _create(double mz, double intensity, int charge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             TheoreticalPeak inst
 */

static struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__create(double __pyx_v_mz, double __pyx_v_intensity, int __pyx_v_charge) {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_inst = 0;
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":808
 *         cdef:
 *             TheoreticalPeak inst
 *         inst = TheoreticalPeak.__new__(TheoreticalPeak)             # <<<<<<<<<<<<<<
 *         inst.mz = mz
 *         inst.intensity = intensity
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak(((PyTypeObject *)__pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":809
 *             TheoreticalPeak inst
 *         inst = TheoreticalPeak.__new__(TheoreticalPeak)
 *         inst.mz = mz             # <<<<<<<<<<<<<<
 *         inst.intensity = intensity
 *         inst.charge = charge
 */
  __pyx_v_inst->mz = __pyx_v_mz;

  /* "brainpy/_c/isotopic_distribution.pyx":810
 *         inst = TheoreticalPeak.__new__(TheoreticalPeak)
 *         inst.mz = mz
 *         inst.intensity = intensity             # <<<<<<<<<<<<<<
 *         inst.charge = charge
 *         return inst
 */
  __pyx_v_inst->intensity = __pyx_v_intensity;

  /* "brainpy/_c/isotopic_distribution.pyx":811
 *         inst.mz = mz
 *         inst.intensity = intensity
 *         inst.charge = charge             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->charge = __pyx_v_charge;

  /* "brainpy/_c/isotopic_distribution.pyx":812
 *         inst.intensity = intensity
 *         inst.charge = charge
 *         return inst             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":805
 * 
 *     @staticmethod
 *     cdef TheoreticalPeak _create(double mz, double intensity, int charge):             # <<<<<<<<<<<<<<
 *         cdef:
 *             TheoreticalPeak inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pxd":65
 * cdef class TheoreticalPeak(object):
 *     cdef:
 *         public double mz             # <<<<<<<<<<<<<<
 *         public double intensity
 *         public int charge
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz___get__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mz); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.mz.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_2__set__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_v_self->mz = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.mz.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pxd":66
 *     cdef:
 *         public double mz
 *         public double intensity             # <<<<<<<<<<<<<<
 *         public int charge
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity___get__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->intensity); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.intensity.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_2__set__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 66, __pyx_L1_error)
  __pyx_v_self->intensity = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.intensity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pxd":67
 *         public double mz
 *         public double intensity
 *         public int charge             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge___get__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge___get__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.charge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_2__set__(((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_2__set__(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 67, __pyx_L1_error)
  __pyx_v_self->charge = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.TheoreticalPeak.charge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":815
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict comp_dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_7main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_6main[] = "main()";
static PyMethodDef __pyx_mdef_7brainpy_2_c_21isotopic_distribution_7main = {"main", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_7main, METH_NOARGS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_6main};
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_7main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("main (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_6main(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_6main(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_comp_dict = 0;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_distribution2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("main", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":822
 *         IsotopicDistribution* distribution2
 * 
 *     comp_dict = dict(H=2, O=1)             # <<<<<<<<<<<<<<
 *     print comp_dict
 *     composition = dict_to_composition(comp_dict)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_H, __pyx_int_2) < 0) __PYX_ERR(0, 822, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_O, __pyx_int_1) < 0) __PYX_ERR(0, 822, __pyx_L1_error)
  __pyx_v_comp_dict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":823
 * 
 *     comp_dict = dict(H=2, O=1)
 *     print comp_dict             # <<<<<<<<<<<<<<
 *     composition = dict_to_composition(comp_dict)
 *     print_composition(composition)
 */
  if (__Pyx_PrintOne(0, __pyx_v_comp_dict) < 0) __PYX_ERR(0, 823, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":824
 *     comp_dict = dict(H=2, O=1)
 *     print comp_dict
 *     composition = dict_to_composition(comp_dict)             # <<<<<<<<<<<<<<
 *     print_composition(composition)
 *     distribution = make_isotopic_distribution(composition, 4)
 */
  __pyx_v_composition = __pyx_f_7brainpy_2_c_11composition_dict_to_composition(__pyx_v_comp_dict);

  /* "brainpy/_c/isotopic_distribution.pyx":825
 *     print comp_dict
 *     composition = dict_to_composition(comp_dict)
 *     print_composition(composition)             # <<<<<<<<<<<<<<
 *     distribution = make_isotopic_distribution(composition, 4)
 *     print "Going to print constants"
 */
  __pyx_f_7brainpy_2_c_11composition_print_composition(__pyx_v_composition);

  /* "brainpy/_c/isotopic_distribution.pyx":826
 *     composition = dict_to_composition(comp_dict)
 *     print_composition(composition)
 *     distribution = make_isotopic_distribution(composition, 4)             # <<<<<<<<<<<<<<
 *     print "Going to print constants"
 *     print_isotopic_constants(distribution._isotopic_constants)
 */
  __pyx_v_distribution = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(__pyx_v_composition, 4, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":827
 *     print_composition(composition)
 *     distribution = make_isotopic_distribution(composition, 4)
 *     print "Going to print constants"             # <<<<<<<<<<<<<<
 *     print_isotopic_constants(distribution._isotopic_constants)
 *     print "Done"
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Going_to_print_constants) < 0) __PYX_ERR(0, 827, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":828
 *     distribution = make_isotopic_distribution(composition, 4)
 *     print "Going to print constants"
 *     print_isotopic_constants(distribution._isotopic_constants)             # <<<<<<<<<<<<<<
 *     print "Done"
 * 
 */
  __pyx_f_7brainpy_2_c_18isotopic_constants_print_isotopic_constants(__pyx_v_distribution->_isotopic_constants);

  /* "brainpy/_c/isotopic_distribution.pyx":829
 *     print "Going to print constants"
 *     print_isotopic_constants(distribution._isotopic_constants)
 *     print "Done"             # <<<<<<<<<<<<<<
 * 
 *     print "Trying to free"
 */
  if (__Pyx_PrintOne(0, __pyx_n_s_Done) < 0) __PYX_ERR(0, 829, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":831
 *     print "Done"
 * 
 *     print "Trying to free"             # <<<<<<<<<<<<<<
 *     free_isotopic_distribution(distribution)
 *     print "Free Done"
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Trying_to_free) < 0) __PYX_ERR(0, 831, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":832
 * 
 *     print "Trying to free"
 *     free_isotopic_distribution(distribution)             # <<<<<<<<<<<<<<
 *     print "Free Done"
 * 
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(__pyx_v_distribution);

  /* "brainpy/_c/isotopic_distribution.pyx":833
 *     print "Trying to free"
 *     free_isotopic_distribution(distribution)
 *     print "Free Done"             # <<<<<<<<<<<<<<
 * 
 *     distribution2 = make_isotopic_distribution(composition, 4)
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Free_Done) < 0) __PYX_ERR(0, 833, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":835
 *     print "Free Done"
 * 
 *     distribution2 = make_isotopic_distribution(composition, 4)             # <<<<<<<<<<<<<<
 *     print "Seconc construction"
 *     print_isotopic_constants(distribution2._isotopic_constants)
 */
  __pyx_v_distribution2 = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(__pyx_v_composition, 4, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":836
 * 
 *     distribution2 = make_isotopic_distribution(composition, 4)
 *     print "Seconc construction"             # <<<<<<<<<<<<<<
 *     print_isotopic_constants(distribution2._isotopic_constants)
 *     print "Second Free"
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Seconc_construction) < 0) __PYX_ERR(0, 836, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":837
 *     distribution2 = make_isotopic_distribution(composition, 4)
 *     print "Seconc construction"
 *     print_isotopic_constants(distribution2._isotopic_constants)             # <<<<<<<<<<<<<<
 *     print "Second Free"
 *     free_isotopic_distribution(distribution2)
 */
  __pyx_f_7brainpy_2_c_18isotopic_constants_print_isotopic_constants(__pyx_v_distribution2->_isotopic_constants);

  /* "brainpy/_c/isotopic_distribution.pyx":838
 *     print "Seconc construction"
 *     print_isotopic_constants(distribution2._isotopic_constants)
 *     print "Second Free"             # <<<<<<<<<<<<<<
 *     free_isotopic_distribution(distribution2)
 *     print comp_dict
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Second_Free) < 0) __PYX_ERR(0, 838, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":839
 *     print_isotopic_constants(distribution2._isotopic_constants)
 *     print "Second Free"
 *     free_isotopic_distribution(distribution2)             # <<<<<<<<<<<<<<
 *     print comp_dict
 *     free_composition(composition)
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(__pyx_v_distribution2);

  /* "brainpy/_c/isotopic_distribution.pyx":840
 *     print "Second Free"
 *     free_isotopic_distribution(distribution2)
 *     print comp_dict             # <<<<<<<<<<<<<<
 *     free_composition(composition)
 *     print "Really done"
 */
  if (__Pyx_PrintOne(0, __pyx_v_comp_dict) < 0) __PYX_ERR(0, 840, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":841
 *     free_isotopic_distribution(distribution2)
 *     print comp_dict
 *     free_composition(composition)             # <<<<<<<<<<<<<<
 *     print "Really done"
 * 
 */
  __pyx_f_7brainpy_2_c_11composition_free_composition(__pyx_v_composition);

  /* "brainpy/_c/isotopic_distribution.pyx":842
 *     print comp_dict
 *     free_composition(composition)
 *     print "Really done"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PrintOne(0, __pyx_kp_s_Really_done) < 0) __PYX_ERR(0, 842, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":815
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict comp_dict
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.main", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_comp_dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/isotopic_distribution.pyx":845
 * 
 * 
 * def test(object composition, int max_npeaks=300):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* composition_struct
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_9test(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7brainpy_2_c_21isotopic_distribution_8test[] = "test(composition, int max_npeaks=300)";
static PyMethodDef __pyx_mdef_7brainpy_2_c_21isotopic_distribution_9test = {"test", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_9test, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_8test};
static PyObject *__pyx_pw_7brainpy_2_c_21isotopic_distribution_9test(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_composition = 0;
  int __pyx_v_max_npeaks;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_composition,&__pyx_n_s_max_npeaks,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_npeaks);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "test") < 0)) __PYX_ERR(0, 845, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_composition = values[0];
    if (values[1]) {
      __pyx_v_max_npeaks = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_max_npeaks == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 845, __pyx_L3_error)
    } else {
      __pyx_v_max_npeaks = ((int)0x12C);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("test", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 845, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.test", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_21isotopic_distribution_8test(__pyx_self, __pyx_v_composition, __pyx_v_max_npeaks);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_21isotopic_distribution_8test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_composition, int __pyx_v_max_npeaks) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_struct;
  int __pyx_v_npeaks;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *__pyx_v_dist;
  struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *__pyx_v_peak_set;
  PyObject *__pyx_v_peaklist = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  __Pyx_RefNannySetupContext("test", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":851
 *         int npeaks, max_n_variants
 * 
 *     composition_struct = dict_to_composition(dict(composition))             # <<<<<<<<<<<<<<
 *     validate_composition(composition_struct)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_composition_struct = __pyx_f_7brainpy_2_c_11composition_dict_to_composition(((PyObject*)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":852
 * 
 *     composition_struct = dict_to_composition(dict(composition))
 *     validate_composition(composition_struct)             # <<<<<<<<<<<<<<
 * 
 *     npeaks = guess_npeaks(composition_struct, max_npeaks)
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_validate_composition(__pyx_v_composition_struct);

  /* "brainpy/_c/isotopic_distribution.pyx":854
 *     validate_composition(composition_struct)
 * 
 *     npeaks = guess_npeaks(composition_struct, max_npeaks)             # <<<<<<<<<<<<<<
 *     print("Guessed # of Peaks: ", npeaks)
 * 
 */
  __pyx_v_npeaks = __pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks(__pyx_v_composition_struct, __pyx_v_max_npeaks, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":855
 * 
 *     npeaks = guess_npeaks(composition_struct, max_npeaks)
 *     print("Guessed # of Peaks: ", npeaks)             # <<<<<<<<<<<<<<
 * 
 *     dist = make_isotopic_distribution(composition_struct, npeaks)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_npeaks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Guessed_of_Peaks);
  __Pyx_GIVEREF(__pyx_kp_s_Guessed_of_Peaks);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_Guessed_of_Peaks);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_2) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":857
 *     print("Guessed # of Peaks: ", npeaks)
 * 
 *     dist = make_isotopic_distribution(composition_struct, npeaks)             # <<<<<<<<<<<<<<
 * 
 *     print("Size of probability vector:", id_probability_vector(dist).used)
 */
  __pyx_v_dist = __pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution(__pyx_v_composition_struct, __pyx_v_npeaks, NULL);

  /* "brainpy/_c/isotopic_distribution.pyx":859
 *     dist = make_isotopic_distribution(composition_struct, npeaks)
 * 
 *     print("Size of probability vector:", id_probability_vector(dist).used)             # <<<<<<<<<<<<<<
 * 
 *     peak_set = id_aggregated_isotopic_variants(dist, 1, PROTON)
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_f_7brainpy_2_c_21isotopic_distribution_id_probability_vector(__pyx_v_dist)->used); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_s_Size_of_probability_vector);
  __Pyx_GIVEREF(__pyx_kp_s_Size_of_probability_vector);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_Size_of_probability_vector);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":861
 *     print("Size of probability vector:", id_probability_vector(dist).used)
 * 
 *     peak_set = id_aggregated_isotopic_variants(dist, 1, PROTON)             # <<<<<<<<<<<<<<
 * 
 *     peaklist = peaklist_to_list(peak_set)
 */
  __pyx_v_peak_set = __pyx_f_7brainpy_2_c_21isotopic_distribution_id_aggregated_isotopic_variants(__pyx_v_dist, 1, __pyx_v_7brainpy_2_c_11composition_PROTON);

  /* "brainpy/_c/isotopic_distribution.pyx":863
 *     peak_set = id_aggregated_isotopic_variants(dist, 1, PROTON)
 * 
 *     peaklist = peaklist_to_list(peak_set)             # <<<<<<<<<<<<<<
 * 
 *     print("Actual # of Peaks Returned:", len(peaklist))
 */
  __pyx_t_1 = __pyx_f_7brainpy_2_c_21isotopic_distribution_peaklist_to_list(__pyx_v_peak_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_peaklist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":865
 *     peaklist = peaklist_to_list(peak_set)
 * 
 *     print("Actual # of Peaks Returned:", len(peaklist))             # <<<<<<<<<<<<<<
 * 
 *     free_peak_list(peak_set)
 */
  if (unlikely(__pyx_v_peaklist == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 865, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_v_peaklist); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 865, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_s_Actual_of_Peaks_Returned);
  __Pyx_GIVEREF(__pyx_kp_s_Actual_of_Peaks_Returned);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_Actual_of_Peaks_Returned);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_2) < 0) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":867
 *     print("Actual # of Peaks Returned:", len(peaklist))
 * 
 *     free_peak_list(peak_set)             # <<<<<<<<<<<<<<
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_peak_list(__pyx_v_peak_set);

  /* "brainpy/_c/isotopic_distribution.pyx":868
 * 
 *     free_peak_list(peak_set)
 *     free_isotopic_distribution(dist)             # <<<<<<<<<<<<<<
 *     free_composition(composition_struct)
 *     return peaklist
 */
  __pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution(__pyx_v_dist);

  /* "brainpy/_c/isotopic_distribution.pyx":869
 *     free_peak_list(peak_set)
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)             # <<<<<<<<<<<<<<
 *     return peaklist
 */
  __pyx_f_7brainpy_2_c_11composition_free_composition(__pyx_v_composition_struct);

  /* "brainpy/_c/isotopic_distribution.pyx":870
 *     free_isotopic_distribution(dist)
 *     free_composition(composition_struct)
 *     return peaklist             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_peaklist);
  __pyx_r = __pyx_v_peaklist;
  goto __pyx_L0;

  /* "brainpy/_c/isotopic_distribution.pyx":845
 * 
 * 
 * def test(object composition, int max_npeaks=300):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* composition_struct
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("brainpy._c.isotopic_distribution.test", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_peaklist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak __pyx_vtable_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;

static struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *__pyx_freelist_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak[1000000];
static int __pyx_freecount_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = 0;

static PyObject *__pyx_tp_new_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak[--__pyx_freecount_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak];
    memset(o, 0, sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak));
    (void) PyObject_INIT(o, t);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)o);
  p->__pyx_vtab = __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
  return o;
}

static void __pyx_tp_dealloc_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak < 1000000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak[__pyx_freecount_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak++] = ((struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyObject *__pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_mz(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_1__get__(o);
}

static int __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_mz(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_2mz_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_intensity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_1__get__(o);
}

static int __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_intensity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9intensity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_charge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_1__get__(o);
}

static int __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_charge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_6charge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak[] = {
  {"_eq", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_5_eq, METH_O, __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_4_eq},
  {"__reduce__", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_11__reduce__, METH_NOARGS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_10__reduce__},
  {"clone", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_13clone, METH_NOARGS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_12clone},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak[] = {
  {(char *)"mz", __pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_mz, __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_mz, (char *)"mz: 'double'", 0},
  {(char *)"intensity", __pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_intensity, __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_intensity, (char *)"intensity: 'double'", 0},
  {(char *)"charge", __pyx_getprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_charge, __pyx_setprop_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_charge, (char *)"charge: 'int'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = {
  PyVarObject_HEAD_INIT(0, 0)
  "brainpy._c.isotopic_distribution.TheoreticalPeak", /*tp_name*/
  sizeof(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_7__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "TheoreticalPeak(mz, intensity, charge)", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"check_composition_non_negative", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_1check_composition_non_negative, METH_O, __pyx_doc_7brainpy_2_c_21isotopic_distribution_check_composition_non_negative},
  {"_isotopic_variants", (PyCFunction)__pyx_pw_7brainpy_2_c_21isotopic_distribution_5_isotopic_variants, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7brainpy_2_c_21isotopic_distribution_4_isotopic_variants},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_isotopic_distribution(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_isotopic_distribution},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "isotopic_distribution",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Actual_of_Peaks_Returned, __pyx_k_Actual_of_Peaks_Returned, sizeof(__pyx_k_Actual_of_Peaks_Returned), 0, 0, 1, 0},
  {&__pyx_n_s_Done, __pyx_k_Done, sizeof(__pyx_k_Done), 0, 0, 1, 1},
  {&__pyx_kp_s_Free_Done, __pyx_k_Free_Done, sizeof(__pyx_k_Free_Done), 0, 0, 1, 0},
  {&__pyx_kp_s_Going_to_print_constants, __pyx_k_Going_to_print_constants, sizeof(__pyx_k_Going_to_print_constants), 0, 0, 1, 0},
  {&__pyx_kp_s_Guessed_of_Peaks, __pyx_k_Guessed_of_Peaks, sizeof(__pyx_k_Guessed_of_Peaks), 0, 0, 1, 0},
  {&__pyx_n_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 1},
  {&__pyx_n_s_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 1, 1},
  {&__pyx_kp_s_Peak_mz_f_intensity_f_charge_d, __pyx_k_Peak_mz_f_intensity_f_charge_d, sizeof(__pyx_k_Peak_mz_f_intensity_f_charge_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Really_done, __pyx_k_Really_done, sizeof(__pyx_k_Really_done), 0, 0, 1, 0},
  {&__pyx_kp_s_Seconc_construction, __pyx_k_Seconc_construction, sizeof(__pyx_k_Seconc_construction), 0, 0, 1, 0},
  {&__pyx_kp_s_Second_Free, __pyx_k_Second_Free, sizeof(__pyx_k_Second_Free), 0, 0, 1, 0},
  {&__pyx_kp_s_Size_of_probability_vector, __pyx_k_Size_of_probability_vector, sizeof(__pyx_k_Size_of_probability_vector), 0, 0, 1, 0},
  {&__pyx_kp_s_Trying_to_free, __pyx_k_Trying_to_free, sizeof(__pyx_k_Trying_to_free), 0, 0, 1, 0},
  {&__pyx_kp_s_brainpy__c_isotopic_distribution, __pyx_k_brainpy__c_isotopic_distribution, sizeof(__pyx_k_brainpy__c_isotopic_distribution), 0, 0, 1, 0},
  {&__pyx_n_s_brainpy__c_isotopic_distribution_2, __pyx_k_brainpy__c_isotopic_distribution_2, sizeof(__pyx_k_brainpy__c_isotopic_distribution_2), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_charge_carrier, __pyx_k_charge_carrier, sizeof(__pyx_k_charge_carrier), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_comp_dict, __pyx_k_comp_dict, sizeof(__pyx_k_comp_dict), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_composition_struct, __pyx_k_composition_struct, sizeof(__pyx_k_composition_struct), 0, 0, 1, 1},
  {&__pyx_n_s_dist, __pyx_k_dist, sizeof(__pyx_k_dist), 0, 0, 1, 1},
  {&__pyx_n_s_distribution, __pyx_k_distribution, sizeof(__pyx_k_distribution), 0, 0, 1, 1},
  {&__pyx_n_s_distribution2, __pyx_k_distribution2, sizeof(__pyx_k_distribution2), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_intensity, __pyx_k_intensity, sizeof(__pyx_k_intensity), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_main_2, __pyx_k_main_2, sizeof(__pyx_k_main_2), 0, 0, 1, 1},
  {&__pyx_n_s_max_n_variants, __pyx_k_max_n_variants, sizeof(__pyx_k_max_n_variants), 0, 0, 1, 1},
  {&__pyx_n_s_max_npeaks, __pyx_k_max_npeaks, sizeof(__pyx_k_max_npeaks), 0, 0, 1, 1},
  {&__pyx_n_s_mz, __pyx_k_mz, sizeof(__pyx_k_mz), 0, 0, 1, 1},
  {&__pyx_n_s_npeaks, __pyx_k_npeaks, sizeof(__pyx_k_npeaks), 0, 0, 1, 1},
  {&__pyx_n_s_peak_set, __pyx_k_peak_set, sizeof(__pyx_k_peak_set), 0, 0, 1, 1},
  {&__pyx_n_s_peaklist, __pyx_k_peaklist, sizeof(__pyx_k_peaklist), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyisotopic_variants, __pyx_k_pyisotopic_variants, sizeof(__pyx_k_pyisotopic_variants), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_test_2, __pyx_k_test_2, sizeof(__pyx_k_test_2), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 183, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "brainpy/_c/isotopic_distribution.pyx":667
 * 
 * 
 * def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,             # <<<<<<<<<<<<<<
 *                         double charge_carrier=PROTON):
 *     '''
 */
  __pyx_tuple__4 = PyTuple_Pack(4, __pyx_n_s_composition, __pyx_n_s_npeaks, __pyx_n_s_charge, __pyx_n_s_charge_carrier); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_isotopic_distribution, __pyx_n_s_pyisotopic_variants, 667, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 667, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":815
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict comp_dict
 */
  __pyx_tuple__6 = PyTuple_Pack(4, __pyx_n_s_comp_dict, __pyx_n_s_composition, __pyx_n_s_distribution, __pyx_n_s_distribution2); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_isotopic_distribution, __pyx_n_s_main_2, 815, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 815, __pyx_L1_error)

  /* "brainpy/_c/isotopic_distribution.pyx":845
 * 
 * 
 * def test(object composition, int max_npeaks=300):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* composition_struct
 */
  __pyx_tuple__8 = PyTuple_Pack(9, __pyx_n_s_composition, __pyx_n_s_max_npeaks, __pyx_n_s_composition_struct, __pyx_n_s_distribution, __pyx_n_s_npeaks, __pyx_n_s_max_n_variants, __pyx_n_s_dist, __pyx_n_s_peak_set, __pyx_n_s_peaklist); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(2, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_isotopic_distribution, __pyx_n_s_test, 845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("make_peak", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *(double, double, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("print_peak", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_print_peak, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_peak_list", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_make_peak_list, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_peak_list", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_free_peak_list, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("resize_peak_list", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_resize_peak_list, "int (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peak_list_append", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_append, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_Peak *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peak_list_reset", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_reset, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peak_list_ignore_below", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_ignore_below, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *, double, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peak_list_truncate_after", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_truncate_after, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *, double, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peak_list_shift", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peak_list_shift, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("peaklist_to_list", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_peaklist_to_list, "PyObject *(struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_element_cache", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_make_element_cache, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_element_cache", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_free_element_cache, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("resize_element_cache", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_resize_element_cache, "int (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_cache_put", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_put, "int (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *, struct __pyx_t_7brainpy_2_c_11composition_Element **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_cache_get", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_element_cache_get, "int (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_isotopic_distribution", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *(struct __pyx_t_7brainpy_2_c_11composition_Composition *, int, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_make_isotopic_distribution *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_isotopic_distribution", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_free_isotopic_distribution, "void (struct __pyx_t_7brainpy_2_c_21isotopic_distribution_IsotopicDistribution *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("max_variants", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_max_variants, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_21isotopic_distribution_ElementCache *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("guess_npeaks", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_guess_npeaks, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, size_t, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_guess_npeaks *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_isotopic_variants", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution__isotopic_variants, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution__isotopic_variants *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("isotopic_variants", (void (*)(void))__pyx_f_7brainpy_2_c_21isotopic_distribution_isotopic_variants, "struct __pyx_t_7brainpy_2_c_21isotopic_distribution_PeakList *(struct __pyx_t_7brainpy_2_c_11composition_Composition *, int, struct __pyx_opt_args_7brainpy_2_c_21isotopic_distribution_isotopic_variants *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = &__pyx_vtable_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
  __pyx_vtable_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak._create = (struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*)(double, double, int))__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__create;
  __pyx_vtable_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak._eq = (int (*)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak__eq;
  __pyx_vtable_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.clone = (struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *(*)(struct __pyx_obj_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_21isotopic_distribution_15TheoreticalPeak_clone;
  if (PyType_Ready(&__pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.tp_dictoffset && __pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak.tp_dict, __pyx_vtabptr_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "TheoreticalPeak", (PyObject *)&__pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_ptype_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak = &__pyx_type_7brainpy_2_c_21isotopic_distribution_TheoreticalPeak;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = __Pyx_ImportType("brainpy._c.composition", "PyComposition", sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), 1); if (unlikely(!__pyx_ptype_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(2, 117, __pyx_L1_error)
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = (struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition*)__Pyx_GetVtable(__pyx_ptype_7brainpy_2_c_11composition_PyComposition->tp_dict); if (unlikely(!__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition)) __PYX_ERR(2, 117, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "nist_mass", (void **)&__pyx_vp_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "PROTON", (void **)&__pyx_vp_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_ElementTable", (void **)&__pyx_vp_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "DEFAULT_ISOTOPIC_CONSTANTS_SIZE", (void **)&__pyx_vp_7brainpy_2_c_18isotopic_constants_DEFAULT_ISOTOPIC_CONSTANTS_SIZE, "size_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("brainpy._c.composition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "mass_charge_ratio", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_mass_charge_ratio, "double (double, int, struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "_parse_isotope_string", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition__parse_isotope_string, "char *(char *, int *, char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_isotope_by_neutron_shift", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift, "struct __pyx_t_7brainpy_2_c_11composition_Isotope *(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "element_monoisotopic_mass", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass, "double (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "element_max_neutron_shift", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_element_max_neutron_shift, "int (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "make_fixed_isotope_element", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element, "struct __pyx_t_7brainpy_2_c_11composition_Element *(struct __pyx_t_7brainpy_2_c_11composition_Element *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "make_element_hash_table", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_make_element_hash_table, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *(size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "element_hash_table_get", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_element_hash_table_get, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "element_hash_table_put", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_element_hash_table_put, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "free_element_hash_table", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_free_element_hash_table, "size_t (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "print_composition", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_print_composition, "void (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "composition_get_element_count", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_composition_get_element_count, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "composition_mass", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_composition_mass, "double (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "free_composition", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_free_composition, "void (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "dict_to_composition", (void (**)(void))&__pyx_f_7brainpy_2_c_11composition_dict_to_composition, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("brainpy._c.double_vector"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "make_double_vector_with_size", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_make_double_vector_with_size, "struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "make_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_make_double_vector, "struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "double_vector_append", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_double_vector_append, "int (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "free_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_free_double_vector, "void (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "print_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_print_double_vector, "void (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "reset_double_vector", (void (**)(void))&__pyx_f_7brainpy_2_c_13double_vector_reset_double_vector, "void (struct __pyx_t_7brainpy_2_c_13double_vector_DoubleVector *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_ImportModule("brainpy._c.isotopic_constants"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "newton", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_newton, "void (__pyx_t_7brainpy_2_c_18isotopic_constants_dvec *, __pyx_t_7brainpy_2_c_18isotopic_constants_dvec *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "make_isotopic_constants", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_make_isotopic_constants, "struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_resize", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_resize, "int (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "free_isotopic_constants", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_free_isotopic_constants, "void (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_add_element", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_add_element, "void (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_get", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_get, "int (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, struct __pyx_t_7brainpy_2_c_18isotopic_constants_PhiConstants **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_update_coefficients", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_update_coefficients, "void (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_nth_element_power_sum", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum, "double (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_nth_modified_element_power_sum", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_modified_element_power_sum, "double (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, char *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "isotopic_constants_nth_element_power_sum_by_index", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_isotopic_constants_nth_element_power_sum_by_index, "double (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *, size_t, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "print_isotopic_constants", (void (**)(void))&__pyx_f_7brainpy_2_c_18isotopic_constants_print_isotopic_constants, "void (struct __pyx_t_7brainpy_2_c_18isotopic_constants_IsotopicConstants *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initisotopic_distribution(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initisotopic_distribution(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_isotopic_distribution(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_isotopic_distribution(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_isotopic_distribution(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_isotopic_distribution(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("isotopic_distribution", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_brainpy___c__isotopic_distribution) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "brainpy._c.isotopic_distribution")) {
      if (unlikely(PyDict_SetItemString(modules, "brainpy._c.isotopic_distribution", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_variable_import_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "brainpy/_c/isotopic_distribution.pyx":668
 * 
 * def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,
 *                         double charge_carrier=PROTON):             # <<<<<<<<<<<<<<
 *     '''
 *     Compute a peak list representing the theoretical isotopic cluster for `composition`.
 */
  __pyx_k_ = __pyx_v_7brainpy_2_c_11composition_PROTON;

  /* "brainpy/_c/isotopic_distribution.pyx":667
 * 
 * 
 * def pyisotopic_variants(object composition not None, object npeaks=None, int charge=0,             # <<<<<<<<<<<<<<
 *                         double charge_carrier=PROTON):
 *     '''
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_21isotopic_distribution_3pyisotopic_variants, NULL, __pyx_n_s_brainpy__c_isotopic_distribution_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyisotopic_variants, __pyx_t_1) < 0) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":692
 * 
 * 
 * cpdef list _isotopic_variants(object composition, object npeaks=None, int charge=0, double charge_carrier=PROTON):             # <<<<<<<<<<<<<<
 *     '''
 *     Compute a peak list representing the theoretical isotopic cluster for `composition`.
 */
  __pyx_k__2 = __pyx_v_7brainpy_2_c_11composition_PROTON;
  __pyx_k__2 = __pyx_v_7brainpy_2_c_11composition_PROTON;

  /* "brainpy/_c/isotopic_distribution.pyx":741
 * 
 * 
 * cdef PeakList* isotopic_variants(Composition* composition, int npeaks, int charge=0, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopicDistribution* dist
 */
  __pyx_k__3 = __pyx_v_7brainpy_2_c_11composition_PROTON;

  /* "brainpy/_c/isotopic_distribution.pyx":815
 * 
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict comp_dict
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_21isotopic_distribution_7main, NULL, __pyx_n_s_brainpy__c_isotopic_distribution_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_main_2, __pyx_t_1) < 0) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":845
 * 
 * 
 * def test(object composition, int max_npeaks=300):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* composition_struct
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_21isotopic_distribution_9test, NULL, __pyx_n_s_brainpy__c_isotopic_distribution_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/isotopic_distribution.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_2, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init brainpy._c.isotopic_distribution", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init brainpy._c.isotopic_distribution");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
  static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
  #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
