/* Generated by Cython 0.28.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "brainpy\\_c\\compat.h"
        ],
        "include_dirs": [
            ".\\brainpy\\_c",
            "brainpy/_c/"
        ],
        "name": "brainpy._c.composition",
        "sources": [
            "brainpy/_c/composition.pyx"
        ]
    },
    "module_name": "brainpy._c.composition"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__brainpy___c__composition
#define __PYX_HAVE_API__brainpy___c__composition
/* Early includes */
#include "compat.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "brainpy\\_c\\composition.pyx",
  "brainpy\\_c\\composition.pxd",
  "stringsource",
  "type.pxd",
};

/* "brainpy/_c/composition.pxd":3
 * from brainpy.mass_dict import nist_mass as __nist_mass
 * 
 * ctypedef int count_type             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
typedef int __pyx_t_7brainpy_2_c_11composition_count_type;

/*--- Type declarations ---*/
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition;
struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__;
struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr;
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass;
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio;
struct __pyx_t_7brainpy_2_c_11composition_Isotope;
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap;
struct __pyx_t_7brainpy_2_c_11composition_Element;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket;
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable;
struct __pyx_t_7brainpy_2_c_11composition_Composition;
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl;

/* "brainpy/_c/composition.pxd":10
 * cdef double PROTON = nist_mass["H+"][0][0]
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":11
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=*) nogil
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=*) nogil             # <<<<<<<<<<<<<<
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil
 * 
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio {
  int __pyx_n;
  double charge_carrier;
};

/* "brainpy/_c/composition.pxd":17
 * # Isotope and IsotopeMap Declarations
 * 
 * cdef struct Isotope:             # <<<<<<<<<<<<<<
 *     double mass
 *     double abundance
 */
struct __pyx_t_7brainpy_2_c_11composition_Isotope {
  double mass;
  double abundance;
  int neutrons;
  int neutron_shift;
};

/* "brainpy/_c/composition.pxd":23
 *     int neutron_shift
 * 
 * cdef struct IsotopeMap:             # <<<<<<<<<<<<<<
 *     Isotope* bins
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap {
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *bins;
  size_t size;
};

/* "brainpy/_c/composition.pxd":35
 * # Element Declarations
 * 
 * cdef struct Element:             # <<<<<<<<<<<<<<
 *     char* symbol
 *     IsotopeMap* isotopes
 */
struct __pyx_t_7brainpy_2_c_11composition_Element {
  char *symbol;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *isotopes;
  int monoisotopic_isotope_index;
};

/* "brainpy/_c/composition.pxd":54
 * # ElementHashTable and ElementHashBucket Declarations
 * 
 * cdef struct ElementHashBucket:             # <<<<<<<<<<<<<<
 *     Element** elements
 *     size_t used
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket {
  struct __pyx_t_7brainpy_2_c_11composition_Element **elements;
  size_t used;
  size_t size;
};

/* "brainpy/_c/composition.pxd":63
 * 
 * 
 * cdef struct ElementHashTable:             # <<<<<<<<<<<<<<
 *     ElementHashBucket* buckets
 *     size_t size
 */
struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *buckets;
  size_t size;
};

/* "brainpy/_c/composition.pxd":92
 * # Composition Declarations
 * 
 * cdef struct Composition:             # <<<<<<<<<<<<<<
 *     char** elements
 *     count_type* counts
 */
struct __pyx_t_7brainpy_2_c_11composition_Composition {
  char **elements;
  __pyx_t_7brainpy_2_c_11composition_count_type *counts;
  size_t size;
  size_t used;
  int max_variants;
};

/* "brainpy/_c/composition.pxd":124
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double mass(self)
 */
struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl {
  int __pyx_n;
  int free_existing;
};

/* "brainpy/_c/composition.pyx":1122
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *       START
 *       ISOTOPE
 */
enum  {
  __pyx_e_7brainpy_2_c_11composition_START,
  __pyx_e_7brainpy_2_c_11composition_ISOTOPE,
  __pyx_e_7brainpy_2_c_11composition_ELEMENT,
  __pyx_e_7brainpy_2_c_11composition_COUNT
};

/* "brainpy/_c/composition.pxd":117
 * cdef Composition* dict_to_composition(dict comp_dict)
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* impl
 */
struct __pyx_obj_7brainpy_2_c_11composition_PyComposition {
  PyObject_HEAD
  struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtab;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *impl;
  double cached_mass;
  int _clean;
};


/* "brainpy/_c/composition.pyx":1075
 *         return self.impl.used
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 */
struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ {
  PyObject_HEAD
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self;
};


/* "brainpy/_c/composition.pyx":1076
 * 
 *     def __repr__(self):
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):
 */
struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *__pyx_outer_scope;
  PyObject *__pyx_v_kv;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "brainpy/_c/composition.pyx":833
 * 
 * 
 * cdef class PyComposition(object):             # <<<<<<<<<<<<<<
 *     '''A mapping representing a chemical composition.
 * 
 */

struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*_create)(struct __pyx_t_7brainpy_2_c_11composition_Composition *);
  void (*_set_impl)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args);
  double (*mass)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_pycomposition)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
  int (*__pyx___equality_dict)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*copy)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *__pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->exc_type, &(self)->exc_value, &(self)->exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(self);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->exc_type, (self)->exc_value, (self)->exc_traceback);\
    (self)->exc_type = (self)->exc_value = (self)->exc_traceback = NULL;\
    }
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__pyx_CoroutineObject *self);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_13PyComposition__create(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_base); /* proto*/
static void __pyx_f_7brainpy_2_c_11composition_13PyComposition__set_impl(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args); /* proto*/
static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_13PyComposition_copy(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_7brainpy_2_c_11composition_13PyComposition_mass(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_pycomposition(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_dict(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'brainpy._c.compat' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc' */

/* Module declarations from 'brainpy._c.composition' */
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition_PyComposition = 0;
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct____repr__ = 0;
static PyTypeObject *__pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr = 0;
static PyObject *__pyx_v_7brainpy_2_c_11composition_nist_mass = 0;
static double __pyx_v_7brainpy_2_c_11composition_PROTON;
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_7brainpy_2_c_11composition__ElementTable;
static PyObject *__pyx_v_7brainpy_2_c_11composition__string_table = 0;
static char *__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(char *, int *, char *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_f_7brainpy_2_c_11composition_make_isotope_map(PyObject *, size_t); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition_free_isotope_map(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition__isotopes_of(char *, struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap **); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_f_7brainpy_2_c_11composition_make_element(char *); /*proto*/
static double __pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass(struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(struct __pyx_t_7brainpy_2_c_11composition_Element *, int); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition_free_element_hash_bucket(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_f_7brainpy_2_c_11composition_make_element_hash_table(size_t); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_insert(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *, struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_find(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_element_hash_table_put(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, struct __pyx_t_7brainpy_2_c_11composition_Element *); /*proto*/
static size_t __pyx_f_7brainpy_2_c_11composition_hash_string(char *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_make_composition(void); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_copy_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition_print_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_set_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type *); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_inc_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_resize(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static double __pyx_f_7brainpy_2_c_11composition_composition_mass(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition_free_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_composition_add(struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, int); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_iadd(struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, int); /*proto*/
static void __pyx_f_7brainpy_2_c_11composition_composition_imul(struct __pyx_t_7brainpy_2_c_11composition_Composition *, int); /*proto*/
static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_parse_formula(PyObject *, int __pyx_skip_dispatch); /*proto*/
static char *__pyx_f_7brainpy_2_c_11composition__make_isotope_string(struct __pyx_t_7brainpy_2_c_11composition_Element *, struct __pyx_t_7brainpy_2_c_11composition_Isotope *, char *); /*proto*/
static char *__pyx_f_7brainpy_2_c_11composition__make_fixed_isotope_string(struct __pyx_t_7brainpy_2_c_11composition_Element *, struct __pyx_t_7brainpy_2_c_11composition_Isotope *, char *); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift_index(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_count(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int); /*proto*/
static double __pyx_f_7brainpy_2_c_11composition_element_isotopic_mass(struct __pyx_t_7brainpy_2_c_11composition_Element *, int); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_resize(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *); /*proto*/
static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_f_7brainpy_2_c_11composition_make_element_hash_table_populated(size_t); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_composition_eq(struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *); /*proto*/
static PyObject *__pyx_f_7brainpy_2_c_11composition__store_string(PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "brainpy._c.composition"
extern int __pyx_module_is_main_brainpy___c__composition;
int __pyx_module_is_main_brainpy___c__composition = 0;

/* Implementation of 'brainpy._c.composition' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_H[] = "H+";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__3[] = ", ";
static const char __pyx_k_arg[] = "arg";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_s_d[] = "\"%s\": %d";
static const char __pyx_k_sym[] = "sym";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_elem[] = "elem";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mass[] = "mass";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_Error[] = "Error, ";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_PROTON[] = "PROTON";
static const char __pyx_k_Summed[] = "Summed";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_main_2[] = "main";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_string[] = "string_";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_Massing[] = "Massing";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_element[] = "element";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_isotope[] = "isotope";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_elem_obj[] = "elem_obj";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_nist_mass[] = "nist_mass";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_select_1_1[] = "__select_1_1";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_isotope_obj[] = "isotope_obj";
static const char __pyx_k_nist_mass_2[] = "__nist_mass";
static const char __pyx_k_ElementTable[] = "_ElementTable";
static const char __pyx_k_equality_dict[] = "__equality_dict";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_element_symbol[] = "element_symbol";
static const char __pyx_k_isotope_string[] = "isotope_string";
static const char __pyx_k_Fetched_element[] = "Fetched element";
static const char __pyx_k_PyComposition_s[] = "PyComposition({%s})";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_sum_composition[] = "sum_composition";
static const char __pyx_k_c_element_symbol[] = "c_element_symbol";
static const char __pyx_k_test_hash_string[] = "test_hash_string";
static const char __pyx_k_brainpy_mass_dict[] = "brainpy.mass_dict";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_isotope_parse_test[] = "isotope_parse_test";
static const char __pyx_k_element_name_buffer[] = "element_name_buffer";
static const char __pyx_k_isotope_update_test[] = "isotope_update_test";
static const char __pyx_k_fixed_isotope_element[] = "fixed_isotope_element";
static const char __pyx_k_repr___locals_genexpr[] = "__repr__.<locals>.genexpr";
static const char __pyx_k_brainpy__c_composition[] = "brainpy._c.composition";
static const char __pyx_k_element_hash_table_get[] = "element_hash_table_get";
static const char __pyx_k_equality_pycomposition[] = "__equality_pycomposition";
static const char __pyx_k_brainpy__c_composition_pyx[] = "brainpy\\_c\\composition.pyx";
static const char __pyx_k_Cannot_add_s_to_PyComposition[] = "Cannot add %s to PyComposition";
static const char __pyx_k_Cannot_subtract_s_from_PyComposi[] = "Cannot subtract %s from PyComposition";
static const char __pyx_k_self_impl_cannot_be_converted_to[] = "self.impl cannot be converted to a Python object for pickling";
static PyObject *__pyx_kp_s_Cannot_add_s_to_PyComposition;
static PyObject *__pyx_kp_s_Cannot_subtract_s_from_PyComposi;
static PyObject *__pyx_n_s_ElementTable;
static PyObject *__pyx_kp_s_Error;
static PyObject *__pyx_kp_s_Fetched_element;
static PyObject *__pyx_kp_s_H;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_Massing;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_PROTON;
static PyObject *__pyx_kp_s_PyComposition_s;
static PyObject *__pyx_n_s_Summed;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_n_s_arg;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_brainpy__c_composition;
static PyObject *__pyx_kp_s_brainpy__c_composition_pyx;
static PyObject *__pyx_n_s_brainpy_mass_dict;
static PyObject *__pyx_n_s_c_element_symbol;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_elem;
static PyObject *__pyx_n_s_elem_obj;
static PyObject *__pyx_n_s_element;
static PyObject *__pyx_n_s_element_hash_table_get;
static PyObject *__pyx_n_s_element_name_buffer;
static PyObject *__pyx_n_s_element_symbol;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_equality_dict;
static PyObject *__pyx_n_s_equality_pycomposition;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_fixed_isotope_element;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_isotope;
static PyObject *__pyx_n_s_isotope_obj;
static PyObject *__pyx_n_s_isotope_parse_test;
static PyObject *__pyx_n_s_isotope_string;
static PyObject *__pyx_n_s_isotope_update_test;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_main_2;
static PyObject *__pyx_n_s_mass;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_nist_mass;
static PyObject *__pyx_n_s_nist_mass_2;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr___locals_genexpr;
static PyObject *__pyx_kp_s_s_d;
static PyObject *__pyx_n_s_select_1_1;
static PyObject *__pyx_kp_s_self_impl_cannot_be_converted_to;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_sum_composition;
static PyObject *__pyx_n_s_sym;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_test_hash_string;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_pf_7brainpy_2_c_11composition___select_1_1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_2test_hash_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string_); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_4main(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_6isotope_update_test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_element_symbol); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_8isotope_parse_test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_element_symbol); /* proto */
static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition___init__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_base, PyObject *__pyx_v_kwargs); /* proto */
static void __pyx_pf_7brainpy_2_c_11composition_13PyComposition_2__dealloc__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_4__getitem__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6__setitem__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key, __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_count); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_8update(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_10keys(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_12values(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_14pop(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default); /* proto */
static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_16__contains__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_18items(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_20copy(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_22__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_24__iadd__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_26__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_28__isub__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_30__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_scale); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_32__imul__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_v_scale); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_34__iter__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_36mass(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_7brainpy_2_c_11composition_13PyComposition_38__len__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_8__repr___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_40__repr__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_42__equality_pycomposition(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_44__equality_dict(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_46__richcmp__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass___get__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass_2__set__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean___get__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean_2__set__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_48__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_50__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7brainpy_2_c_11composition_10parse_formula(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_formula); /* proto */
static PyObject *__pyx_tp_new_7brainpy_2_c_11composition_PyComposition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct____repr__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyString_Type_encode = {0, &__pyx_n_s_encode, 0, 0, 0};
static PyObject *__pyx_float_0_;
static PyObject *__pyx_int_0;
static double __pyx_k_;
static double __pyx_k__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
/* Late includes */

/* "brainpy/_c/composition.pyx":34
 * 
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (mz * fabs(z)) - (z * charge_carrier)
 * 
 */

static double __pyx_f_7brainpy_2_c_11composition_neutral_mass(double __pyx_v_mz, int __pyx_v_z, struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k_;
  double __pyx_r;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
    }
  }

  /* "brainpy/_c/composition.pyx":35
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON) nogil:
 *     return (mz * fabs(z)) - (z * charge_carrier)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((__pyx_v_mz * fabs(__pyx_v_z)) - (__pyx_v_z * __pyx_v_charge_carrier));
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":34
 * 
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (mz * fabs(z)) - (z * charge_carrier)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":39
 * 
 * @cython.cdivision
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (neutral_mass + (z * charge_carrier)) / fabs(z)
 * 
 */

static double __pyx_f_7brainpy_2_c_11composition_mass_charge_ratio(double __pyx_v_neutral_mass, int __pyx_v_z, struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio *__pyx_optional_args) {
  double __pyx_v_charge_carrier = __pyx_k__2;
  double __pyx_r;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_charge_carrier = __pyx_optional_args->charge_carrier;
    }
  }

  /* "brainpy/_c/composition.pyx":40
 * @cython.cdivision
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON) nogil:
 *     return (neutral_mass + (z * charge_carrier)) / fabs(z)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((__pyx_v_neutral_mass + (__pyx_v_z * __pyx_v_charge_carrier)) / fabs(__pyx_v_z));
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":39
 * 
 * @cython.cdivision
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (neutral_mass + (z * charge_carrier)) / fabs(z)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":43
 * 
 * 
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i = 0
 */

static char *__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(char *__pyx_v_label, int *__pyx_v_isotope_num, char *__pyx_v_element_name) {
  size_t __pyx_v_i;
  int __pyx_v_in_bracket;
  char __pyx_v_current;
  size_t __pyx_v_name_end;
  size_t __pyx_v_num_start;
  size_t __pyx_v_num_end;
  size_t __pyx_v_size;
  char __pyx_v_number_part[4];
  char *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/composition.pyx":45
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil:
 *     cdef:
 *         size_t i = 0             # <<<<<<<<<<<<<<
 *         bint in_bracket = False
 *         char current
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":46
 *     cdef:
 *         size_t i = 0
 *         bint in_bracket = False             # <<<<<<<<<<<<<<
 *         char current
 *         size_t name_end
 */
  __pyx_v_in_bracket = 0;

  /* "brainpy/_c/composition.pyx":55
 *         char[4] number_part
 * 
 *     name_end = 0             # <<<<<<<<<<<<<<
 *     num_start = 0
 *     num_end = 0
 */
  __pyx_v_name_end = 0;

  /* "brainpy/_c/composition.pyx":56
 * 
 *     name_end = 0
 *     num_start = 0             # <<<<<<<<<<<<<<
 *     num_end = 0
 *     size = strlen(label)
 */
  __pyx_v_num_start = 0;

  /* "brainpy/_c/composition.pyx":57
 *     name_end = 0
 *     num_start = 0
 *     num_end = 0             # <<<<<<<<<<<<<<
 *     size = strlen(label)
 *     for i in range(size):
 */
  __pyx_v_num_end = 0;

  /* "brainpy/_c/composition.pyx":58
 *     num_start = 0
 *     num_end = 0
 *     size = strlen(label)             # <<<<<<<<<<<<<<
 *     for i in range(size):
 *         current = label[i]
 */
  __pyx_v_size = strlen(__pyx_v_label);

  /* "brainpy/_c/composition.pyx":59
 *     num_end = 0
 *     size = strlen(label)
 *     for i in range(size):             # <<<<<<<<<<<<<<
 *         current = label[i]
 *         if in_bracket:
 */
  __pyx_t_1 = __pyx_v_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":60
 *     size = strlen(label)
 *     for i in range(size):
 *         current = label[i]             # <<<<<<<<<<<<<<
 *         if in_bracket:
 *             if current == 93:
 */
    __pyx_v_current = (__pyx_v_label[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":61
 *     for i in range(size):
 *         current = label[i]
 *         if in_bracket:             # <<<<<<<<<<<<<<
 *             if current == 93:
 *                 break
 */
    __pyx_t_4 = (__pyx_v_in_bracket != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":62
 *         current = label[i]
 *         if in_bracket:
 *             if current == 93:             # <<<<<<<<<<<<<<
 *                 break
 *             num_end += 1
 */
      __pyx_t_4 = ((__pyx_v_current == 93) != 0);
      if (__pyx_t_4) {

        /* "brainpy/_c/composition.pyx":63
 *         if in_bracket:
 *             if current == 93:
 *                 break             # <<<<<<<<<<<<<<
 *             num_end += 1
 *         elif current == 91:
 */
        goto __pyx_L4_break;

        /* "brainpy/_c/composition.pyx":62
 *         current = label[i]
 *         if in_bracket:
 *             if current == 93:             # <<<<<<<<<<<<<<
 *                 break
 *             num_end += 1
 */
      }

      /* "brainpy/_c/composition.pyx":64
 *             if current == 93:
 *                 break
 *             num_end += 1             # <<<<<<<<<<<<<<
 *         elif current == 91:
 *             in_bracket = True
 */
      __pyx_v_num_end = (__pyx_v_num_end + 1);

      /* "brainpy/_c/composition.pyx":61
 *     for i in range(size):
 *         current = label[i]
 *         if in_bracket:             # <<<<<<<<<<<<<<
 *             if current == 93:
 *                 break
 */
      goto __pyx_L5;
    }

    /* "brainpy/_c/composition.pyx":65
 *                 break
 *             num_end += 1
 *         elif current == 91:             # <<<<<<<<<<<<<<
 *             in_bracket = True
 *             name_end = i
 */
    __pyx_t_4 = ((__pyx_v_current == 91) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":66
 *             num_end += 1
 *         elif current == 91:
 *             in_bracket = True             # <<<<<<<<<<<<<<
 *             name_end = i
 *             num_start = i + 1
 */
      __pyx_v_in_bracket = 1;

      /* "brainpy/_c/composition.pyx":67
 *         elif current == 91:
 *             in_bracket = True
 *             name_end = i             # <<<<<<<<<<<<<<
 *             num_start = i + 1
 *             num_end = num_start
 */
      __pyx_v_name_end = __pyx_v_i;

      /* "brainpy/_c/composition.pyx":68
 *             in_bracket = True
 *             name_end = i
 *             num_start = i + 1             # <<<<<<<<<<<<<<
 *             num_end = num_start
 *         else:
 */
      __pyx_v_num_start = (__pyx_v_i + 1);

      /* "brainpy/_c/composition.pyx":69
 *             name_end = i
 *             num_start = i + 1
 *             num_end = num_start             # <<<<<<<<<<<<<<
 *         else:
 *             name_end += 1
 */
      __pyx_v_num_end = __pyx_v_num_start;

      /* "brainpy/_c/composition.pyx":65
 *                 break
 *             num_end += 1
 *         elif current == 91:             # <<<<<<<<<<<<<<
 *             in_bracket = True
 *             name_end = i
 */
      goto __pyx_L5;
    }

    /* "brainpy/_c/composition.pyx":71
 *             num_end = num_start
 *         else:
 *             name_end += 1             # <<<<<<<<<<<<<<
 * 
 *     if num_start > 0:
 */
    /*else*/ {
      __pyx_v_name_end = (__pyx_v_name_end + 1);
    }
    __pyx_L5:;
  }
  __pyx_L4_break:;

  /* "brainpy/_c/composition.pyx":73
 *             name_end += 1
 * 
 *     if num_start > 0:             # <<<<<<<<<<<<<<
 *         memcpy(number_part, label + num_start, num_end - num_start)
 *         number_part[num_end - num_start] = '\0'
 */
  __pyx_t_4 = ((__pyx_v_num_start > 0) != 0);
  if (__pyx_t_4) {

    /* "brainpy/_c/composition.pyx":74
 * 
 *     if num_start > 0:
 *         memcpy(number_part, label + num_start, num_end - num_start)             # <<<<<<<<<<<<<<
 *         number_part[num_end - num_start] = '\0'
 *         isotope_num[0] = atoi(number_part)
 */
    (void)(memcpy(__pyx_v_number_part, (__pyx_v_label + __pyx_v_num_start), (__pyx_v_num_end - __pyx_v_num_start)));

    /* "brainpy/_c/composition.pyx":75
 *     if num_start > 0:
 *         memcpy(number_part, label + num_start, num_end - num_start)
 *         number_part[num_end - num_start] = '\0'             # <<<<<<<<<<<<<<
 *         isotope_num[0] = atoi(number_part)
 *     else:
 */
    (__pyx_v_number_part[(__pyx_v_num_end - __pyx_v_num_start)]) = '\x00';

    /* "brainpy/_c/composition.pyx":76
 *         memcpy(number_part, label + num_start, num_end - num_start)
 *         number_part[num_end - num_start] = '\0'
 *         isotope_num[0] = atoi(number_part)             # <<<<<<<<<<<<<<
 *     else:
 *         isotope_num[0] = 0
 */
    (__pyx_v_isotope_num[0]) = atoi(__pyx_v_number_part);

    /* "brainpy/_c/composition.pyx":73
 *             name_end += 1
 * 
 *     if num_start > 0:             # <<<<<<<<<<<<<<
 *         memcpy(number_part, label + num_start, num_end - num_start)
 *         number_part[num_end - num_start] = '\0'
 */
    goto __pyx_L7;
  }

  /* "brainpy/_c/composition.pyx":78
 *         isotope_num[0] = atoi(number_part)
 *     else:
 *         isotope_num[0] = 0             # <<<<<<<<<<<<<<
 * 
 *     element_name[name_end] = '\0'
 */
  /*else*/ {
    (__pyx_v_isotope_num[0]) = 0;
  }
  __pyx_L7:;

  /* "brainpy/_c/composition.pyx":80
 *         isotope_num[0] = 0
 * 
 *     element_name[name_end] = '\0'             # <<<<<<<<<<<<<<
 *     memcpy(element_name, label, name_end)
 * 
 */
  (__pyx_v_element_name[__pyx_v_name_end]) = '\x00';

  /* "brainpy/_c/composition.pyx":81
 * 
 *     element_name[name_end] = '\0'
 *     memcpy(element_name, label, name_end)             # <<<<<<<<<<<<<<
 * 
 *     return element_name
 */
  (void)(memcpy(__pyx_v_element_name, __pyx_v_label, __pyx_v_name_end));

  /* "brainpy/_c/composition.pyx":83
 *     memcpy(element_name, label, name_end)
 * 
 *     return element_name             # <<<<<<<<<<<<<<
 * 
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 */
  __pyx_r = __pyx_v_element_name;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":43
 * 
 * 
 * cdef char* _parse_isotope_string(char* label, int* isotope_num, char* element_name) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i = 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":85
 *     return element_name
 * 
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:             # <<<<<<<<<<<<<<
 *     if isotope.neutron_shift == 0:
 *         sprintf(out, "%s", element.symbol)
 */

static char *__pyx_f_7brainpy_2_c_11composition__make_isotope_string(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element, struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_isotope, char *__pyx_v_out) {
  char *__pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":86
 * 
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 *     if isotope.neutron_shift == 0:             # <<<<<<<<<<<<<<
 *         sprintf(out, "%s", element.symbol)
 *         return out
 */
  __pyx_t_1 = ((__pyx_v_isotope->neutron_shift == 0) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":87
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 *     if isotope.neutron_shift == 0:
 *         sprintf(out, "%s", element.symbol)             # <<<<<<<<<<<<<<
 *         return out
 *     else:
 */
    (void)(sprintf(__pyx_v_out, ((char *)"%s"), __pyx_v_element->symbol));

    /* "brainpy/_c/composition.pyx":88
 *     if isotope.neutron_shift == 0:
 *         sprintf(out, "%s", element.symbol)
 *         return out             # <<<<<<<<<<<<<<
 *     else:
 *         sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
 */
    __pyx_r = __pyx_v_out;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":86
 * 
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 *     if isotope.neutron_shift == 0:             # <<<<<<<<<<<<<<
 *         sprintf(out, "%s", element.symbol)
 *         return out
 */
  }

  /* "brainpy/_c/composition.pyx":90
 *         return out
 *     else:
 *         sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  /*else*/ {
    (void)(sprintf(__pyx_v_out, ((char *)"%s[%d]"), __pyx_v_element->symbol, __pyx_v_isotope->neutrons));

    /* "brainpy/_c/composition.pyx":91
 *     else:
 *         sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
 *         return out             # <<<<<<<<<<<<<<
 * 
 * cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 */
    __pyx_r = __pyx_v_out;
    goto __pyx_L0;
  }

  /* "brainpy/_c/composition.pyx":85
 *     return element_name
 * 
 * cdef char* _make_isotope_string(Element* element, Isotope* isotope, char* out) nogil:             # <<<<<<<<<<<<<<
 *     if isotope.neutron_shift == 0:
 *         sprintf(out, "%s", element.symbol)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":93
 *         return out
 * 
 * cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:             # <<<<<<<<<<<<<<
 *     sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
 *     return out
 */

static char *__pyx_f_7brainpy_2_c_11composition__make_fixed_isotope_string(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element, struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_isotope, char *__pyx_v_out) {
  char *__pyx_r;

  /* "brainpy/_c/composition.pyx":94
 * 
 * cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 *     sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
  (void)(sprintf(__pyx_v_out, ((char *)"%s[%d]"), __pyx_v_element->symbol, __pyx_v_isotope->neutrons));

  /* "brainpy/_c/composition.pyx":95
 * cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:
 *     sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
 *     return out             # <<<<<<<<<<<<<<
 * 
 * # -----------------------------------------------------------------------------
 */
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":93
 *         return out
 * 
 * cdef char* _make_fixed_isotope_string(Element* element, Isotope* isotope, char* out) nogil:             # <<<<<<<<<<<<<<
 *     sprintf(out, "%s[%d]", element.symbol, isotope.neutrons)
 *     return out
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":100
 * # Isotope and IsotopeMap Methods
 * 
 * cdef IsotopeMap* make_isotope_map(list organized_isotope_data, size_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopeMap* result
 */

static struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_f_7brainpy_2_c_11composition_make_isotope_map(PyObject *__pyx_v_organized_isotope_data, size_t __pyx_v_size) {
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_result;
  size_t __pyx_v_i;
  CYTHON_UNUSED int __pyx_v_k;
  PyObject *__pyx_v_values = 0;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  double __pyx_t_9;
  __Pyx_RefNannySetupContext("make_isotope_map", 0);

  /* "brainpy/_c/composition.pyx":107
 *         tuple values
 * 
 *     result = <IsotopeMap*>malloc(sizeof(IsotopeMap))             # <<<<<<<<<<<<<<
 *     result.bins = <Isotope*>malloc(sizeof(Isotope) * size)
 *     result.size = size
 */
  __pyx_v_result = ((struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap))));

  /* "brainpy/_c/composition.pyx":108
 * 
 *     result = <IsotopeMap*>malloc(sizeof(IsotopeMap))
 *     result.bins = <Isotope*>malloc(sizeof(Isotope) * size)             # <<<<<<<<<<<<<<
 *     result.size = size
 * 
 */
  __pyx_v_result->bins = ((struct __pyx_t_7brainpy_2_c_11composition_Isotope *)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Isotope)) * __pyx_v_size)));

  /* "brainpy/_c/composition.pyx":109
 *     result = <IsotopeMap*>malloc(sizeof(IsotopeMap))
 *     result.bins = <Isotope*>malloc(sizeof(Isotope) * size)
 *     result.size = size             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __pyx_v_result->size = __pyx_v_size;

  /* "brainpy/_c/composition.pyx":111
 *     result.size = size
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     for k, values in organized_isotope_data:
 *         result.bins[i].mass = values[0]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":112
 * 
 *     i = 0
 *     for k, values in organized_isotope_data:             # <<<<<<<<<<<<<<
 *         result.bins[i].mass = values[0]
 *         result.bins[i].abundance = values[1]
 */
  if (unlikely(__pyx_v_organized_isotope_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 112, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_organized_isotope_data; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 112, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 112, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 112, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyTuple_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 112, __pyx_L1_error)
    __pyx_v_k = __pyx_t_8;
    __Pyx_XDECREF_SET(__pyx_v_values, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "brainpy/_c/composition.pyx":113
 *     i = 0
 *     for k, values in organized_isotope_data:
 *         result.bins[i].mass = values[0]             # <<<<<<<<<<<<<<
 *         result.bins[i].abundance = values[1]
 *         result.bins[i].neutron_shift = values[2]
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 113, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_values, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_result->bins[__pyx_v_i]).mass = __pyx_t_9;

    /* "brainpy/_c/composition.pyx":114
 *     for k, values in organized_isotope_data:
 *         result.bins[i].mass = values[0]
 *         result.bins[i].abundance = values[1]             # <<<<<<<<<<<<<<
 *         result.bins[i].neutron_shift = values[2]
 *         result.bins[i].neutrons = values[3]
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_values, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_result->bins[__pyx_v_i]).abundance = __pyx_t_9;

    /* "brainpy/_c/composition.pyx":115
 *         result.bins[i].mass = values[0]
 *         result.bins[i].abundance = values[1]
 *         result.bins[i].neutron_shift = values[2]             # <<<<<<<<<<<<<<
 *         result.bins[i].neutrons = values[3]
 *         i += 1
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 115, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_values, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_result->bins[__pyx_v_i]).neutron_shift = __pyx_t_8;

    /* "brainpy/_c/composition.pyx":116
 *         result.bins[i].abundance = values[1]
 *         result.bins[i].neutron_shift = values[2]
 *         result.bins[i].neutrons = values[3]             # <<<<<<<<<<<<<<
 *         i += 1
 *     return result
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 116, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_values, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    (__pyx_v_result->bins[__pyx_v_i]).neutrons = __pyx_t_8;

    /* "brainpy/_c/composition.pyx":117
 *         result.bins[i].neutron_shift = values[2]
 *         result.bins[i].neutrons = values[3]
 *         i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":112
 * 
 *     i = 0
 *     for k, values in organized_isotope_data:             # <<<<<<<<<<<<<<
 *         result.bins[i].mass = values[0]
 *         result.bins[i].abundance = values[1]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":118
 *         result.bins[i].neutrons = values[3]
 *         i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":100
 * # Isotope and IsotopeMap Methods
 * 
 * cdef IsotopeMap* make_isotope_map(list organized_isotope_data, size_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         IsotopeMap* result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("brainpy._c.composition.make_isotope_map", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":121
 * 
 * 
 * cdef Isotope* get_isotope_by_neutron_shift(IsotopeMap* isotopes, int neutron_shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotopes, int __pyx_v_neutron_shift) {
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_isotope_item;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/composition.pyx":126
 *         Isotope* isotope_item
 * 
 *     for i in range(isotopes.size):             # <<<<<<<<<<<<<<
 *         isotope_item = &(isotopes.bins[i])
 *         # if isotope_item.neutron_shift == neutron_shift:
 */
  __pyx_t_1 = __pyx_v_isotopes->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":127
 * 
 *     for i in range(isotopes.size):
 *         isotope_item = &(isotopes.bins[i])             # <<<<<<<<<<<<<<
 *         # if isotope_item.neutron_shift == neutron_shift:
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 */
    __pyx_v_isotope_item = (&(__pyx_v_isotopes->bins[__pyx_v_i]));

    /* "brainpy/_c/composition.pyx":129
 *         isotope_item = &(isotopes.bins[i])
 *         # if isotope_item.neutron_shift == neutron_shift:
 *         if isotopes.bins[i].neutron_shift == neutron_shift:             # <<<<<<<<<<<<<<
 *             return isotope_item
 *     return NULL
 */
    __pyx_t_4 = (((__pyx_v_isotopes->bins[__pyx_v_i]).neutron_shift == __pyx_v_neutron_shift) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":130
 *         # if isotope_item.neutron_shift == neutron_shift:
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 *             return isotope_item             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
      __pyx_r = __pyx_v_isotope_item;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":129
 *         isotope_item = &(isotopes.bins[i])
 *         # if isotope_item.neutron_shift == neutron_shift:
 *         if isotopes.bins[i].neutron_shift == neutron_shift:             # <<<<<<<<<<<<<<
 *             return isotope_item
 *     return NULL
 */
    }
  }

  /* "brainpy/_c/composition.pyx":131
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 *             return isotope_item
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":121
 * 
 * 
 * cdef Isotope* get_isotope_by_neutron_shift(IsotopeMap* isotopes, int neutron_shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":134
 * 
 * 
 * cdef int get_isotope_by_neutron_shift_index(IsotopeMap* isotopes, int neutron_shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift_index(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotopes, int __pyx_v_neutron_shift) {
  size_t __pyx_v_i;
  int __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/composition.pyx":138
 *         size_t i
 * 
 *     for i in range(isotopes.size):             # <<<<<<<<<<<<<<
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 *             return i
 */
  __pyx_t_1 = __pyx_v_isotopes->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":139
 * 
 *     for i in range(isotopes.size):
 *         if isotopes.bins[i].neutron_shift == neutron_shift:             # <<<<<<<<<<<<<<
 *             return i
 *     return -1
 */
    __pyx_t_4 = (((__pyx_v_isotopes->bins[__pyx_v_i]).neutron_shift == __pyx_v_neutron_shift) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":140
 *     for i in range(isotopes.size):
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 *             return i             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
      __pyx_r = __pyx_v_i;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":139
 * 
 *     for i in range(isotopes.size):
 *         if isotopes.bins[i].neutron_shift == neutron_shift:             # <<<<<<<<<<<<<<
 *             return i
 *     return -1
 */
    }
  }

  /* "brainpy/_c/composition.pyx":141
 *         if isotopes.bins[i].neutron_shift == neutron_shift:
 *             return i
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = -1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":134
 * 
 * 
 * cdef int get_isotope_by_neutron_shift_index(IsotopeMap* isotopes, int neutron_shift) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":144
 * 
 * 
 * cdef Isotope* get_isotope_by_neutron_count(IsotopeMap* isotopes, int neutrons) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_count(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotopes, int __pyx_v_neutrons) {
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_isotope_item;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/composition.pyx":149
 *         Isotope* isotope_item
 * 
 *     for i in range(isotopes.size):             # <<<<<<<<<<<<<<
 *         isotope_item = &(isotopes.bins[i])
 *         if isotope_item.neutrons == neutrons:
 */
  __pyx_t_1 = __pyx_v_isotopes->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":150
 * 
 *     for i in range(isotopes.size):
 *         isotope_item = &(isotopes.bins[i])             # <<<<<<<<<<<<<<
 *         if isotope_item.neutrons == neutrons:
 *             return isotope_item
 */
    __pyx_v_isotope_item = (&(__pyx_v_isotopes->bins[__pyx_v_i]));

    /* "brainpy/_c/composition.pyx":151
 *     for i in range(isotopes.size):
 *         isotope_item = &(isotopes.bins[i])
 *         if isotope_item.neutrons == neutrons:             # <<<<<<<<<<<<<<
 *             return isotope_item
 *     return NULL
 */
    __pyx_t_4 = ((__pyx_v_isotope_item->neutrons == __pyx_v_neutrons) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":152
 *         isotope_item = &(isotopes.bins[i])
 *         if isotope_item.neutrons == neutrons:
 *             return isotope_item             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
      __pyx_r = __pyx_v_isotope_item;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":151
 *     for i in range(isotopes.size):
 *         isotope_item = &(isotopes.bins[i])
 *         if isotope_item.neutrons == neutrons:             # <<<<<<<<<<<<<<
 *             return isotope_item
 *     return NULL
 */
    }
  }

  /* "brainpy/_c/composition.pyx":153
 *         if isotope_item.neutrons == neutrons:
 *             return isotope_item
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":144
 * 
 * 
 * cdef Isotope* get_isotope_by_neutron_count(IsotopeMap* isotopes, int neutrons) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":156
 * 
 * 
 * cdef void print_isotope_map(IsotopeMap* isotope_map):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_11composition_print_isotope_map(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotope_map) {
  size_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  __Pyx_RefNannySetupContext("print_isotope_map", 0);

  /* "brainpy/_c/composition.pyx":160
 *         size_t i
 * 
 *     for i in range(isotope_map.size):             # <<<<<<<<<<<<<<
 *         printf("%f, %f, %d\n", isotope_map.bins[i].mass, isotope_map.bins[i].abundance, isotope_map.bins[i].neutron_shift)
 * 
 */
  __pyx_t_1 = __pyx_v_isotope_map->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":161
 * 
 *     for i in range(isotope_map.size):
 *         printf("%f, %f, %d\n", isotope_map.bins[i].mass, isotope_map.bins[i].abundance, isotope_map.bins[i].neutron_shift)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (void)(printf(((char const *)"%f, %f, %d\n"), (__pyx_v_isotope_map->bins[__pyx_v_i]).mass, (__pyx_v_isotope_map->bins[__pyx_v_i]).abundance, (__pyx_v_isotope_map->bins[__pyx_v_i]).neutron_shift));
  }

  /* "brainpy/_c/composition.pyx":156
 * 
 * 
 * cdef void print_isotope_map(IsotopeMap* isotope_map):             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "brainpy/_c/composition.pyx":164
 * 
 * 
 * cdef void free_isotope_map(IsotopeMap* isotopes) nogil:             # <<<<<<<<<<<<<<
 *     free(isotopes.bins)
 *     free(isotopes)
 */

static void __pyx_f_7brainpy_2_c_11composition_free_isotope_map(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotopes) {

  /* "brainpy/_c/composition.pyx":165
 * 
 * cdef void free_isotope_map(IsotopeMap* isotopes) nogil:
 *     free(isotopes.bins)             # <<<<<<<<<<<<<<
 *     free(isotopes)
 * 
 */
  free(__pyx_v_isotopes->bins);

  /* "brainpy/_c/composition.pyx":166
 * cdef void free_isotope_map(IsotopeMap* isotopes) nogil:
 *     free(isotopes.bins)
 *     free(isotopes)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_isotopes);

  /* "brainpy/_c/composition.pyx":164
 * 
 * 
 * cdef void free_isotope_map(IsotopeMap* isotopes) nogil:             # <<<<<<<<<<<<<<
 *     free(isotopes.bins)
 *     free(isotopes)
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":172
 * # Element Methods
 * 
 * cdef double element_monoisotopic_mass(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     if element.monoisotopic_isotope_index < 0:
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass
 */

static double __pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {
  double __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":173
 * 
 * cdef double element_monoisotopic_mass(Element* element) nogil:
 *     if element.monoisotopic_isotope_index < 0:             # <<<<<<<<<<<<<<
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_element->monoisotopic_isotope_index < 0) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":174
 * cdef double element_monoisotopic_mass(Element* element) nogil:
 *     if element.monoisotopic_isotope_index < 0:
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass             # <<<<<<<<<<<<<<
 *     else:
 *         return element.isotopes.bins[element.monoisotopic_isotope_index].mass
 */
    __pyx_r = __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift(__pyx_v_element->isotopes, 0)->mass;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":173
 * 
 * cdef double element_monoisotopic_mass(Element* element) nogil:
 *     if element.monoisotopic_isotope_index < 0:             # <<<<<<<<<<<<<<
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass
 *     else:
 */
  }

  /* "brainpy/_c/composition.pyx":176
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass
 *     else:
 *         return element.isotopes.bins[element.monoisotopic_isotope_index].mass             # <<<<<<<<<<<<<<
 * 
 * cdef double element_isotopic_mass(Element* element, int isotope_number) nogil:
 */
  /*else*/ {
    __pyx_r = (__pyx_v_element->isotopes->bins[__pyx_v_element->monoisotopic_isotope_index]).mass;
    goto __pyx_L0;
  }

  /* "brainpy/_c/composition.pyx":172
 * # Element Methods
 * 
 * cdef double element_monoisotopic_mass(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     if element.monoisotopic_isotope_index < 0:
 *         return get_isotope_by_neutron_shift(element.isotopes, 0).mass
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":178
 *         return element.isotopes.bins[element.monoisotopic_isotope_index].mass
 * 
 * cdef double element_isotopic_mass(Element* element, int isotope_number) nogil:             # <<<<<<<<<<<<<<
 *     return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass
 * 
 */

static double __pyx_f_7brainpy_2_c_11composition_element_isotopic_mass(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element, int __pyx_v_isotope_number) {
  double __pyx_r;

  /* "brainpy/_c/composition.pyx":179
 * 
 * cdef double element_isotopic_mass(Element* element, int isotope_number) nogil:
 *     return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass             # <<<<<<<<<<<<<<
 * 
 * cdef int element_min_neutron_shift(Element* element) nogil:
 */
  __pyx_r = __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_count(__pyx_v_element->isotopes, __pyx_v_isotope_number)->mass;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":178
 *         return element.isotopes.bins[element.monoisotopic_isotope_index].mass
 * 
 * cdef double element_isotopic_mass(Element* element, int isotope_number) nogil:             # <<<<<<<<<<<<<<
 *     return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":181
 *     return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass
 * 
 * cdef int element_min_neutron_shift(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     return element.isotopes.bins[0].neutron_shift
 * 
 */

static int __pyx_f_7brainpy_2_c_11composition_element_min_neutron_shift(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {
  int __pyx_r;

  /* "brainpy/_c/composition.pyx":182
 * 
 * cdef int element_min_neutron_shift(Element* element) nogil:
 *     return element.isotopes.bins[0].neutron_shift             # <<<<<<<<<<<<<<
 * 
 * cdef int element_max_neutron_shift(Element* element) nogil:
 */
  __pyx_r = (__pyx_v_element->isotopes->bins[0]).neutron_shift;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":181
 *     return get_isotope_by_neutron_count(element.isotopes, isotope_number).mass
 * 
 * cdef int element_min_neutron_shift(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     return element.isotopes.bins[0].neutron_shift
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":184
 *     return element.isotopes.bins[0].neutron_shift
 * 
 * cdef int element_max_neutron_shift(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 */

static int __pyx_f_7brainpy_2_c_11composition_element_max_neutron_shift(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {
  int __pyx_r;

  /* "brainpy/_c/composition.pyx":185
 * 
 * cdef int element_max_neutron_shift(Element* element) nogil:
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift             # <<<<<<<<<<<<<<
 * 
 * def __select_1_1(object x):
 */
  __pyx_r = (__pyx_v_element->isotopes->bins[(__pyx_v_element->isotopes->size - 1)]).neutron_shift;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":184
 *     return element.isotopes.bins[0].neutron_shift
 * 
 * cdef int element_max_neutron_shift(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":187
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 * def __select_1_1(object x):             # <<<<<<<<<<<<<<
 *     return x[1][1]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_1__select_1_1(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition___select_1_1[] = "__select_1_1(x)";
static PyMethodDef __pyx_mdef_7brainpy_2_c_11composition_1__select_1_1 = {"__select_1_1", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_1__select_1_1, METH_O, __pyx_doc_7brainpy_2_c_11composition___select_1_1};
static PyObject *__pyx_pw_7brainpy_2_c_11composition_1__select_1_1(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__select_1_1 (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition___select_1_1(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition___select_1_1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__select_1_1", 0);

  /* "brainpy/_c/composition.pyx":188
 * 
 * def __select_1_1(object x):
 *     return x[1][1]             # <<<<<<<<<<<<<<
 * 
 * cdef void _isotopes_of(char* element_symbol, IsotopeMap** isotope_frequencies):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_x, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":187
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 * def __select_1_1(object x):             # <<<<<<<<<<<<<<
 *     return x[1][1]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("brainpy._c.composition.__select_1_1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":190
 *     return x[1][1]
 * 
 * cdef void _isotopes_of(char* element_symbol, IsotopeMap** isotope_frequencies):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict freqs, element_data
 */

static void __pyx_f_7brainpy_2_c_11composition__isotopes_of(char *__pyx_v_element_symbol, struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap **__pyx_v_isotope_frequencies) {
  PyObject *__pyx_v_freqs = 0;
  PyObject *__pyx_v_element_data = 0;
  PyObject *__pyx_v_freq_list = 0;
  int __pyx_v_i;
  PyObject *__pyx_v_bunches = 0;
  PyObject *__pyx_v_mass_freqs = 0;
  int __pyx_v_mono_neutrons;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_py_element_symbol = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  __Pyx_RefNannySetupContext("_isotopes_of", 0);

  /* "brainpy/_c/composition.pyx":203
 *         str py_element_symbol
 * 
 *     freqs = dict()             # <<<<<<<<<<<<<<
 *     py_element_symbol = PyString_FromString(element_symbol)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_freqs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":204
 * 
 *     freqs = dict()
 *     py_element_symbol = PyString_FromString(element_symbol)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = PyStr_FromString(__pyx_v_element_symbol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_element_symbol = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":206
 *     py_element_symbol = PyString_FromString(element_symbol)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         element_data = nist_mass[py_element_symbol]
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "brainpy/_c/composition.pyx":207
 * 
 *     try:
 *         element_data = nist_mass[py_element_symbol]             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         element_data = nist_mass[py_element_symbol.encode('utf-8')]
 */
      if (unlikely(__pyx_v_7brainpy_2_c_11composition_nist_mass == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 207, __pyx_L3_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_7brainpy_2_c_11composition_nist_mass, __pyx_v_py_element_symbol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 207, __pyx_L3_error)
      __pyx_v_element_data = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "brainpy/_c/composition.pyx":206
 *     py_element_symbol = PyString_FromString(element_symbol)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         element_data = nist_mass[py_element_symbol]
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":208
 *     try:
 *         element_data = nist_mass[py_element_symbol]
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         element_data = nist_mass[py_element_symbol.encode('utf-8')]
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("brainpy._c.composition._isotopes_of", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 208, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "brainpy/_c/composition.pyx":209
 *         element_data = nist_mass[py_element_symbol]
 *     except KeyError:
 *         element_data = nist_mass[py_element_symbol.encode('utf-8')]             # <<<<<<<<<<<<<<
 * 
 *     for i, mass_freqs in element_data.items():
 */
      if (unlikely(__pyx_v_7brainpy_2_c_11composition_nist_mass == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 209, __pyx_L5_except_error)
      }
      __pyx_t_8 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyString_Type_encode, __pyx_v_py_element_symbol, __pyx_kp_s_utf_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 209, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_7brainpy_2_c_11composition_nist_mass, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 209, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (!(likely(PyDict_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 209, __pyx_L5_except_error)
      __Pyx_XDECREF_SET(__pyx_v_element_data, ((PyObject*)__pyx_t_9));
      __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "brainpy/_c/composition.pyx":206
 *     py_element_symbol = PyString_FromString(element_symbol)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         element_data = nist_mass[py_element_symbol]
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "brainpy/_c/composition.pyx":211
 *         element_data = nist_mass[py_element_symbol.encode('utf-8')]
 * 
 *     for i, mass_freqs in element_data.items():             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             continue
 */
  if (unlikely(__pyx_v_element_data == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 211, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyDict_Items(__pyx_v_element_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_6 = __pyx_t_7; __Pyx_INCREF(__pyx_t_6); __pyx_t_10 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 211, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 211, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 211, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_11(__pyx_t_6);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 211, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 211, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_1)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_9)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_8), 2) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L14_unpacking_done;
      __pyx_L13_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 211, __pyx_L1_error)
      __pyx_L14_unpacking_done:;
    }
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyTuple_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 211, __pyx_L1_error)
    __pyx_v_i = __pyx_t_5;
    __Pyx_XDECREF_SET(__pyx_v_mass_freqs, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "brainpy/_c/composition.pyx":212
 * 
 *     for i, mass_freqs in element_data.items():
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         if mass_freqs[1] > 0:
 */
    __pyx_t_13 = ((__pyx_v_i == 0) != 0);
    if (__pyx_t_13) {

      /* "brainpy/_c/composition.pyx":213
 *     for i, mass_freqs in element_data.items():
 *         if i == 0:
 *             continue             # <<<<<<<<<<<<<<
 *         if mass_freqs[1] > 0:
 *             freqs[i] = mass_freqs
 */
      goto __pyx_L11_continue;

      /* "brainpy/_c/composition.pyx":212
 * 
 *     for i, mass_freqs in element_data.items():
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         if mass_freqs[1] > 0:
 */
    }

    /* "brainpy/_c/composition.pyx":214
 *         if i == 0:
 *             continue
 *         if mass_freqs[1] > 0:             # <<<<<<<<<<<<<<
 *             freqs[i] = mass_freqs
 * 
 */
    if (unlikely(__pyx_v_mass_freqs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 214, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v_mass_freqs, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_13) {

      /* "brainpy/_c/composition.pyx":215
 *             continue
 *         if mass_freqs[1] > 0:
 *             freqs[i] = mass_freqs             # <<<<<<<<<<<<<<
 * 
 *     if len(freqs) == 0:
 */
      __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (unlikely(PyDict_SetItem(__pyx_v_freqs, __pyx_t_9, __pyx_v_mass_freqs) < 0)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "brainpy/_c/composition.pyx":214
 *         if i == 0:
 *             continue
 *         if mass_freqs[1] > 0:             # <<<<<<<<<<<<<<
 *             freqs[i] = mass_freqs
 * 
 */
    }

    /* "brainpy/_c/composition.pyx":211
 *         element_data = nist_mass[py_element_symbol.encode('utf-8')]
 * 
 *     for i, mass_freqs in element_data.items():             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             continue
 */
    __pyx_L11_continue:;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "brainpy/_c/composition.pyx":217
 *             freqs[i] = mass_freqs
 * 
 *     if len(freqs) == 0:             # <<<<<<<<<<<<<<
 *         isotope_frequencies[0] = make_isotope_map([], 0)
 *         return
 */
  __pyx_t_10 = PyDict_Size(__pyx_v_freqs); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_t_13 = ((__pyx_t_10 == 0) != 0);
  if (__pyx_t_13) {

    /* "brainpy/_c/composition.pyx":218
 * 
 *     if len(freqs) == 0:
 *         isotope_frequencies[0] = make_isotope_map([], 0)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    (__pyx_v_isotope_frequencies[0]) = __pyx_f_7brainpy_2_c_11composition_make_isotope_map(((PyObject*)__pyx_t_6), 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "brainpy/_c/composition.pyx":219
 *     if len(freqs) == 0:
 *         isotope_frequencies[0] = make_isotope_map([], 0)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     freq_list = list(freqs.items())
 */
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":217
 *             freqs[i] = mass_freqs
 * 
 *     if len(freqs) == 0:             # <<<<<<<<<<<<<<
 *         isotope_frequencies[0] = make_isotope_map([], 0)
 *         return
 */
  }

  /* "brainpy/_c/composition.pyx":221
 *         return
 * 
 *     freq_list = list(freqs.items())             # <<<<<<<<<<<<<<
 *     freq_list.sort(key=__select_1_1)
 *     mono_neutrons = freq_list[-1][0]
 */
  __pyx_t_6 = __Pyx_PyDict_Items(__pyx_v_freqs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_freq_list = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "brainpy/_c/composition.pyx":222
 * 
 *     freq_list = list(freqs.items())
 *     freq_list.sort(key=__select_1_1)             # <<<<<<<<<<<<<<
 *     mono_neutrons = freq_list[-1][0]
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_freq_list, __pyx_n_s_sort); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_select_1_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_key, __pyx_t_7) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "brainpy/_c/composition.pyx":223
 *     freq_list = list(freqs.items())
 *     freq_list.sort(key=__select_1_1)
 *     mono_neutrons = freq_list[-1][0]             # <<<<<<<<<<<<<<
 * 
 *     bunches = []
 */
  __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_freq_list, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_mono_neutrons = __pyx_t_5;

  /* "brainpy/_c/composition.pyx":225
 *     mono_neutrons = freq_list[-1][0]
 * 
 *     bunches = []             # <<<<<<<<<<<<<<
 *     for k, v in freqs.items():
 *         bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_bunches = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "brainpy/_c/composition.pyx":226
 * 
 *     bunches = []
 *     for k, v in freqs.items():             # <<<<<<<<<<<<<<
 *         bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_Items(__pyx_v_freqs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
    __pyx_t_7 = __pyx_t_6; __Pyx_INCREF(__pyx_t_7); __pyx_t_10 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 226, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_6); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 226, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_6); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 226, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_11(__pyx_t_7);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 226, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 226, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_9)) goto __pyx_L20_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_1 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_1)) goto __pyx_L20_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_8), 2) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L21_unpacking_done;
      __pyx_L20_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 226, __pyx_L1_error)
      __pyx_L21_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":227
 *     bunches = []
 *     for k, v in freqs.items():
 *         bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))             # <<<<<<<<<<<<<<
 * 
 *     bunches.sort()
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_mono_neutrons); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_k, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_v, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_mono_neutrons); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_14 = PyNumber_Subtract(__pyx_v_k, __pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_14);
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_v_k);
    __pyx_t_6 = 0;
    __pyx_t_9 = 0;
    __pyx_t_14 = 0;
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_8);
    __pyx_t_1 = 0;
    __pyx_t_8 = 0;
    __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_bunches, __pyx_t_14); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "brainpy/_c/composition.pyx":226
 * 
 *     bunches = []
 *     for k, v in freqs.items():             # <<<<<<<<<<<<<<
 *         bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "brainpy/_c/composition.pyx":229
 *         bunches.append((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k)))
 * 
 *     bunches.sort()             # <<<<<<<<<<<<<<
 * 
 *     # bunches = sorted([((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k))
 */
  __pyx_t_15 = PyList_Sort(__pyx_v_bunches); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 229, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":233
 *     # bunches = sorted([((k - mono_neutrons, (v[0], v[1], k - mono_neutrons, k))
 *     #                   for k, v in freqs.items())], key=lambda x: x[0])
 *     isotope_frequencies[0] = make_isotope_map(bunches, len(bunches))             # <<<<<<<<<<<<<<
 * 
 * cdef Element* make_element(char* symbol):
 */
  __pyx_t_10 = PyList_GET_SIZE(__pyx_v_bunches); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 233, __pyx_L1_error)
  (__pyx_v_isotope_frequencies[0]) = __pyx_f_7brainpy_2_c_11composition_make_isotope_map(__pyx_v_bunches, __pyx_t_10);

  /* "brainpy/_c/composition.pyx":190
 *     return x[1][1]
 * 
 * cdef void _isotopes_of(char* element_symbol, IsotopeMap** isotope_frequencies):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict freqs, element_data
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_WriteUnraisable("brainpy._c.composition._isotopes_of", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_freqs);
  __Pyx_XDECREF(__pyx_v_element_data);
  __Pyx_XDECREF(__pyx_v_freq_list);
  __Pyx_XDECREF(__pyx_v_bunches);
  __Pyx_XDECREF(__pyx_v_mass_freqs);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_py_element_symbol);
  __Pyx_RefNannyFinishContext();
}

/* "brainpy/_c/composition.pyx":235
 *     isotope_frequencies[0] = make_isotope_map(bunches, len(bunches))
 * 
 * cdef Element* make_element(char* symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element* element
 */

static struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_f_7brainpy_2_c_11composition_make_element(char *__pyx_v_symbol) {
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_element", 0);

  /* "brainpy/_c/composition.pyx":238
 *     cdef:
 *         Element* element
 *     element = <Element*>malloc(sizeof(Element))             # <<<<<<<<<<<<<<
 *     element.symbol = symbol
 *     _isotopes_of(symbol, &element.isotopes)
 */
  __pyx_v_element = ((struct __pyx_t_7brainpy_2_c_11composition_Element *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element))));

  /* "brainpy/_c/composition.pyx":239
 *         Element* element
 *     element = <Element*>malloc(sizeof(Element))
 *     element.symbol = symbol             # <<<<<<<<<<<<<<
 *     _isotopes_of(symbol, &element.isotopes)
 *     element.monoisotopic_isotope_index = -1
 */
  __pyx_v_element->symbol = __pyx_v_symbol;

  /* "brainpy/_c/composition.pyx":240
 *     element = <Element*>malloc(sizeof(Element))
 *     element.symbol = symbol
 *     _isotopes_of(symbol, &element.isotopes)             # <<<<<<<<<<<<<<
 *     element.monoisotopic_isotope_index = -1
 *     element.monoisotopic_isotope_index = get_isotope_by_neutron_shift_index(
 */
  __pyx_f_7brainpy_2_c_11composition__isotopes_of(__pyx_v_symbol, (&__pyx_v_element->isotopes));

  /* "brainpy/_c/composition.pyx":241
 *     element.symbol = symbol
 *     _isotopes_of(symbol, &element.isotopes)
 *     element.monoisotopic_isotope_index = -1             # <<<<<<<<<<<<<<
 *     element.monoisotopic_isotope_index = get_isotope_by_neutron_shift_index(
 *         element.isotopes, 0)
 */
  __pyx_v_element->monoisotopic_isotope_index = -1;

  /* "brainpy/_c/composition.pyx":242
 *     _isotopes_of(symbol, &element.isotopes)
 *     element.monoisotopic_isotope_index = -1
 *     element.monoisotopic_isotope_index = get_isotope_by_neutron_shift_index(             # <<<<<<<<<<<<<<
 *         element.isotopes, 0)
 *     return element
 */
  __pyx_v_element->monoisotopic_isotope_index = __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift_index(__pyx_v_element->isotopes, 0);

  /* "brainpy/_c/composition.pyx":244
 *     element.monoisotopic_isotope_index = get_isotope_by_neutron_shift_index(
 *         element.isotopes, 0)
 *     return element             # <<<<<<<<<<<<<<
 * 
 * cdef void free_element(Element* element) nogil:
 */
  __pyx_r = __pyx_v_element;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":235
 *     isotope_frequencies[0] = make_isotope_map(bunches, len(bunches))
 * 
 * cdef Element* make_element(char* symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element* element
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":246
 *     return element
 * 
 * cdef void free_element(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     free_isotope_map(element.isotopes)
 *     free(element)
 */

static void __pyx_f_7brainpy_2_c_11composition_free_element(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {

  /* "brainpy/_c/composition.pyx":247
 * 
 * cdef void free_element(Element* element) nogil:
 *     free_isotope_map(element.isotopes)             # <<<<<<<<<<<<<<
 *     free(element)
 * 
 */
  __pyx_f_7brainpy_2_c_11composition_free_isotope_map(__pyx_v_element->isotopes);

  /* "brainpy/_c/composition.pyx":248
 * cdef void free_element(Element* element) nogil:
 *     free_isotope_map(element.isotopes)
 *     free(element)             # <<<<<<<<<<<<<<
 * 
 * cdef Element* make_fixed_isotope_element(Element* element, int neutron_count) nogil:
 */
  free(__pyx_v_element);

  /* "brainpy/_c/composition.pyx":246
 *     return element
 * 
 * cdef void free_element(Element* element) nogil:             # <<<<<<<<<<<<<<
 *     free_isotope_map(element.isotopes)
 *     free(element)
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":250
 *     free(element)
 * 
 * cdef Element* make_fixed_isotope_element(Element* element, int neutron_count) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i
 */

static struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element, int __pyx_v_neutron_count) {
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_out;
  struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *__pyx_v_isotope_map;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_reference;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_r;
  int __pyx_t_1;
  double __pyx_t_2;

  /* "brainpy/_c/composition.pyx":258
 *         Isotope* reference
 *         char* symbol
 *     out = <Element*>malloc(sizeof(Element))             # <<<<<<<<<<<<<<
 *     isotope_map = <IsotopeMap*>malloc(sizeof(IsotopeMap))
 *     isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)
 */
  __pyx_v_out = ((struct __pyx_t_7brainpy_2_c_11composition_Element *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element))));

  /* "brainpy/_c/composition.pyx":259
 *         char* symbol
 *     out = <Element*>malloc(sizeof(Element))
 *     isotope_map = <IsotopeMap*>malloc(sizeof(IsotopeMap))             # <<<<<<<<<<<<<<
 *     isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)
 *     isotope_map.size = 1
 */
  __pyx_v_isotope_map = ((struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap))));

  /* "brainpy/_c/composition.pyx":260
 *     out = <Element*>malloc(sizeof(Element))
 *     isotope_map = <IsotopeMap*>malloc(sizeof(IsotopeMap))
 *     isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)             # <<<<<<<<<<<<<<
 *     isotope_map.size = 1
 *     out.isotopes = isotope_map
 */
  __pyx_v_isotope_map->bins = ((struct __pyx_t_7brainpy_2_c_11composition_Isotope *)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Isotope)) * 1)));

  /* "brainpy/_c/composition.pyx":261
 *     isotope_map = <IsotopeMap*>malloc(sizeof(IsotopeMap))
 *     isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)
 *     isotope_map.size = 1             # <<<<<<<<<<<<<<
 *     out.isotopes = isotope_map
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
 */
  __pyx_v_isotope_map->size = 1;

  /* "brainpy/_c/composition.pyx":262
 *     isotope_map.bins = <Isotope*>malloc(sizeof(Isotope) * 1)
 *     isotope_map.size = 1
 *     out.isotopes = isotope_map             # <<<<<<<<<<<<<<
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
 *     if reference == NULL:
 */
  __pyx_v_out->isotopes = __pyx_v_isotope_map;

  /* "brainpy/_c/composition.pyx":263
 *     isotope_map.size = 1
 *     out.isotopes = isotope_map
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)             # <<<<<<<<<<<<<<
 *     if reference == NULL:
 *         printf("reference was NULL!\n")
 */
  __pyx_v_reference = __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_count(__pyx_v_element->isotopes, __pyx_v_neutron_count);

  /* "brainpy/_c/composition.pyx":264
 *     out.isotopes = isotope_map
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
 *     if reference == NULL:             # <<<<<<<<<<<<<<
 *         printf("reference was NULL!\n")
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_reference == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":265
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
 *     if reference == NULL:
 *         printf("reference was NULL!\n")             # <<<<<<<<<<<<<<
 *         return NULL
 *     isotope_map.bins[0].mass = reference.mass
 */
    (void)(printf(((char const *)"reference was NULL!\n")));

    /* "brainpy/_c/composition.pyx":266
 *     if reference == NULL:
 *         printf("reference was NULL!\n")
 *         return NULL             # <<<<<<<<<<<<<<
 *     isotope_map.bins[0].mass = reference.mass
 *     isotope_map.bins[0].abundance = 1.0
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":264
 *     out.isotopes = isotope_map
 *     reference = get_isotope_by_neutron_count(element.isotopes, neutron_count)
 *     if reference == NULL:             # <<<<<<<<<<<<<<
 *         printf("reference was NULL!\n")
 *         return NULL
 */
  }

  /* "brainpy/_c/composition.pyx":267
 *         printf("reference was NULL!\n")
 *         return NULL
 *     isotope_map.bins[0].mass = reference.mass             # <<<<<<<<<<<<<<
 *     isotope_map.bins[0].abundance = 1.0
 *     isotope_map.bins[0].neutron_shift = 0
 */
  __pyx_t_2 = __pyx_v_reference->mass;
  (__pyx_v_isotope_map->bins[0]).mass = __pyx_t_2;

  /* "brainpy/_c/composition.pyx":268
 *         return NULL
 *     isotope_map.bins[0].mass = reference.mass
 *     isotope_map.bins[0].abundance = 1.0             # <<<<<<<<<<<<<<
 *     isotope_map.bins[0].neutron_shift = 0
 *     isotope_map.bins[0].neutrons = neutron_count
 */
  (__pyx_v_isotope_map->bins[0]).abundance = 1.0;

  /* "brainpy/_c/composition.pyx":269
 *     isotope_map.bins[0].mass = reference.mass
 *     isotope_map.bins[0].abundance = 1.0
 *     isotope_map.bins[0].neutron_shift = 0             # <<<<<<<<<<<<<<
 *     isotope_map.bins[0].neutrons = neutron_count
 *     out.symbol = <char*>malloc(sizeof(char) * 10)
 */
  (__pyx_v_isotope_map->bins[0]).neutron_shift = 0;

  /* "brainpy/_c/composition.pyx":270
 *     isotope_map.bins[0].abundance = 1.0
 *     isotope_map.bins[0].neutron_shift = 0
 *     isotope_map.bins[0].neutrons = neutron_count             # <<<<<<<<<<<<<<
 *     out.symbol = <char*>malloc(sizeof(char) * 10)
 *     _make_fixed_isotope_string(element, reference, out.symbol)
 */
  (__pyx_v_isotope_map->bins[0]).neutrons = __pyx_v_neutron_count;

  /* "brainpy/_c/composition.pyx":271
 *     isotope_map.bins[0].neutron_shift = 0
 *     isotope_map.bins[0].neutrons = neutron_count
 *     out.symbol = <char*>malloc(sizeof(char) * 10)             # <<<<<<<<<<<<<<
 *     _make_fixed_isotope_string(element, reference, out.symbol)
 *     out.monoisotopic_isotope_index = 0
 */
  __pyx_v_out->symbol = ((char *)malloc(((sizeof(char)) * 10)));

  /* "brainpy/_c/composition.pyx":272
 *     isotope_map.bins[0].neutrons = neutron_count
 *     out.symbol = <char*>malloc(sizeof(char) * 10)
 *     _make_fixed_isotope_string(element, reference, out.symbol)             # <<<<<<<<<<<<<<
 *     out.monoisotopic_isotope_index = 0
 *     return out
 */
  (void)(__pyx_f_7brainpy_2_c_11composition__make_fixed_isotope_string(__pyx_v_element, __pyx_v_reference, __pyx_v_out->symbol));

  /* "brainpy/_c/composition.pyx":273
 *     out.symbol = <char*>malloc(sizeof(char) * 10)
 *     _make_fixed_isotope_string(element, reference, out.symbol)
 *     out.monoisotopic_isotope_index = 0             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
  __pyx_v_out->monoisotopic_isotope_index = 0;

  /* "brainpy/_c/composition.pyx":274
 *     _make_fixed_isotope_string(element, reference, out.symbol)
 *     out.monoisotopic_isotope_index = 0
 *     return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":250
 *     free(element)
 * 
 * cdef Element* make_fixed_isotope_element(Element* element, int neutron_count) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":280
 * # ElementHashTable and ElementHashBucket Methods
 * 
 * cdef size_t hash_string(char *string_) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

static size_t __pyx_f_7brainpy_2_c_11composition_hash_string(char *__pyx_v_string_) {
  size_t __pyx_v_hash_value;
  size_t __pyx_v_i;
  int __pyx_v_c;
  size_t __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":285
 *         size_t i
 *         int c
 *     hash_value = 5381             # <<<<<<<<<<<<<<
 *     i = 0
 *     c = string_[i]
 */
  __pyx_v_hash_value = 0x1505;

  /* "brainpy/_c/composition.pyx":286
 *         int c
 *     hash_value = 5381
 *     i = 0             # <<<<<<<<<<<<<<
 *     c = string_[i]
 *     while (c):
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":287
 *     hash_value = 5381
 *     i = 0
 *     c = string_[i]             # <<<<<<<<<<<<<<
 *     while (c):
 *         hash_value = ((hash_value << 5) + hash_value) + c
 */
  __pyx_v_c = (__pyx_v_string_[__pyx_v_i]);

  /* "brainpy/_c/composition.pyx":288
 *     i = 0
 *     c = string_[i]
 *     while (c):             # <<<<<<<<<<<<<<
 *         hash_value = ((hash_value << 5) + hash_value) + c
 *         i += 1
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_c != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":289
 *     c = string_[i]
 *     while (c):
 *         hash_value = ((hash_value << 5) + hash_value) + c             # <<<<<<<<<<<<<<
 *         i += 1
 *         c = string_[i]
 */
    __pyx_v_hash_value = (((__pyx_v_hash_value << 5) + __pyx_v_hash_value) + __pyx_v_c);

    /* "brainpy/_c/composition.pyx":290
 *     while (c):
 *         hash_value = ((hash_value << 5) + hash_value) + c
 *         i += 1             # <<<<<<<<<<<<<<
 *         c = string_[i]
 *     return hash_value;
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":291
 *         hash_value = ((hash_value << 5) + hash_value) + c
 *         i += 1
 *         c = string_[i]             # <<<<<<<<<<<<<<
 *     return hash_value;
 * 
 */
    __pyx_v_c = (__pyx_v_string_[__pyx_v_i]);
  }

  /* "brainpy/_c/composition.pyx":292
 *         i += 1
 *         c = string_[i]
 *     return hash_value;             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_hash_value;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":280
 * # ElementHashTable and ElementHashBucket Methods
 * 
 * cdef size_t hash_string(char *string_) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":295
 * 
 * 
 * def test_hash_string(str string_):             # <<<<<<<<<<<<<<
 *     return hash_string(PyString_AsString(string_))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_3test_hash_string(PyObject *__pyx_self, PyObject *__pyx_v_string_); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_2test_hash_string[] = "test_hash_string(str string_)";
static PyMethodDef __pyx_mdef_7brainpy_2_c_11composition_3test_hash_string = {"test_hash_string", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_3test_hash_string, METH_O, __pyx_doc_7brainpy_2_c_11composition_2test_hash_string};
static PyObject *__pyx_pw_7brainpy_2_c_11composition_3test_hash_string(PyObject *__pyx_self, PyObject *__pyx_v_string_) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("test_hash_string (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_string_), (&PyString_Type), 1, "string_", 1))) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_2test_hash_string(__pyx_self, ((PyObject*)__pyx_v_string_));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_2test_hash_string(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string_) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("test_hash_string", 0);

  /* "brainpy/_c/composition.pyx":296
 * 
 * def test_hash_string(str string_):
 *     return hash_string(PyString_AsString(string_))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_7brainpy_2_c_11composition_hash_string(PyStr_AsString(__pyx_v_string_))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":295
 * 
 * 
 * def test_hash_string(str string_):             # <<<<<<<<<<<<<<
 *     return hash_string(PyString_AsString(string_))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.test_hash_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":299
 * 
 * 
 * cdef ElementHashTable* make_element_hash_table(size_t size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         ElementHashTable* table
 */

static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_f_7brainpy_2_c_11composition_make_element_hash_table(size_t __pyx_v_size) {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;

  /* "brainpy/_c/composition.pyx":304
 *         size_t i
 * 
 *     table = <ElementHashTable*>malloc(sizeof(ElementHashTable))             # <<<<<<<<<<<<<<
 *     table.buckets = <ElementHashBucket*>malloc(sizeof(ElementHashBucket) * size)
 *     table.size = size
 */
  __pyx_v_table = ((struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable))));

  /* "brainpy/_c/composition.pyx":305
 * 
 *     table = <ElementHashTable*>malloc(sizeof(ElementHashTable))
 *     table.buckets = <ElementHashBucket*>malloc(sizeof(ElementHashBucket) * size)             # <<<<<<<<<<<<<<
 *     table.size = size
 * 
 */
  __pyx_v_table->buckets = ((struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket)) * __pyx_v_size)));

  /* "brainpy/_c/composition.pyx":306
 *     table = <ElementHashTable*>malloc(sizeof(ElementHashTable))
 *     table.buckets = <ElementHashBucket*>malloc(sizeof(ElementHashBucket) * size)
 *     table.size = size             # <<<<<<<<<<<<<<
 * 
 *     for i in range(size):
 */
  __pyx_v_table->size = __pyx_v_size;

  /* "brainpy/_c/composition.pyx":308
 *     table.size = size
 * 
 *     for i in range(size):             # <<<<<<<<<<<<<<
 *         table.buckets[i].size = 6
 *         table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)
 */
  __pyx_t_1 = __pyx_v_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":309
 * 
 *     for i in range(size):
 *         table.buckets[i].size = 6             # <<<<<<<<<<<<<<
 *         table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)
 *         table.buckets[i].used = 0
 */
    (__pyx_v_table->buckets[__pyx_v_i]).size = 6;

    /* "brainpy/_c/composition.pyx":310
 *     for i in range(size):
 *         table.buckets[i].size = 6
 *         table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)             # <<<<<<<<<<<<<<
 *         table.buckets[i].used = 0
 *     return table
 */
    (__pyx_v_table->buckets[__pyx_v_i]).elements = ((struct __pyx_t_7brainpy_2_c_11composition_Element **)malloc(((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element *)) * (__pyx_v_table->buckets[__pyx_v_i]).size)));

    /* "brainpy/_c/composition.pyx":311
 *         table.buckets[i].size = 6
 *         table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)
 *         table.buckets[i].used = 0             # <<<<<<<<<<<<<<
 *     return table
 * 
 */
    (__pyx_v_table->buckets[__pyx_v_i]).used = 0;
  }

  /* "brainpy/_c/composition.pyx":312
 *         table.buckets[i].elements = <Element**>malloc(sizeof(Element*) * table.buckets[i].size)
 *         table.buckets[i].used = 0
 *     return table             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_table;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":299
 * 
 * 
 * cdef ElementHashTable* make_element_hash_table(size_t size) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         ElementHashTable* table
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":315
 * 
 * 
 * cdef int element_hash_bucket_resize(ElementHashBucket* bucket) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element** elements
 */

static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_resize(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *__pyx_v_bucket) {
  struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_elements;
  size_t __pyx_v_new_size;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":320
 *         size_t new_size
 * 
 *     new_size = bucket.size * 2             # <<<<<<<<<<<<<<
 *     elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)
 *     if elements == NULL:
 */
  __pyx_v_new_size = (__pyx_v_bucket->size * 2);

  /* "brainpy/_c/composition.pyx":321
 * 
 *     new_size = bucket.size * 2
 *     elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)             # <<<<<<<<<<<<<<
 *     if elements == NULL:
 *         printf("element_hash_bucket_resize failed\n")
 */
  __pyx_v_elements = ((struct __pyx_t_7brainpy_2_c_11composition_Element **)realloc(__pyx_v_bucket->elements, ((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Element *)) * __pyx_v_new_size)));

  /* "brainpy/_c/composition.pyx":322
 *     new_size = bucket.size * 2
 *     elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)
 *     if elements == NULL:             # <<<<<<<<<<<<<<
 *         printf("element_hash_bucket_resize failed\n")
 *         return -1
 */
  __pyx_t_1 = ((__pyx_v_elements == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":323
 *     elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)
 *     if elements == NULL:
 *         printf("element_hash_bucket_resize failed\n")             # <<<<<<<<<<<<<<
 *         return -1
 *     bucket.elements = elements
 */
    (void)(printf(((char const *)"element_hash_bucket_resize failed\n")));

    /* "brainpy/_c/composition.pyx":324
 *     if elements == NULL:
 *         printf("element_hash_bucket_resize failed\n")
 *         return -1             # <<<<<<<<<<<<<<
 *     bucket.elements = elements
 *     return 0
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":322
 *     new_size = bucket.size * 2
 *     elements = <Element**>realloc(bucket.elements, sizeof(Element*) * new_size)
 *     if elements == NULL:             # <<<<<<<<<<<<<<
 *         printf("element_hash_bucket_resize failed\n")
 *         return -1
 */
  }

  /* "brainpy/_c/composition.pyx":325
 *         printf("element_hash_bucket_resize failed\n")
 *         return -1
 *     bucket.elements = elements             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_bucket->elements = __pyx_v_elements;

  /* "brainpy/_c/composition.pyx":326
 *         return -1
 *     bucket.elements = elements
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":315
 * 
 * 
 * cdef int element_hash_bucket_resize(ElementHashBucket* bucket) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element** elements
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":329
 * 
 * 
 * cdef int element_hash_bucket_insert(ElementHashBucket* bucket, Element* element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_insert(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *__pyx_v_bucket, struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {
  int __pyx_v_status;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":334
 *         int status
 *         Element* el
 *     if (bucket.used + 1) == bucket.size:             # <<<<<<<<<<<<<<
 *         status = element_hash_bucket_resize(bucket)
 *         if status != 0:
 */
  __pyx_t_1 = (((__pyx_v_bucket->used + 1) == __pyx_v_bucket->size) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":335
 *         Element* el
 *     if (bucket.used + 1) == bucket.size:
 *         status = element_hash_bucket_resize(bucket)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             printf("element_hash_bucket_insert failed with %s\n", element.symbol)
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_resize(__pyx_v_bucket);

    /* "brainpy/_c/composition.pyx":336
 *     if (bucket.used + 1) == bucket.size:
 *         status = element_hash_bucket_resize(bucket)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_bucket_insert failed with %s\n", element.symbol)
 *             return -1
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":337
 *         status = element_hash_bucket_resize(bucket)
 *         if status != 0:
 *             printf("element_hash_bucket_insert failed with %s\n", element.symbol)             # <<<<<<<<<<<<<<
 *             return -1
 *     # printf("Inserting %s into bucket with total capacity %d\n", element.symbol, bucket.size)
 */
      (void)(printf(((char const *)"element_hash_bucket_insert failed with %s\n"), __pyx_v_element->symbol));

      /* "brainpy/_c/composition.pyx":338
 *         if status != 0:
 *             printf("element_hash_bucket_insert failed with %s\n", element.symbol)
 *             return -1             # <<<<<<<<<<<<<<
 *     # printf("Inserting %s into bucket with total capacity %d\n", element.symbol, bucket.size)
 *     bucket.elements[bucket.used] = element
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":336
 *     if (bucket.used + 1) == bucket.size:
 *         status = element_hash_bucket_resize(bucket)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_bucket_insert failed with %s\n", element.symbol)
 *             return -1
 */
    }

    /* "brainpy/_c/composition.pyx":334
 *         int status
 *         Element* el
 *     if (bucket.used + 1) == bucket.size:             # <<<<<<<<<<<<<<
 *         status = element_hash_bucket_resize(bucket)
 *         if status != 0:
 */
  }

  /* "brainpy/_c/composition.pyx":340
 *             return -1
 *     # printf("Inserting %s into bucket with total capacity %d\n", element.symbol, bucket.size)
 *     bucket.elements[bucket.used] = element             # <<<<<<<<<<<<<<
 *     # printf("Element %s put in slot %d\n", bucket.elements[bucket.used].symbol, bucket.used)
 *     bucket.used += 1
 */
  (__pyx_v_bucket->elements[__pyx_v_bucket->used]) = __pyx_v_element;

  /* "brainpy/_c/composition.pyx":342
 *     bucket.elements[bucket.used] = element
 *     # printf("Element %s put in slot %d\n", bucket.elements[bucket.used].symbol, bucket.used)
 *     bucket.used += 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_bucket->used = (__pyx_v_bucket->used + 1);

  /* "brainpy/_c/composition.pyx":343
 *     # printf("Element %s put in slot %d\n", bucket.elements[bucket.used].symbol, bucket.used)
 *     bucket.used += 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":329
 * 
 * 
 * cdef int element_hash_bucket_insert(ElementHashBucket* bucket, Element* element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":346
 * 
 * 
 * cdef int element_hash_bucket_find(ElementHashBucket* bucket, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static int __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_find(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *__pyx_v_bucket, char *__pyx_v_symbol, struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_out) {
  size_t __pyx_v_i;
  int __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;

  /* "brainpy/_c/composition.pyx":349
 *     cdef:
 *         size_t i
 *     for i in range(bucket.used):             # <<<<<<<<<<<<<<
 *         if strcmp(bucket.elements[i].symbol, symbol) == 0:
 *             out[0] = bucket.elements[i]
 */
  __pyx_t_1 = __pyx_v_bucket->used;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":350
 *         size_t i
 *     for i in range(bucket.used):
 *         if strcmp(bucket.elements[i].symbol, symbol) == 0:             # <<<<<<<<<<<<<<
 *             out[0] = bucket.elements[i]
 *             return 0
 */
    __pyx_t_4 = ((strcmp((__pyx_v_bucket->elements[__pyx_v_i])->symbol, __pyx_v_symbol) == 0) != 0);
    if (__pyx_t_4) {

      /* "brainpy/_c/composition.pyx":351
 *     for i in range(bucket.used):
 *         if strcmp(bucket.elements[i].symbol, symbol) == 0:
 *             out[0] = bucket.elements[i]             # <<<<<<<<<<<<<<
 *             return 0
 *     return -1
 */
      (__pyx_v_out[0]) = (__pyx_v_bucket->elements[__pyx_v_i]);

      /* "brainpy/_c/composition.pyx":352
 *         if strcmp(bucket.elements[i].symbol, symbol) == 0:
 *             out[0] = bucket.elements[i]
 *             return 0             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":350
 *         size_t i
 *     for i in range(bucket.used):
 *         if strcmp(bucket.elements[i].symbol, symbol) == 0:             # <<<<<<<<<<<<<<
 *             out[0] = bucket.elements[i]
 *             return 0
 */
    }
  }

  /* "brainpy/_c/composition.pyx":353
 *             out[0] = bucket.elements[i]
 *             return 0
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = -1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":346
 * 
 * 
 * cdef int element_hash_bucket_find(ElementHashBucket* bucket, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":356
 * 
 * 
 * cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil:             # <<<<<<<<<<<<<<
 *     free(bucket.elements)
 *     free(bucket)
 */

static void __pyx_f_7brainpy_2_c_11composition_free_element_hash_bucket(struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *__pyx_v_bucket) {

  /* "brainpy/_c/composition.pyx":357
 * 
 * cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil:
 *     free(bucket.elements)             # <<<<<<<<<<<<<<
 *     free(bucket)
 * 
 */
  free(__pyx_v_bucket->elements);

  /* "brainpy/_c/composition.pyx":358
 * cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil:
 *     free(bucket.elements)
 *     free(bucket)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_bucket);

  /* "brainpy/_c/composition.pyx":356
 * 
 * 
 * cdef void free_element_hash_bucket(ElementHashBucket* bucket) nogil:             # <<<<<<<<<<<<<<
 *     free(bucket.elements)
 *     free(bucket)
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":362
 * 
 * @cython.cdivision(True)
 * cdef int element_hash_table_get(ElementHashTable* table, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

static int __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table, char *__pyx_v_symbol, struct __pyx_t_7brainpy_2_c_11composition_Element **__pyx_v_out) {
  size_t __pyx_v_hash_value;
  size_t __pyx_v_position;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket __pyx_v_bucket;
  int __pyx_v_status;
  int __pyx_r;

  /* "brainpy/_c/composition.pyx":369
 *         int status
 * 
 *     hash_value = hash_string(symbol)             # <<<<<<<<<<<<<<
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])
 */
  __pyx_v_hash_value = __pyx_f_7brainpy_2_c_11composition_hash_string(__pyx_v_symbol);

  /* "brainpy/_c/composition.pyx":370
 * 
 *     hash_value = hash_string(symbol)
 *     position = hash_value % table.size             # <<<<<<<<<<<<<<
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_find(&bucket, symbol, out)
 */
  __pyx_v_position = (__pyx_v_hash_value % __pyx_v_table->size);

  /* "brainpy/_c/composition.pyx":371
 *     hash_value = hash_string(symbol)
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])             # <<<<<<<<<<<<<<
 *     status = element_hash_bucket_find(&bucket, symbol, out)
 *     return status
 */
  __pyx_v_bucket = (__pyx_v_table->buckets[__pyx_v_position]);

  /* "brainpy/_c/composition.pyx":372
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_find(&bucket, symbol, out)             # <<<<<<<<<<<<<<
 *     return status
 * 
 */
  __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_find((&__pyx_v_bucket), __pyx_v_symbol, __pyx_v_out);

  /* "brainpy/_c/composition.pyx":373
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_find(&bucket, symbol, out)
 *     return status             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_status;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":362
 * 
 * @cython.cdivision(True)
 * cdef int element_hash_table_get(ElementHashTable* table, char* symbol, Element** out) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":377
 * 
 * @cython.cdivision(True)
 * cdef int element_hash_table_put(ElementHashTable* table, Element* element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

static int __pyx_f_7brainpy_2_c_11composition_element_hash_table_put(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table, struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element) {
  size_t __pyx_v_hash_value;
  size_t __pyx_v_position;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket __pyx_v_bucket;
  int __pyx_v_status;
  int __pyx_r;

  /* "brainpy/_c/composition.pyx":384
 *         int status
 * 
 *     hash_value = hash_string(element.symbol)             # <<<<<<<<<<<<<<
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])
 */
  __pyx_v_hash_value = __pyx_f_7brainpy_2_c_11composition_hash_string(__pyx_v_element->symbol);

  /* "brainpy/_c/composition.pyx":385
 * 
 *     hash_value = hash_string(element.symbol)
 *     position = hash_value % table.size             # <<<<<<<<<<<<<<
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_insert(&bucket, element)
 */
  __pyx_v_position = (__pyx_v_hash_value % __pyx_v_table->size);

  /* "brainpy/_c/composition.pyx":386
 *     hash_value = hash_string(element.symbol)
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])             # <<<<<<<<<<<<<<
 *     status = element_hash_bucket_insert(&bucket, element)
 *     table.buckets[position] = bucket
 */
  __pyx_v_bucket = (__pyx_v_table->buckets[__pyx_v_position]);

  /* "brainpy/_c/composition.pyx":387
 *     position = hash_value % table.size
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_insert(&bucket, element)             # <<<<<<<<<<<<<<
 *     table.buckets[position] = bucket
 *     return status
 */
  __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_bucket_insert((&__pyx_v_bucket), __pyx_v_element);

  /* "brainpy/_c/composition.pyx":388
 *     bucket = (table.buckets[position])
 *     status = element_hash_bucket_insert(&bucket, element)
 *     table.buckets[position] = bucket             # <<<<<<<<<<<<<<
 *     return status
 * 
 */
  (__pyx_v_table->buckets[__pyx_v_position]) = __pyx_v_bucket;

  /* "brainpy/_c/composition.pyx":389
 *     status = element_hash_bucket_insert(&bucket, element)
 *     table.buckets[position] = bucket
 *     return status             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_status;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":377
 * 
 * @cython.cdivision(True)
 * cdef int element_hash_table_put(ElementHashTable* table, Element* element) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t hash_value
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":392
 * 
 * 
 * cdef size_t free_element_hash_table(ElementHashTable* table) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static size_t __pyx_f_7brainpy_2_c_11composition_free_element_hash_table(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table) {
  size_t __pyx_v_i;
  size_t __pyx_r;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;

  /* "brainpy/_c/composition.pyx":396
 *         size_t i
 * 
 *     for i in range(table.size):             # <<<<<<<<<<<<<<
 *         free_element_hash_bucket(&(table.buckets[i]))
 *     free(table)
 */
  __pyx_t_1 = __pyx_v_table->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "brainpy/_c/composition.pyx":397
 * 
 *     for i in range(table.size):
 *         free_element_hash_bucket(&(table.buckets[i]))             # <<<<<<<<<<<<<<
 *     free(table)
 * 
 */
    __pyx_f_7brainpy_2_c_11composition_free_element_hash_bucket((&(__pyx_v_table->buckets[__pyx_v_i])));
  }

  /* "brainpy/_c/composition.pyx":398
 *     for i in range(table.size):
 *         free_element_hash_bucket(&(table.buckets[i]))
 *     free(table)             # <<<<<<<<<<<<<<
 * 
 * cdef ElementHashTable* make_element_hash_table_populated(size_t size):
 */
  free(__pyx_v_table);

  /* "brainpy/_c/composition.pyx":392
 * 
 * 
 * cdef size_t free_element_hash_table(ElementHashTable* table) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":400
 *     free(table)
 * 
 * cdef ElementHashTable* make_element_hash_table_populated(size_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element* element
 */

static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_f_7brainpy_2_c_11composition_make_element_hash_table_populated(size_t __pyx_v_size) {
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_out_test;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table;
  PyObject *__pyx_v_pk = 0;
  char *__pyx_v_k;
  int __pyx_v_status;
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("make_element_hash_table_populated", 0);

  /* "brainpy/_c/composition.pyx":410
 *         int status
 * 
 *     table = make_element_hash_table(size)             # <<<<<<<<<<<<<<
 * 
 *     for pk in nist_mass:
 */
  __pyx_v_table = __pyx_f_7brainpy_2_c_11composition_make_element_hash_table(__pyx_v_size);

  /* "brainpy/_c/composition.pyx":412
 *     table = make_element_hash_table(size)
 * 
 *     for pk in nist_mass:             # <<<<<<<<<<<<<<
 *         k = PyString_AsString(pk)
 *         element = make_element(k)
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_7brainpy_2_c_11composition_nist_mass == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 412, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_7brainpy_2_c_11composition_nist_mass, 1, ((PyObject *)NULL), (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(PyString_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_pk, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "brainpy/_c/composition.pyx":413
 * 
 *     for pk in nist_mass:
 *         k = PyString_AsString(pk)             # <<<<<<<<<<<<<<
 *         element = make_element(k)
 *         status = element_hash_table_put(table, element)
 */
    __pyx_v_k = PyStr_AsString(__pyx_v_pk);

    /* "brainpy/_c/composition.pyx":414
 *     for pk in nist_mass:
 *         k = PyString_AsString(pk)
 *         element = make_element(k)             # <<<<<<<<<<<<<<
 *         status = element_hash_table_put(table, element)
 *         if status != 0:
 */
    __pyx_v_element = __pyx_f_7brainpy_2_c_11composition_make_element(__pyx_v_k);

    /* "brainpy/_c/composition.pyx":415
 *         k = PyString_AsString(pk)
 *         element = make_element(k)
 *         status = element_hash_table_put(table, element)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             printf("element_hash_table_put exited with status %d\n", status)
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_put(__pyx_v_table, __pyx_v_element);

    /* "brainpy/_c/composition.pyx":416
 *         element = make_element(k)
 *         status = element_hash_table_put(table, element)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_table_put exited with status %d\n", status)
 *         status = element_hash_table_get(table, k, &out_test)
 */
    __pyx_t_7 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_7) {

      /* "brainpy/_c/composition.pyx":417
 *         status = element_hash_table_put(table, element)
 *         if status != 0:
 *             printf("element_hash_table_put exited with status %d\n", status)             # <<<<<<<<<<<<<<
 *         status = element_hash_table_get(table, k, &out_test)
 *         if status != 0:
 */
      (void)(printf(((char const *)"element_hash_table_put exited with status %d\n"), __pyx_v_status));

      /* "brainpy/_c/composition.pyx":416
 *         element = make_element(k)
 *         status = element_hash_table_put(table, element)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_table_put exited with status %d\n", status)
 *         status = element_hash_table_get(table, k, &out_test)
 */
    }

    /* "brainpy/_c/composition.pyx":418
 *         if status != 0:
 *             printf("element_hash_table_put exited with status %d\n", status)
 *         status = element_hash_table_get(table, k, &out_test)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             printf("element_hash_table_get exited with status %d\n", status)
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_table, __pyx_v_k, (&__pyx_v_out_test));

    /* "brainpy/_c/composition.pyx":419
 *             printf("element_hash_table_put exited with status %d\n", status)
 *         status = element_hash_table_get(table, k, &out_test)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_table_get exited with status %d\n", status)
 *     return table
 */
    __pyx_t_7 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_7) {

      /* "brainpy/_c/composition.pyx":420
 *         status = element_hash_table_get(table, k, &out_test)
 *         if status != 0:
 *             printf("element_hash_table_get exited with status %d\n", status)             # <<<<<<<<<<<<<<
 *     return table
 * 
 */
      (void)(printf(((char const *)"element_hash_table_get exited with status %d\n"), __pyx_v_status));

      /* "brainpy/_c/composition.pyx":419
 *             printf("element_hash_table_put exited with status %d\n", status)
 *         status = element_hash_table_get(table, k, &out_test)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("element_hash_table_get exited with status %d\n", status)
 *     return table
 */
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":421
 *         if status != 0:
 *             printf("element_hash_table_get exited with status %d\n", status)
 *     return table             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_table;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":400
 *     free(table)
 * 
 * cdef ElementHashTable* make_element_hash_table_populated(size_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Element* element
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("brainpy._c.composition.make_element_hash_table_populated", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pk);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":427
 * 
 * 
 * cdef ElementHashTable* get_system_element_hash_table() nogil:             # <<<<<<<<<<<<<<
 *     return _ElementTable
 * 
 */

static struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_f_7brainpy_2_c_11composition_get_system_element_hash_table(void) {
  struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_r;

  /* "brainpy/_c/composition.pyx":428
 * 
 * cdef ElementHashTable* get_system_element_hash_table() nogil:
 *     return _ElementTable             # <<<<<<<<<<<<<<
 * 
 * cdef int set_system_element_hash_table(ElementHashTable* table) nogil:
 */
  __pyx_r = __pyx_v_7brainpy_2_c_11composition__ElementTable;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":427
 * 
 * 
 * cdef ElementHashTable* get_system_element_hash_table() nogil:             # <<<<<<<<<<<<<<
 *     return _ElementTable
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":430
 *     return _ElementTable
 * 
 * cdef int set_system_element_hash_table(ElementHashTable* table) nogil:             # <<<<<<<<<<<<<<
 *     _ElementTable[0] = table[0]
 * 
 */

static int __pyx_f_7brainpy_2_c_11composition_set_system_element_hash_table(struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *__pyx_v_table) {
  int __pyx_r;

  /* "brainpy/_c/composition.pyx":431
 * 
 * cdef int set_system_element_hash_table(ElementHashTable* table) nogil:
 *     _ElementTable[0] = table[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_7brainpy_2_c_11composition__ElementTable[0]) = (__pyx_v_table[0]);

  /* "brainpy/_c/composition.pyx":430
 *     return _ElementTable
 * 
 * cdef int set_system_element_hash_table(ElementHashTable* table) nogil:             # <<<<<<<<<<<<<<
 *     _ElementTable[0] = table[0]
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":437
 * # Composition Methods
 * 
 * cdef Composition* make_composition() nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Create a new, empty Composition struct
 */

static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_make_composition(void) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_r;

  /* "brainpy/_c/composition.pyx":443
 *     cdef:
 *         Composition* composition
 *     composition = <Composition*>malloc(sizeof(Composition))             # <<<<<<<<<<<<<<
 *     composition.elements = <char**>calloc(7, sizeof(char*))
 *     composition.counts = <count_type*>calloc(7, sizeof(count_type))
 */
  __pyx_v_composition = ((struct __pyx_t_7brainpy_2_c_11composition_Composition *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Composition))));

  /* "brainpy/_c/composition.pyx":444
 *         Composition* composition
 *     composition = <Composition*>malloc(sizeof(Composition))
 *     composition.elements = <char**>calloc(7, sizeof(char*))             # <<<<<<<<<<<<<<
 *     composition.counts = <count_type*>calloc(7, sizeof(count_type))
 *     composition.size = 7
 */
  __pyx_v_composition->elements = ((char **)calloc(7, (sizeof(char *))));

  /* "brainpy/_c/composition.pyx":445
 *     composition = <Composition*>malloc(sizeof(Composition))
 *     composition.elements = <char**>calloc(7, sizeof(char*))
 *     composition.counts = <count_type*>calloc(7, sizeof(count_type))             # <<<<<<<<<<<<<<
 *     composition.size = 7
 *     composition.used = 0
 */
  __pyx_v_composition->counts = ((__pyx_t_7brainpy_2_c_11composition_count_type *)calloc(7, (sizeof(__pyx_t_7brainpy_2_c_11composition_count_type))));

  /* "brainpy/_c/composition.pyx":446
 *     composition.elements = <char**>calloc(7, sizeof(char*))
 *     composition.counts = <count_type*>calloc(7, sizeof(count_type))
 *     composition.size = 7             # <<<<<<<<<<<<<<
 *     composition.used = 0
 *     composition.max_variants = 0
 */
  __pyx_v_composition->size = 7;

  /* "brainpy/_c/composition.pyx":447
 *     composition.counts = <count_type*>calloc(7, sizeof(count_type))
 *     composition.size = 7
 *     composition.used = 0             # <<<<<<<<<<<<<<
 *     composition.max_variants = 0
 *     return composition
 */
  __pyx_v_composition->used = 0;

  /* "brainpy/_c/composition.pyx":448
 *     composition.size = 7
 *     composition.used = 0
 *     composition.max_variants = 0             # <<<<<<<<<<<<<<
 *     return composition
 * 
 */
  __pyx_v_composition->max_variants = 0;

  /* "brainpy/_c/composition.pyx":449
 *     composition.used = 0
 *     composition.max_variants = 0
 *     return composition             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_eq(Composition* composition_1, Composition* composition_2) nogil:
 */
  __pyx_r = __pyx_v_composition;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":437
 * # Composition Methods
 * 
 * cdef Composition* make_composition() nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Create a new, empty Composition struct
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":451
 *     return composition
 * 
 * cdef int composition_eq(Composition* composition_1, Composition* composition_2) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Test two Composition instances for element-wise equality
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_eq(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_1, struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_2) {
  int __pyx_v_status;
  size_t __pyx_v_i;
  char *__pyx_v_symbol;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value_1;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value_2;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":461
 *         count_type value_1, value_2
 * 
 *     if composition_1.used != composition_2.used:             # <<<<<<<<<<<<<<
 *         return 0
 *     i = 0
 */
  __pyx_t_1 = ((__pyx_v_composition_1->used != __pyx_v_composition_2->used) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":462
 * 
 *     if composition_1.used != composition_2.used:
 *         return 0             # <<<<<<<<<<<<<<
 *     i = 0
 *     while i < composition_2.used:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":461
 *         count_type value_1, value_2
 * 
 *     if composition_1.used != composition_2.used:             # <<<<<<<<<<<<<<
 *         return 0
 *     i = 0
 */
  }

  /* "brainpy/_c/composition.pyx":463
 *     if composition_1.used != composition_2.used:
 *         return 0
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":464
 *         return 0
 *     i = 0
 *     while i < composition_2.used:             # <<<<<<<<<<<<<<
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value_2)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition_2->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":465
 *     i = 0
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]             # <<<<<<<<<<<<<<
 *         status = composition_get_element_count(composition_2, symbol, &value_2)
 *         if status != 0:
 */
    __pyx_v_symbol = (__pyx_v_composition_2->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":466
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value_2)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 0
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition_2, __pyx_v_symbol, (&__pyx_v_value_2));

    /* "brainpy/_c/composition.pyx":467
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value_2)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":468
 *         status = composition_get_element_count(composition_2, symbol, &value_2)
 *         if status != 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 *         if status != 0:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":467
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value_2)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 */
    }

    /* "brainpy/_c/composition.pyx":469
 *         if status != 0:
 *             return 0
 *         status = composition_get_element_count(composition_1, symbol, &value_1)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             return 0
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition_1, __pyx_v_symbol, (&__pyx_v_value_1));

    /* "brainpy/_c/composition.pyx":470
 *             return 0
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if value_1 != value_2:
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":471
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 *         if status != 0:
 *             return 0             # <<<<<<<<<<<<<<
 *         if value_1 != value_2:
 *             return 0
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":470
 *             return 0
 *         status = composition_get_element_count(composition_1, symbol, &value_1)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             return 0
 *         if value_1 != value_2:
 */
    }

    /* "brainpy/_c/composition.pyx":472
 *         if status != 0:
 *             return 0
 *         if value_1 != value_2:             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
    __pyx_t_1 = ((__pyx_v_value_1 != __pyx_v_value_2) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":473
 *             return 0
 *         if value_1 != value_2:
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":472
 *         if status != 0:
 *             return 0
 *         if value_1 != value_2:             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
    }

    /* "brainpy/_c/composition.pyx":474
 *         if value_1 != value_2:
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":476
 *         i += 1
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef Composition* copy_composition(Composition* composition) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":451
 *     return composition
 * 
 * cdef int composition_eq(Composition* composition_1, Composition* composition_2) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Test two Composition instances for element-wise equality
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":478
 *     return 1
 * 
 * cdef Composition* copy_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Create a new Composition instance whose element counts are copied from
 */

static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_copy_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_result;
  CYTHON_UNUSED int __pyx_v_status;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;

  /* "brainpy/_c/composition.pyx":487
 *         int status
 *         size_t i
 *     result = <Composition*>malloc(sizeof(Composition))             # <<<<<<<<<<<<<<
 *     result.elements = <char**>malloc(sizeof(char*) * composition.size)
 *     result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)
 */
  __pyx_v_result = ((struct __pyx_t_7brainpy_2_c_11composition_Composition *)malloc((sizeof(struct __pyx_t_7brainpy_2_c_11composition_Composition))));

  /* "brainpy/_c/composition.pyx":488
 *         size_t i
 *     result = <Composition*>malloc(sizeof(Composition))
 *     result.elements = <char**>malloc(sizeof(char*) * composition.size)             # <<<<<<<<<<<<<<
 *     result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)
 *     result.size = composition.size
 */
  __pyx_v_result->elements = ((char **)malloc(((sizeof(char *)) * __pyx_v_composition->size)));

  /* "brainpy/_c/composition.pyx":489
 *     result = <Composition*>malloc(sizeof(Composition))
 *     result.elements = <char**>malloc(sizeof(char*) * composition.size)
 *     result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)             # <<<<<<<<<<<<<<
 *     result.size = composition.size
 *     result.used = 0
 */
  __pyx_v_result->counts = ((__pyx_t_7brainpy_2_c_11composition_count_type *)malloc(((sizeof(__pyx_t_7brainpy_2_c_11composition_count_type)) * __pyx_v_composition->size)));

  /* "brainpy/_c/composition.pyx":490
 *     result.elements = <char**>malloc(sizeof(char*) * composition.size)
 *     result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)
 *     result.size = composition.size             # <<<<<<<<<<<<<<
 *     result.used = 0
 * 
 */
  __pyx_t_1 = __pyx_v_composition->size;
  __pyx_v_result->size = __pyx_t_1;

  /* "brainpy/_c/composition.pyx":491
 *     result.counts = <count_type*>malloc(sizeof(count_type) * composition.size)
 *     result.size = composition.size
 *     result.used = 0             # <<<<<<<<<<<<<<
 * 
 *     i = 0
 */
  __pyx_v_result->used = 0;

  /* "brainpy/_c/composition.pyx":493
 *     result.used = 0
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < composition.used:
 *         status = composition_set_element_count(result, composition.elements[i], composition.counts[i])
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":494
 * 
 *     i = 0
 *     while i < composition.used:             # <<<<<<<<<<<<<<
 *         status = composition_set_element_count(result, composition.elements[i], composition.counts[i])
 *         i += 1
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/composition.pyx":495
 *     i = 0
 *     while i < composition.used:
 *         status = composition_set_element_count(result, composition.elements[i], composition.counts[i])             # <<<<<<<<<<<<<<
 *         i += 1
 *     return result
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_result, (__pyx_v_composition->elements[__pyx_v_i]), (__pyx_v_composition->counts[__pyx_v_i]));

    /* "brainpy/_c/composition.pyx":496
 *     while i < composition.used:
 *         status = composition_set_element_count(result, composition.elements[i], composition.counts[i])
 *         i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":497
 *         status = composition_set_element_count(result, composition.elements[i], composition.counts[i])
 *         i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef void print_composition(Composition* composition) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":478
 *     return 1
 * 
 * cdef Composition* copy_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Create a new Composition instance whose element counts are copied from
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":499
 *     return result
 * 
 * cdef void print_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_11composition_print_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  size_t __pyx_v_i;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":503
 *         size_t i
 *         int status
 *     i = 0             # <<<<<<<<<<<<<<
 *     printf("Addr %d %d\n", <int>composition, <int>(composition.elements))
 *     printf("Size: %d, Used: %d\n", composition.size, composition.used)
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":504
 *         int status
 *     i = 0
 *     printf("Addr %d %d\n", <int>composition, <int>(composition.elements))             # <<<<<<<<<<<<<<
 *     printf("Size: %d, Used: %d\n", composition.size, composition.used)
 *     printf("{\n")
 */
  (void)(printf(((char const *)"Addr %d %d\n"), ((int)__pyx_v_composition), ((int)__pyx_v_composition->elements)));

  /* "brainpy/_c/composition.pyx":505
 *     i = 0
 *     printf("Addr %d %d\n", <int>composition, <int>(composition.elements))
 *     printf("Size: %d, Used: %d\n", composition.size, composition.used)             # <<<<<<<<<<<<<<
 *     printf("{\n")
 *     while i < composition.used:
 */
  (void)(printf(((char const *)"Size: %d, Used: %d\n"), __pyx_v_composition->size, __pyx_v_composition->used));

  /* "brainpy/_c/composition.pyx":506
 *     printf("Addr %d %d\n", <int>composition, <int>(composition.elements))
 *     printf("Size: %d, Used: %d\n", composition.size, composition.used)
 *     printf("{\n")             # <<<<<<<<<<<<<<
 *     while i < composition.used:
 *         printf("  %s: %d\n", composition.elements[i], composition.counts[i])
 */
  (void)(printf(((char const *)"{\n")));

    /* "brainpy/_c/composition.pyx":507
 *     printf("Size: %d, Used: %d\n", composition.size, composition.used)
 *     printf("{\n")
 *     while i < composition.used:             # <<<<<<<<<<<<<<
 *         printf("  %s: %d\n", composition.elements[i], composition.counts[i])
 *         i += 1
 */
    while (1) {
      __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
      if (!__pyx_t_1) break;

      /* "brainpy/_c/composition.pyx":508
 *     printf("{\n")
 *     while i < composition.used:
 *         printf("  %s: %d\n", composition.elements[i], composition.counts[i])             # <<<<<<<<<<<<<<
 *         i += 1
 *     printf("}\n\n")
 */
      (void)(printf(((char const *)"  %s: %d\n"), (__pyx_v_composition->elements[__pyx_v_i]), (__pyx_v_composition->counts[__pyx_v_i])));

      /* "brainpy/_c/composition.pyx":509
 *     while i < composition.used:
 *         printf("  %s: %d\n", composition.elements[i], composition.counts[i])
 *         i += 1             # <<<<<<<<<<<<<<
 *     printf("}\n\n")
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);
    }

    /* "brainpy/_c/composition.pyx":510
 *         printf("  %s: %d\n", composition.elements[i], composition.counts[i])
 *         i += 1
 *     printf("}\n\n")             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_set_element_count(Composition* composition, char* element, count_type count) nogil:
 */
  (void)(printf(((char const *)"}\n\n")));

  /* "brainpy/_c/composition.pyx":499
 *     return result
 * 
 * cdef void print_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":512
 *     printf("}\n\n")
 * 
 * cdef int composition_set_element_count(Composition* composition, char* element, count_type count) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Set the count for `element` in `composition`
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_set_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, char *__pyx_v_element, __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_count) {
  size_t __pyx_v_i;
  int __pyx_v_status;
  int __pyx_v_done;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":525
 *         int status
 *         bint done
 *     done = False             # <<<<<<<<<<<<<<
 *     i = 0
 *     while (i < composition.used):
 */
  __pyx_v_done = 0;

  /* "brainpy/_c/composition.pyx":526
 *         bint done
 *     done = False
 *     i = 0             # <<<<<<<<<<<<<<
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":527
 *     done = False
 *     i = 0
 *     while (i < composition.used):             # <<<<<<<<<<<<<<
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":528
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             composition.counts[i] = count
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element, (__pyx_v_composition->elements[__pyx_v_i])) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":529
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True             # <<<<<<<<<<<<<<
 *             composition.counts[i] = count
 *             return 0
 */
      __pyx_v_done = 1;

      /* "brainpy/_c/composition.pyx":530
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 *             composition.counts[i] = count             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
      (__pyx_v_composition->counts[__pyx_v_i]) = __pyx_v_count;

      /* "brainpy/_c/composition.pyx":531
 *             done = True
 *             composition.counts[i] = count
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":528
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             composition.counts[i] = count
 */
    }

    /* "brainpy/_c/composition.pyx":532
 *             composition.counts[i] = count
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     if not done:
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":534
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         composition.used += 1
 *         if composition.used >= composition.size:
 */
  __pyx_t_1 = ((!(__pyx_v_done != 0)) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":535
 * 
 *     if not done:
 *         composition.used += 1             # <<<<<<<<<<<<<<
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 */
    __pyx_v_composition->used = (__pyx_v_composition->used + 1);

    /* "brainpy/_c/composition.pyx":536
 *     if not done:
 *         composition.used += 1
 *         if composition.used >= composition.size:             # <<<<<<<<<<<<<<
 *             status = composition_resize(composition)
 *             if status != 0:
 */
    __pyx_t_1 = ((__pyx_v_composition->used >= __pyx_v_composition->size) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":537
 *         composition.used += 1
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)             # <<<<<<<<<<<<<<
 *             if status != 0:
 *                 return -1
 */
      __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_resize(__pyx_v_composition);

      /* "brainpy/_c/composition.pyx":538
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 *             if status != 0:             # <<<<<<<<<<<<<<
 *                 return -1
 *         composition.elements[i] = element
 */
      __pyx_t_1 = ((__pyx_v_status != 0) != 0);
      if (__pyx_t_1) {

        /* "brainpy/_c/composition.pyx":539
 *             status = composition_resize(composition)
 *             if status != 0:
 *                 return -1             # <<<<<<<<<<<<<<
 *         composition.elements[i] = element
 *         composition.counts[i] = count
 */
        __pyx_r = -1;
        goto __pyx_L0;

        /* "brainpy/_c/composition.pyx":538
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 *             if status != 0:             # <<<<<<<<<<<<<<
 *                 return -1
 *         composition.elements[i] = element
 */
      }

      /* "brainpy/_c/composition.pyx":536
 *     if not done:
 *         composition.used += 1
 *         if composition.used >= composition.size:             # <<<<<<<<<<<<<<
 *             status = composition_resize(composition)
 *             if status != 0:
 */
    }

    /* "brainpy/_c/composition.pyx":540
 *             if status != 0:
 *                 return -1
 *         composition.elements[i] = element             # <<<<<<<<<<<<<<
 *         composition.counts[i] = count
 *         return 0
 */
    (__pyx_v_composition->elements[__pyx_v_i]) = __pyx_v_element;

    /* "brainpy/_c/composition.pyx":541
 *                 return -1
 *         composition.elements[i] = element
 *         composition.counts[i] = count             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
    (__pyx_v_composition->counts[__pyx_v_i]) = __pyx_v_count;

    /* "brainpy/_c/composition.pyx":542
 *         composition.elements[i] = element
 *         composition.counts[i] = count
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":534
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         composition.used += 1
 *         if composition.used >= composition.size:
 */
  }

  /* "brainpy/_c/composition.pyx":543
 *         composition.counts[i] = count
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_get_element_count(Composition* composition, char* element, count_type* count) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":512
 *     printf("}\n\n")
 * 
 * cdef int composition_set_element_count(Composition* composition, char* element, count_type count) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Set the count for `element` in `composition`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":545
 *     return 1
 * 
 * cdef int composition_get_element_count(Composition* composition, char* element, count_type* count) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Get the count of `element` in `composition`. The count is 0 if `element` is not in `composition`
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, char *__pyx_v_element, __pyx_t_7brainpy_2_c_11composition_count_type *__pyx_v_count) {
  size_t __pyx_v_i;
  int __pyx_v_done;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":557
 *         int status
 *         bint done
 *     done = False             # <<<<<<<<<<<<<<
 *     i = 0
 *     while (i < composition.used):
 */
  __pyx_v_done = 0;

  /* "brainpy/_c/composition.pyx":558
 *         bint done
 *     done = False
 *     i = 0             # <<<<<<<<<<<<<<
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":559
 *     done = False
 *     i = 0
 *     while (i < composition.used):             # <<<<<<<<<<<<<<
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":560
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             count[0] = composition.counts[i]
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element, (__pyx_v_composition->elements[__pyx_v_i])) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":561
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True             # <<<<<<<<<<<<<<
 *             count[0] = composition.counts[i]
 *             return 0
 */
      __pyx_v_done = 1;

      /* "brainpy/_c/composition.pyx":562
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 *             count[0] = composition.counts[i]             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
      (__pyx_v_count[0]) = (__pyx_v_composition->counts[__pyx_v_i]);

      /* "brainpy/_c/composition.pyx":563
 *             done = True
 *             count[0] = composition.counts[i]
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":560
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             count[0] = composition.counts[i]
 */
    }

    /* "brainpy/_c/composition.pyx":564
 *             count[0] = composition.counts[i]
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     if not done:
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":566
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         count[0] = 0
 *     return 1
 */
  __pyx_t_1 = ((!(__pyx_v_done != 0)) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":567
 * 
 *     if not done:
 *         count[0] = 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    (__pyx_v_count[0]) = 0;

    /* "brainpy/_c/composition.pyx":566
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         count[0] = 0
 *     return 1
 */
  }

  /* "brainpy/_c/composition.pyx":568
 *     if not done:
 *         count[0] = 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_inc_element_count(Composition* composition, char* element, count_type increment) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":545
 *     return 1
 * 
 * cdef int composition_get_element_count(Composition* composition, char* element, count_type* count) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Get the count of `element` in `composition`. The count is 0 if `element` is not in `composition`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":570
 *     return 1
 * 
 * cdef int composition_inc_element_count(Composition* composition, char* element, count_type increment) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Increase the count for `element` in `composition` by `increment`.
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_inc_element_count(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, char *__pyx_v_element, __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_increment) {
  size_t __pyx_v_i;
  int __pyx_v_status;
  int __pyx_v_done;
  int __pyx_r;
  int __pyx_t_1;
  size_t __pyx_t_2;

  /* "brainpy/_c/composition.pyx":583
 *         int status
 *         bint done
 *     done = False             # <<<<<<<<<<<<<<
 *     i = 0
 *     while (i < composition.used):
 */
  __pyx_v_done = 0;

  /* "brainpy/_c/composition.pyx":584
 *         bint done
 *     done = False
 *     i = 0             # <<<<<<<<<<<<<<
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":585
 *     done = False
 *     i = 0
 *     while (i < composition.used):             # <<<<<<<<<<<<<<
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":586
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             composition.counts[i] += increment
 */
    __pyx_t_1 = ((strcmp(__pyx_v_element, (__pyx_v_composition->elements[__pyx_v_i])) == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":587
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True             # <<<<<<<<<<<<<<
 *             composition.counts[i] += increment
 *             return 0
 */
      __pyx_v_done = 1;

      /* "brainpy/_c/composition.pyx":588
 *         if strcmp(element, composition.elements[i]) == 0:
 *             done = True
 *             composition.counts[i] += increment             # <<<<<<<<<<<<<<
 *             return 0
 *         i += 1
 */
      __pyx_t_2 = __pyx_v_i;
      (__pyx_v_composition->counts[__pyx_t_2]) = ((__pyx_v_composition->counts[__pyx_t_2]) + __pyx_v_increment);

      /* "brainpy/_c/composition.pyx":589
 *             done = True
 *             composition.counts[i] += increment
 *             return 0             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":586
 *     i = 0
 *     while (i < composition.used):
 *         if strcmp(element, composition.elements[i]) == 0:             # <<<<<<<<<<<<<<
 *             done = True
 *             composition.counts[i] += increment
 */
    }

    /* "brainpy/_c/composition.pyx":590
 *             composition.counts[i] += increment
 *             return 0
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     if not done:
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":592
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         composition.used += 1
 *         if composition.used >= composition.size:
 */
  __pyx_t_1 = ((!(__pyx_v_done != 0)) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":593
 * 
 *     if not done:
 *         composition.used += 1             # <<<<<<<<<<<<<<
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 */
    __pyx_v_composition->used = (__pyx_v_composition->used + 1);

    /* "brainpy/_c/composition.pyx":594
 *     if not done:
 *         composition.used += 1
 *         if composition.used >= composition.size:             # <<<<<<<<<<<<<<
 *             status = composition_resize(composition)
 *             if status != 0:
 */
    __pyx_t_1 = ((__pyx_v_composition->used >= __pyx_v_composition->size) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":595
 *         composition.used += 1
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)             # <<<<<<<<<<<<<<
 *             if status != 0:
 *                 return -1
 */
      __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_resize(__pyx_v_composition);

      /* "brainpy/_c/composition.pyx":596
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 *             if status != 0:             # <<<<<<<<<<<<<<
 *                 return -1
 *         composition.elements[i] = element
 */
      __pyx_t_1 = ((__pyx_v_status != 0) != 0);
      if (__pyx_t_1) {

        /* "brainpy/_c/composition.pyx":597
 *             status = composition_resize(composition)
 *             if status != 0:
 *                 return -1             # <<<<<<<<<<<<<<
 *         composition.elements[i] = element
 *         composition.counts[i] = increment
 */
        __pyx_r = -1;
        goto __pyx_L0;

        /* "brainpy/_c/composition.pyx":596
 *         if composition.used >= composition.size:
 *             status = composition_resize(composition)
 *             if status != 0:             # <<<<<<<<<<<<<<
 *                 return -1
 *         composition.elements[i] = element
 */
      }

      /* "brainpy/_c/composition.pyx":594
 *     if not done:
 *         composition.used += 1
 *         if composition.used >= composition.size:             # <<<<<<<<<<<<<<
 *             status = composition_resize(composition)
 *             if status != 0:
 */
    }

    /* "brainpy/_c/composition.pyx":598
 *             if status != 0:
 *                 return -1
 *         composition.elements[i] = element             # <<<<<<<<<<<<<<
 *         composition.counts[i] = increment
 *         return 0
 */
    (__pyx_v_composition->elements[__pyx_v_i]) = __pyx_v_element;

    /* "brainpy/_c/composition.pyx":599
 *                 return -1
 *         composition.elements[i] = element
 *         composition.counts[i] = increment             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
    (__pyx_v_composition->counts[__pyx_v_i]) = __pyx_v_increment;

    /* "brainpy/_c/composition.pyx":600
 *         composition.elements[i] = element
 *         composition.counts[i] = increment
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":592
 *         i += 1
 * 
 *     if not done:             # <<<<<<<<<<<<<<
 *         composition.used += 1
 *         if composition.used >= composition.size:
 */
  }

  /* "brainpy/_c/composition.pyx":601
 *         composition.counts[i] = increment
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_resize(Composition* composition) nogil:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":570
 *     return 1
 * 
 * cdef int composition_inc_element_count(Composition* composition, char* element, count_type increment) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Increase the count for `element` in `composition` by `increment`.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":603
 *     return 1
 * 
 * cdef int composition_resize(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Increases the size of the parallel arrays in `composition`, doubling them in length
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_resize(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":611
 *     -1: Failure due to Out-of-Memory
 *     '''
 *     composition.elements = <char**>realloc(composition.elements, sizeof(char*) * composition.size * 2)             # <<<<<<<<<<<<<<
 *     composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)
 *     composition.size *= 2
 */
  __pyx_v_composition->elements = ((char **)realloc(__pyx_v_composition->elements, (((sizeof(char *)) * __pyx_v_composition->size) * 2)));

  /* "brainpy/_c/composition.pyx":612
 *     '''
 *     composition.elements = <char**>realloc(composition.elements, sizeof(char*) * composition.size * 2)
 *     composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)             # <<<<<<<<<<<<<<
 *     composition.size *= 2
 *     if composition.counts == NULL:
 */
  __pyx_v_composition->counts = ((__pyx_t_7brainpy_2_c_11composition_count_type *)realloc(__pyx_v_composition->counts, (((sizeof(__pyx_t_7brainpy_2_c_11composition_count_type)) * __pyx_v_composition->size) * 2)));

  /* "brainpy/_c/composition.pyx":613
 *     composition.elements = <char**>realloc(composition.elements, sizeof(char*) * composition.size * 2)
 *     composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)
 *     composition.size *= 2             # <<<<<<<<<<<<<<
 *     if composition.counts == NULL:
 *         return -1
 */
  __pyx_v_composition->size = (__pyx_v_composition->size * 2);

  /* "brainpy/_c/composition.pyx":614
 *     composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)
 *     composition.size *= 2
 *     if composition.counts == NULL:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 0
 */
  __pyx_t_1 = ((__pyx_v_composition->counts == NULL) != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":615
 *     composition.size *= 2
 *     if composition.counts == NULL:
 *         return -1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":614
 *     composition.counts = <count_type*>realloc(composition.counts, sizeof(count_type) * composition.size * 2)
 *     composition.size *= 2
 *     if composition.counts == NULL:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 0
 */
  }

  /* "brainpy/_c/composition.pyx":616
 *     if composition.counts == NULL:
 *         return -1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef double composition_mass(Composition* composition) nogil:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":603
 *     return 1
 * 
 * cdef int composition_resize(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Increases the size of the parallel arrays in `composition`, doubling them in length
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":618
 *     return 0
 * 
 * cdef double composition_mass(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Calculates the monoisotopic mass of `composition`
 */

static double __pyx_f_7brainpy_2_c_11composition_composition_mass(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  double __pyx_v_mass;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_element;
  char *__pyx_v_element_label;
  char __pyx_v_element_name[10];
  int __pyx_v_isotope_number;
  int __pyx_v_status;
  size_t __pyx_v_i;
  double __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":630
 *         int status
 *         size_t i
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < composition.used:
 *         element_label = composition.elements[i]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":631
 *         size_t i
 *     i = 0
 *     while i < composition.used:             # <<<<<<<<<<<<<<
 *         element_label = composition.elements[i]
 *         _parse_isotope_string(element_label, &isotope_number, element_name)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":632
 *     i = 0
 *     while i < composition.used:
 *         element_label = composition.elements[i]             # <<<<<<<<<<<<<<
 *         _parse_isotope_string(element_label, &isotope_number, element_name)
 *         status = element_hash_table_get(_ElementTable, element_name, &element)
 */
    __pyx_v_element_label = (__pyx_v_composition->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":633
 *     while i < composition.used:
 *         element_label = composition.elements[i]
 *         _parse_isotope_string(element_label, &isotope_number, element_name)             # <<<<<<<<<<<<<<
 *         status = element_hash_table_get(_ElementTable, element_name, &element)
 *         if status != 0:
 */
    (void)(__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(__pyx_v_element_label, (&__pyx_v_isotope_number), __pyx_v_element_name));

    /* "brainpy/_c/composition.pyx":634
 *         element_label = composition.elements[i]
 *         _parse_isotope_string(element_label, &isotope_number, element_name)
 *         status = element_hash_table_get(_ElementTable, element_name, &element)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             printf("Could not find element %s\n", element_label)
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_element_name, (&__pyx_v_element));

    /* "brainpy/_c/composition.pyx":635
 *         _parse_isotope_string(element_label, &isotope_number, element_name)
 *         status = element_hash_table_get(_ElementTable, element_name, &element)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("Could not find element %s\n", element_label)
 *         if isotope_number == 0:
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":636
 *         status = element_hash_table_get(_ElementTable, element_name, &element)
 *         if status != 0:
 *             printf("Could not find element %s\n", element_label)             # <<<<<<<<<<<<<<
 *         if isotope_number == 0:
 *             mass += element_monoisotopic_mass(element) * composition.counts[i]
 */
      (void)(printf(((char const *)"Could not find element %s\n"), __pyx_v_element_label));

      /* "brainpy/_c/composition.pyx":635
 *         _parse_isotope_string(element_label, &isotope_number, element_name)
 *         status = element_hash_table_get(_ElementTable, element_name, &element)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             printf("Could not find element %s\n", element_label)
 *         if isotope_number == 0:
 */
    }

    /* "brainpy/_c/composition.pyx":637
 *         if status != 0:
 *             printf("Could not find element %s\n", element_label)
 *         if isotope_number == 0:             # <<<<<<<<<<<<<<
 *             mass += element_monoisotopic_mass(element) * composition.counts[i]
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_isotope_number == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":638
 *             printf("Could not find element %s\n", element_label)
 *         if isotope_number == 0:
 *             mass += element_monoisotopic_mass(element) * composition.counts[i]             # <<<<<<<<<<<<<<
 *         else:
 *             mass += element_isotopic_mass(element, isotope_number) * composition.counts[i]
 */
      __pyx_v_mass = (__pyx_v_mass + (__pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass(__pyx_v_element) * (__pyx_v_composition->counts[__pyx_v_i])));

      /* "brainpy/_c/composition.pyx":637
 *         if status != 0:
 *             printf("Could not find element %s\n", element_label)
 *         if isotope_number == 0:             # <<<<<<<<<<<<<<
 *             mass += element_monoisotopic_mass(element) * composition.counts[i]
 *         else:
 */
      goto __pyx_L6;
    }

    /* "brainpy/_c/composition.pyx":640
 *             mass += element_monoisotopic_mass(element) * composition.counts[i]
 *         else:
 *             mass += element_isotopic_mass(element, isotope_number) * composition.counts[i]             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
    /*else*/ {
      __pyx_v_mass = (__pyx_v_mass + (__pyx_f_7brainpy_2_c_11composition_element_isotopic_mass(__pyx_v_element, __pyx_v_isotope_number) * (__pyx_v_composition->counts[__pyx_v_i])));
    }
    __pyx_L6:;

    /* "brainpy/_c/composition.pyx":641
 *         else:
 *             mass += element_isotopic_mass(element, isotope_number) * composition.counts[i]
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 *     return mass
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":643
 *         i += 1
 * 
 *     return mass             # <<<<<<<<<<<<<<
 * 
 * cdef void free_composition(Composition* composition) nogil:
 */
  __pyx_r = __pyx_v_mass;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":618
 *     return 0
 * 
 * cdef double composition_mass(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     '''
 *     Calculates the monoisotopic mass of `composition`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":645
 *     return mass
 * 
 * cdef void free_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     free(composition.elements)
 *     free(composition.counts)
 */

static void __pyx_f_7brainpy_2_c_11composition_free_composition(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {

  /* "brainpy/_c/composition.pyx":646
 * 
 * cdef void free_composition(Composition* composition) nogil:
 *     free(composition.elements)             # <<<<<<<<<<<<<<
 *     free(composition.counts)
 *     free(composition)
 */
  free(__pyx_v_composition->elements);

  /* "brainpy/_c/composition.pyx":647
 * cdef void free_composition(Composition* composition) nogil:
 *     free(composition.elements)
 *     free(composition.counts)             # <<<<<<<<<<<<<<
 *     free(composition)
 * 
 */
  free(__pyx_v_composition->counts);

  /* "brainpy/_c/composition.pyx":648
 *     free(composition.elements)
 *     free(composition.counts)
 *     free(composition)             # <<<<<<<<<<<<<<
 * 
 * cdef Composition* composition_add(Composition* composition_1, Composition* composition_2, int sign) nogil:
 */
  free(__pyx_v_composition);

  /* "brainpy/_c/composition.pyx":645
 *     return mass
 * 
 * cdef void free_composition(Composition* composition) nogil:             # <<<<<<<<<<<<<<
 *     free(composition.elements)
 *     free(composition.counts)
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":650
 *     free(composition)
 * 
 * cdef Composition* composition_add(Composition* composition_1, Composition* composition_2, int sign) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_composition_add(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_1, struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_2, int __pyx_v_sign) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_result;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  char *__pyx_v_symbol;
  CYTHON_UNUSED int __pyx_v_status;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":658
 *         size_t i
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     result = copy_composition(composition_1)
 * 
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":659
 * 
 *     i = 0
 *     result = copy_composition(composition_1)             # <<<<<<<<<<<<<<
 * 
 *     while i < composition_2.used:
 */
  __pyx_v_result = __pyx_f_7brainpy_2_c_11composition_copy_composition(__pyx_v_composition_1);

  /* "brainpy/_c/composition.pyx":661
 *     result = copy_composition(composition_1)
 * 
 *     while i < composition_2.used:             # <<<<<<<<<<<<<<
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition_2->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":662
 * 
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]             # <<<<<<<<<<<<<<
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 */
    __pyx_v_symbol = (__pyx_v_composition_2->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":663
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)             # <<<<<<<<<<<<<<
 *         i += 1
 *         if value == 0:
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition_2, __pyx_v_symbol, (&__pyx_v_value));

    /* "brainpy/_c/composition.pyx":664
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             continue
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":665
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         status = composition_inc_element_count(result, symbol, sign * value)
 */
    __pyx_t_1 = ((__pyx_v_value == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":666
 *         i += 1
 *         if value == 0:
 *             continue             # <<<<<<<<<<<<<<
 *         status = composition_inc_element_count(result, symbol, sign * value)
 *     return result
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":665
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         status = composition_inc_element_count(result, symbol, sign * value)
 */
    }

    /* "brainpy/_c/composition.pyx":667
 *         if value == 0:
 *             continue
 *         status = composition_inc_element_count(result, symbol, sign * value)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_inc_element_count(__pyx_v_result, __pyx_v_symbol, (__pyx_v_sign * __pyx_v_value));
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":668
 *             continue
 *         status = composition_inc_element_count(result, symbol, sign * value)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef int composition_iadd(Composition* composition_1, Composition* composition_2, int sign) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":650
 *     free(composition)
 * 
 * cdef Composition* composition_add(Composition* composition_1, Composition* composition_2, int sign) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":670
 *     return result
 * 
 * cdef int composition_iadd(Composition* composition_1, Composition* composition_2, int sign) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* symbol
 */

static int __pyx_f_7brainpy_2_c_11composition_composition_iadd(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_1, struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition_2, int __pyx_v_sign) {
  char *__pyx_v_symbol;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  int __pyx_v_status;
  size_t __pyx_v_i;
  int __pyx_r;
  int __pyx_t_1;

  /* "brainpy/_c/composition.pyx":676
 *         int status
 *         size_t i
 *     i = 0             # <<<<<<<<<<<<<<
 *     status = 0
 *     while i < composition_2.used:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":677
 *         size_t i
 *     i = 0
 *     status = 0             # <<<<<<<<<<<<<<
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]
 */
  __pyx_v_status = 0;

  /* "brainpy/_c/composition.pyx":678
 *     i = 0
 *     status = 0
 *     while i < composition_2.used:             # <<<<<<<<<<<<<<
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition_2->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":679
 *     status = 0
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]             # <<<<<<<<<<<<<<
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 */
    __pyx_v_symbol = (__pyx_v_composition_2->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":680
 *     while i < composition_2.used:
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)             # <<<<<<<<<<<<<<
 *         i += 1
 *         if value == 0:
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition_2, __pyx_v_symbol, (&__pyx_v_value));

    /* "brainpy/_c/composition.pyx":681
 *         symbol = composition_2.elements[i]
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             continue
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":682
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         status = composition_inc_element_count(composition_1, symbol, sign * value)
 */
    __pyx_t_1 = ((__pyx_v_value == 0) != 0);
    if (__pyx_t_1) {

      /* "brainpy/_c/composition.pyx":683
 *         i += 1
 *         if value == 0:
 *             continue             # <<<<<<<<<<<<<<
 *         status = composition_inc_element_count(composition_1, symbol, sign * value)
 *         if status != 0:
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":682
 *         status = composition_get_element_count(composition_2, symbol, &value)
 *         i += 1
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         status = composition_inc_element_count(composition_1, symbol, sign * value)
 */
    }

    /* "brainpy/_c/composition.pyx":684
 *         if value == 0:
 *             continue
 *         status = composition_inc_element_count(composition_1, symbol, sign * value)             # <<<<<<<<<<<<<<
 *         if status != 0:
 *             pass
 */
    __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_inc_element_count(__pyx_v_composition_1, __pyx_v_symbol, (__pyx_v_sign * __pyx_v_value));

    /* "brainpy/_c/composition.pyx":685
 *             continue
 *         status = composition_inc_element_count(composition_1, symbol, sign * value)
 *         if status != 0:             # <<<<<<<<<<<<<<
 *             pass
 *     return status
 */
    __pyx_t_1 = ((__pyx_v_status != 0) != 0);
    if (__pyx_t_1) {
    }
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":687
 *         if status != 0:
 *             pass
 *     return status             # <<<<<<<<<<<<<<
 * 
 * cdef Composition* composition_mul(Composition* composition, int scale) nogil:
 */
  __pyx_r = __pyx_v_status;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":670
 *     return result
 * 
 * cdef int composition_iadd(Composition* composition_1, Composition* composition_2, int sign) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* symbol
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":689
 *     return status
 * 
 * cdef Composition* composition_mul(Composition* composition, int scale) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_composition_mul(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, int __pyx_v_scale) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_result;
  size_t __pyx_v_i;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_r;
  int __pyx_t_1;
  size_t __pyx_t_2;

  /* "brainpy/_c/composition.pyx":694
 *         size_t i
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     result = copy_composition(composition)
 *     while i < result.used:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":695
 * 
 *     i = 0
 *     result = copy_composition(composition)             # <<<<<<<<<<<<<<
 *     while i < result.used:
 *         result.counts[i] *= scale
 */
  __pyx_v_result = __pyx_f_7brainpy_2_c_11composition_copy_composition(__pyx_v_composition);

  /* "brainpy/_c/composition.pyx":696
 *     i = 0
 *     result = copy_composition(composition)
 *     while i < result.used:             # <<<<<<<<<<<<<<
 *         result.counts[i] *= scale
 *         i += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_result->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":697
 *     result = copy_composition(composition)
 *     while i < result.used:
 *         result.counts[i] *= scale             # <<<<<<<<<<<<<<
 *         i += 1
 *     return result
 */
    __pyx_t_2 = __pyx_v_i;
    (__pyx_v_result->counts[__pyx_t_2]) = ((__pyx_v_result->counts[__pyx_t_2]) * __pyx_v_scale);

    /* "brainpy/_c/composition.pyx":698
 *     while i < result.used:
 *         result.counts[i] *= scale
 *         i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":699
 *         result.counts[i] *= scale
 *         i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef void composition_imul(Composition* composition, int scale) nogil:
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":689
 *     return status
 * 
 * cdef Composition* composition_mul(Composition* composition, int scale) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":701
 *     return result
 * 
 * cdef void composition_imul(Composition* composition, int scale) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

static void __pyx_f_7brainpy_2_c_11composition_composition_imul(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, int __pyx_v_scale) {
  size_t __pyx_v_i;
  int __pyx_t_1;
  size_t __pyx_t_2;

  /* "brainpy/_c/composition.pyx":705
 *         size_t i
 * 
 *     i = 0             # <<<<<<<<<<<<<<
 *     while i < composition.used:
 *         composition.counts[i] *= scale
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":706
 * 
 *     i = 0
 *     while i < composition.used:             # <<<<<<<<<<<<<<
 *         composition.counts[i] *= scale
 *         i += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_1) break;

    /* "brainpy/_c/composition.pyx":707
 *     i = 0
 *     while i < composition.used:
 *         composition.counts[i] *= scale             # <<<<<<<<<<<<<<
 *         i += 1
 * 
 */
    __pyx_t_2 = __pyx_v_i;
    (__pyx_v_composition->counts[__pyx_t_2]) = ((__pyx_v_composition->counts[__pyx_t_2]) * __pyx_v_scale);

    /* "brainpy/_c/composition.pyx":708
 *     while i < composition.used:
 *         composition.counts[i] *= scale
 *         i += 1             # <<<<<<<<<<<<<<
 * 
 * cdef dict composition_to_dict(Composition* composition):
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "brainpy/_c/composition.pyx":701
 *     return result
 * 
 * cdef void composition_imul(Composition* composition, int scale) nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         size_t i
 */

  /* function exit code */
}

/* "brainpy/_c/composition.pyx":710
 *         i += 1
 * 
 * cdef dict composition_to_dict(Composition* composition):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict result
 */

static PyObject *__pyx_f_7brainpy_2_c_11composition_composition_to_dict(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition) {
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_symbol = 0;
  size_t __pyx_v_i;
  char *__pyx_v_symbol_c;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("composition_to_dict", 0);

  /* "brainpy/_c/composition.pyx":717
 *         char* symbol_c
 *         count_type value
 *     i = 0             # <<<<<<<<<<<<<<
 *     result = dict()
 *     while i < composition.used:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":718
 *         count_type value
 *     i = 0
 *     result = dict()             # <<<<<<<<<<<<<<
 *     while i < composition.used:
 *         symbol_c = composition.elements[i]
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":719
 *     i = 0
 *     result = dict()
 *     while i < composition.used:             # <<<<<<<<<<<<<<
 *         symbol_c = composition.elements[i]
 *         composition_get_element_count(composition, symbol_c, &value)
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_composition->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/composition.pyx":720
 *     result = dict()
 *     while i < composition.used:
 *         symbol_c = composition.elements[i]             # <<<<<<<<<<<<<<
 *         composition_get_element_count(composition, symbol_c, &value)
 *         if value == 0:
 */
    __pyx_v_symbol_c = (__pyx_v_composition->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":721
 *     while i < composition.used:
 *         symbol_c = composition.elements[i]
 *         composition_get_element_count(composition, symbol_c, &value)             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             continue
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition, __pyx_v_symbol_c, (&__pyx_v_value)));

    /* "brainpy/_c/composition.pyx":722
 *         symbol_c = composition.elements[i]
 *         composition_get_element_count(composition, symbol_c, &value)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         symbol = PyString_FromString(symbol_c)
 */
    __pyx_t_2 = ((__pyx_v_value == 0) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/composition.pyx":723
 *         composition_get_element_count(composition, symbol_c, &value)
 *         if value == 0:
 *             continue             # <<<<<<<<<<<<<<
 *         symbol = PyString_FromString(symbol_c)
 *         result[symbol] = value
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":722
 *         symbol_c = composition.elements[i]
 *         composition_get_element_count(composition, symbol_c, &value)
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             continue
 *         symbol = PyString_FromString(symbol_c)
 */
    }

    /* "brainpy/_c/composition.pyx":724
 *         if value == 0:
 *             continue
 *         symbol = PyString_FromString(symbol_c)             # <<<<<<<<<<<<<<
 *         result[symbol] = value
 *         i += 1
 */
    __pyx_t_1 = PyStr_FromString(__pyx_v_symbol_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_symbol, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":725
 *             continue
 *         symbol = PyString_FromString(symbol_c)
 *         result[symbol] = value             # <<<<<<<<<<<<<<
 *         i += 1
 *     return result
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_symbol, __pyx_t_1) < 0)) __PYX_ERR(0, 725, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":726
 *         symbol = PyString_FromString(symbol_c)
 *         result[symbol] = value
 *         i += 1             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":727
 *         result[symbol] = value
 *         i += 1
 *     return result             # <<<<<<<<<<<<<<
 * 
 * cdef Composition* dict_to_composition(dict comp_dict):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":710
 *         i += 1
 * 
 * cdef dict composition_to_dict(Composition* composition):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.composition_to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":729
 *     return result
 * 
 * cdef Composition* dict_to_composition(dict comp_dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

static struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_f_7brainpy_2_c_11composition_dict_to_composition(PyObject *__pyx_v_comp_dict) {
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_result;
  PyObject *__pyx_v_symbol = 0;
  char *__pyx_v_symbol_c;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_t_9;
  __Pyx_RefNannySetupContext("dict_to_composition", 0);

  /* "brainpy/_c/composition.pyx":736
 *         char* symbol_c
 *         count_type value
 *     result = make_composition()             # <<<<<<<<<<<<<<
 *     for symbol, value in comp_dict.items():
 *         symbol_c = PyString_AsString(symbol)
 */
  __pyx_v_result = __pyx_f_7brainpy_2_c_11composition_make_composition();

  /* "brainpy/_c/composition.pyx":737
 *         count_type value
 *     result = make_composition()
 *     for symbol, value in comp_dict.items():             # <<<<<<<<<<<<<<
 *         symbol_c = PyString_AsString(symbol)
 *         composition_set_element_count(result, symbol_c, value)
 */
  if (unlikely(__pyx_v_comp_dict == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 737, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Items(__pyx_v_comp_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 737, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 737, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 737, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 737, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 737, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 737, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 737, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(PyString_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 737, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_symbol, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __pyx_v_value = __pyx_t_9;

    /* "brainpy/_c/composition.pyx":738
 *     result = make_composition()
 *     for symbol, value in comp_dict.items():
 *         symbol_c = PyString_AsString(symbol)             # <<<<<<<<<<<<<<
 *         composition_set_element_count(result, symbol_c, value)
 *     return result
 */
    __pyx_v_symbol_c = PyStr_AsString(__pyx_v_symbol);

    /* "brainpy/_c/composition.pyx":739
 *     for symbol, value in comp_dict.items():
 *         symbol_c = PyString_AsString(symbol)
 *         composition_set_element_count(result, symbol_c, value)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_result, __pyx_v_symbol_c, __pyx_v_value));

    /* "brainpy/_c/composition.pyx":737
 *         count_type value
 *     result = make_composition()
 *     for symbol, value in comp_dict.items():             # <<<<<<<<<<<<<<
 *         symbol_c = PyString_AsString(symbol)
 *         composition_set_element_count(result, symbol_c, value)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":740
 *         symbol_c = PyString_AsString(symbol)
 *         composition_set_element_count(result, symbol_c, value)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":729
 *     return result
 * 
 * cdef Composition* dict_to_composition(dict comp_dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Composition* result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("brainpy._c.composition.dict_to_composition", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":746
 * # Prototyping and Testing
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef Element* elem
 *     cdef Composition* composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_5main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_4main[] = "main()";
static PyMethodDef __pyx_mdef_7brainpy_2_c_11composition_5main = {"main", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_5main, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_4main};
static PyObject *__pyx_pw_7brainpy_2_c_11composition_5main(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("main (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_4main(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_4main(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_elem;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition;
  struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_sum_composition;
  CYTHON_UNUSED int __pyx_v_j;
  CYTHON_UNUSED char *__pyx_v_sym;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("main", 0);

  /* "brainpy/_c/composition.pyx":756
 *     cdef dict copy
 * 
 *     sym = "O"             # <<<<<<<<<<<<<<
 * 
 *     j = element_hash_table_get(_ElementTable, "C", &elem)
 */
  __pyx_v_sym = ((char *)"O");

  /* "brainpy/_c/composition.pyx":758
 *     sym = "O"
 * 
 *     j = element_hash_table_get(_ElementTable, "C", &elem)             # <<<<<<<<<<<<<<
 *     # print elem.symbol, element_monoisotopic_mass(elem)
 * 
 */
  __pyx_v_j = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, ((char *)"C"), (&__pyx_v_elem));

  /* "brainpy/_c/composition.pyx":761
 *     # print elem.symbol, element_monoisotopic_mass(elem)
 * 
 *     composition = make_composition()             # <<<<<<<<<<<<<<
 *     composition_set_element_count(composition, "O[18]", 1)
 *     composition_set_element_count(composition, "H", 2)
 */
  __pyx_v_composition = __pyx_f_7brainpy_2_c_11composition_make_composition();

  /* "brainpy/_c/composition.pyx":762
 * 
 *     composition = make_composition()
 *     composition_set_element_count(composition, "O[18]", 1)             # <<<<<<<<<<<<<<
 *     composition_set_element_count(composition, "H", 2)
 *     print_composition(composition)
 */
  (void)(__pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_composition, ((char *)"O[18]"), 1));

  /* "brainpy/_c/composition.pyx":763
 *     composition = make_composition()
 *     composition_set_element_count(composition, "O[18]", 1)
 *     composition_set_element_count(composition, "H", 2)             # <<<<<<<<<<<<<<
 *     print_composition(composition)
 *     print("Massing")
 */
  (void)(__pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_composition, ((char *)"H"), 2));

  /* "brainpy/_c/composition.pyx":764
 *     composition_set_element_count(composition, "O[18]", 1)
 *     composition_set_element_count(composition, "H", 2)
 *     print_composition(composition)             # <<<<<<<<<<<<<<
 *     print("Massing")
 *     print(composition_mass(composition))
 */
  __pyx_f_7brainpy_2_c_11composition_print_composition(__pyx_v_composition);

  /* "brainpy/_c/composition.pyx":765
 *     composition_set_element_count(composition, "H", 2)
 *     print_composition(composition)
 *     print("Massing")             # <<<<<<<<<<<<<<
 *     print(composition_mass(composition))
 * 
 */
  if (__Pyx_PrintOne(0, __pyx_n_s_Massing) < 0) __PYX_ERR(0, 765, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":766
 *     print_composition(composition)
 *     print("Massing")
 *     print(composition_mass(composition))             # <<<<<<<<<<<<<<
 * 
 *     sum_composition = composition_add(composition, composition, 1)
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_composition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":768
 *     print(composition_mass(composition))
 * 
 *     sum_composition = composition_add(composition, composition, 1)             # <<<<<<<<<<<<<<
 *     print("Summed")
 *     print(composition_mass(sum_composition))
 */
  __pyx_v_sum_composition = __pyx_f_7brainpy_2_c_11composition_composition_add(__pyx_v_composition, __pyx_v_composition, 1);

  /* "brainpy/_c/composition.pyx":769
 * 
 *     sum_composition = composition_add(composition, composition, 1)
 *     print("Summed")             # <<<<<<<<<<<<<<
 *     print(composition_mass(sum_composition))
 *     print(composition_mass(composition))
 */
  if (__Pyx_PrintOne(0, __pyx_n_s_Summed) < 0) __PYX_ERR(0, 769, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":770
 *     sum_composition = composition_add(composition, composition, 1)
 *     print("Summed")
 *     print(composition_mass(sum_composition))             # <<<<<<<<<<<<<<
 *     print(composition_mass(composition))
 *     composition_iadd(sum_composition, composition, 1)
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_sum_composition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":771
 *     print("Summed")
 *     print(composition_mass(sum_composition))
 *     print(composition_mass(composition))             # <<<<<<<<<<<<<<
 *     composition_iadd(sum_composition, composition, 1)
 *     print(composition_mass(sum_composition))
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_composition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":772
 *     print(composition_mass(sum_composition))
 *     print(composition_mass(composition))
 *     composition_iadd(sum_composition, composition, 1)             # <<<<<<<<<<<<<<
 *     print(composition_mass(sum_composition))
 * 
 */
  (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_sum_composition, __pyx_v_composition, 1));

  /* "brainpy/_c/composition.pyx":773
 *     print(composition_mass(composition))
 *     composition_iadd(sum_composition, composition, 1)
 *     print(composition_mass(sum_composition))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_sum_composition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":746
 * # Prototyping and Testing
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef Element* elem
 *     cdef Composition* composition
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.main", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":776
 * 
 * 
 * def isotope_update_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_7isotope_update_test(PyObject *__pyx_self, PyObject *__pyx_v_element_symbol); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_6isotope_update_test[] = "isotope_update_test(str element_symbol)";
static PyMethodDef __pyx_mdef_7brainpy_2_c_11composition_7isotope_update_test = {"isotope_update_test", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_7isotope_update_test, METH_O, __pyx_doc_7brainpy_2_c_11composition_6isotope_update_test};
static PyObject *__pyx_pw_7brainpy_2_c_11composition_7isotope_update_test(PyObject *__pyx_self, PyObject *__pyx_v_element_symbol) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isotope_update_test (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_element_symbol), (&PyString_Type), 1, "element_symbol", 1))) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_6isotope_update_test(__pyx_self, ((PyObject*)__pyx_v_element_symbol));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_6isotope_update_test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_element_symbol) {
  char *__pyx_v_c_element_symbol;
  char *__pyx_v_element;
  char __pyx_v_isotope_string[10];
  int __pyx_v_isotope;
  int __pyx_v_status;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_elem_obj;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_fixed_isotope_element;
  struct __pyx_t_7brainpy_2_c_11composition_Isotope *__pyx_v_isotope_obj;
  CYTHON_UNUSED PyObject *__pyx_v_out = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("isotope_update_test", 0);

  /* "brainpy/_c/composition.pyx":788
 *         str out
 * 
 *     element = <char*>malloc(sizeof(char) * 10)             # <<<<<<<<<<<<<<
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     printf("c_element_symbol %s\n", c_element_symbol)
 */
  __pyx_v_element = ((char *)malloc(((sizeof(char)) * 10)));

  /* "brainpy/_c/composition.pyx":789
 * 
 *     element = <char*>malloc(sizeof(char) * 10)
 *     c_element_symbol = PyString_AsString(element_symbol)             # <<<<<<<<<<<<<<
 *     printf("c_element_symbol %s\n", c_element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 */
  __pyx_v_c_element_symbol = PyStr_AsString(__pyx_v_element_symbol);

  /* "brainpy/_c/composition.pyx":790
 *     element = <char*>malloc(sizeof(char) * 10)
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     printf("c_element_symbol %s\n", c_element_symbol)             # <<<<<<<<<<<<<<
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     printf("status %s\n", element)
 */
  (void)(printf(((char const *)"c_element_symbol %s\n"), __pyx_v_c_element_symbol));

  /* "brainpy/_c/composition.pyx":791
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     printf("c_element_symbol %s\n", c_element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)             # <<<<<<<<<<<<<<
 *     printf("status %s\n", element)
 *     out = PyString_FromString(element)
 */
  (void)(__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(__pyx_v_c_element_symbol, (&__pyx_v_isotope), __pyx_v_element));

  /* "brainpy/_c/composition.pyx":792
 *     printf("c_element_symbol %s\n", c_element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     printf("status %s\n", element)             # <<<<<<<<<<<<<<
 *     out = PyString_FromString(element)
 *     printf("Element Symbol: %s, Isotope: %d\n", element, isotope)
 */
  (void)(printf(((char const *)"status %s\n"), __pyx_v_element));

  /* "brainpy/_c/composition.pyx":793
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     printf("status %s\n", element)
 *     out = PyString_FromString(element)             # <<<<<<<<<<<<<<
 *     printf("Element Symbol: %s, Isotope: %d\n", element, isotope)
 *     print("element_hash_table_get")
 */
  __pyx_t_1 = PyStr_FromString(__pyx_v_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":794
 *     printf("status %s\n", element)
 *     out = PyString_FromString(element)
 *     printf("Element Symbol: %s, Isotope: %d\n", element, isotope)             # <<<<<<<<<<<<<<
 *     print("element_hash_table_get")
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)
 */
  (void)(printf(((char const *)"Element Symbol: %s, Isotope: %d\n"), __pyx_v_element, __pyx_v_isotope));

  /* "brainpy/_c/composition.pyx":795
 *     out = PyString_FromString(element)
 *     printf("Element Symbol: %s, Isotope: %d\n", element, isotope)
 *     print("element_hash_table_get")             # <<<<<<<<<<<<<<
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)
 *     if status != 0:
 */
  if (__Pyx_PrintOne(0, __pyx_n_s_element_hash_table_get) < 0) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":796
 *     printf("Element Symbol: %s, Isotope: %d\n", element, isotope)
 *     print("element_hash_table_get")
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         print("Error, ", element_symbol, element, isotope)
 */
  __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_element, (&__pyx_v_elem_obj));

  /* "brainpy/_c/composition.pyx":797
 *     print("element_hash_table_get")
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         print("Error, ", element_symbol, element, isotope)
 *     print("Fetched element", elem_obj.symbol)
 */
  __pyx_t_2 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":798
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)
 *     if status != 0:
 *         print("Error, ", element_symbol, element, isotope)             # <<<<<<<<<<<<<<
 *     print("Fetched element", elem_obj.symbol)
 *     isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_isotope); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_kp_s_Error);
    __Pyx_GIVEREF(__pyx_kp_s_Error);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_s_Error);
    __Pyx_INCREF(__pyx_v_element_symbol);
    __Pyx_GIVEREF(__pyx_v_element_symbol);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_element_symbol);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    if (__Pyx_PrintOne(0, __pyx_t_4) < 0) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":797
 *     print("element_hash_table_get")
 *     status = element_hash_table_get(_ElementTable, element, &elem_obj)
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         print("Error, ", element_symbol, element, isotope)
 *     print("Fetched element", elem_obj.symbol)
 */
  }

  /* "brainpy/_c/composition.pyx":799
 *     if status != 0:
 *         print("Error, ", element_symbol, element, isotope)
 *     print("Fetched element", elem_obj.symbol)             # <<<<<<<<<<<<<<
 *     isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)
 *     print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)
 */
  __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_elem_obj->symbol); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_Fetched_element);
  __Pyx_GIVEREF(__pyx_kp_s_Fetched_element);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_Fetched_element);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_3) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "brainpy/_c/composition.pyx":800
 *         print("Error, ", element_symbol, element, isotope)
 *     print("Fetched element", elem_obj.symbol)
 *     isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)             # <<<<<<<<<<<<<<
 *     print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)
 *     _make_isotope_string(elem_obj, isotope_obj, isotope_string)
 */
  __pyx_v_isotope_obj = __pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_count(__pyx_v_elem_obj->isotopes, __pyx_v_isotope);

  /* "brainpy/_c/composition.pyx":801
 *     print("Fetched element", elem_obj.symbol)
 *     isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)
 *     print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)             # <<<<<<<<<<<<<<
 *     _make_isotope_string(elem_obj, isotope_obj, isotope_string)
 *     print(isotope_string)
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_isotope_obj->mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_isotope_obj->neutrons); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_isotope_obj->neutron_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_5) < 0) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "brainpy/_c/composition.pyx":802
 *     isotope_obj = get_isotope_by_neutron_count(elem_obj.isotopes, isotope)
 *     print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)
 *     _make_isotope_string(elem_obj, isotope_obj, isotope_string)             # <<<<<<<<<<<<<<
 *     print(isotope_string)
 *     fixed_isotope_element = make_fixed_isotope_element(elem_obj, isotope)
 */
  (void)(__pyx_f_7brainpy_2_c_11composition__make_isotope_string(__pyx_v_elem_obj, __pyx_v_isotope_obj, __pyx_v_isotope_string));

  /* "brainpy/_c/composition.pyx":803
 *     print(isotope_obj.mass, isotope_obj.neutrons, isotope_obj.neutron_shift)
 *     _make_isotope_string(elem_obj, isotope_obj, isotope_string)
 *     print(isotope_string)             # <<<<<<<<<<<<<<
 *     fixed_isotope_element = make_fixed_isotope_element(elem_obj, isotope)
 *     printf("Fixed: %s, %f", fixed_isotope_element.symbol, element_monoisotopic_mass(fixed_isotope_element))
 */
  __pyx_t_5 = __Pyx_PyObject_FromString(__pyx_v_isotope_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_PrintOne(0, __pyx_t_5) < 0) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "brainpy/_c/composition.pyx":804
 *     _make_isotope_string(elem_obj, isotope_obj, isotope_string)
 *     print(isotope_string)
 *     fixed_isotope_element = make_fixed_isotope_element(elem_obj, isotope)             # <<<<<<<<<<<<<<
 *     printf("Fixed: %s, %f", fixed_isotope_element.symbol, element_monoisotopic_mass(fixed_isotope_element))
 * 
 */
  __pyx_v_fixed_isotope_element = __pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(__pyx_v_elem_obj, __pyx_v_isotope);

  /* "brainpy/_c/composition.pyx":805
 *     print(isotope_string)
 *     fixed_isotope_element = make_fixed_isotope_element(elem_obj, isotope)
 *     printf("Fixed: %s, %f", fixed_isotope_element.symbol, element_monoisotopic_mass(fixed_isotope_element))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(printf(((char const *)"Fixed: %s, %f"), __pyx_v_fixed_isotope_element->symbol, __pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass(__pyx_v_fixed_isotope_element)));

  /* "brainpy/_c/composition.pyx":776
 * 
 * 
 * def isotope_update_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("brainpy._c.composition.isotope_update_test", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":808
 * 
 * 
 * def isotope_parse_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_9isotope_parse_test(PyObject *__pyx_self, PyObject *__pyx_v_element_symbol); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_8isotope_parse_test[] = "isotope_parse_test(str element_symbol)";
static PyMethodDef __pyx_mdef_7brainpy_2_c_11composition_9isotope_parse_test = {"isotope_parse_test", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_9isotope_parse_test, METH_O, __pyx_doc_7brainpy_2_c_11composition_8isotope_parse_test};
static PyObject *__pyx_pw_7brainpy_2_c_11composition_9isotope_parse_test(PyObject *__pyx_self, PyObject *__pyx_v_element_symbol) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isotope_parse_test (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_element_symbol), (&PyString_Type), 1, "element_symbol", 1))) __PYX_ERR(0, 808, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_8isotope_parse_test(__pyx_self, ((PyObject*)__pyx_v_element_symbol));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_8isotope_parse_test(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_element_symbol) {
  char *__pyx_v_c_element_symbol;
  char *__pyx_v_element;
  int __pyx_v_isotope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("isotope_parse_test", 0);

  /* "brainpy/_c/composition.pyx":814
 *         char[10] isotope_string
 *         int isotope
 *     element = <char*>malloc(sizeof(char) * 10)             # <<<<<<<<<<<<<<
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 */
  __pyx_v_element = ((char *)malloc(((sizeof(char)) * 10)));

  /* "brainpy/_c/composition.pyx":815
 *         int isotope
 *     element = <char*>malloc(sizeof(char) * 10)
 *     c_element_symbol = PyString_AsString(element_symbol)             # <<<<<<<<<<<<<<
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     print(element, isotope)
 */
  __pyx_v_c_element_symbol = PyStr_AsString(__pyx_v_element_symbol);

  /* "brainpy/_c/composition.pyx":816
 *     element = <char*>malloc(sizeof(char) * 10)
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)             # <<<<<<<<<<<<<<
 *     print(element, isotope)
 *     free(element)
 */
  (void)(__pyx_f_7brainpy_2_c_11composition__parse_isotope_string(__pyx_v_c_element_symbol, (&__pyx_v_isotope), __pyx_v_element));

  /* "brainpy/_c/composition.pyx":817
 *     c_element_symbol = PyString_AsString(element_symbol)
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     print(element, isotope)             # <<<<<<<<<<<<<<
 *     free(element)
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_isotope); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_3) < 0) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "brainpy/_c/composition.pyx":818
 *     _parse_isotope_string(c_element_symbol, &isotope, element)
 *     print(element, isotope)
 *     free(element)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_element);

  /* "brainpy/_c/composition.pyx":808
 * 
 * 
 * def isotope_parse_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("brainpy._c.composition.isotope_parse_test", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":823
 * cdef dict _string_table = {}
 * 
 * cdef str _store_string(str symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject* po
 */

static PyObject *__pyx_f_7brainpy_2_c_11composition__store_string(PyObject *__pyx_v_symbol) {
  PyObject *__pyx_v_po;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("_store_string", 0);

  /* "brainpy/_c/composition.pyx":826
 *     cdef:
 *         PyObject* po
 *     po = PyDict_GetItem(_string_table, symbol)             # <<<<<<<<<<<<<<
 *     if po == NULL:
 *         PyDict_SetItem(_string_table, symbol, symbol)
 */
  __pyx_t_1 = __pyx_v_7brainpy_2_c_11composition__string_table;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_po = PyDict_GetItem(__pyx_t_1, __pyx_v_symbol);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":827
 *         PyObject* po
 *     po = PyDict_GetItem(_string_table, symbol)
 *     if po == NULL:             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(_string_table, symbol, symbol)
 *         po = PyDict_GetItem(_string_table, symbol)
 */
  __pyx_t_2 = ((__pyx_v_po == NULL) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":828
 *     po = PyDict_GetItem(_string_table, symbol)
 *     if po == NULL:
 *         PyDict_SetItem(_string_table, symbol, symbol)             # <<<<<<<<<<<<<<
 *         po = PyDict_GetItem(_string_table, symbol)
 *     return <str>po
 */
    __pyx_t_1 = __pyx_v_7brainpy_2_c_11composition__string_table;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyDict_SetItem(__pyx_t_1, __pyx_v_symbol, __pyx_v_symbol); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 828, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":829
 *     if po == NULL:
 *         PyDict_SetItem(_string_table, symbol, symbol)
 *         po = PyDict_GetItem(_string_table, symbol)             # <<<<<<<<<<<<<<
 *     return <str>po
 * 
 */
    __pyx_t_1 = __pyx_v_7brainpy_2_c_11composition__string_table;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_po = PyDict_GetItem(__pyx_t_1, __pyx_v_symbol);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":827
 *         PyObject* po
 *     po = PyDict_GetItem(_string_table, symbol)
 *     if po == NULL:             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(_string_table, symbol, symbol)
 *         po = PyDict_GetItem(_string_table, symbol)
 */
  }

  /* "brainpy/_c/composition.pyx":830
 *         PyDict_SetItem(_string_table, symbol, symbol)
 *         po = PyDict_GetItem(_string_table, symbol)
 *     return <str>po             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject*)__pyx_v_po));
  __pyx_r = ((PyObject*)__pyx_v_po);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":823
 * cdef dict _string_table = {}
 * 
 * cdef str _store_string(str symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         PyObject* po
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition._store_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":843
 * 
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition inst
 */

static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_13PyComposition__create(struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_base) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_inst = 0;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "brainpy/_c/composition.pyx":847
 *             PyComposition inst
 * 
 *         inst = PyComposition.__new__(PyComposition)             # <<<<<<<<<<<<<<
 *         if base != NULL:
 *             composition_iadd(inst.impl, base, 1)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_7brainpy_2_c_11composition_PyComposition(((PyTypeObject *)__pyx_ptype_7brainpy_2_c_11composition_PyComposition), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_inst = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":848
 * 
 *         inst = PyComposition.__new__(PyComposition)
 *         if base != NULL:             # <<<<<<<<<<<<<<
 *             composition_iadd(inst.impl, base, 1)
 *         inst._clean = False
 */
  __pyx_t_2 = ((__pyx_v_base != NULL) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":849
 *         inst = PyComposition.__new__(PyComposition)
 *         if base != NULL:
 *             composition_iadd(inst.impl, base, 1)             # <<<<<<<<<<<<<<
 *         inst._clean = False
 *         return inst
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_inst->impl, __pyx_v_base, 1));

    /* "brainpy/_c/composition.pyx":848
 * 
 *         inst = PyComposition.__new__(PyComposition)
 *         if base != NULL:             # <<<<<<<<<<<<<<
 *             composition_iadd(inst.impl, base, 1)
 *         inst._clean = False
 */
  }

  /* "brainpy/_c/composition.pyx":850
 *         if base != NULL:
 *             composition_iadd(inst.impl, base, 1)
 *         inst._clean = False             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->_clean = 0;

  /* "brainpy/_c/composition.pyx":851
 *             composition_iadd(inst.impl, base, 1)
 *         inst._clean = False
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":843
 * 
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":853
 *         return inst
 * 
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):             # <<<<<<<<<<<<<<
 *         if free_existing:
 *             free_composition(self.impl)
 */

static void __pyx_f_7brainpy_2_c_11composition_13PyComposition__set_impl(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_t_7brainpy_2_c_11composition_Composition *__pyx_v_composition, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args) {
  int __pyx_v_free_existing = ((int)1);
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_set_impl", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_free_existing = __pyx_optional_args->free_existing;
    }
  }

  /* "brainpy/_c/composition.pyx":854
 * 
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):
 *         if free_existing:             # <<<<<<<<<<<<<<
 *             free_composition(self.impl)
 *         self.impl = composition
 */
  __pyx_t_1 = (__pyx_v_free_existing != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":855
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):
 *         if free_existing:
 *             free_composition(self.impl)             # <<<<<<<<<<<<<<
 *         self.impl = composition
 *         self._clean = False
 */
    __pyx_f_7brainpy_2_c_11composition_free_composition(__pyx_v_self->impl);

    /* "brainpy/_c/composition.pyx":854
 * 
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):
 *         if free_existing:             # <<<<<<<<<<<<<<
 *             free_composition(self.impl)
 *         self.impl = composition
 */
  }

  /* "brainpy/_c/composition.pyx":856
 *         if free_existing:
 *             free_composition(self.impl)
 *         self.impl = composition             # <<<<<<<<<<<<<<
 *         self._clean = False
 * 
 */
  __pyx_v_self->impl = __pyx_v_composition;

  /* "brainpy/_c/composition.pyx":857
 *             free_composition(self.impl)
 *         self.impl = composition
 *         self._clean = False             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, base=None, **kwargs):
 */
  __pyx_v_self->_clean = 0;

  /* "brainpy/_c/composition.pyx":853
 *         return inst
 * 
 *     cdef void _set_impl(self, Composition* composition, bint free_existing=1):             # <<<<<<<<<<<<<<
 *         if free_existing:
 *             free_composition(self.impl)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "brainpy/_c/composition.pyx":859
 *         self._clean = False
 * 
 *     def __init__(self, base=None, **kwargs):             # <<<<<<<<<<<<<<
 *         self.impl = make_composition()
 *         self._clean = False
 */

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_base = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_base,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 859, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_base = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 859, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition___init__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), __pyx_v_base, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition___init__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_base, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_py_element = NULL;
  PyObject *__pyx_v_count = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "brainpy/_c/composition.pyx":860
 * 
 *     def __init__(self, base=None, **kwargs):
 *         self.impl = make_composition()             # <<<<<<<<<<<<<<
 *         self._clean = False
 *         self.cached_mass = 0.
 */
  __pyx_v_self->impl = __pyx_f_7brainpy_2_c_11composition_make_composition();

  /* "brainpy/_c/composition.pyx":861
 *     def __init__(self, base=None, **kwargs):
 *         self.impl = make_composition()
 *         self._clean = False             # <<<<<<<<<<<<<<
 *         self.cached_mass = 0.
 * 
 */
  __pyx_v_self->_clean = 0;

  /* "brainpy/_c/composition.pyx":862
 *         self.impl = make_composition()
 *         self._clean = False
 *         self.cached_mass = 0.             # <<<<<<<<<<<<<<
 * 
 *         if base is not None and isinstance(base, dict):
 */
  __pyx_v_self->cached_mass = 0.;

  /* "brainpy/_c/composition.pyx":864
 *         self.cached_mass = 0.
 * 
 *         if base is not None and isinstance(base, dict):             # <<<<<<<<<<<<<<
 *             self.update(base)
 * 
 */
  __pyx_t_2 = (__pyx_v_base != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PyDict_Check(__pyx_v_base); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":865
 * 
 *         if base is not None and isinstance(base, dict):
 *             self.update(base)             # <<<<<<<<<<<<<<
 * 
 *         for py_element, count in kwargs.items():
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_base); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_base};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_base};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_base);
        __Pyx_GIVEREF(__pyx_v_base);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_base);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":864
 *         self.cached_mass = 0.
 * 
 *         if base is not None and isinstance(base, dict):             # <<<<<<<<<<<<<<
 *             self.update(base)
 * 
 */
  }

  /* "brainpy/_c/composition.pyx":867
 *             self.update(base)
 * 
 *         for py_element, count in kwargs.items():             # <<<<<<<<<<<<<<
 *             self[py_element] = count
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 867, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 867, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 867, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 867, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 867, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 867, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 867, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_7)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_6 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_6)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 867, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_py_element, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_count, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "brainpy/_c/composition.pyx":868
 * 
 *         for py_element, count in kwargs.items():
 *             self[py_element] = count             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_v_py_element, __pyx_v_count) < 0)) __PYX_ERR(0, 868, __pyx_L1_error)

    /* "brainpy/_c/composition.pyx":867
 *             self.update(base)
 * 
 *         for py_element, count in kwargs.items():             # <<<<<<<<<<<<<<
 *             self[py_element] = count
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "brainpy/_c/composition.pyx":859
 *         self._clean = False
 * 
 *     def __init__(self, base=None, **kwargs):             # <<<<<<<<<<<<<<
 *         self.impl = make_composition()
 *         self._clean = False
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_element);
  __Pyx_XDECREF(__pyx_v_count);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":870
 *             self[py_element] = count
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_composition(self.impl)
 * 
 */

/* Python wrapper */
static void __pyx_pw_7brainpy_2_c_11composition_13PyComposition_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7brainpy_2_c_11composition_13PyComposition_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7brainpy_2_c_11composition_13PyComposition_2__dealloc__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7brainpy_2_c_11composition_13PyComposition_2__dealloc__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "brainpy/_c/composition.pyx":871
 * 
 *     def __dealloc__(self):
 *         free_composition(self.impl)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, str key):
 */
  __pyx_f_7brainpy_2_c_11composition_free_composition(__pyx_v_self->impl);

  /* "brainpy/_c/composition.pyx":870
 *             self[py_element] = count
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_composition(self.impl)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "brainpy/_c/composition.pyx":873
 *         free_composition(self.impl)
 * 
 *     def __getitem__(self, str key):             # <<<<<<<<<<<<<<
 *         cdef:
 *             count_type count
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 873, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_4__getitem__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject*)__pyx_v_key));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_4__getitem__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key) {
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_count;
  int __pyx_v_status;
  char *__pyx_v_ckey;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__getitem__", 0);
  __Pyx_INCREF(__pyx_v_key);

  /* "brainpy/_c/composition.pyx":878
 *             int status
 *             char* ckey
 *         key = _store_string(key)             # <<<<<<<<<<<<<<
 *         ckey = PyString_AsString(key)
 *         status = composition_get_element_count(self.impl, ckey, &count)
 */
  __pyx_t_1 = __pyx_f_7brainpy_2_c_11composition__store_string(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":879
 *             char* ckey
 *         key = _store_string(key)
 *         ckey = PyString_AsString(key)             # <<<<<<<<<<<<<<
 *         status = composition_get_element_count(self.impl, ckey, &count)
 *         if status == 0:
 */
  __pyx_v_ckey = PyStr_AsString(__pyx_v_key);

  /* "brainpy/_c/composition.pyx":880
 *         key = _store_string(key)
 *         ckey = PyString_AsString(key)
 *         status = composition_get_element_count(self.impl, ckey, &count)             # <<<<<<<<<<<<<<
 *         if status == 0:
 *             return count
 */
  __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_self->impl, __pyx_v_ckey, (&__pyx_v_count));

  /* "brainpy/_c/composition.pyx":881
 *         ckey = PyString_AsString(key)
 *         status = composition_get_element_count(self.impl, ckey, &count)
 *         if status == 0:             # <<<<<<<<<<<<<<
 *             return count
 *         else:
 */
  __pyx_t_2 = ((__pyx_v_status == 0) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":882
 *         status = composition_get_element_count(self.impl, ckey, &count)
 *         if status == 0:
 *             return count             # <<<<<<<<<<<<<<
 *         else:
 *             # composition_set_element_count(self.impl, ckey, 0)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":881
 *         ckey = PyString_AsString(key)
 *         status = composition_get_element_count(self.impl, ckey, &count)
 *         if status == 0:             # <<<<<<<<<<<<<<
 *             return count
 *         else:
 */
  }

  /* "brainpy/_c/composition.pyx":885
 *         else:
 *             # composition_set_element_count(self.impl, ckey, 0)
 *             return 0.             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, str key, count_type count):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_float_0_);
    __pyx_r = __pyx_float_0_;
    goto __pyx_L0;
  }

  /* "brainpy/_c/composition.pyx":873
 *         free_composition(self.impl)
 * 
 *     def __getitem__(self, str key):             # <<<<<<<<<<<<<<
 *         cdef:
 *             count_type count
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":887
 *             return 0.
 * 
 *     def __setitem__(self, str key, count_type count):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int status
 */

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_arg_count); /*proto*/
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_7__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_arg_count) {
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_count;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_count); {
    __pyx_v_count = __Pyx_PyInt_As_int(__pyx_arg_count); if (unlikely((__pyx_v_count == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 887, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6__setitem__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject*)__pyx_v_key), ((__pyx_t_7brainpy_2_c_11composition_count_type)__pyx_v_count));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6__setitem__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key, __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_count) {
  CYTHON_UNUSED int __pyx_v_status;
  char *__pyx_v_ckey;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_key);

  /* "brainpy/_c/composition.pyx":891
 *             int status
 *             char* ckey
 *         key = _store_string(key)             # <<<<<<<<<<<<<<
 *         ckey = PyString_AsString(key)
 *         status = composition_set_element_count(self.impl, ckey, count)
 */
  __pyx_t_1 = __pyx_f_7brainpy_2_c_11composition__store_string(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":892
 *             char* ckey
 *         key = _store_string(key)
 *         ckey = PyString_AsString(key)             # <<<<<<<<<<<<<<
 *         status = composition_set_element_count(self.impl, ckey, count)
 *         self._clean = False
 */
  __pyx_v_ckey = PyStr_AsString(__pyx_v_key);

  /* "brainpy/_c/composition.pyx":893
 *         key = _store_string(key)
 *         ckey = PyString_AsString(key)
 *         status = composition_set_element_count(self.impl, ckey, count)             # <<<<<<<<<<<<<<
 *         self._clean = False
 * 
 */
  __pyx_v_status = __pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_self->impl, __pyx_v_ckey, __pyx_v_count);

  /* "brainpy/_c/composition.pyx":894
 *         ckey = PyString_AsString(key)
 *         status = composition_set_element_count(self.impl, ckey, count)
 *         self._clean = False             # <<<<<<<<<<<<<<
 * 
 *     def update(self, arg=None, **kwargs):
 */
  __pyx_v_self->_clean = 0;

  /* "brainpy/_c/composition.pyx":887
 *             return 0.
 * 
 *     def __setitem__(self, str key, count_type count):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int status
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":896
 *         self._clean = False
 * 
 *     def update(self, arg=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if arg is not None:
 *             self.update(**arg)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_9update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_8update[] = "PyComposition.update(self, arg=None, **kwargs)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_9update(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "update") < 0)) __PYX_ERR(0, 896, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_arg = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 896, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_8update(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), __pyx_v_arg, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_8update(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("update", 0);

  /* "brainpy/_c/composition.pyx":897
 * 
 *     def update(self, arg=None, **kwargs):
 *         if arg is not None:             # <<<<<<<<<<<<<<
 *             self.update(**arg)
 *         for key, value in kwargs.items():
 */
  __pyx_t_1 = (__pyx_v_arg != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":898
 *     def update(self, arg=None, **kwargs):
 *         if arg is not None:
 *             self.update(**arg)             # <<<<<<<<<<<<<<
 *         for key, value in kwargs.items():
 *             self[key] = value
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_arg == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
      __PYX_ERR(0, 898, __pyx_L1_error)
    }
    if (likely(PyDict_CheckExact(__pyx_v_arg))) {
      __pyx_t_4 = PyDict_Copy(__pyx_v_arg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    } else {
      __pyx_t_4 = PyObject_CallFunctionObjArgs((PyObject*)&PyDict_Type, __pyx_v_arg, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "brainpy/_c/composition.pyx":897
 * 
 *     def update(self, arg=None, **kwargs):
 *         if arg is not None:             # <<<<<<<<<<<<<<
 *             self.update(**arg)
 *         for key, value in kwargs.items():
 */
  }

  /* "brainpy/_c/composition.pyx":899
 *         if arg is not None:
 *             self.update(**arg)
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             self[key] = value
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 899, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 899, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 899, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 899, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 899, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 899, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 899, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 899, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "brainpy/_c/composition.pyx":900
 *             self.update(**arg)
 *         for key, value in kwargs.items():
 *             self[key] = value             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
    if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 900, __pyx_L1_error)

    /* "brainpy/_c/composition.pyx":899
 *         if arg is not None:
 *             self.update(**arg)
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             self[key] = value
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "brainpy/_c/composition.pyx":896
 *         self._clean = False
 * 
 *     def update(self, arg=None, **kwargs):             # <<<<<<<<<<<<<<
 *         if arg is not None:
 *             self.update(**arg)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":902
 *             self[key] = value
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_11keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_10keys[] = "PyComposition.keys(self)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_11keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_10keys(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_10keys(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  size_t __pyx_v_i;
  int __pyx_v_count;
  char *__pyx_v_elem;
  PyObject *__pyx_v_keys = 0;
  PyObject *__pyx_v_key_str = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "brainpy/_c/composition.pyx":909
 *             list keys
 *             str key_str
 *         i = 0             # <<<<<<<<<<<<<<
 *         keys = []
 *         while i < self.impl.used:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":910
 *             str key_str
 *         i = 0
 *         keys = []             # <<<<<<<<<<<<<<
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":911
 *         i = 0
 *         keys = []
 *         while i < self.impl.used:             # <<<<<<<<<<<<<<
 *             elem = self.impl.elements[i]
 *             count = self.impl.counts[i]
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_self->impl->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/composition.pyx":912
 *         keys = []
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]             # <<<<<<<<<<<<<<
 *             count = self.impl.counts[i]
 *             i += 1
 */
    __pyx_v_elem = (__pyx_v_self->impl->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":913
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]
 *             count = self.impl.counts[i]             # <<<<<<<<<<<<<<
 *             i += 1
 *             if count == 0:
 */
    __pyx_v_count = (__pyx_v_self->impl->counts[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":914
 *             elem = self.impl.elements[i]
 *             count = self.impl.counts[i]
 *             i += 1             # <<<<<<<<<<<<<<
 *             if count == 0:
 *                 continue
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":915
 *             count = self.impl.counts[i]
 *             i += 1
 *             if count == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             key_str = PyString_FromString(elem)
 */
    __pyx_t_2 = ((__pyx_v_count == 0) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/composition.pyx":916
 *             i += 1
 *             if count == 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             key_str = PyString_FromString(elem)
 *             PyList_Append(keys, key_str)
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":915
 *             count = self.impl.counts[i]
 *             i += 1
 *             if count == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             key_str = PyString_FromString(elem)
 */
    }

    /* "brainpy/_c/composition.pyx":917
 *             if count == 0:
 *                 continue
 *             key_str = PyString_FromString(elem)             # <<<<<<<<<<<<<<
 *             PyList_Append(keys, key_str)
 *         return keys
 */
    __pyx_t_1 = PyStr_FromString(__pyx_v_elem); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_key_str, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "brainpy/_c/composition.pyx":918
 *                 continue
 *             key_str = PyString_FromString(elem)
 *             PyList_Append(keys, key_str)             # <<<<<<<<<<<<<<
 *         return keys
 * 
 */
    __pyx_t_3 = PyList_Append(__pyx_v_keys, __pyx_v_key_str); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 918, __pyx_L1_error)
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":919
 *             key_str = PyString_FromString(elem)
 *             PyList_Append(keys, key_str)
 *         return keys             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_keys);
  __pyx_r = __pyx_v_keys;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":902
 *             self[key] = value
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_key_str);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":921
 *         return keys
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_13values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_12values[] = "PyComposition.values(self)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_13values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_12values(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_12values(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  size_t __pyx_v_i;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  PyObject *__pyx_v_counts = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("values", 0);

  /* "brainpy/_c/composition.pyx":926
 *             count_type value
 *             list counts
 *         counts = []             # <<<<<<<<<<<<<<
 *         i = 0
 *         while i < self.impl.used:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_counts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":927
 *             list counts
 *         counts = []
 *         i = 0             # <<<<<<<<<<<<<<
 *         while i < self.impl.used:
 *             value = self.impl.counts[i]
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":928
 *         counts = []
 *         i = 0
 *         while i < self.impl.used:             # <<<<<<<<<<<<<<
 *             value = self.impl.counts[i]
 *             i += 1
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_self->impl->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/composition.pyx":929
 *         i = 0
 *         while i < self.impl.used:
 *             value = self.impl.counts[i]             # <<<<<<<<<<<<<<
 *             i += 1
 *             if value == 0:
 */
    __pyx_v_value = (__pyx_v_self->impl->counts[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":930
 *         while i < self.impl.used:
 *             value = self.impl.counts[i]
 *             i += 1             # <<<<<<<<<<<<<<
 *             if value == 0:
 *                 continue
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":931
 *             value = self.impl.counts[i]
 *             i += 1
 *             if value == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             counts.append(value)
 */
    __pyx_t_2 = ((__pyx_v_value == 0) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/composition.pyx":932
 *             i += 1
 *             if value == 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             counts.append(value)
 *         return counts
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":931
 *             value = self.impl.counts[i]
 *             i += 1
 *             if value == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             counts.append(value)
 */
    }

    /* "brainpy/_c/composition.pyx":933
 *             if value == 0:
 *                 continue
 *             counts.append(value)             # <<<<<<<<<<<<<<
 *         return counts
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_counts, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":934
 *                 continue
 *             counts.append(value)
 *         return counts             # <<<<<<<<<<<<<<
 * 
 *     def pop(self, str key, object default=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_counts);
  __pyx_r = __pyx_v_counts;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":921
 *         return keys
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_counts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":936
 *         return counts
 * 
 *     def pop(self, str key, object default=None):             # <<<<<<<<<<<<<<
 *         value = self[key]
 *         self[key] = 0
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_15pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_14pop[] = "PyComposition.pop(self, str key, default=None)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_15pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_default = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_default,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pop") < 0)) __PYX_ERR(0, 936, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_key = ((PyObject*)values[0]);
    __pyx_v_default = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 936, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 936, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_14pop(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), __pyx_v_key, __pyx_v_default);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_14pop(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default) {
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("pop", 0);

  /* "brainpy/_c/composition.pyx":937
 * 
 *     def pop(self, str key, object default=None):
 *         value = self[key]             # <<<<<<<<<<<<<<
 *         self[key] = 0
 *         if value == 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":938
 *     def pop(self, str key, object default=None):
 *         value = self[key]
 *         self[key] = 0             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return default
 */
  if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_v_key, __pyx_int_0) < 0)) __PYX_ERR(0, 938, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":939
 *         value = self[key]
 *         self[key] = 0
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         else:
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_value, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":940
 *         self[key] = 0
 *         if value == 0:
 *             return default             # <<<<<<<<<<<<<<
 *         else:
 *             return value
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_default);
    __pyx_r = __pyx_v_default;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":939
 *         value = self[key]
 *         self[key] = 0
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return default
 *         else:
 */
  }

  /* "brainpy/_c/composition.pyx":942
 *             return default
 *         else:
 *             return value             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, str key):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
  }

  /* "brainpy/_c/composition.pyx":936
 *         return counts
 * 
 *     def pop(self, str key, object default=None):             # <<<<<<<<<<<<<<
 *         value = self[key]
 *         self[key] = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":944
 *             return value
 * 
 *     def __contains__(self, str key):             # <<<<<<<<<<<<<<
 *         return self[key] != 0
 * 
 */

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_17__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_17__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key), (&PyString_Type), 1, "key", 1))) __PYX_ERR(0, 944, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_16__contains__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject*)__pyx_v_key));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_16__contains__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "brainpy/_c/composition.pyx":945
 * 
 *     def __contains__(self, str key):
 *         return self[key] != 0             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":944
 *             return value
 * 
 *     def __contains__(self, str key):             # <<<<<<<<<<<<<<
 *         return self[key] != 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":947
 *         return self[key] != 0
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_19items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_18items[] = "PyComposition.items(self)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_19items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_18items(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_18items(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  size_t __pyx_v_i;
  char *__pyx_v_elem;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("items", 0);

  /* "brainpy/_c/composition.pyx":954
 *             list items
 * 
 *         i = 0             # <<<<<<<<<<<<<<
 *         items = []
 *         while i < self.impl.used:
 */
  __pyx_v_i = 0;

  /* "brainpy/_c/composition.pyx":955
 * 
 *         i = 0
 *         items = []             # <<<<<<<<<<<<<<
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_items = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":956
 *         i = 0
 *         items = []
 *         while i < self.impl.used:             # <<<<<<<<<<<<<<
 *             elem = self.impl.elements[i]
 *             value = self.impl.counts[i]
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_self->impl->used) != 0);
    if (!__pyx_t_2) break;

    /* "brainpy/_c/composition.pyx":957
 *         items = []
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]             # <<<<<<<<<<<<<<
 *             value = self.impl.counts[i]
 *             i += 1
 */
    __pyx_v_elem = (__pyx_v_self->impl->elements[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":958
 *         while i < self.impl.used:
 *             elem = self.impl.elements[i]
 *             value = self.impl.counts[i]             # <<<<<<<<<<<<<<
 *             i += 1
 *             if value == 0:
 */
    __pyx_v_value = (__pyx_v_self->impl->counts[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":959
 *             elem = self.impl.elements[i]
 *             value = self.impl.counts[i]
 *             i += 1             # <<<<<<<<<<<<<<
 *             if value == 0:
 *                 continue
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":960
 *             value = self.impl.counts[i]
 *             i += 1
 *             if value == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             items.append((PyString_FromString(elem), value))
 */
    __pyx_t_2 = ((__pyx_v_value == 0) != 0);
    if (__pyx_t_2) {

      /* "brainpy/_c/composition.pyx":961
 *             i += 1
 *             if value == 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             items.append((PyString_FromString(elem), value))
 *         return items
 */
      goto __pyx_L3_continue;

      /* "brainpy/_c/composition.pyx":960
 *             value = self.impl.counts[i]
 *             i += 1
 *             if value == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             items.append((PyString_FromString(elem), value))
 */
    }

    /* "brainpy/_c/composition.pyx":962
 *             if value == 0:
 *                 continue
 *             items.append((PyString_FromString(elem), value))             # <<<<<<<<<<<<<<
 *         return items
 * 
 */
    __pyx_t_1 = PyStr_FromString(__pyx_v_elem); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_items, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L3_continue:;
  }

  /* "brainpy/_c/composition.pyx":963
 *                 continue
 *             items.append((PyString_FromString(elem), value))
 *         return items             # <<<<<<<<<<<<<<
 * 
 *     cpdef PyComposition copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_items);
  __pyx_r = __pyx_v_items;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":947
 *         return self[key] != 0
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":965
 *         return items
 * 
 *     cpdef PyComposition copy(self):             # <<<<<<<<<<<<<<
 *         inst = PyComposition()
 *         composition_iadd(inst.impl, self.impl, 1)
 */

static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_21copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_13PyComposition_copy(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_inst = NULL;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_21copy)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 965, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/composition.pyx":966
 * 
 *     cpdef PyComposition copy(self):
 *         inst = PyComposition()             # <<<<<<<<<<<<<<
 *         composition_iadd(inst.impl, self.impl, 1)
 *         return inst
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7brainpy_2_c_11composition_PyComposition)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inst = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":967
 *     cpdef PyComposition copy(self):
 *         inst = PyComposition()
 *         composition_iadd(inst.impl, self.impl, 1)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_inst->impl, __pyx_v_self->impl, 1));

  /* "brainpy/_c/composition.pyx":968
 *         inst = PyComposition()
 *         composition_iadd(inst.impl, self.impl, 1)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, other):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":965
 *         return items
 * 
 *     cpdef PyComposition copy(self):             # <<<<<<<<<<<<<<
 *         inst = PyComposition()
 *         composition_iadd(inst.impl, self.impl, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_21copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_20copy[] = "PyComposition.copy(self) -> PyComposition";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_21copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_20copy(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_20copy(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7brainpy_2_c_11composition_13PyComposition_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":970
 *         return inst
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_23__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_23__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_22__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_22__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_result = 0;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v__other = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "brainpy/_c/composition.pyx":975
 *             PyComposition _other
 * 
 *         result = self.copy()             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, dict):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 975, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":977
 *         result = self.copy()
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 result[key] += value
 */
  __pyx_t_4 = PyDict_Check(__pyx_v_other); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "brainpy/_c/composition.pyx":978
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 result[key] += value
 *         elif isinstance(other, PyComposition):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 978, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 978, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 978, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 978, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 978, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 978, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 978, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 978, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "brainpy/_c/composition.pyx":979
 *         if isinstance(other, dict):
 *             for key, value in other.items():
 *                 result[key] += value             # <<<<<<<<<<<<<<
 *         elif isinstance(other, PyComposition):
 *             _other = other
 */
      __Pyx_INCREF(__pyx_v_key);
      __pyx_t_1 = __pyx_v_key;
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_result), __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_8, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_result), __pyx_t_1, __pyx_t_3) < 0)) __PYX_ERR(0, 979, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "brainpy/_c/composition.pyx":978
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 result[key] += value
 *         elif isinstance(other, PyComposition):
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "brainpy/_c/composition.pyx":977
 *         result = self.copy()
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 result[key] += value
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":980
 *             for key, value in other.items():
 *                 result[key] += value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, 1)
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "brainpy/_c/composition.pyx":981
 *                 result[key] += value
 *         elif isinstance(other, PyComposition):
 *             _other = other             # <<<<<<<<<<<<<<
 *             composition_iadd(result.impl, _other.impl, 1)
 *         else:
 */
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 981, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v__other = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "brainpy/_c/composition.pyx":982
 *         elif isinstance(other, PyComposition):
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, 1)             # <<<<<<<<<<<<<<
 *         else:
 *             return NotImplemented
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_result->impl, __pyx_v__other->impl, 1));

    /* "brainpy/_c/composition.pyx":980
 *             for key, value in other.items():
 *                 result[key] += value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, 1)
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":984
 *             composition_iadd(result.impl, _other.impl, 1)
 *         else:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "brainpy/_c/composition.pyx":985
 *         else:
 *             return NotImplemented
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __iadd__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":970
 *         return inst
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v__other);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":987
 *         return result
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition _other
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_25__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_25__iadd__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_24__iadd__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_24__iadd__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v__other = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("__iadd__", 0);

  /* "brainpy/_c/composition.pyx":991
 *             PyComposition _other
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 self[key] += value
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":992
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 self[key] += value
 *         elif isinstance(other, PyComposition):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 992, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 992, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 992, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 992, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 992, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 992, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "brainpy/_c/composition.pyx":993
 *         if isinstance(other, dict):
 *             for key, value in other.items():
 *                 self[key] += value             # <<<<<<<<<<<<<<
 *         elif isinstance(other, PyComposition):
 *             _other = other
 */
      __Pyx_INCREF(__pyx_v_key);
      __pyx_t_3 = __pyx_v_key;
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_8, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "brainpy/_c/composition.pyx":992
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 self[key] += value
 *         elif isinstance(other, PyComposition):
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":991
 *             PyComposition _other
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 self[key] += value
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":994
 *             for key, value in other.items():
 *                 self[key] += value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, 1)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "brainpy/_c/composition.pyx":995
 *                 self[key] += value
 *         elif isinstance(other, PyComposition):
 *             _other = other             # <<<<<<<<<<<<<<
 *             composition_iadd(self.impl, _other.impl, 1)
 *         else:
 */
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 995, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v__other = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":996
 *         elif isinstance(other, PyComposition):
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, 1)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("Cannot add %s to PyComposition" % type(other))
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_self->impl, __pyx_v__other->impl, 1));

    /* "brainpy/_c/composition.pyx":994
 *             for key, value in other.items():
 *                 self[key] += value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, 1)
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":998
 *             composition_iadd(self.impl, _other.impl, 1)
 *         else:
 *             raise TypeError("Cannot add %s to PyComposition" % type(other))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_add_s_to_PyComposition, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 998, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "brainpy/_c/composition.pyx":999
 *         else:
 *             raise TypeError("Cannot add %s to PyComposition" % type(other))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":987
 *         return result
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition _other
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v__other);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1001
 *         return self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_27__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_27__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_26__sub__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_26__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_result = 0;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v__other = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("__sub__", 0);

  /* "brainpy/_c/composition.pyx":1006
 *             PyComposition _other
 * 
 *         result = self.copy()             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, dict):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1006, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "brainpy/_c/composition.pyx":1008
 *         result = self.copy()
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 result[key] -= value
 */
  __pyx_t_4 = PyDict_Check(__pyx_v_other); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "brainpy/_c/composition.pyx":1009
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 result[key] -= value
 *         elif isinstance(other, PyComposition):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1009, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1009, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1009, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1009, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1009, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1009, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "brainpy/_c/composition.pyx":1010
 *         if isinstance(other, dict):
 *             for key, value in other.items():
 *                 result[key] -= value             # <<<<<<<<<<<<<<
 *         elif isinstance(other, PyComposition):
 *             _other = other
 */
      __Pyx_INCREF(__pyx_v_key);
      __pyx_t_1 = __pyx_v_key;
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_result), __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PyNumber_InPlaceSubtract(__pyx_t_8, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_result), __pyx_t_1, __pyx_t_3) < 0)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "brainpy/_c/composition.pyx":1009
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 result[key] -= value
 *         elif isinstance(other, PyComposition):
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "brainpy/_c/composition.pyx":1008
 *         result = self.copy()
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 result[key] -= value
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":1011
 *             for key, value in other.items():
 *                 result[key] -= value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, -1)
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "brainpy/_c/composition.pyx":1012
 *                 result[key] -= value
 *         elif isinstance(other, PyComposition):
 *             _other = other             # <<<<<<<<<<<<<<
 *             composition_iadd(result.impl, _other.impl, -1)
 *         else:
 */
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1012, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v__other = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "brainpy/_c/composition.pyx":1013
 *         elif isinstance(other, PyComposition):
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, -1)             # <<<<<<<<<<<<<<
 *         else:
 *             return NotImplemented
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_result->impl, __pyx_v__other->impl, -1));

    /* "brainpy/_c/composition.pyx":1011
 *             for key, value in other.items():
 *                 result[key] -= value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(result.impl, _other.impl, -1)
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":1015
 *             composition_iadd(result.impl, _other.impl, -1)
 *         else:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "brainpy/_c/composition.pyx":1016
 *         else:
 *             return NotImplemented
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __isub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1001
 *         return self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v__other);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1018
 *         return result
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition _other
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_29__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_29__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__isub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_28__isub__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_28__isub__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v__other = 0;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("__isub__", 0);

  /* "brainpy/_c/composition.pyx":1022
 *             PyComposition _other
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 self[key] -= value
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":1023
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 self[key] -= value
 *         elif isinstance(other, PyComposition):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1023, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1023, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1023, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1023, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1023, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1023, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "brainpy/_c/composition.pyx":1024
 *         if isinstance(other, dict):
 *             for key, value in other.items():
 *                 self[key] -= value             # <<<<<<<<<<<<<<
 *         elif isinstance(other, PyComposition):
 *             _other = other
 */
      __Pyx_INCREF(__pyx_v_key);
      __pyx_t_3 = __pyx_v_key;
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1024, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = PyNumber_InPlaceSubtract(__pyx_t_8, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1024, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(((PyObject *)__pyx_v_self), __pyx_t_3, __pyx_t_5) < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "brainpy/_c/composition.pyx":1023
 * 
 *         if isinstance(other, dict):
 *             for key, value in other.items():             # <<<<<<<<<<<<<<
 *                 self[key] -= value
 *         elif isinstance(other, PyComposition):
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":1022
 *             PyComposition _other
 * 
 *         if isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             for key, value in other.items():
 *                 self[key] -= value
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":1025
 *             for key, value in other.items():
 *                 self[key] -= value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, -1)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "brainpy/_c/composition.pyx":1026
 *                 self[key] -= value
 *         elif isinstance(other, PyComposition):
 *             _other = other             # <<<<<<<<<<<<<<
 *             composition_iadd(self.impl, _other.impl, -1)
 *         else:
 */
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1026, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v__other = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":1027
 *         elif isinstance(other, PyComposition):
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, -1)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("Cannot subtract %s from PyComposition" % type(other))
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_iadd(__pyx_v_self->impl, __pyx_v__other->impl, -1));

    /* "brainpy/_c/composition.pyx":1025
 *             for key, value in other.items():
 *                 self[key] -= value
 *         elif isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             _other = other
 *             composition_iadd(self.impl, _other.impl, -1)
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":1029
 *             composition_iadd(self.impl, _other.impl, -1)
 *         else:
 *             raise TypeError("Cannot subtract %s from PyComposition" % type(other))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_subtract_s_from_PyComposi, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1029, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "brainpy/_c/composition.pyx":1030
 *         else:
 *             raise TypeError("Cannot subtract %s from PyComposition" % type(other))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, scale):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1018
 *         return result
 * 
 *     def __isub__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition _other
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__isub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v__other);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1032
 *         return self
 * 
 *     def __mul__(self, scale):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_31__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_scale); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_31__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_scale) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_30__mul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_scale));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_30__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_scale) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_result = 0;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_inst = 0;
  int __pyx_v_scale_factor;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__mul__", 0);

  /* "brainpy/_c/composition.pyx":1038
 *             int scale_factor
 * 
 *         if isinstance(self, PyComposition):             # <<<<<<<<<<<<<<
 *             inst = self
 *             scale_factor = PyInt_AsLong(scale)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_self, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":1039
 * 
 *         if isinstance(self, PyComposition):
 *             inst = self             # <<<<<<<<<<<<<<
 *             scale_factor = PyInt_AsLong(scale)
 *         else:
 */
    if (!(likely(((__pyx_v_self) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1039, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_self;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_inst = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "brainpy/_c/composition.pyx":1040
 *         if isinstance(self, PyComposition):
 *             inst = self
 *             scale_factor = PyInt_AsLong(scale)             # <<<<<<<<<<<<<<
 *         else:
 *             inst = scale
 */
    __pyx_v_scale_factor = PyInt_AsLong(__pyx_v_scale);

    /* "brainpy/_c/composition.pyx":1038
 *             int scale_factor
 * 
 *         if isinstance(self, PyComposition):             # <<<<<<<<<<<<<<
 *             inst = self
 *             scale_factor = PyInt_AsLong(scale)
 */
    goto __pyx_L3;
  }

  /* "brainpy/_c/composition.pyx":1042
 *             scale_factor = PyInt_AsLong(scale)
 *         else:
 *             inst = scale             # <<<<<<<<<<<<<<
 *             scale_factor = PyInt_AsLong(self)
 * 
 */
  /*else*/ {
    if (!(likely(((__pyx_v_scale) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_scale, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1042, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_scale;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_inst = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "brainpy/_c/composition.pyx":1043
 *         else:
 *             inst = scale
 *             scale_factor = PyInt_AsLong(self)             # <<<<<<<<<<<<<<
 * 
 *         result = inst.copy()
 */
    __pyx_v_scale_factor = PyInt_AsLong(__pyx_v_self);
  }
  __pyx_L3:;

  /* "brainpy/_c/composition.pyx":1045
 *             scale_factor = PyInt_AsLong(self)
 * 
 *         result = inst.copy()             # <<<<<<<<<<<<<<
 * 
 *         composition_imul(result.impl, scale_factor)
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_inst->__pyx_vtab)->copy(__pyx_v_inst, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "brainpy/_c/composition.pyx":1047
 *         result = inst.copy()
 * 
 *         composition_imul(result.impl, scale_factor)             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
  __pyx_f_7brainpy_2_c_11composition_composition_imul(__pyx_v_result->impl, __pyx_v_scale_factor);

  /* "brainpy/_c/composition.pyx":1049
 *         composition_imul(result.impl, scale_factor)
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __imul__(self, int scale):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1032
 *         return self
 * 
 *     def __mul__(self, scale):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyComposition result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1051
 *         return result
 * 
 *     def __imul__(self, int scale):             # <<<<<<<<<<<<<<
 *         composition_imul(self.impl, scale)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_33__imul__(PyObject *__pyx_v_self, PyObject *__pyx_arg_scale); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_33__imul__(PyObject *__pyx_v_self, PyObject *__pyx_arg_scale) {
  int __pyx_v_scale;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imul__ (wrapper)", 0);
  assert(__pyx_arg_scale); {
    __pyx_v_scale = __Pyx_PyInt_As_int(__pyx_arg_scale); if (unlikely((__pyx_v_scale == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1051, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__imul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_32__imul__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((int)__pyx_v_scale));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_32__imul__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_v_scale) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__imul__", 0);

  /* "brainpy/_c/composition.pyx":1052
 * 
 *     def __imul__(self, int scale):
 *         composition_imul(self.impl, scale)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_f_7brainpy_2_c_11composition_composition_imul(__pyx_v_self->impl, __pyx_v_scale);

  /* "brainpy/_c/composition.pyx":1053
 *     def __imul__(self, int scale):
 *         composition_imul(self.impl, scale)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1051
 *         return result
 * 
 *     def __imul__(self, int scale):             # <<<<<<<<<<<<<<
 *         composition_imul(self.impl, scale)
 *         return self
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1055
 *         return self
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_35__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_35__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_34__iter__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_34__iter__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "brainpy/_c/composition.pyx":1056
 * 
 *     def __iter__(self):
 *         return iter(self.keys())             # <<<<<<<<<<<<<<
 * 
 *     cpdef double mass(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1055
 *         return self
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.keys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1058
 *         return iter(self.keys())
 * 
 *     cpdef double mass(self):             # <<<<<<<<<<<<<<
 *         '''Calculate the monoisotopic mass of this chemical composition
 * 
 */

static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_37mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_7brainpy_2_c_11composition_13PyComposition_mass(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("mass", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_37mass)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/composition.pyx":1065
 *         float
 *         '''
 *         if self._clean:             # <<<<<<<<<<<<<<
 *             return self.cached_mass
 *         else:
 */
  __pyx_t_6 = (__pyx_v_self->_clean != 0);
  if (__pyx_t_6) {

    /* "brainpy/_c/composition.pyx":1066
 *         '''
 *         if self._clean:
 *             return self.cached_mass             # <<<<<<<<<<<<<<
 *         else:
 *             self.cached_mass = composition_mass(self.impl)
 */
    __pyx_r = __pyx_v_self->cached_mass;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":1065
 *         float
 *         '''
 *         if self._clean:             # <<<<<<<<<<<<<<
 *             return self.cached_mass
 *         else:
 */
  }

  /* "brainpy/_c/composition.pyx":1068
 *             return self.cached_mass
 *         else:
 *             self.cached_mass = composition_mass(self.impl)             # <<<<<<<<<<<<<<
 *             self._clean = True
 *             return self.cached_mass
 */
  /*else*/ {
    __pyx_v_self->cached_mass = __pyx_f_7brainpy_2_c_11composition_composition_mass(__pyx_v_self->impl);

    /* "brainpy/_c/composition.pyx":1069
 *         else:
 *             self.cached_mass = composition_mass(self.impl)
 *             self._clean = True             # <<<<<<<<<<<<<<
 *             return self.cached_mass
 * 
 */
    __pyx_v_self->_clean = 1;

    /* "brainpy/_c/composition.pyx":1070
 *             self.cached_mass = composition_mass(self.impl)
 *             self._clean = True
 *             return self.cached_mass             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_r = __pyx_v_self->cached_mass;
    goto __pyx_L0;
  }

  /* "brainpy/_c/composition.pyx":1058
 *         return iter(self.keys())
 * 
 *     cpdef double mass(self):             # <<<<<<<<<<<<<<
 *         '''Calculate the monoisotopic mass of this chemical composition
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("brainpy._c.composition.PyComposition.mass", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_37mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_36mass[] = "PyComposition.mass(self) -> double\nCalculate the monoisotopic mass of this chemical composition\n\n        Returns\n        -------\n        float\n        ";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_37mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mass (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_36mass(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_36mass(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("mass", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7brainpy_2_c_11composition_13PyComposition_mass(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1072
 *             return self.cached_mass
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.impl.used
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_7brainpy_2_c_11composition_13PyComposition_39__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_7brainpy_2_c_11composition_13PyComposition_39__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_38__len__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_7brainpy_2_c_11composition_13PyComposition_38__len__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "brainpy/_c/composition.pyx":1073
 * 
 *     def __len__(self):
 *         return self.impl.used             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->impl->used;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1072
 *             return self.cached_mass
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.impl.used
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1075
 *         return self.impl.used
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_41__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_41__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_40__repr__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7brainpy_2_c_11composition_13PyComposition_8__repr___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "brainpy/_c/composition.pyx":1076
 * 
 *     def __repr__(self):
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):
 */

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_8__repr___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr(__pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1076, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7brainpy_2_c_11composition_13PyComposition_8__repr___2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_brainpy__c_composition); if (unlikely(!gen)) __PYX_ERR(0, 1076, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7brainpy_2_c_11composition_13PyComposition_8__repr___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1076, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 1076, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1076, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1076, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1076, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1076, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_kv);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_kv, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_d, __pyx_cur_scope->__pyx_v_kv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1076, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1075
 *         return self.impl.used
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 */

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_40__repr__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __pyx_cur_scope = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct____repr__(__pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct____repr__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1075, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "brainpy/_c/composition.pyx":1076
 * 
 *     def __repr__(self):
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_8__repr___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_PyComposition_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1075
 *         return self.impl.used
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1078
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):             # <<<<<<<<<<<<<<
 *         return composition_eq(self.impl, other.impl)
 * 
 */

static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_43__equality_pycomposition(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_pycomposition(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__equality_pycomposition", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equality_pycomposition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_43__equality_pycomposition)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1078, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/composition.pyx":1079
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):
 *         return composition_eq(self.impl, other.impl)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint __equality_dict(self, dict other):
 */
  __pyx_r = __pyx_f_7brainpy_2_c_11composition_composition_eq(__pyx_v_self->impl, __pyx_v_other->impl);
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1078
 *         return "PyComposition({%s})" % ', '.join("\"%s\": %d" % kv for kv in self.items())
 * 
 *     cpdef bint __equality_pycomposition(self, PyComposition other):             # <<<<<<<<<<<<<<
 *         return composition_eq(self.impl, other.impl)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("brainpy._c.composition.PyComposition.__equality_pycomposition", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_43__equality_pycomposition(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_42__equality_pycomposition[] = "PyComposition.__equality_pycomposition(self, PyComposition other) -> bool";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_43__equality_pycomposition(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__equality_pycomposition (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7brainpy_2_c_11composition_PyComposition, 1, "other", 0))) __PYX_ERR(0, 1078, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_42__equality_pycomposition(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_42__equality_pycomposition(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__equality_pycomposition", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_pycomposition(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__equality_pycomposition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1081
 *         return composition_eq(self.impl, other.impl)
 * 
 *     cpdef bint __equality_dict(self, dict other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_45__equality_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_dict(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  size_t __pyx_v_n;
  PyObject *__pyx_v_key = 0;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_value;
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_v_my_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *(*__pyx_t_9)(PyObject *);
  __pyx_t_7brainpy_2_c_11composition_count_type __pyx_t_10;
  __Pyx_RefNannySetupContext("__equality_dict", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equality_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_45__equality_dict)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1081, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1081, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "brainpy/_c/composition.pyx":1088
 *             count_type my_value
 * 
 *         n = len(other)             # <<<<<<<<<<<<<<
 *         if n != self.impl.used:
 *             return False
 */
  if (unlikely(__pyx_v_other == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1088, __pyx_L1_error)
  }
  __pyx_t_7 = PyDict_Size(__pyx_v_other); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1088, __pyx_L1_error)
  __pyx_v_n = __pyx_t_7;

  /* "brainpy/_c/composition.pyx":1089
 * 
 *         n = len(other)
 *         if n != self.impl.used:             # <<<<<<<<<<<<<<
 *             return False
 *         for key, value in other.items():
 */
  __pyx_t_6 = ((__pyx_v_n != __pyx_v_self->impl->used) != 0);
  if (__pyx_t_6) {

    /* "brainpy/_c/composition.pyx":1090
 *         n = len(other)
 *         if n != self.impl.used:
 *             return False             # <<<<<<<<<<<<<<
 *         for key, value in other.items():
 *             my_value = self[key]
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "brainpy/_c/composition.pyx":1089
 * 
 *         n = len(other)
 *         if n != self.impl.used:             # <<<<<<<<<<<<<<
 *             return False
 *         for key, value in other.items():
 */
  }

  /* "brainpy/_c/composition.pyx":1091
 *         if n != self.impl.used:
 *             return False
 *         for key, value in other.items():             # <<<<<<<<<<<<<<
 *             my_value = self[key]
 *             if my_value != value:
 */
  if (unlikely(__pyx_v_other == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1091, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Items(__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1091, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1091, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1091, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1091, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1091, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1091, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1091, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1091, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_5 = __pyx_t_9(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1091, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1091, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 1091, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;
    __pyx_v_value = __pyx_t_10;

    /* "brainpy/_c/composition.pyx":1092
 *             return False
 *         for key, value in other.items():
 *             my_value = self[key]             # <<<<<<<<<<<<<<
 *             if my_value != value:
 *                 return False
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_my_value = __pyx_t_10;

    /* "brainpy/_c/composition.pyx":1093
 *         for key, value in other.items():
 *             my_value = self[key]
 *             if my_value != value:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    __pyx_t_6 = ((__pyx_v_my_value != __pyx_v_value) != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/composition.pyx":1094
 *             my_value = self[key]
 *             if my_value != value:
 *                 return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
      __pyx_r = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1093
 *         for key, value in other.items():
 *             my_value = self[key]
 *             if my_value != value:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    }

    /* "brainpy/_c/composition.pyx":1091
 *         if n != self.impl.used:
 *             return False
 *         for key, value in other.items():             # <<<<<<<<<<<<<<
 *             my_value = self[key]
 *             if my_value != value:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":1095
 *             if my_value != value:
 *                 return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, object other, int code):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1081
 *         return composition_eq(self.impl, other.impl)
 * 
 *     cpdef bint __equality_dict(self, dict other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("brainpy._c.composition.PyComposition.__equality_dict", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_45__equality_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_44__equality_dict[] = "PyComposition.__equality_dict(self, dict other) -> bool";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_45__equality_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__equality_dict (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), (&PyDict_Type), 1, "other", 1))) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_44__equality_dict(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject*)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_44__equality_dict(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__equality_dict", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_dict(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__equality_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1097
 *         return True
 * 
 *     def __richcmp__(self, object other, int code):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, PyComposition):
 *             other, self = self, other
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_47__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_47__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_46__richcmp__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((int)__pyx_v_code));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_46__richcmp__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_code) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);

  /* "brainpy/_c/composition.pyx":1098
 * 
 *     def __richcmp__(self, object other, int code):
 *         if not isinstance(self, PyComposition):             # <<<<<<<<<<<<<<
 *             other, self = self, other
 *         if isinstance(other, PyComposition):
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":1099
 *     def __richcmp__(self, object other, int code):
 *         if not isinstance(self, PyComposition):
 *             other, self = self, other             # <<<<<<<<<<<<<<
 *         if isinstance(other, PyComposition):
 *             if code == 2:
 */
    __pyx_t_3 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_3);
    if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1099, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_self, ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":1098
 * 
 *     def __richcmp__(self, object other, int code):
 *         if not isinstance(self, PyComposition):             # <<<<<<<<<<<<<<
 *             other, self = self, other
 *         if isinstance(other, PyComposition):
 */
  }

  /* "brainpy/_c/composition.pyx":1100
 *         if not isinstance(self, PyComposition):
 *             other, self = self, other
 *         if isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self.__equality_pycomposition(other)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "brainpy/_c/composition.pyx":1101
 *             other, self = self, other
 *         if isinstance(other, PyComposition):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_pycomposition(other)
 *             elif code == 3:
 */
    switch (__pyx_v_code) {
      case 2:

      /* "brainpy/_c/composition.pyx":1102
 *         if isinstance(other, PyComposition):
 *             if code == 2:
 *                 return self.__equality_pycomposition(other)             # <<<<<<<<<<<<<<
 *             elif code == 3:
 *                 return not self.__equality_pycomposition(other)
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1102, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_pycomposition(__pyx_v_self, ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_other), 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1101
 *             other, self = self, other
 *         if isinstance(other, PyComposition):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_pycomposition(other)
 *             elif code == 3:
 */
      break;

      /* "brainpy/_c/composition.pyx":1103
 *             if code == 2:
 *                 return self.__equality_pycomposition(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):
 */
      case 3:

      /* "brainpy/_c/composition.pyx":1104
 *                 return self.__equality_pycomposition(other)
 *             elif code == 3:
 *                 return not self.__equality_pycomposition(other)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, dict):
 *             if code == 2:
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(((__pyx_v_other) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_other, __pyx_ptype_7brainpy_2_c_11composition_PyComposition))))) __PYX_ERR(0, 1104, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_pycomposition(__pyx_v_self, ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1103
 *             if code == 2:
 *                 return self.__equality_pycomposition(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):
 */
      break;
      default: break;
    }

    /* "brainpy/_c/composition.pyx":1100
 *         if not isinstance(self, PyComposition):
 *             other, self = self, other
 *         if isinstance(other, PyComposition):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self.__equality_pycomposition(other)
 */
    goto __pyx_L4;
  }

  /* "brainpy/_c/composition.pyx":1105
 *             elif code == 3:
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self.__equality_dict(other)
 */
  __pyx_t_1 = PyDict_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "brainpy/_c/composition.pyx":1106
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 */
    switch (__pyx_v_code) {
      case 2:

      /* "brainpy/_c/composition.pyx":1107
 *         elif isinstance(other, dict):
 *             if code == 2:
 *                 return self.__equality_dict(other)             # <<<<<<<<<<<<<<
 *             elif code == 3:
 *                 return not self.__equality_dict(other)
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(PyDict_CheckExact(__pyx_v_other))||((__pyx_v_other) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_other)->tp_name), 0))) __PYX_ERR(0, 1107, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_dict(__pyx_v_self, ((PyObject*)__pyx_v_other), 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1106
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 */
      break;

      /* "brainpy/_c/composition.pyx":1108
 *             if code == 2:
 *                 return self.__equality_dict(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_dict(other)
 *         else:
 */
      case 3:

      /* "brainpy/_c/composition.pyx":1109
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 *                 return not self.__equality_dict(other)             # <<<<<<<<<<<<<<
 *         else:
 *             other = dict(other)
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(PyDict_CheckExact(__pyx_v_other))||((__pyx_v_other) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_other)->tp_name), 0))) __PYX_ERR(0, 1109, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_dict(__pyx_v_self, ((PyObject*)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1108
 *             if code == 2:
 *                 return self.__equality_dict(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_dict(other)
 *         else:
 */
      break;
      default: break;
    }

    /* "brainpy/_c/composition.pyx":1105
 *             elif code == 3:
 *                 return not self.__equality_pycomposition(other)
 *         elif isinstance(other, dict):             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self.__equality_dict(other)
 */
    goto __pyx_L4;
  }

  /* "brainpy/_c/composition.pyx":1111
 *                 return not self.__equality_dict(other)
 *         else:
 *             other = dict(other)             # <<<<<<<<<<<<<<
 *             if code == 2:
 *                 return self.__equality_dict(other)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_v_other); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "brainpy/_c/composition.pyx":1112
 *         else:
 *             other = dict(other)
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 */
    switch (__pyx_v_code) {
      case 2:

      /* "brainpy/_c/composition.pyx":1113
 *             other = dict(other)
 *             if code == 2:
 *                 return self.__equality_dict(other)             # <<<<<<<<<<<<<<
 *             elif code == 3:
 *                 return not self.__equality_dict(other)
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(PyDict_CheckExact(__pyx_v_other))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_other)->tp_name), 0))) __PYX_ERR(0, 1113, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_dict(__pyx_v_self, ((PyObject*)__pyx_v_other), 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1113, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1112
 *         else:
 *             other = dict(other)
 *             if code == 2:             # <<<<<<<<<<<<<<
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 */
      break;

      /* "brainpy/_c/composition.pyx":1114
 *             if code == 2:
 *                 return self.__equality_dict(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_dict(other)
 * 
 */
      case 3:

      /* "brainpy/_c/composition.pyx":1115
 *                 return self.__equality_dict(other)
 *             elif code == 3:
 *                 return not self.__equality_dict(other)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      if (!(likely(PyDict_CheckExact(__pyx_v_other))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_other)->tp_name), 0))) __PYX_ERR(0, 1115, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyBool_FromLong((!(((struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self->__pyx_vtab)->__pyx___equality_dict(__pyx_v_self, ((PyObject*)__pyx_v_other), 0) != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "brainpy/_c/composition.pyx":1114
 *             if code == 2:
 *                 return self.__equality_dict(other)
 *             elif code == 3:             # <<<<<<<<<<<<<<
 *                 return not self.__equality_dict(other)
 * 
 */
      break;
      default: break;
    }
  }
  __pyx_L4:;

  /* "brainpy/_c/composition.pyx":1097
 *         return True
 * 
 *     def __richcmp__(self, object other, int code):             # <<<<<<<<<<<<<<
 *         if not isinstance(self, PyComposition):
 *             other, self = self, other
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pxd":120
 *     cdef:
 *         Composition* impl
 *         public double cached_mass             # <<<<<<<<<<<<<<
 *         public bint _clean
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass___get__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass___get__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->cached_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.cached_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass_2__set__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_11cached_mass_2__set__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 120, __pyx_L1_error)
  __pyx_v_self->cached_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.cached_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pxd":121
 *         Composition* impl
 *         public double cached_mass
 *         public bint _clean             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef PyComposition _create(Composition* base)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean___get__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean___get__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_clean); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition._clean.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean_2__set__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7brainpy_2_c_11composition_13PyComposition_6_clean_2__set__(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 121, __pyx_L1_error)
  __pyx_v_self->_clean = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition._clean.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_49__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_48__reduce_cython__[] = "PyComposition.__reduce_cython__(self)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_49__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_48__reduce_cython__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_48__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_51__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_13PyComposition_50__setstate_cython__[] = "PyComposition.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_13PyComposition_51__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_13PyComposition_50__setstate_cython__(((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_13PyComposition_50__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.PyComposition.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "brainpy/_c/composition.pyx":1129
 * 
 * 
 * cpdef PyComposition parse_formula(str formula):             # <<<<<<<<<<<<<<
 *     """Parse a chemical formula and construct a :class:`PyComposition` object
 * 
 */

static PyObject *__pyx_pw_7brainpy_2_c_11composition_11parse_formula(PyObject *__pyx_self, PyObject *__pyx_v_formula); /*proto*/
static struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_f_7brainpy_2_c_11composition_parse_formula(PyObject *__pyx_v_formula, CYTHON_UNUSED int __pyx_skip_dispatch) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_n;
  Py_ssize_t __pyx_v_elstart;
  Py_ssize_t __pyx_v_elend;
  Py_ssize_t __pyx_v_numstart;
  Py_ssize_t __pyx_v_numend;
  Py_ssize_t __pyx_v_isostart;
  Py_ssize_t __pyx_v_isoend;
  char *__pyx_v_cstr;
  char *__pyx_v_temp;
  char __pyx_v_a;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_elem;
  struct __pyx_t_7brainpy_2_c_11composition_Element *__pyx_v_fixed_isotope_elem;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_v_composition = 0;
  int __pyx_v_state;
  int __pyx_v_count;
  int __pyx_v_prev_count;
  int __pyx_v_fixed_isotope;
  int __pyx_v_found;
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("parse_formula", 0);

  /* "brainpy/_c/composition.pyx":1157
 *         PyComposition composition
 *         int state, count, prev_count, fixed_isotope, status, found
 *     prev_count = 0             # <<<<<<<<<<<<<<
 *     state = ELEMENT
 *     cstr = PyString_AsString(formula)
 */
  __pyx_v_prev_count = 0;

  /* "brainpy/_c/composition.pyx":1158
 *         int state, count, prev_count, fixed_isotope, status, found
 *     prev_count = 0
 *     state = ELEMENT             # <<<<<<<<<<<<<<
 *     cstr = PyString_AsString(formula)
 *     temp = <char*>malloc(sizeof(char) * 10)
 */
  __pyx_v_state = __pyx_e_7brainpy_2_c_11composition_ELEMENT;

  /* "brainpy/_c/composition.pyx":1159
 *     prev_count = 0
 *     state = ELEMENT
 *     cstr = PyString_AsString(formula)             # <<<<<<<<<<<<<<
 *     temp = <char*>malloc(sizeof(char) * 10)
 *     for i in range(10):
 */
  __pyx_v_cstr = PyStr_AsString(__pyx_v_formula);

  /* "brainpy/_c/composition.pyx":1160
 *     state = ELEMENT
 *     cstr = PyString_AsString(formula)
 *     temp = <char*>malloc(sizeof(char) * 10)             # <<<<<<<<<<<<<<
 *     for i in range(10):
 *         temp[i] = "\0"
 */
  __pyx_v_temp = ((char *)malloc(((sizeof(char)) * 10)));

  /* "brainpy/_c/composition.pyx":1161
 *     cstr = PyString_AsString(formula)
 *     temp = <char*>malloc(sizeof(char) * 10)
 *     for i in range(10):             # <<<<<<<<<<<<<<
 *         temp[i] = "\0"
 *     n = len(formula)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 10; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "brainpy/_c/composition.pyx":1162
 *     temp = <char*>malloc(sizeof(char) * 10)
 *     for i in range(10):
 *         temp[i] = "\0"             # <<<<<<<<<<<<<<
 *     n = len(formula)
 *     elstart = isostart = isoend = 0
 */
    (__pyx_v_temp[__pyx_v_i]) = '\x00';
  }

  /* "brainpy/_c/composition.pyx":1163
 *     for i in range(10):
 *         temp[i] = "\0"
 *     n = len(formula)             # <<<<<<<<<<<<<<
 *     elstart = isostart = isoend = 0
 *     composition = PyComposition()
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_formula); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1163, __pyx_L1_error)
  __pyx_v_n = __pyx_t_2;

  /* "brainpy/_c/composition.pyx":1164
 *         temp[i] = "\0"
 *     n = len(formula)
 *     elstart = isostart = isoend = 0             # <<<<<<<<<<<<<<
 *     composition = PyComposition()
 *     for i in range(n):
 */
  __pyx_v_elstart = 0;
  __pyx_v_isostart = 0;
  __pyx_v_isoend = 0;

  /* "brainpy/_c/composition.pyx":1165
 *     n = len(formula)
 *     elstart = isostart = isoend = 0
 *     composition = PyComposition()             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         a = cstr[i]
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7brainpy_2_c_11composition_PyComposition)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_composition = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "brainpy/_c/composition.pyx":1166
 *     elstart = isostart = isoend = 0
 *     composition = PyComposition()
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         a = cstr[i]
 *         if a == ']':
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_4 = __pyx_t_1;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "brainpy/_c/composition.pyx":1167
 *     composition = PyComposition()
 *     for i in range(n):
 *         a = cstr[i]             # <<<<<<<<<<<<<<
 *         if a == ']':
 *             if state == ISOTOPE:
 */
    __pyx_v_a = (__pyx_v_cstr[__pyx_v_i]);

    /* "brainpy/_c/composition.pyx":1168
 *     for i in range(n):
 *         a = cstr[i]
 *         if a == ']':             # <<<<<<<<<<<<<<
 *             if state == ISOTOPE:
 *                 elend = i
 */
    __pyx_t_6 = ((__pyx_v_a == ']') != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/composition.pyx":1169
 *         a = cstr[i]
 *         if a == ']':
 *             if state == ISOTOPE:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 isoend = i
 */
      __pyx_t_6 = ((__pyx_v_state == __pyx_e_7brainpy_2_c_11composition_ISOTOPE) != 0);
      if (likely(__pyx_t_6)) {

        /* "brainpy/_c/composition.pyx":1170
 *         if a == ']':
 *             if state == ISOTOPE:
 *                 elend = i             # <<<<<<<<<<<<<<
 *                 isoend = i
 *                 numstart = i + 1
 */
        __pyx_v_elend = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1171
 *             if state == ISOTOPE:
 *                 elend = i
 *                 isoend = i             # <<<<<<<<<<<<<<
 *                 numstart = i + 1
 *                 state = COUNT
 */
        __pyx_v_isoend = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1172
 *                 elend = i
 *                 isoend = i
 *                 numstart = i + 1             # <<<<<<<<<<<<<<
 *                 state = COUNT
 *             else:
 */
        __pyx_v_numstart = (__pyx_v_i + 1);

        /* "brainpy/_c/composition.pyx":1173
 *                 isoend = i
 *                 numstart = i + 1
 *                 state = COUNT             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError((formula, i, state))
 */
        __pyx_v_state = __pyx_e_7brainpy_2_c_11composition_COUNT;

        /* "brainpy/_c/composition.pyx":1169
 *         a = cstr[i]
 *         if a == ']':
 *             if state == ISOTOPE:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 isoend = i
 */
        goto __pyx_L8;
      }

      /* "brainpy/_c/composition.pyx":1175
 *                 state = COUNT
 *             else:
 *                 raise ValueError((formula, i, state))             # <<<<<<<<<<<<<<
 *         elif a == '[':
 *             if state == ELEMENT:
 */
      /*else*/ {
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_state); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_formula);
        __Pyx_GIVEREF(__pyx_v_formula);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_formula);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_3 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1175, __pyx_L1_error)
      }
      __pyx_L8:;

      /* "brainpy/_c/composition.pyx":1168
 *     for i in range(n):
 *         a = cstr[i]
 *         if a == ']':             # <<<<<<<<<<<<<<
 *             if state == ISOTOPE:
 *                 elend = i
 */
      goto __pyx_L7;
    }

    /* "brainpy/_c/composition.pyx":1176
 *             else:
 *                 raise ValueError((formula, i, state))
 *         elif a == '[':             # <<<<<<<<<<<<<<
 *             if state == ELEMENT:
 *                 elend = i
 */
    __pyx_t_6 = ((__pyx_v_a == '[') != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/composition.pyx":1177
 *                 raise ValueError((formula, i, state))
 *         elif a == '[':
 *             if state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 isostart = i
 */
      __pyx_t_6 = ((__pyx_v_state == __pyx_e_7brainpy_2_c_11composition_ELEMENT) != 0);
      if (likely(__pyx_t_6)) {

        /* "brainpy/_c/composition.pyx":1178
 *         elif a == '[':
 *             if state == ELEMENT:
 *                 elend = i             # <<<<<<<<<<<<<<
 *                 isostart = i
 *                 state = ISOTOPE
 */
        __pyx_v_elend = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1179
 *             if state == ELEMENT:
 *                 elend = i
 *                 isostart = i             # <<<<<<<<<<<<<<
 *                 state = ISOTOPE
 *             else:
 */
        __pyx_v_isostart = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1180
 *                 elend = i
 *                 isostart = i
 *                 state = ISOTOPE             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError((formula, i, state))
 */
        __pyx_v_state = __pyx_e_7brainpy_2_c_11composition_ISOTOPE;

        /* "brainpy/_c/composition.pyx":1177
 *                 raise ValueError((formula, i, state))
 *         elif a == '[':
 *             if state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 isostart = i
 */
        goto __pyx_L9;
      }

      /* "brainpy/_c/composition.pyx":1182
 *                 state = ISOTOPE
 *             else:
 *                 raise ValueError((formula, i, state))             # <<<<<<<<<<<<<<
 *         elif isdigit(a):
 *             if state == ISOTOPE or state == COUNT:
 */
      /*else*/ {
        __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_state); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_formula);
        __Pyx_GIVEREF(__pyx_v_formula);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_formula);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_8);
        __pyx_t_7 = 0;
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 1182, __pyx_L1_error)
      }
      __pyx_L9:;

      /* "brainpy/_c/composition.pyx":1176
 *             else:
 *                 raise ValueError((formula, i, state))
 *         elif a == '[':             # <<<<<<<<<<<<<<
 *             if state == ELEMENT:
 *                 elend = i
 */
      goto __pyx_L7;
    }

    /* "brainpy/_c/composition.pyx":1183
 *             else:
 *                 raise ValueError((formula, i, state))
 *         elif isdigit(a):             # <<<<<<<<<<<<<<
 *             if state == ISOTOPE or state == COUNT:
 *                 pass
 */
    __pyx_t_6 = (isdigit(__pyx_v_a) != 0);
    if (__pyx_t_6) {

      /* "brainpy/_c/composition.pyx":1184
 *                 raise ValueError((formula, i, state))
 *         elif isdigit(a):
 *             if state == ISOTOPE or state == COUNT:             # <<<<<<<<<<<<<<
 *                 pass
 *             elif state == ELEMENT:
 */
      switch (__pyx_v_state) {
        case __pyx_e_7brainpy_2_c_11composition_ISOTOPE:
        case __pyx_e_7brainpy_2_c_11composition_COUNT:
        break;

        /* "brainpy/_c/composition.pyx":1186
 *             if state == ISOTOPE or state == COUNT:
 *                 pass
 *             elif state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 numstart = i
 */
        case __pyx_e_7brainpy_2_c_11composition_ELEMENT:

        /* "brainpy/_c/composition.pyx":1187
 *                 pass
 *             elif state == ELEMENT:
 *                 elend = i             # <<<<<<<<<<<<<<
 *                 numstart = i
 *                 state = COUNT
 */
        __pyx_v_elend = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1188
 *             elif state == ELEMENT:
 *                 elend = i
 *                 numstart = i             # <<<<<<<<<<<<<<
 *                 state = COUNT
 *             else:
 */
        __pyx_v_numstart = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1189
 *                 elend = i
 *                 numstart = i
 *                 state = COUNT             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError((formula, i, state))
 */
        __pyx_v_state = __pyx_e_7brainpy_2_c_11composition_COUNT;

        /* "brainpy/_c/composition.pyx":1186
 *             if state == ISOTOPE or state == COUNT:
 *                 pass
 *             elif state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 elend = i
 *                 numstart = i
 */
        break;
        default:

        /* "brainpy/_c/composition.pyx":1191
 *                 state = COUNT
 *             else:
 *                 raise ValueError((formula, i, state))             # <<<<<<<<<<<<<<
 *         else:
 *             if state == ELEMENT:
 */
        __pyx_t_8 = PyInt_FromSsize_t(__pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_state); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_formula);
        __Pyx_GIVEREF(__pyx_v_formula);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_formula);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
        __pyx_t_8 = 0;
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 1191, __pyx_L1_error)
        break;
      }

      /* "brainpy/_c/composition.pyx":1183
 *             else:
 *                 raise ValueError((formula, i, state))
 *         elif isdigit(a):             # <<<<<<<<<<<<<<
 *             if state == ISOTOPE or state == COUNT:
 *                 pass
 */
      goto __pyx_L7;
    }

    /* "brainpy/_c/composition.pyx":1193
 *                 raise ValueError((formula, i, state))
 *         else:
 *             if state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 continue
 *             elif state == COUNT:
 */
    /*else*/ {

      /* "brainpy/_c/composition.pyx":1195
 *             if state == ELEMENT:
 *                 continue
 *             elif state == COUNT:             # <<<<<<<<<<<<<<
 *                 numend = i
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 */
      switch (__pyx_v_state) {

        /* "brainpy/_c/composition.pyx":1193
 *                 raise ValueError((formula, i, state))
 *         else:
 *             if state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 continue
 *             elif state == COUNT:
 */
        case __pyx_e_7brainpy_2_c_11composition_ELEMENT:

        /* "brainpy/_c/composition.pyx":1194
 *         else:
 *             if state == ELEMENT:
 *                 continue             # <<<<<<<<<<<<<<
 *             elif state == COUNT:
 *                 numend = i
 */
        goto __pyx_L5_continue;

        /* "brainpy/_c/composition.pyx":1193
 *                 raise ValueError((formula, i, state))
 *         else:
 *             if state == ELEMENT:             # <<<<<<<<<<<<<<
 *                 continue
 *             elif state == COUNT:
 */
        break;

        /* "brainpy/_c/composition.pyx":1195
 *             if state == ELEMENT:
 *                 continue
 *             elif state == COUNT:             # <<<<<<<<<<<<<<
 *                 numend = i
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 */
        case __pyx_e_7brainpy_2_c_11composition_COUNT:

        /* "brainpy/_c/composition.pyx":1196
 *                 continue
 *             elif state == COUNT:
 *                 numend = i             # <<<<<<<<<<<<<<
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 *                 temp[numend - numstart] = 0
 */
        __pyx_v_numend = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1197
 *             elif state == COUNT:
 *                 numend = i
 *                 strncpy(temp, cstr + numstart, numend - numstart)             # <<<<<<<<<<<<<<
 *                 temp[numend - numstart] = 0
 *                 count = atoi(temp)
 */
        (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_numstart), (__pyx_v_numend - __pyx_v_numstart)));

        /* "brainpy/_c/composition.pyx":1198
 *                 numend = i
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 *                 temp[numend - numstart] = 0             # <<<<<<<<<<<<<<
 *                 count = atoi(temp)
 *                 strncpy(temp, cstr + elstart, elend - elstart)
 */
        (__pyx_v_temp[(__pyx_v_numend - __pyx_v_numstart)]) = 0;

        /* "brainpy/_c/composition.pyx":1199
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 *                 temp[numend - numstart] = 0
 *                 count = atoi(temp)             # <<<<<<<<<<<<<<
 *                 strncpy(temp, cstr + elstart, elend - elstart)
 *                 temp[elend - elstart] = 0
 */
        __pyx_v_count = atoi(__pyx_v_temp);

        /* "brainpy/_c/composition.pyx":1200
 *                 temp[numend - numstart] = 0
 *                 count = atoi(temp)
 *                 strncpy(temp, cstr + elstart, elend - elstart)             # <<<<<<<<<<<<<<
 *                 temp[elend - elstart] = 0
 *                 found = element_hash_table_get(_ElementTable, temp, &elem)
 */
        (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_elstart), (__pyx_v_elend - __pyx_v_elstart)));

        /* "brainpy/_c/composition.pyx":1201
 *                 count = atoi(temp)
 *                 strncpy(temp, cstr + elstart, elend - elstart)
 *                 temp[elend - elstart] = 0             # <<<<<<<<<<<<<<
 *                 found = element_hash_table_get(_ElementTable, temp, &elem)
 *                 # something in this loop is suspicious and causes periodic crashes.
 */
        (__pyx_v_temp[(__pyx_v_elend - __pyx_v_elstart)]) = 0;

        /* "brainpy/_c/composition.pyx":1202
 *                 strncpy(temp, cstr + elstart, elend - elstart)
 *                 temp[elend - elstart] = 0
 *                 found = element_hash_table_get(_ElementTable, temp, &elem)             # <<<<<<<<<<<<<<
 *                 # something in this loop is suspicious and causes periodic crashes.
 *                 if isostart != 0 and isoend != 0 and found != 0:
 */
        __pyx_v_found = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_temp, (&__pyx_v_elem));

        /* "brainpy/_c/composition.pyx":1204
 *                 found = element_hash_table_get(_ElementTable, temp, &elem)
 *                 # something in this loop is suspicious and causes periodic crashes.
 *                 if isostart != 0 and isoend != 0 and found != 0:             # <<<<<<<<<<<<<<
 *                     isostart += 1
 *                     strncpy(temp, cstr + isostart, isoend - isostart)
 */
        __pyx_t_9 = ((__pyx_v_isostart != 0) != 0);
        if (__pyx_t_9) {
        } else {
          __pyx_t_6 = __pyx_t_9;
          goto __pyx_L11_bool_binop_done;
        }
        __pyx_t_9 = ((__pyx_v_isoend != 0) != 0);
        if (__pyx_t_9) {
        } else {
          __pyx_t_6 = __pyx_t_9;
          goto __pyx_L11_bool_binop_done;
        }
        __pyx_t_9 = ((__pyx_v_found != 0) != 0);
        __pyx_t_6 = __pyx_t_9;
        __pyx_L11_bool_binop_done:;
        if (__pyx_t_6) {

          /* "brainpy/_c/composition.pyx":1205
 *                 # something in this loop is suspicious and causes periodic crashes.
 *                 if isostart != 0 and isoend != 0 and found != 0:
 *                     isostart += 1             # <<<<<<<<<<<<<<
 *                     strncpy(temp, cstr + isostart, isoend - isostart)
 *                     temp[isoend - isostart] = 0
 */
          __pyx_v_isostart = (__pyx_v_isostart + 1);

          /* "brainpy/_c/composition.pyx":1206
 *                 if isostart != 0 and isoend != 0 and found != 0:
 *                     isostart += 1
 *                     strncpy(temp, cstr + isostart, isoend - isostart)             # <<<<<<<<<<<<<<
 *                     temp[isoend - isostart] = 0
 *                     fixed_isotope = atoi(temp)
 */
          (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_isostart), (__pyx_v_isoend - __pyx_v_isostart)));

          /* "brainpy/_c/composition.pyx":1207
 *                     isostart += 1
 *                     strncpy(temp, cstr + isostart, isoend - isostart)
 *                     temp[isoend - isostart] = 0             # <<<<<<<<<<<<<<
 *                     fixed_isotope = atoi(temp)
 * 
 */
          (__pyx_v_temp[(__pyx_v_isoend - __pyx_v_isostart)]) = 0;

          /* "brainpy/_c/composition.pyx":1208
 *                     strncpy(temp, cstr + isostart, isoend - isostart)
 *                     temp[isoend - isostart] = 0
 *                     fixed_isotope = atoi(temp)             # <<<<<<<<<<<<<<
 * 
 *                     strncpy(temp, cstr + elstart, isoend - elstart + 1)
 */
          __pyx_v_fixed_isotope = atoi(__pyx_v_temp);

          /* "brainpy/_c/composition.pyx":1210
 *                     fixed_isotope = atoi(temp)
 * 
 *                     strncpy(temp, cstr + elstart, isoend - elstart + 1)             # <<<<<<<<<<<<<<
 *                     temp[isoend - elstart + 2] = 0
 * 
 */
          (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_elstart), ((__pyx_v_isoend - __pyx_v_elstart) + 1)));

          /* "brainpy/_c/composition.pyx":1211
 * 
 *                     strncpy(temp, cstr + elstart, isoend - elstart + 1)
 *                     temp[isoend - elstart + 2] = 0             # <<<<<<<<<<<<<<
 * 
 *                     found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 */
          (__pyx_v_temp[((__pyx_v_isoend - __pyx_v_elstart) + 2)]) = 0;

          /* "brainpy/_c/composition.pyx":1213
 *                     temp[isoend - elstart + 2] = 0
 * 
 *                     found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)             # <<<<<<<<<<<<<<
 *                     if found != 0:
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 */
          __pyx_v_found = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_temp, (&__pyx_v_fixed_isotope_elem));

          /* "brainpy/_c/composition.pyx":1214
 * 
 *                     found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *                     if found != 0:             # <<<<<<<<<<<<<<
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                         element_hash_table_put(_ElementTable, fixed_isotope_elem)
 */
          __pyx_t_6 = ((__pyx_v_found != 0) != 0);
          if (__pyx_t_6) {

            /* "brainpy/_c/composition.pyx":1215
 *                     found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *                     if found != 0:
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)             # <<<<<<<<<<<<<<
 *                         element_hash_table_put(_ElementTable, fixed_isotope_elem)
 *                         elem = fixed_isotope_elem
 */
            __pyx_v_fixed_isotope_elem = __pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(__pyx_v_elem, __pyx_v_fixed_isotope);

            /* "brainpy/_c/composition.pyx":1216
 *                     if found != 0:
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                         element_hash_table_put(_ElementTable, fixed_isotope_elem)             # <<<<<<<<<<<<<<
 *                         elem = fixed_isotope_elem
 *                     else:
 */
            (void)(__pyx_f_7brainpy_2_c_11composition_element_hash_table_put(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_fixed_isotope_elem));

            /* "brainpy/_c/composition.pyx":1217
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                         element_hash_table_put(_ElementTable, fixed_isotope_elem)
 *                         elem = fixed_isotope_elem             # <<<<<<<<<<<<<<
 *                     else:
 *                         elem = fixed_isotope_elem
 */
            __pyx_v_elem = __pyx_v_fixed_isotope_elem;

            /* "brainpy/_c/composition.pyx":1214
 * 
 *                     found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *                     if found != 0:             # <<<<<<<<<<<<<<
 *                         fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                         element_hash_table_put(_ElementTable, fixed_isotope_elem)
 */
            goto __pyx_L14;
          }

          /* "brainpy/_c/composition.pyx":1219
 *                         elem = fixed_isotope_elem
 *                     else:
 *                         elem = fixed_isotope_elem             # <<<<<<<<<<<<<<
 *                 prev_count = 0
 *                 composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 */
          /*else*/ {
            __pyx_v_elem = __pyx_v_fixed_isotope_elem;
          }
          __pyx_L14:;

          /* "brainpy/_c/composition.pyx":1204
 *                 found = element_hash_table_get(_ElementTable, temp, &elem)
 *                 # something in this loop is suspicious and causes periodic crashes.
 *                 if isostart != 0 and isoend != 0 and found != 0:             # <<<<<<<<<<<<<<
 *                     isostart += 1
 *                     strncpy(temp, cstr + isostart, isoend - isostart)
 */
        }

        /* "brainpy/_c/composition.pyx":1220
 *                     else:
 *                         elem = fixed_isotope_elem
 *                 prev_count = 0             # <<<<<<<<<<<<<<
 *                 composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 *                 composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 */
        __pyx_v_prev_count = 0;

        /* "brainpy/_c/composition.pyx":1221
 *                         elem = fixed_isotope_elem
 *                 prev_count = 0
 *                 composition_get_element_count(composition.impl, elem.symbol, &prev_count)             # <<<<<<<<<<<<<<
 *                 composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 *                 state = ELEMENT
 */
        (void)(__pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition->impl, __pyx_v_elem->symbol, (&__pyx_v_prev_count)));

        /* "brainpy/_c/composition.pyx":1222
 *                 prev_count = 0
 *                 composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 *                 composition_set_element_count(composition.impl, elem.symbol, count + prev_count)             # <<<<<<<<<<<<<<
 *                 state = ELEMENT
 *                 elstart = i
 */
        (void)(__pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_composition->impl, __pyx_v_elem->symbol, (__pyx_v_count + __pyx_v_prev_count)));

        /* "brainpy/_c/composition.pyx":1223
 *                 composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 *                 composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 *                 state = ELEMENT             # <<<<<<<<<<<<<<
 *                 elstart = i
 *                 isostart = 0
 */
        __pyx_v_state = __pyx_e_7brainpy_2_c_11composition_ELEMENT;

        /* "brainpy/_c/composition.pyx":1224
 *                 composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 *                 state = ELEMENT
 *                 elstart = i             # <<<<<<<<<<<<<<
 *                 isostart = 0
 *                 isoend = 0
 */
        __pyx_v_elstart = __pyx_v_i;

        /* "brainpy/_c/composition.pyx":1225
 *                 state = ELEMENT
 *                 elstart = i
 *                 isostart = 0             # <<<<<<<<<<<<<<
 *                 isoend = 0
 * 
 */
        __pyx_v_isostart = 0;

        /* "brainpy/_c/composition.pyx":1226
 *                 elstart = i
 *                 isostart = 0
 *                 isoend = 0             # <<<<<<<<<<<<<<
 * 
 *     if state == COUNT:
 */
        __pyx_v_isoend = 0;

        /* "brainpy/_c/composition.pyx":1195
 *             if state == ELEMENT:
 *                 continue
 *             elif state == COUNT:             # <<<<<<<<<<<<<<
 *                 numend = i
 *                 strncpy(temp, cstr + numstart, numend - numstart)
 */
        break;
        default: break;
      }
    }
    __pyx_L7:;
    __pyx_L5_continue:;
  }

  /* "brainpy/_c/composition.pyx":1228
 *                 isoend = 0
 * 
 *     if state == COUNT:             # <<<<<<<<<<<<<<
 *         numend = i + 1
 *         strncpy(temp, cstr + numstart, numend - numstart)
 */
  __pyx_t_6 = ((__pyx_v_state == __pyx_e_7brainpy_2_c_11composition_COUNT) != 0);
  if (__pyx_t_6) {

    /* "brainpy/_c/composition.pyx":1229
 * 
 *     if state == COUNT:
 *         numend = i + 1             # <<<<<<<<<<<<<<
 *         strncpy(temp, cstr + numstart, numend - numstart)
 *         temp[numend - numstart] = 0
 */
    __pyx_v_numend = (__pyx_v_i + 1);

    /* "brainpy/_c/composition.pyx":1230
 *     if state == COUNT:
 *         numend = i + 1
 *         strncpy(temp, cstr + numstart, numend - numstart)             # <<<<<<<<<<<<<<
 *         temp[numend - numstart] = 0
 *         count = atoi(temp)
 */
    (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_numstart), (__pyx_v_numend - __pyx_v_numstart)));

    /* "brainpy/_c/composition.pyx":1231
 *         numend = i + 1
 *         strncpy(temp, cstr + numstart, numend - numstart)
 *         temp[numend - numstart] = 0             # <<<<<<<<<<<<<<
 *         count = atoi(temp)
 *         strncpy(temp, cstr + elstart, elend - elstart)
 */
    (__pyx_v_temp[(__pyx_v_numend - __pyx_v_numstart)]) = 0;

    /* "brainpy/_c/composition.pyx":1232
 *         strncpy(temp, cstr + numstart, numend - numstart)
 *         temp[numend - numstart] = 0
 *         count = atoi(temp)             # <<<<<<<<<<<<<<
 *         strncpy(temp, cstr + elstart, elend - elstart)
 *         temp[elend - elstart] = 0
 */
    __pyx_v_count = atoi(__pyx_v_temp);

    /* "brainpy/_c/composition.pyx":1233
 *         temp[numend - numstart] = 0
 *         count = atoi(temp)
 *         strncpy(temp, cstr + elstart, elend - elstart)             # <<<<<<<<<<<<<<
 *         temp[elend - elstart] = 0
 *         found = element_hash_table_get(_ElementTable, temp, &elem)
 */
    (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_elstart), (__pyx_v_elend - __pyx_v_elstart)));

    /* "brainpy/_c/composition.pyx":1234
 *         count = atoi(temp)
 *         strncpy(temp, cstr + elstart, elend - elstart)
 *         temp[elend - elstart] = 0             # <<<<<<<<<<<<<<
 *         found = element_hash_table_get(_ElementTable, temp, &elem)
 *         if isostart != 0 and isoend != 0 and found != 0:
 */
    (__pyx_v_temp[(__pyx_v_elend - __pyx_v_elstart)]) = 0;

    /* "brainpy/_c/composition.pyx":1235
 *         strncpy(temp, cstr + elstart, elend - elstart)
 *         temp[elend - elstart] = 0
 *         found = element_hash_table_get(_ElementTable, temp, &elem)             # <<<<<<<<<<<<<<
 *         if isostart != 0 and isoend != 0 and found != 0:
 *             isostart += 1
 */
    __pyx_v_found = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_temp, (&__pyx_v_elem));

    /* "brainpy/_c/composition.pyx":1236
 *         temp[elend - elstart] = 0
 *         found = element_hash_table_get(_ElementTable, temp, &elem)
 *         if isostart != 0 and isoend != 0 and found != 0:             # <<<<<<<<<<<<<<
 *             isostart += 1
 *             strncpy(temp, cstr + isostart, isoend - isostart)
 */
    __pyx_t_9 = ((__pyx_v_isostart != 0) != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_6 = __pyx_t_9;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_9 = ((__pyx_v_isoend != 0) != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_6 = __pyx_t_9;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_9 = ((__pyx_v_found != 0) != 0);
    __pyx_t_6 = __pyx_t_9;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_6) {

      /* "brainpy/_c/composition.pyx":1237
 *         found = element_hash_table_get(_ElementTable, temp, &elem)
 *         if isostart != 0 and isoend != 0 and found != 0:
 *             isostart += 1             # <<<<<<<<<<<<<<
 *             strncpy(temp, cstr + isostart, isoend - isostart)
 *             temp[isoend - isostart] = 0
 */
      __pyx_v_isostart = (__pyx_v_isostart + 1);

      /* "brainpy/_c/composition.pyx":1238
 *         if isostart != 0 and isoend != 0 and found != 0:
 *             isostart += 1
 *             strncpy(temp, cstr + isostart, isoend - isostart)             # <<<<<<<<<<<<<<
 *             temp[isoend - isostart] = 0
 * 
 */
      (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_isostart), (__pyx_v_isoend - __pyx_v_isostart)));

      /* "brainpy/_c/composition.pyx":1239
 *             isostart += 1
 *             strncpy(temp, cstr + isostart, isoend - isostart)
 *             temp[isoend - isostart] = 0             # <<<<<<<<<<<<<<
 * 
 *             fixed_isotope = atoi(temp)
 */
      (__pyx_v_temp[(__pyx_v_isoend - __pyx_v_isostart)]) = 0;

      /* "brainpy/_c/composition.pyx":1241
 *             temp[isoend - isostart] = 0
 * 
 *             fixed_isotope = atoi(temp)             # <<<<<<<<<<<<<<
 * 
 *             strncpy(temp, cstr + elstart, isoend - elstart + 1)
 */
      __pyx_v_fixed_isotope = atoi(__pyx_v_temp);

      /* "brainpy/_c/composition.pyx":1243
 *             fixed_isotope = atoi(temp)
 * 
 *             strncpy(temp, cstr + elstart, isoend - elstart + 1)             # <<<<<<<<<<<<<<
 *             temp[isoend - elstart + 2] = 0
 * 
 */
      (void)(strncpy(__pyx_v_temp, (__pyx_v_cstr + __pyx_v_elstart), ((__pyx_v_isoend - __pyx_v_elstart) + 1)));

      /* "brainpy/_c/composition.pyx":1244
 * 
 *             strncpy(temp, cstr + elstart, isoend - elstart + 1)
 *             temp[isoend - elstart + 2] = 0             # <<<<<<<<<<<<<<
 * 
 *             found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 */
      (__pyx_v_temp[((__pyx_v_isoend - __pyx_v_elstart) + 2)]) = 0;

      /* "brainpy/_c/composition.pyx":1246
 *             temp[isoend - elstart + 2] = 0
 * 
 *             found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)             # <<<<<<<<<<<<<<
 *             if found != 0:
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 */
      __pyx_v_found = __pyx_f_7brainpy_2_c_11composition_element_hash_table_get(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_temp, (&__pyx_v_fixed_isotope_elem));

      /* "brainpy/_c/composition.pyx":1247
 * 
 *             found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *             if found != 0:             # <<<<<<<<<<<<<<
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                 element_hash_table_put(_ElementTable, fixed_isotope_elem)
 */
      __pyx_t_6 = ((__pyx_v_found != 0) != 0);
      if (__pyx_t_6) {

        /* "brainpy/_c/composition.pyx":1248
 *             found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *             if found != 0:
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)             # <<<<<<<<<<<<<<
 *                 element_hash_table_put(_ElementTable, fixed_isotope_elem)
 *                 elem = fixed_isotope_elem
 */
        __pyx_v_fixed_isotope_elem = __pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element(__pyx_v_elem, __pyx_v_fixed_isotope);

        /* "brainpy/_c/composition.pyx":1249
 *             if found != 0:
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                 element_hash_table_put(_ElementTable, fixed_isotope_elem)             # <<<<<<<<<<<<<<
 *                 elem = fixed_isotope_elem
 *             else:
 */
        (void)(__pyx_f_7brainpy_2_c_11composition_element_hash_table_put(__pyx_v_7brainpy_2_c_11composition__ElementTable, __pyx_v_fixed_isotope_elem));

        /* "brainpy/_c/composition.pyx":1250
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                 element_hash_table_put(_ElementTable, fixed_isotope_elem)
 *                 elem = fixed_isotope_elem             # <<<<<<<<<<<<<<
 *             else:
 *                 elem = fixed_isotope_elem
 */
        __pyx_v_elem = __pyx_v_fixed_isotope_elem;

        /* "brainpy/_c/composition.pyx":1247
 * 
 *             found = element_hash_table_get(_ElementTable, temp, &fixed_isotope_elem)
 *             if found != 0:             # <<<<<<<<<<<<<<
 *                 fixed_isotope_elem = make_fixed_isotope_element(elem, fixed_isotope)
 *                 element_hash_table_put(_ElementTable, fixed_isotope_elem)
 */
        goto __pyx_L20;
      }

      /* "brainpy/_c/composition.pyx":1252
 *                 elem = fixed_isotope_elem
 *             else:
 *                 elem = fixed_isotope_elem             # <<<<<<<<<<<<<<
 *         prev_count = 0
 *         composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 */
      /*else*/ {
        __pyx_v_elem = __pyx_v_fixed_isotope_elem;
      }
      __pyx_L20:;

      /* "brainpy/_c/composition.pyx":1236
 *         temp[elend - elstart] = 0
 *         found = element_hash_table_get(_ElementTable, temp, &elem)
 *         if isostart != 0 and isoend != 0 and found != 0:             # <<<<<<<<<<<<<<
 *             isostart += 1
 *             strncpy(temp, cstr + isostart, isoend - isostart)
 */
    }

    /* "brainpy/_c/composition.pyx":1253
 *             else:
 *                 elem = fixed_isotope_elem
 *         prev_count = 0             # <<<<<<<<<<<<<<
 *         composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 *         composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 */
    __pyx_v_prev_count = 0;

    /* "brainpy/_c/composition.pyx":1254
 *                 elem = fixed_isotope_elem
 *         prev_count = 0
 *         composition_get_element_count(composition.impl, elem.symbol, &prev_count)             # <<<<<<<<<<<<<<
 *         composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 * 
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_get_element_count(__pyx_v_composition->impl, __pyx_v_elem->symbol, (&__pyx_v_prev_count)));

    /* "brainpy/_c/composition.pyx":1255
 *         prev_count = 0
 *         composition_get_element_count(composition.impl, elem.symbol, &prev_count)
 *         composition_set_element_count(composition.impl, elem.symbol, count + prev_count)             # <<<<<<<<<<<<<<
 * 
 *     free(temp)
 */
    (void)(__pyx_f_7brainpy_2_c_11composition_composition_set_element_count(__pyx_v_composition->impl, __pyx_v_elem->symbol, (__pyx_v_count + __pyx_v_prev_count)));

    /* "brainpy/_c/composition.pyx":1228
 *                 isoend = 0
 * 
 *     if state == COUNT:             # <<<<<<<<<<<<<<
 *         numend = i + 1
 *         strncpy(temp, cstr + numstart, numend - numstart)
 */
  }

  /* "brainpy/_c/composition.pyx":1257
 *         composition_set_element_count(composition.impl, elem.symbol, count + prev_count)
 * 
 *     free(temp)             # <<<<<<<<<<<<<<
 *     return composition
 */
  free(__pyx_v_temp);

  /* "brainpy/_c/composition.pyx":1258
 * 
 *     free(temp)
 *     return composition             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_composition));
  __pyx_r = __pyx_v_composition;
  goto __pyx_L0;

  /* "brainpy/_c/composition.pyx":1129
 * 
 * 
 * cpdef PyComposition parse_formula(str formula):             # <<<<<<<<<<<<<<
 *     """Parse a chemical formula and construct a :class:`PyComposition` object
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("brainpy._c.composition.parse_formula", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_composition);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7brainpy_2_c_11composition_11parse_formula(PyObject *__pyx_self, PyObject *__pyx_v_formula); /*proto*/
static char __pyx_doc_7brainpy_2_c_11composition_10parse_formula[] = "parse_formula(str formula) -> PyComposition\nParse a chemical formula and construct a :class:`PyComposition` object\n\n    Parameters\n    ----------\n    formula : :class:`str`\n\n    Returns\n    -------\n    :class:`PyComposition`\n\n    Raises\n    ------\n    ValueError\n        If the formula doesn't match the expected pattern\n    ";
static PyObject *__pyx_pw_7brainpy_2_c_11composition_11parse_formula(PyObject *__pyx_self, PyObject *__pyx_v_formula) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse_formula (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_formula), (&PyString_Type), 1, "formula", 1))) __PYX_ERR(0, 1129, __pyx_L1_error)
  __pyx_r = __pyx_pf_7brainpy_2_c_11composition_10parse_formula(__pyx_self, ((PyObject*)__pyx_v_formula));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7brainpy_2_c_11composition_10parse_formula(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_formula) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("parse_formula", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7brainpy_2_c_11composition_parse_formula(__pyx_v_formula, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("brainpy._c.composition.parse_formula", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_7brainpy_2_c_11composition_PyComposition __pyx_vtable_7brainpy_2_c_11composition_PyComposition;

static PyObject *__pyx_tp_new_7brainpy_2_c_11composition_PyComposition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)o);
  p->__pyx_vtab = __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition;
  return o;
}

static void __pyx_tp_dealloc_7brainpy_2_c_11composition_PyComposition(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7brainpy_2_c_11composition_13PyComposition_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_7brainpy_2_c_11composition_PyComposition(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_7brainpy_2_c_11composition_PyComposition(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_11composition_13PyComposition_7__setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop_7brainpy_2_c_11composition_13PyComposition_cached_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_1__get__(o);
}

static int __pyx_setprop_7brainpy_2_c_11composition_13PyComposition_cached_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_11composition_13PyComposition_11cached_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7brainpy_2_c_11composition_13PyComposition__clean(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_1__get__(o);
}

static int __pyx_setprop_7brainpy_2_c_11composition_13PyComposition__clean(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7brainpy_2_c_11composition_13PyComposition_6_clean_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7brainpy_2_c_11composition_PyComposition[] = {
  {"update", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_9update, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_8update},
  {"keys", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_11keys, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_10keys},
  {"values", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_13values, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_12values},
  {"pop", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_15pop, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_14pop},
  {"items", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_19items, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_18items},
  {"copy", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_21copy, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_20copy},
  {"mass", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_37mass, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_36mass},
  {"__equality_pycomposition", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_43__equality_pycomposition, METH_O, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_42__equality_pycomposition},
  {"__equality_dict", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_45__equality_dict, METH_O, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_44__equality_dict},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_49__reduce_cython__, METH_NOARGS, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_48__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_13PyComposition_51__setstate_cython__, METH_O, __pyx_doc_7brainpy_2_c_11composition_13PyComposition_50__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7brainpy_2_c_11composition_PyComposition[] = {
  {(char *)"cached_mass", __pyx_getprop_7brainpy_2_c_11composition_13PyComposition_cached_mass, __pyx_setprop_7brainpy_2_c_11composition_13PyComposition_cached_mass, (char *)"cached_mass: 'double'", 0},
  {(char *)"_clean", __pyx_getprop_7brainpy_2_c_11composition_13PyComposition__clean, __pyx_setprop_7brainpy_2_c_11composition_13PyComposition__clean, (char *)"_clean: 'bool'", 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_PyComposition = {
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_23__add__, /*nb_add*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_27__sub__, /*nb_subtract*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_31__mul__, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_25__iadd__, /*nb_inplace_add*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_29__isub__, /*nb_inplace_subtract*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_33__imul__, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_PyComposition = {
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_39__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_7brainpy_2_c_11composition_PyComposition, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_17__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PyComposition = {
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_39__len__, /*mp_length*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_5__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_7brainpy_2_c_11composition_PyComposition, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_7brainpy_2_c_11composition_PyComposition = {
  PyVarObject_HEAD_INIT(0, 0)
  "brainpy._c.composition.PyComposition", /*tp_name*/
  sizeof(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7brainpy_2_c_11composition_PyComposition, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_41__repr__, /*tp_repr*/
  &__pyx_tp_as_number_PyComposition, /*tp_as_number*/
  &__pyx_tp_as_sequence_PyComposition, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PyComposition, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "PyComposition(base=None, **kwargs)\nA mapping representing a chemical composition.\n\n    Implements arithmetic operations, +/- is defined\n    between a :class:`PyComposition` and a :class:`Mapping`-like\n    object, and * is defined between a :class:`PyComposition` and\n    an integer.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_47__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_35__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7brainpy_2_c_11composition_PyComposition, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7brainpy_2_c_11composition_PyComposition, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_7brainpy_2_c_11composition_13PyComposition_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7brainpy_2_c_11composition_PyComposition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *__pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct____repr__[8];
static int __pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct____repr__ = 0;

static PyObject *__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct____repr__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct____repr__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__)))) {
    o = (PyObject*)__pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct____repr__[--__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct____repr__];
    memset(o, 0, sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7brainpy_2_c_11composition___pyx_scope_struct____repr__(PyObject *o) {
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *p = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct____repr__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__)))) {
    __pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct____repr__[__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct____repr__++] = ((struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7brainpy_2_c_11composition___pyx_scope_struct____repr__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *p = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7brainpy_2_c_11composition___pyx_scope_struct____repr__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *p = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "brainpy._c.composition.__pyx_scope_struct____repr__", /*tp_name*/
  sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct____repr__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7brainpy_2_c_11composition___pyx_scope_struct____repr__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7brainpy_2_c_11composition___pyx_scope_struct____repr__, /*tp_traverse*/
  __pyx_tp_clear_7brainpy_2_c_11composition___pyx_scope_struct____repr__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct____repr__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *__pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr[--__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_kv);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr[__pyx_freecount_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_kv) {
    e = (*v)(p->__pyx_v_kv, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "brainpy._c.composition.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"parse_formula", (PyCFunction)__pyx_pw_7brainpy_2_c_11composition_11parse_formula, METH_O, __pyx_doc_7brainpy_2_c_11composition_10parse_formula},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_composition(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_composition},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "composition",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Cannot_add_s_to_PyComposition, __pyx_k_Cannot_add_s_to_PyComposition, sizeof(__pyx_k_Cannot_add_s_to_PyComposition), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_subtract_s_from_PyComposi, __pyx_k_Cannot_subtract_s_from_PyComposi, sizeof(__pyx_k_Cannot_subtract_s_from_PyComposi), 0, 0, 1, 0},
  {&__pyx_n_s_ElementTable, __pyx_k_ElementTable, sizeof(__pyx_k_ElementTable), 0, 0, 1, 1},
  {&__pyx_kp_s_Error, __pyx_k_Error, sizeof(__pyx_k_Error), 0, 0, 1, 0},
  {&__pyx_kp_s_Fetched_element, __pyx_k_Fetched_element, sizeof(__pyx_k_Fetched_element), 0, 0, 1, 0},
  {&__pyx_kp_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_Massing, __pyx_k_Massing, sizeof(__pyx_k_Massing), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_PROTON, __pyx_k_PROTON, sizeof(__pyx_k_PROTON), 0, 0, 1, 1},
  {&__pyx_kp_s_PyComposition_s, __pyx_k_PyComposition_s, sizeof(__pyx_k_PyComposition_s), 0, 0, 1, 0},
  {&__pyx_n_s_Summed, __pyx_k_Summed, sizeof(__pyx_k_Summed), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_n_s_arg, __pyx_k_arg, sizeof(__pyx_k_arg), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_brainpy__c_composition, __pyx_k_brainpy__c_composition, sizeof(__pyx_k_brainpy__c_composition), 0, 0, 1, 1},
  {&__pyx_kp_s_brainpy__c_composition_pyx, __pyx_k_brainpy__c_composition_pyx, sizeof(__pyx_k_brainpy__c_composition_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_brainpy_mass_dict, __pyx_k_brainpy_mass_dict, sizeof(__pyx_k_brainpy_mass_dict), 0, 0, 1, 1},
  {&__pyx_n_s_c_element_symbol, __pyx_k_c_element_symbol, sizeof(__pyx_k_c_element_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_elem, __pyx_k_elem, sizeof(__pyx_k_elem), 0, 0, 1, 1},
  {&__pyx_n_s_elem_obj, __pyx_k_elem_obj, sizeof(__pyx_k_elem_obj), 0, 0, 1, 1},
  {&__pyx_n_s_element, __pyx_k_element, sizeof(__pyx_k_element), 0, 0, 1, 1},
  {&__pyx_n_s_element_hash_table_get, __pyx_k_element_hash_table_get, sizeof(__pyx_k_element_hash_table_get), 0, 0, 1, 1},
  {&__pyx_n_s_element_name_buffer, __pyx_k_element_name_buffer, sizeof(__pyx_k_element_name_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_element_symbol, __pyx_k_element_symbol, sizeof(__pyx_k_element_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_equality_dict, __pyx_k_equality_dict, sizeof(__pyx_k_equality_dict), 0, 0, 1, 1},
  {&__pyx_n_s_equality_pycomposition, __pyx_k_equality_pycomposition, sizeof(__pyx_k_equality_pycomposition), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_fixed_isotope_element, __pyx_k_fixed_isotope_element, sizeof(__pyx_k_fixed_isotope_element), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_isotope, __pyx_k_isotope, sizeof(__pyx_k_isotope), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_obj, __pyx_k_isotope_obj, sizeof(__pyx_k_isotope_obj), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_parse_test, __pyx_k_isotope_parse_test, sizeof(__pyx_k_isotope_parse_test), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_string, __pyx_k_isotope_string, sizeof(__pyx_k_isotope_string), 0, 0, 1, 1},
  {&__pyx_n_s_isotope_update_test, __pyx_k_isotope_update_test, sizeof(__pyx_k_isotope_update_test), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_main_2, __pyx_k_main_2, sizeof(__pyx_k_main_2), 0, 0, 1, 1},
  {&__pyx_n_s_mass, __pyx_k_mass, sizeof(__pyx_k_mass), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_nist_mass, __pyx_k_nist_mass, sizeof(__pyx_k_nist_mass), 0, 0, 1, 1},
  {&__pyx_n_s_nist_mass_2, __pyx_k_nist_mass_2, sizeof(__pyx_k_nist_mass_2), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr___locals_genexpr, __pyx_k_repr___locals_genexpr, sizeof(__pyx_k_repr___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_kp_s_s_d, __pyx_k_s_d, sizeof(__pyx_k_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_select_1_1, __pyx_k_select_1_1, sizeof(__pyx_k_select_1_1), 0, 0, 1, 1},
  {&__pyx_kp_s_self_impl_cannot_be_converted_to, __pyx_k_self_impl_cannot_be_converted_to, sizeof(__pyx_k_self_impl_cannot_be_converted_to), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_sum_composition, __pyx_k_sum_composition, sizeof(__pyx_k_sum_composition), 0, 0, 1, 1},
  {&__pyx_n_s_sym, __pyx_k_sym, sizeof(__pyx_k_sym), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_test_hash_string, __pyx_k_test_hash_string, sizeof(__pyx_k_test_hash_string), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 984, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 998, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 1175, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_self_impl_cannot_be_converted_to); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":4
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.impl cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self_impl_cannot_be_converted_to); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "brainpy/_c/composition.pyx":187
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 * def __select_1_1(object x):             # <<<<<<<<<<<<<<
 *     return x[1][1]
 * 
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_n_s_x); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_composition_pyx, __pyx_n_s_select_1_1, 187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 187, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":295
 * 
 * 
 * def test_hash_string(str string_):             # <<<<<<<<<<<<<<
 *     return hash_string(PyString_AsString(string_))
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_n_s_string); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_composition_pyx, __pyx_n_s_test_hash_string, 295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 295, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":746
 * # Prototyping and Testing
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef Element* elem
 *     cdef Composition* composition
 */
  __pyx_tuple__10 = PyTuple_Pack(9, __pyx_n_s_elem, __pyx_n_s_composition, __pyx_n_s_sum_composition, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_sym, __pyx_n_s_k, __pyx_n_s_element_name_buffer, __pyx_n_s_copy); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_composition_pyx, __pyx_n_s_main_2, 746, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 746, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":776
 * 
 * 
 * def isotope_update_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */
  __pyx_tuple__12 = PyTuple_Pack(10, __pyx_n_s_element_symbol, __pyx_n_s_c_element_symbol, __pyx_n_s_element, __pyx_n_s_isotope_string, __pyx_n_s_isotope, __pyx_n_s_status, __pyx_n_s_elem_obj, __pyx_n_s_fixed_isotope_element, __pyx_n_s_isotope_obj, __pyx_n_s_out); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(1, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_composition_pyx, __pyx_n_s_isotope_update_test, 776, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 776, __pyx_L1_error)

  /* "brainpy/_c/composition.pyx":808
 * 
 * 
 * def isotope_parse_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */
  __pyx_tuple__14 = PyTuple_Pack(5, __pyx_n_s_element_symbol, __pyx_n_s_c_element_symbol, __pyx_n_s_element, __pyx_n_s_isotope_string, __pyx_n_s_isotope); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_brainpy__c_composition_pyx, __pyx_n_s_isotope_parse_test, 808, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyString_Type_encode.type = (PyObject*)&PyString_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_ = PyFloat_FromDouble(0.); if (unlikely(!__pyx_float_0_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_7brainpy_2_c_11composition_nist_mass = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_7brainpy_2_c_11composition__string_table = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_nist_mass, (void *)&__pyx_v_7brainpy_2_c_11composition_nist_mass, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_PROTON, (void *)&__pyx_v_7brainpy_2_c_11composition_PROTON, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_ElementTable, (void *)&__pyx_v_7brainpy_2_c_11composition__ElementTable, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("neutral_mass", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_neutral_mass, "double (double, int, struct __pyx_opt_args_7brainpy_2_c_11composition_neutral_mass *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("mass_charge_ratio", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_mass_charge_ratio, "double (double, int, struct __pyx_opt_args_7brainpy_2_c_11composition_mass_charge_ratio *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_parse_isotope_string", (void (*)(void))__pyx_f_7brainpy_2_c_11composition__parse_isotope_string, "char *(char *, int *, char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_isotope_map", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_make_isotope_map, "struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *(PyObject *, size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_isotope_by_neutron_shift", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_get_isotope_by_neutron_shift, "struct __pyx_t_7brainpy_2_c_11composition_Isotope *(struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_isotope_map", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_free_isotope_map, "void (struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_isotopes_of", (void (*)(void))__pyx_f_7brainpy_2_c_11composition__isotopes_of, "void (char *, struct __pyx_t_7brainpy_2_c_11composition_IsotopeMap **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_element", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_make_element, "struct __pyx_t_7brainpy_2_c_11composition_Element *(char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_monoisotopic_mass", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_monoisotopic_mass, "double (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_min_neutron_shift", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_min_neutron_shift, "int (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_max_neutron_shift", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_max_neutron_shift, "int (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_element", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_free_element, "void (struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_fixed_isotope_element", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_make_fixed_isotope_element, "struct __pyx_t_7brainpy_2_c_11composition_Element *(struct __pyx_t_7brainpy_2_c_11composition_Element *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_element_hash_bucket", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_free_element_hash_bucket, "void (struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_element_hash_table", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_make_element_hash_table, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *(size_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_hash_bucket_insert", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_hash_bucket_insert, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *, struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_hash_bucket_find", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_hash_bucket_find, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashBucket *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_hash_table_get", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_hash_table_get, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, char *, struct __pyx_t_7brainpy_2_c_11composition_Element **)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("element_hash_table_put", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_element_hash_table_put, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *, struct __pyx_t_7brainpy_2_c_11composition_Element *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("hash_string", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_hash_string, "size_t (char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_element_hash_table", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_free_element_hash_table, "size_t (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_system_element_hash_table", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_get_system_element_hash_table, "struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("set_system_element_hash_table", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_set_system_element_hash_table, "int (struct __pyx_t_7brainpy_2_c_11composition_ElementHashTable *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("make_composition", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_make_composition, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("copy_composition", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_copy_composition, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("print_composition", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_print_composition, "void (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_set_element_count", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_set_element_count, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_get_element_count", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_get_element_count, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_inc_element_count", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_inc_element_count, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, char *, __pyx_t_7brainpy_2_c_11composition_count_type)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_resize", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_resize, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_mass", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_mass, "double (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("free_composition", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_free_composition, "void (struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_add", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_add, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_iadd", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_iadd, "int (struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_mul", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_mul, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(struct __pyx_t_7brainpy_2_c_11composition_Composition *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_imul", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_imul, "void (struct __pyx_t_7brainpy_2_c_11composition_Composition *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("composition_to_dict", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_composition_to_dict, "PyObject *(struct __pyx_t_7brainpy_2_c_11composition_Composition *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dict_to_composition", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_dict_to_composition, "struct __pyx_t_7brainpy_2_c_11composition_Composition *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("parse_formula", (void (*)(void))__pyx_f_7brainpy_2_c_11composition_parse_formula, "struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition = &__pyx_vtable_7brainpy_2_c_11composition_PyComposition;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition._create = (struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*)(struct __pyx_t_7brainpy_2_c_11composition_Composition *))__pyx_f_7brainpy_2_c_11composition_13PyComposition__create;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition._set_impl = (void (*)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_t_7brainpy_2_c_11composition_Composition *, struct __pyx_opt_args_7brainpy_2_c_11composition_13PyComposition__set_impl *__pyx_optional_args))__pyx_f_7brainpy_2_c_11composition_13PyComposition__set_impl;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition.mass = (double (*)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_11composition_13PyComposition_mass;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition.__pyx___equality_pycomposition = (int (*)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_pycomposition;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition.__pyx___equality_dict = (int (*)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_11composition_13PyComposition___equality_dict;
  __pyx_vtable_7brainpy_2_c_11composition_PyComposition.copy = (struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *(*)(struct __pyx_obj_7brainpy_2_c_11composition_PyComposition *, int __pyx_skip_dispatch))__pyx_f_7brainpy_2_c_11composition_13PyComposition_copy;
  if (PyType_Ready(&__pyx_type_7brainpy_2_c_11composition_PyComposition) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
  __pyx_type_7brainpy_2_c_11composition_PyComposition.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7brainpy_2_c_11composition_PyComposition.tp_dictoffset && __pyx_type_7brainpy_2_c_11composition_PyComposition.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7brainpy_2_c_11composition_PyComposition.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7brainpy_2_c_11composition_PyComposition.tp_dict, __pyx_vtabptr_7brainpy_2_c_11composition_PyComposition) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "PyComposition", (PyObject *)&__pyx_type_7brainpy_2_c_11composition_PyComposition) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7brainpy_2_c_11composition_PyComposition) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
  __pyx_ptype_7brainpy_2_c_11composition_PyComposition = &__pyx_type_7brainpy_2_c_11composition_PyComposition;
  if (PyType_Ready(&__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
  __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__.tp_dictoffset && __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct____repr__ = &__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct____repr__;
  if (PyType_Ready(&__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 1076, __pyx_L1_error)
  __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr = &__pyx_type_7brainpy_2_c_11composition___pyx_scope_struct_1_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcomposition(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcomposition(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_composition(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_composition(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_composition(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_composition(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("composition", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_brainpy___c__composition) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "brainpy._c.composition")) {
      if (unlikely(PyDict_SetItemString(modules, "brainpy._c.composition", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "brainpy/_c/composition.pyx":25
 * 
 * 
 * from brainpy.mass_dict import nist_mass as __nist_mass             # <<<<<<<<<<<<<<
 * 
 * cdef dict nist_mass
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_nist_mass);
  __Pyx_GIVEREF(__pyx_n_s_nist_mass);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_nist_mass);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_brainpy_mass_dict, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_nist_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nist_mass_2, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":28
 * 
 * cdef dict nist_mass
 * nist_mass = __nist_mass             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_nist_mass_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_7brainpy_2_c_11composition_nist_mass);
  __Pyx_DECREF_SET(__pyx_v_7brainpy_2_c_11composition_nist_mass, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":31
 * 
 * 
 * cdef double PROTON = nist_mass["H+"][0][0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_7brainpy_2_c_11composition_nist_mass == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 31, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_7brainpy_2_c_11composition_nist_mass, __pyx_kp_s_H); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_7brainpy_2_c_11composition_PROTON = __pyx_t_3;

  /* "brainpy/_c/composition.pyx":34
 * 
 * 
 * cdef double neutral_mass(double mz,  int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (mz * fabs(z)) - (z * charge_carrier)
 * 
 */
  __pyx_k_ = __pyx_v_7brainpy_2_c_11composition_PROTON;

  /* "brainpy/_c/composition.pyx":39
 * 
 * @cython.cdivision
 * cdef double mass_charge_ratio(double neutral_mass, int z, double charge_carrier=PROTON) nogil:             # <<<<<<<<<<<<<<
 *     return (neutral_mass + (z * charge_carrier)) / fabs(z)
 * 
 */
  __pyx_k__2 = __pyx_v_7brainpy_2_c_11composition_PROTON;

  /* "brainpy/_c/composition.pyx":187
 *     return element.isotopes.bins[element.isotopes.size - 1].neutron_shift
 * 
 * def __select_1_1(object x):             # <<<<<<<<<<<<<<
 *     return x[1][1]
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_11composition_1__select_1_1, NULL, __pyx_n_s_brainpy__c_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_select_1_1, __pyx_t_2) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":295
 * 
 * 
 * def test_hash_string(str string_):             # <<<<<<<<<<<<<<
 *     return hash_string(PyString_AsString(string_))
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_11composition_3test_hash_string, NULL, __pyx_n_s_brainpy__c_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_hash_string, __pyx_t_2) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":424
 * 
 * 
 * _ElementTable = make_element_hash_table_populated(256)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_7brainpy_2_c_11composition__ElementTable = __pyx_f_7brainpy_2_c_11composition_make_element_hash_table_populated(0x100);

  /* "brainpy/_c/composition.pyx":746
 * # Prototyping and Testing
 * 
 * def main():             # <<<<<<<<<<<<<<
 *     cdef Element* elem
 *     cdef Composition* composition
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_11composition_5main, NULL, __pyx_n_s_brainpy__c_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_main_2, __pyx_t_2) < 0) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":776
 * 
 * 
 * def isotope_update_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_11composition_7isotope_update_test, NULL, __pyx_n_s_brainpy__c_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isotope_update_test, __pyx_t_2) < 0) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":808
 * 
 * 
 * def isotope_parse_test(str element_symbol):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char* c_element_symbol
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7brainpy_2_c_11composition_9isotope_parse_test, NULL, __pyx_n_s_brainpy__c_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isotope_parse_test, __pyx_t_2) < 0) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":821
 * 
 * 
 * cdef dict _string_table = {}             # <<<<<<<<<<<<<<
 * 
 * cdef str _store_string(str symbol):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(__pyx_v_7brainpy_2_c_11composition__string_table);
  __Pyx_DECREF_SET(__pyx_v_7brainpy_2_c_11composition__string_table, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "brainpy/_c/composition.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init brainpy._c.composition", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init brainpy._c.composition");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if PY_VERSION_HEX >= 0x030700A3
    *type = tstate->exc_state.exc_type;
    *value = tstate->exc_state.exc_value;
    *tb = tstate->exc_state.exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = type;
    tstate->exc_state.exc_value = value;
    tstate->exc_state.exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = local_type;
    tstate->exc_state.exc_value = local_value;
    tstate->exc_state.exc_traceback = local_tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* UnpackUnboundCMethod */
  static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod0 */
  static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
  static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* GetModuleGlobalName */
  static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* ArgTypeTest */
      static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyFunctionFastCall */
      #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
        #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
        #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallMethod0 */
        static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *function = PyMethod_GET_FUNCTION(method);
            result = __Pyx_PyObject_CallOneArg(function, self);
            Py_DECREF(method);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
        static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
        static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
        static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
        static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* RaiseDoubleKeywords */
        static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
        static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
        static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ObjectGetItem */
        #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

/* ExtTypeTest */
        static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
        static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* StringJoin */
        #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* PyObject_GenericGetAttrNoDict */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
        static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
        static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* Import */
        static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
        static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* Print */
        #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* PrintOne */
        #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
        static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = *type;
    tstate->exc_state.exc_value = *value;
    tstate->exc_state.exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
        static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = NULL;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
    goto done;
done:
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) return NULL;
    result = __Pyx__PyObject_CallMethod1(method, arg);
    Py_DECREF(method);
    return result;
}

/* CoroutineBase */
        #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if (0) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if (0) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    if (self->exc_type) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
        __Pyx_ExceptionSave(&self->exc_type, &self->exc_value, &self->exc_traceback);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__pyx_CoroutineObject *self) {
    if (likely(self->exc_traceback)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
            static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
            #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
            static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
            #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
            static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
            static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
            #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
            #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
